package CDG::Utils;

use strict;
use CDG;
use vars qw(@ISA @EXPORT @EXPORT_OK %EXPORT_TAGS
            %words %features %labels %levels $edgecounter);

require Exporter;
@ISA = qw(Exporter);

# Export this by default
@EXPORT = ();

# Export this upon request
@EXPORT_OK = qw(%words %features %labels %levels $edgecounter
                AnnoEntry2hash import_annos format_anno write_anno
                format_string parse_string);

# Make it easier to pose this request
%EXPORT_TAGS =
  ( statistics =>
    [qw(%words %features %labels %levels $edgecounter)],
    functions =>
    [qw(AnnoEntry2hash import_annos format_anno write_anno
        format_string parse_string)]);


=head1 NAME

CDG::Utils - utilities for processing CDG structures from perl

=head1 REQUIRES

CDG

=head1 EXPORTS

Nothing, by default.

:statistics -- hashes of counts of how often each word, label etc. was
seen in the input.

:functions -- utility routines, e.g. translating from AnnoEntryStruct
to a more accessible perl hash.

=head1 DESCRIPTION

This module contains functions that are likely to be needed in more
than one perl utility for manipulating CDG structures.

=head1 VARIABLES

The global variables %levels, %labels etc. contain a count of how
often each level, label etc. was seen in the loaded annotations. The
hash %features contains nested counts of how often a value was seen
for each feature.

=cut

our %levels;                       # all levels mentioned in the annotations
our %labels;                       # all labels mentioned in the annotations
our %words;                        # all word forms in the input
our %features = ( word=>\%words ); # all features in the input
our $edgecounter;                  # counts all edges seen in the input

=head1 METHODS
=cut

=head2 AnnoEntry2hash $ae

Takes an AnnoEntry and creates a full perl representation of this
structure. Returns a reference to a hash with keys such as `lattice',
`words' and `edges' that contains all information in $ae. This
structure is retained and can be safely accessed even if the AnnoEntry
that it was created from is deleted.

=cut
sub AnnoEntry2hash($) {

  my $ae = shift;
  my $id = AnnoEntryStruct_id_get($ae);
  my $lattice = AnnoEntryStruct_lattice_get($ae);

  my %anno;
  my @words;
  my @edges;
  my $pos = 0;

  # insert dummy entry so we can count from 1 like cdgp does
  push @words, { no => 0, word => 'NIL'};

  my $i = 0;
  for (my $k = AnnoEntryStruct_annos_get $ae;$k;$k = listNext $k) {
    my $a = listItem $k;

    # note word
    my %word = ( no => ++$i, word => AnnotationStruct_word_get $a );

    $words{$word{word}}++;

    for (my $m = AnnotationStruct_specs_get $a;$m;$m = listNext $m) {
      my $s = listItem $m;
      my $t = SpecificationStruct_type_get $s;

      # note feature
      if ($t eq $CDG::STTag) {
        my $a = SpecificationStruct_kind_get $s;
        my $v = SpecificationStruct_name_get $s;
        $word{$a} = $v;
        $features{$a}{$v}++;
      }

      # note edge
      else {
        my $level = SpecificationStruct_kind_get $s;
        my $label = SpecificationStruct_name_get $s;
        push @edges, {
                      anno => \%anno,
                      level => $level,
                      label => $label,
                      child => $i,
                      parent => SpecificationStruct_position_get ($s),
                      pos => $pos++,
                      no => $edgecounter++
                     };
        $labels{all}{$label}++;
        $labels{$level}{$label}++;
        $levels{$level}++;
      }
    }
    push @words, \%word;
  }
  # store pointers to the words themselves in each edge
  for my $e (@edges) {
    $e->{'@'} = @words[$e->{child}];
    $e->{'^'} = @words[$e->{parent}];
  }

  %anno = ( id => $id,
            lattice => $lattice,
            words => [@words],
            edges => [@edges] );
  return \%anno;

}

=head2 parse_string $s

Recognizes escapes, quotes and pseudo-umlauts in a string and returns
the normalized version.

Example: "'B\"artram\'s Wundert"ute'" (including the single quotes)
is mapped to "Brtram's Wundertte".

=cut
sub parse_string {
  my $s = shift;
  if ($s =~ /^([\'\"])(.*)\1$/) {
    $s = $2;
    $s =~ s|\\||g;
  }

  $s =~ s/\"a//g;
  $s =~ s/\"o//g;
  $s =~ s/\"u//g;
  $s =~ s/\"A//g;
  $s =~ s/\"O//g;
  $s =~ s/\"U//g;
  $s =~ s/\"s//g;

  return $s;
}



=head2 format_string $s

Insert escapes into a string so that it is valid CDG read syntax.

=cut

sub format_string {
  my $s = shift;
  return "''" if $s eq '';
  if ($s !~ /^[0-9]/ &&
      $s !~ /[ \@\#\$\%^&*()\-=+\\|\`~\"\';:\/!?.>,<\[\]\{\}]/) {
    return $s;
  }
  $s =~ s/([\`\'\"])/\\$1/g;
  return "'$s'";
}



# Return a hash of perl representations of all Annotations in the input.
sub import_annos {
  my $annotations = InputStruct_annotations_get($CDG::inputCurrentGrammar);
  my $n = hashSize($annotations);
  my %result;
  print "Importing $n annotations to perl...\n" if $n;
  for (my $l = hashListOfKeys($annotations); $l; $l = listNext $l) {
    my $ae = hashGet($annotations, listItem($l));
    my $h = AnnoEntry2hash $ae;
    my $id = AnnoEntryStruct_id_get $ae;
    $result{$id} = $h;
  }
  return %result;
}


=head2 format_anno $a

Takes a perl annotation and returns a string describing it, much like
writeAnnotation() does. Although the function tries to copy the output
of writeAnnotation() verbatim, a byte-perfect copy is not guaranteed,
only one that is equivalent.

=cut

sub format_anno {
  my $annotation = shift;
  my $result = '';

  $result .= "\n'$$annotation{id}' : '$$annotation{lattice}' <->\n";
  my $firsttime = 1;
  foreach my $w (@{$annotation->{words}}) {
    my $i = $$w{no};
    next unless $i;
    $result .= ",\n" unless $firsttime;
    $firsttime = 0;
    $result .= sprintf "  %d %d %s\n", $i-1, $i, format_string($$w{word});
    foreach (sort keys %{$w}) {
      next if /\b(no|word|le)\b/;
      $result .= sprintf "'%s' / '%s'\n", $_, $$w{$_};
    }
    foreach my $e (@{$annotation->{edges}}) {
      my $c = $$e{child};
      next unless $c == $i;
      $result .=
        sprintf "'%s' -> '%s' -> %s", $$e{level}, $$e{label}, $$e{parent};
      if ($$e{parent}) {
        my $w2 = $annotation->{words}[$$e{parent}];
        $result .= sprintf " /* ('%s') */", $$w2{word};
      }
      $result .= "\n";
    }
  }
  $result .= ";\n\n";
  return $result;
}


=head1 write_anno $a

Takes a perl annotation and writes a representation of it
to a file of the appropriate name. Overwrites the source
file if that came from the cwd.

=cut
sub write_anno {
  my $a = shift;
  my $name = $$a{id};

  open FILE, ">$name.cda" or warn "Can't write $name.cda: $!\n";
  print FILE format_anno $a;
  close FILE;
}




1;
