/*
 * File: cdg.c
 * Author: Ingo Schroeder
 * Purpose: Main program
 * Birth: 24/2/97
 */

/* ---------------------------------------------------------------------- */
#include <configure.h>
#include <my_errno.h>
#include <my_time.h>
#include <my_unistd.h>

#include <stdio.h>
#include <signal.h>
#include <sys/resource.h>

#include "cdg.h"
#include "command.h"
#include "hook.h"
#include "write.h"
#include "netsearch.h"
#include "statistics.h"

/* ---------------------------------------------------------------------- */

/* ---------------------------------------------------------------------- */
/* -- VARIABLES --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * signal handler
 */
void signalHandler( int sig )
{
  signal(sig, SIG_IGN);

  switch (sig) {
    case SIGINT:
      if (cdgCtrlCAllowed) {
	if (cdgCtrlCTrapped)
	  cdgPrintf(CDG_DEFAULT,
		    "Interrupt already trapped, PLEASE wait a moment...\n");
	else
	  cdgCtrlCTrapped = TRUE;
      } else {
	/* Clear this line. */
	rl_delete_text(0, rl_end);
	/* rl_point is the offset of the cursor position in rl_line_buffer.
	 * It must be set to 0. */
	rl_point = 0;
	cdgPrintf(CDG_DEFAULT, "\n");
	/* Let readline know that we are on a new line. */
	rl_on_new_line();
	/* Display the line stored in rl_line_buffer (which is empty except
	 * for the prompt). */
	rl_redisplay();
      }
      break;
    case SIGTERM:
      cdgPrintf(CDG_WARNING, "WARNING: term signal\n");
      cdgFinalize();
      exit(0);
      break;
    case SIGXCPU:
      cdgPrintf(CDG_ERROR, "ERROR: CPU limit exceeded\n");
      cdgFinalize();
      exit(0);
      break;
    default:
      cdgPrintf(CDG_WARNING, "WARNING: no handler installed for signal `%d'\n",
		sig);
  }

  signal(sig, signalHandler);
}

/* ----------------------------------------------------------------------
 * main
 */
int main( int argc, char **argv )
{
  int i;
  char c;
  String initFile = NULL;
  Boolean ignoreInitFile = FALSE;

  cdgInitialize();
  cdgXCDG = FALSE;

  while ((c = getopt(argc, argv, "b:c:qsdemni:x:")) != EOF) {
    switch (c) {
      case 'b':
	if (1 != sscanf(optarg, "%d", &i)) {
	  cdgPrintf(CDG_WARNING, "WARNING: memory limit: %s", optarg);
	} else {

#ifdef SOLARIS
	  struct rlimit rl;

	  if (getrlimit(RLIMIT_AS, &rl)) {
	    perror("ERROR: ");
	    abort();
	  }
	  rl.rlim_cur = i;
	  if (setrlimit(RLIMIT_AS, &rl)) {
	    perror("ERROR: ");
	    abort();
	  }
#elif DARWIN
	  struct rlimit rl;

	  if (getrlimit(RLIMIT_MEMLOCK, &rl)) {
	    perror("ERROR: ");
	    abort();
	  }
	  rl.rlim_cur = i;
	  if (setrlimit(RLIMIT_MEMLOCK, &rl)) {
	    perror("ERROR: ");
	    abort();
	  }
#else
	  cdgPrintf(CDG_INFO, "INFO: memory limit not supported\n");
#endif
	}
	break;
      case 'c':
	if (1 != sscanf(optarg, "%d", &i)) {
	  cdgPrintf(CDG_WARNING, "WARNING: cpu limit: %s", optarg);
	} else {
#ifndef __CYGWIN__
	  struct rlimit rl;

	  if (getrlimit(RLIMIT_CPU, &rl)) {
	    perror("ERROR: ");
	    abort();
	  }
	  rl.rlim_cur = i;
	  if (setrlimit(RLIMIT_CPU, &rl)) {
	    perror("ERROR: ");
	    abort();
	  }
#else
	  cdgPrintf(CDG_INFO, "INFO: cpu limit not supported\n");
#endif
	}
	break;
      case 'd':
	cnShowDeletedFlag = TRUE;
	break;
      case 'e':
	cnEdgesFlag = cnEdgesOff;
	break;
      case 'i':		       /* initial input */
	strFree(initFile);
	initFile = strRegister(optarg);
	if (strcmp(initFile, "noinit") == 0 ||
	    strcmp(initFile, "NULL") == 0 || strcmp(initFile, "none") == 0)
	  ignoreInitFile = TRUE;
	break;
      case 'm':
	nsSearchModifiesNetFlag = TRUE;
	break;
      case 'n':
	nsNormalization = NSNLinear;
	break;
      case 'q':		       /* initial flags */
	hkVerbosity = CDG_WARNING;
	break;
      case 's':
	statUseStatisticsFlag = TRUE;
	break;
      case 'x':		       /* xml output */
	hkVerbosity |= CDG_XML;
	writeXmlFileSet(optarg);
	writeXmlHeader();
	break;
      case '?':
	abort();
    }
  }

  cdgPrintf(CDG_DEFAULT, STARTUPMSG);

  /* read the init file */
  if (ignoreInitFile) {
    cdgPrintf(CDG_INFO, "INFO: ignoring init files\n");
  } else {
    if (!inputReadInitFile(initFile) && initFile) {
      cdgPrintf(CDG_ERROR, "ERROR: can't read initialization file `%s'\n",
		initFile);
    }
  }

  for (i = optind; i < argc; i++) {
    /* This is a little hacky. */
    cmdLoad(1, &argv[i]);
  }

  /* install private signal handlers */
  if (signal(SIGINT, SIG_IGN) != SIG_IGN)
    signal(SIGINT, signalHandler);
  if (signal(SIGTERM, SIG_IGN) != SIG_IGN)
    signal(SIGTERM, signalHandler);
  if (signal(SIGXCPU, SIG_IGN) != SIG_IGN)
    signal(SIGXCPU, signalHandler);


  commandLoop("cdgp> ");
  cdgFinalize();
  return 0;
}
  

/* ---------------------------------------------------------------------- */
