# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# author: Michael Daum (see also THANKS and AUTHORS)
# initial revision: 2003-09-24
#
# Plugin for the treebank2cdg tool to convert the wsj of the penn treebank.
#
# TODO
# - pod this plugin
# - put it into a separate namespace
#------------------------------------------------------------------------------

=head1 NAME

PennPlugin

=head1 DESCRIPTION

This is a plugin for the C<treebank2cdg> tool to convert the Wallstreet Journal
Corpus of the Penn Treebank.

=head1 IMPLEMENTATION

=head2 Callback Tables

=over 10

=item C<%callbacks>

This table defines implementations of used general callbacks in this plugin.

=cut

%callbacks = (
  "initialize" => \&initialize,
  "finalize" => \&finalize,
);

=item C<%headTable>

This is the head table definition.

format:

   <nodeTag> => ["r|l", "<nodeTag>-<edgeTag>", "<nodeTag>-<edgeTag>", ...]

=cut

if (1) { # set to 0 to use collins head table
%headTable = (
  "ADJP"	=> ["l", '^NNS-', '^QP-', '^NN-', '^ADVP-', '^JJ-', '^VBN-', 
			 '^VBG-', '^ADJP-', '^JJR-', '^NP-', '^JJS-', '^DT-', '^FW-', 
			 '^RBR-', '^RBS-', '^SBAR-', '^RB-',],
			 
  "ADVP"	=> ["r", '^RB-', '^RBR-', '^RBS-', '^FW-', '^ADVP-', '^TO-', '^CD-', 
                         '^JJR-', '^JJ-', '^IN-', '^NP-', '^JJS-', '^NN-', ],
			 
  "CONJP"	=> ["r", '^CC-', '^RB-', '^IN-', ],
  
  "FRAG"	=> ["l", ".*" ], # collins searches from right
  
  "INTJ"	=> ["l"],

  "LST"		=> ["l", '^LS-', '^:-',], # collins searches from right
  
  "NAC"		=> ["l", '^NN-', '^NS-', '^NNP-', '^NNPS-', '^NP-', '^NAAC-', '^EX-', 
		         '^CD-', '^QP-', '^PRP-', '^VBG-', '^JJ-', '^JJS-', 
			 '^JJR-', '^ADJP-', '^FW-', ],
			 
  "NP"		=> ["r", # collins likes POS and JJR to be the head 
			 '^(NN|NNP|NNPS|NNS|NX)-', 
			 '^NP-HD', '^NP-', 
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
			 # see npHeadDetection also

  "NX"		=> ["r", # collins likes POS and JJR to be the head :(
			 '^(NN|NNP|NNPS|NNS|NX)-', 
			 '^NP-HD', '^NP-', 
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
  
  "PP"		=> ["l", '^IN-', '^TO-', '^VBG-', '^VBN-', '^RP-', '^FW-',], 
			 # collins searches from right
  
  "PRN"		=> ["l", ""],
  
  "PRT"		=> ["l", '^RP-'], # collins searches from right
  
  "QP"		=> ["l", '^\$-', '^IN-', '^NNS-', '^NN-', '^JJ-', '^RB-', '^DT-', 
			 '^CD-', '^NCD-', '^QP-', '^JJR-', '^JJS-', ],

  "RRC"		=> ["l", '^VP-', '^NP-', "ADVP", "ADJP", "PP",], #collins searches from right

  # diff to collins: TO or IN is never a head of an S, it's the finit verb
  "S"		=> ["l", '^VP-', '^S-', '^SBAR-', '^ADJP-', '^UCP-', 
			 '^NP-', ],

  "SBAR"	=> ["l", '^WHNP-', '^WHPP-', '^WHADVP-', '^WHADJP-', '^IN-', '^DT-', 
			 '^S-', '^SQ-', '^SINV-', '^SBAR-', '^FRAG-', ],
  
  "SBARQ"	=> ["l", '^SQ-', '^S-', '^SINV-', '^SBARQ-', '^FRAG-', ],

  "SINV"	=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^S-', 
			 '^SINV-', '^ADJP-', '^NP-', ],

  "SQ"		=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^SQ-', ],
  
  "UCP"		=> ["l", ".*"], # collins searches from right

  # diff to collins: TO is never a head of a VPm it's the verb
  "VP"		=> ["l", '^VBD', '^VBN', '^MD', '^VBZ', '^VB', '^VBG', 
			 '^VBP', '^VP', '^ADJP', '^NN', '^NNS', '^NP'], 

  "WHADJP"	=> ["l", '^CC-', '^WRB-', '^JJ-', '^ADJP-',],

  "WHADVP"	=> ["r", '^CC-', '^WRB-',],

  "WHNP"	=> ["l", '^WDT-', '^WP-', '^WP\$-', '^WADJP-', '^WHPP-', '^WHNP-', ],

  "WHPP"	=> ["l", '^IN-', '^TO-', '^FW-'],
  "X"		=> ["l"],
);
# collins' head table 
} else {
%headTable = (
  "ADJP"	=> ["l", '^NNS-', '^QP-', '^NN-', '^\$-', '^ADVP-', '^JJ-', '^VBN-', 
			 '^VBG-', '^ADJP-', '^JJR-', '^NP-', '^JJS-', '^DT-', '^FW-', 
			 '^RBR-', '^RBS-', '^SBAR-', '^RB-',],
			 
  "ADVP"	=> ["r", '^RB-', '^RBR-', '^RBS-', '^FW-', '^ADVP-', '^TO-', '^CD-', 
                         '^JJR-', '^JJ-', '^IN-', '^NP-', '^JJS-', '^NN-', ],
			 
  "CONJP"	=> ["r", '^CC-', '^RB-', '^IN-', ],
  
  "FRAG"	=> ["r", ".*" ], # why from right
  
  "INTJ"	=> ["l"],

  "LST"		=> ["r", '^LS-', '^:-',], # why from right
  
  "NAC"		=> ["l", '^NN-', '^NS-', '^NNP-', '^NNPS-', '^NP-', '^NAAC-', '^EX-', 
		         '^\$-', '^CD-', '^QP-', '^PRP-', '^VBG-', '^JJ-', '^JJS-', 
			 '^JJR-', '^ADJP-', '^FW-', ],
			 
		    
  "NP"		=> ["r", # if the last word is POS return this TODO write a conversion
			 '^(NN|NNP|NNPS|NNS|NX|POS|JJR)-', 
			 '^NP-', # search this NP _left to right_
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
			 # see npHeadDetection also

  "NX"		=> ["r", # if the last word is POS return this TODO write a conversion
			 '^(NN|NNP|NNPS|NNS|NX|POS|JJR)-', 
			 '^NP-', # search this NP _left to right_
			 '^(\$|ADJP|PRN)-', '^CD-', '^(JJ|JJS|RB|QP)-',], 
  
  "PP"		=> ["r", '^IN-', '^TO-', '^VBG-', '^VBN-', '^RP-', '^FW-',], # why from right
  
  "PRN"		=> ["l", ".*"],
  
  "PRT"		=> ["r", '^RP-'], # why from right
  
  "QP"		=> ["l", '^\$-', '^IN-', '^NNS-', '^NN-', '^JJ-', '^RB-', '^DT-', 
			 '^CD-', '^NCD-', '^QP-', '^JJR-', '^JJS-', ],

  "RRC"		=> ["r", '^VP-', '^NP-', "ADVP", "ADJP", "PP",], # why from right

  "S"		=> ["l", '^TO-', '^IN-', '^VP-', '^S-', '^SBAR-', '^ADJP-', '^UCP-', 
			 '^NP-', ],

  "SBAR"	=> ["l", '^WHNP-', '^WHPP-', '^WHADVP-', '^WHADJP-', '^IN-', '^DT-', 
			 '^S-', '^SQ-', '^SINV-', '^SBAR-', '^FRAG-', ],
  
  "SBARQ"	=> ["l", '^SQ-', '^S-', '^SINV-', '^SBARQ-', '^FRAG-', ],

  "SINV"	=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^S-', 
			 '^SINV-', '^ADJP-', '^NP-', ],

  "SQ"		=> ["l", '^VBZ-', '^VBD-', '^VBP-', '^VB-', '^MD-', '^VP-', '^SQ-', ],
  
  "UCP"		=> ["r", ".*"], # why from right

  "VP"		=> ["l", '^TO-', '^VBD', '^VBN', '^MD', '^VBZ', '^VB', '^VBG', 
			 '^VBP', '^VP', '^ADJP', '^NN', '^NNS', '^NP'], 

  "WHADJP"	=> ["l", '^CC-', '^WRB-', '^JJ-', '^ADJP-',],

  "WHADVP"	=> ["r", '^CC-', '^WRB-',],

  "WHNP"	=> ["l", '^WDT-', '^WP-', '^WP\$-', '^WADJP-', '^WHPP-', '^WHNP-', ],

  "WHPP"	=> ["l", '^IN-', '^TO-', '^FW-'],
  "X"		=> ["l"],
);
}

=item C<@labelTable>

This maps penn treebank to dependency edge labels. Note that the trigger-label
C<_root> will not become part of the syntax level of the dependency tree but
trigger the dependency conversion C<attachToRoot()>. So these rules plus
C<attachToRoot()> are a way to select which edges become a 
ROOT of the dependency tree.

format:

   1: <dependency label>
   2: <modifier-posTag>-<modifier-edgeTag>
   3: <modifiee-posTag>-<modifiee-edgeTag>
   4: <domiNode-nodeTag>-<domiNode-edgeTag>

=cut

@labelTable = (
 [ '_root', ",", ".*", ".*"],
 [ '_root', '\.', '.*', '.*'],
 [ '_root', ':', '.*', '.*'],
 [ '_root', '\'\'', '.*', '.*'],
 [ '_root', '``', '.*', '.*'],
 [ 'S', ".*", "ROOT-ROOT", ".*"], 
 [ 'S', "-S", ".*", ".*"], 
 [ 'SBJ', '.*', '.*', '-SBJ'],
 [ 'NP', '-NP', '.*', '.*'],
 [ 'VP', '-VP', '.*', '.*'],
 [ 'PP', '-PP', '.*', '.*'],
 [ 'ADJP', '-ADJP', '.*', '.*'],
 [ 'ADVP', '-ADVP', '.*', '.*'],
 [ '', '.*', '.*', '.*'],
);

=item C<%phraseConvTable>

phrase tree conversion rules

=cut

%phraseConvTable = (
  "SBARQ" => \&setEdgeTag, ## TODO need a list of callbacks here
  "SBAR" => \&setEdgeTag,
  "S" => \&setEdgeTag,
  "NP" => \&npHeadDetection,
  "VP" => \&setEdgeTag,
  "PP" => \&setEdgeTag,
  "ADJP" => \&setEdgeTag,
  "ADVP" => \&setEdgeTag,
);

=item C<%depConvTable>

dependency tree conversion rules

=back

=cut

%depConvTable = (
  "_root" => \&attachToRoot,
);

# -----------------------------------------------------------------------------

=head2 Functions

=over 10

=cut

# -----------------------------------------------------------------------------

=item C<ccPhraseConversion()>

=cut

sub ccPhraseConversion {
}

# -----------------------------------------------------------------------------

=item C<npHeadDetection()>

=cut
sub npHeadDetection {
  my $node = shift;

  # detect two competing NP children
  my $headNP;
  foreach my $child (@{$node->{children}}) {
    if ($child->{nodeTag} eq "NP") {
      if (!$headNP || $headNP->{id} > $child->{id}) {
	$headNP = $child;
      }
    }
  }
  if ($headNP) {
    $headNP->{edgeTag} = "HD";
  }

  # chain
  return &setEdgeTag($node);
}

# -----------------------------------------------------------------------------

=item C<setEdgeTag()>

phrase tree conversion to propagate the nodeTag if heads to its edgeTag

=cut

sub setEdgeTag {
  my $node = shift;

  foreach my $child (@{$node->{children}}) {
    if ($child->{isLexHead}) {
      $child->{edgeTag} = $node->{nodeTag};
      return 1
    }
  }

  return 0;
}

#------------------------------------------------------------------------------

=item C<attachToRoot()>

attach a 

=cut

sub attachToRoot {
  my $edge = shift;

  $$edge[1] = "";
  $$edge[2] = $rootNode;

  return 1;
}

# -----------------------------------------------------------------------------

=item C<initializer()>

plugin initializer

=cut

sub initialize {
}

# -----------------------------------------------------------------------------

=item finalize

plugin finalizer

=cut

sub finalize {
}

=back

=cut

# -----------------------------------------------------------------------------

=head1 AUTHORS

Michael Daum E<lt>micha@nats.informatik.uni-hamburg.deE<gt>

=head1 COPYRIGHT

Copyright (C) 1997-2003 The CDG Team E<lt>cdg@nats.informatik.uni-hamburg.deE<gt>

This file is free software; as a special exception the author gives
unlimited permission to copy and/or distribute it, with or without 
modifications, as long as this notice is preserved.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

=cut

