
/* 
 * Copyright (C) 1997-2003 The CDG Team <cdg@nats.inwordatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: cdgdb.c,v 1.13 2003/12/03 13:56:28 foth Exp $ 
 */

/* -------------------------------------------------------------------------
 * @addtogroup Cdgdb Cdgdb - Get lexical entries out of a berkeley database
 * @author maurer
 * @date unknown
 *
 * This file provides an interface for retrieving lexical entries out of a
 * berkeley database. The database file contains only references to an
 * underlying cdg file which actually holds the entries. For loading the
 * requested lexical entries, a temp file will be created out of the cdg
 * file, containing only these entries. This file will then be parsed with
 * \b inpuLoad() into the memory.
 * @{
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <configure.h>
#include <my_unistd.h>
#include <my_time.h>

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/resource.h>
#include <blah.h>
#include "cdg.h"
#include "input.h"
#include "lexemgraph.h"
#include "constraintnet.h"
#include "eval.h"
#include "statistics.h"
#include "netsearch.h"
#include "frobbing.h"
#include "command.h"
#include "increment.h"
#include "scache.h"
#include "hook.h"
#include "timer.h"
#ifndef DARWIN
#include "compile.h"
#endif
#include "gls.h"
#include "incrementalcompletion.h"
#include "parse.h"
#include "parsing.h"
#include "write.h"
#include "search.h"
#include "testing.h"
#include "set.h"
#include "tagger.h"
#include "chunker.h"
#include "db.h"

/* -- VARIABLES ------------------------------------------------------------ */

/** database handle. The handle to the database struct */
static DB *database;

/** the underlying cdg file. This is actually the file, where the lexical
  * entries are stored in. */
static FILE *cdgstream;

/** the filename of the temporary file. This temporary file will contain
  * a set of lexical entries that can then be parsed into the memory. */
static String tmpFilename = NULL;

/** table of words already loaded. Holds the word which are yet retrieved
  * from the database, so that the database won't be questioned for them */
static Hashtable done = NULL;

/** a database entry. 
  * This is an entry for the database which holds a key and the byte position
  * of its lexical entry in the underlying cdg file. */
typedef struct { 
  String key;  /**< the key of the lexical entry. */
  int oBegin; /**< offset of its Beginning in the file */
  int oEnd; /**< offset of its End in the file */
} entryStruct;

/** a pointer to the entryStruct */
typedef entryStruct *dbEntry;


/* -- FUNCTIONS ------------------------------------------------------------ */
Boolean dbOpen(String filename);
Boolean dbClose(void);
Boolean dbLoad(String key); 
Boolean dbLoadEntries(List entries);
List dbGetEntries(String key);
void dbFinalize(void);
Boolean dbOpenCdgFile(String filename); 
dbEntry newDbEntry(void); 
Boolean dbAvailable(void);

/* -- IMPLEMENTATION ------------------------------------------------------- */

/* ------------------------------------------------------------------------- 
 * opens a database.
 * this function opens a database, specified by the parameter 'filename'
 * and stores the handle in the static variable 'database'. If there is no
 * database file with this filename, a new database is created.
 * @param filename the filename of the database file
 * @returns TRUE if success, FALSE otherwise
 */
Boolean dbOpen(String filename)
{
  int ret = 0;		     // return code for database operations

  // create database handle
  if ((ret = db_create(&database, NULL, 0)) != 0) {
    cdgPrintf(CDG_ERROR, "ERROR (db_create): %s\n", db_strerror(ret));
    return FALSE;
  }
  // setting database to accept more than one data item for one key
  // sort order is lexical (this is the default)
  database->set_flags(database, DB_DUP);

  /* open database 'filename' */
  if ((ret =
       database->open(database, filename, NULL, DB_BTREE, DB_RDONLY, 0664)) != 0) {
    cdgPrintf(CDG_ERROR, "ERROR (database->open): %s\n", db_strerror(ret));
    database = NULL;
    return FALSE;
  }
  return TRUE;
}


/* ------------------------------------------------------------------------- 
 * close the database.
 * writes database to disk and closes it.
 * @returns TRUE if success, FALSE otherwise
 */

Boolean dbClose(void)
{
  int ret = 0;		     // return code for database operations

  if (database == NULL) {
    // cdgPrintf(CDG_ERROR, "ERROR: No open database found for closing\n");
    return FALSE;
  }

  if ((ret = database->close(database, 0)) != 0) {
    cdgPrintf(CDG_ERROR, "ERROR (database->close): %s\n", db_strerror(ret));
    return FALSE;
  }
  database = NULL;
  return TRUE;
}


/* ------------------------------------------------------------------------- 
 * gets lexical entries from the database.
 * gets a result list of LexiconItem matching the given word (the key) from 
 * the database.

   This function succeeds only once for a given word form.

 * @param word the word which is the key
 * @returns a list of dbEntry struct, all of them having the same key.
 * @returns NULL if there was an Error
 */

List dbGetEntries(String word)
{
  int ret = 0;		     // return code for database operations
  DBT key, value;	     // DBT structs for key and value
  DBC *cursorp;		     // cursor for getting duplicates
  dbEntry ent;
  List entries = NULL;	     // the return list
  
  if (!database) {
    cdgPrintf(CDG_ERROR, "ERROR: No database opened\n");
    return entries;
  }

  // create cursor for getting duplicates
  database->cursor(database, NULL, &cursorp, 0);

  memset(&key, 0, sizeof (key));
  memset(&value, 0, sizeof (value));
  key.data = word;
  key.size = strlen(word);

  // position the cursor at the right key and get the first item
  if ((ret = cursorp->c_get(cursorp, &key, &value, DB_SET)) != 0) {
    if (ret == DB_NOTFOUND) {
      return NULL;
    } else {
      cdgPrintf(CDG_ERROR, "ERROR (cursorp->c_get): %s\n", db_strerror(ret));
      return NULL;
    }
  }
  ent = newDbEntry();
  ent->key = strCopy(key.data);
  ent->oBegin = *((int *)value.data);
  ent->oEnd = *((int *)value.data + 1);
  entries = listAppendItem(entries, ent);
  // now get all duplicates
  while ((ret =
	  cursorp->c_get(cursorp, &key, &value, DB_NEXT_DUP)) != DB_NOTFOUND) {
    if (ret != 0) {
      cdgPrintf(CDG_ERROR, "ERROR (cursorp->c_get): %s\n", db_strerror(ret));
      break;
    }
    
    ent = newDbEntry();
    ent->key = strCopy(word);
    ent->oBegin = *((int *)value.data);
    ent->oEnd = *((int *)value.data + 1);
    entries = listAppendItem(entries, ent);
  }
  return entries;
}

/* ------------------------------------------------------------------------- 
 * loads lexical entries into the memory.
 * generates a temp file which holds lexical entries and parses them into
 * the memory
 * @param entries a list of dbEntry structs you want to parse
 * @returns True if success, else FALSE
 */
Boolean dbLoadEntries(List entries)
{
  dbEntry ent;
  List l;
  FILE *tmpstream;
  unsigned long mode;
  tmpFilename = strPrintf("/tmp/dbload.%d.cdg", getpid());

  /* if there are no entries for loading*/
  if(!entries) {
    return FALSE;
  }
  
  /* open a temporary file */
  if((tmpstream = fopen(tmpFilename, "w")) == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR (dbLoadEntries): cannot generate tempfile\n");
    return FALSE;
  }
  for (l = entries; l != NULL; l = listNext(l)) {
    ent  =  (dbEntry)listItem(l); 
    fseek(cdgstream, (ent->oBegin -1), SEEK_SET); // seek to begin of entry
    while(ftell(cdgstream) != ent->oEnd) {
      putc(getc(cdgstream), tmpstream);
    }
    putc((int)'\n', tmpstream);
  }
  listForeachFree(entries, free);
  fclose(tmpstream);

  /* suppress uninteresting `loading 58749857.cdg' message */
  mode = hkVerbosity;
  hkVerbosity = hkVerbosity & (~CDG_INFO);
  inputLoad(tmpFilename);
  hkVerbosity = mode;
      
  /* promote new data to current data */
  mergeInput(inputCurrentGrammar, parseResult);
  inputCacheInput(inputCurrentGrammar);
  memFree(parseResult);
  parseResult = NULL;
  
  return TRUE;
}


/* ------------------------------------------------------------------------- 
 * constructor for dbEntry objects.
 * @returns a new dbEntry 
 */
dbEntry newDbEntry() {
  dbEntry entry = memMalloc(sizeof(entryStruct));
  entry->key = NULL;
  entry->oBegin = 0;
  entry->oEnd = 0;
  return entry;
}

/* ------------------------------------------------------------------------- 
 * wrapper for dbLoadEntries().
 * generates a temp file which holds lexical entries and parses them into
 * the memory
 * @param key the key of the lex entries which have to be parsed 
 * @returns TRUE if success, else FALSE
 */

Boolean dbLoad(String key) {
  List records = NULL;
  if(!dbAvailable()) {
    return FALSE;
  }

  /* detect duplicate query */
  if(hashGet(done, strRegister(key))) {
    return TRUE;
  }
  records = dbGetEntries(key);

  /* provide interesting `loading form "Schnabeltier"' message */
  if(records) {
    cdgPrintf(CDG_INFO, "INFO: loading form `%s'\n", key);

    /* remember never to read this key again */
    hashPut(done, strRegister(key), (Pointer)TRUE);
    return dbLoadEntries(records);
  }
  return TRUE;
}

/* ------------------------------------------------------------------------- 
   Another wrapper for dbLoadEntries().

   It loads all known lexicon items for words in the list FORMS. This has
   exactly the same effect as calling dbLoad() repeatedly, but is faster
   because there will be only one file system operation instead of n.
*/
Boolean dbLoadAll(List forms) {
  List records = NULL;
  List l;
  int i = 0;
  
  if(!dbAvailable()) {
    return FALSE;
  }
  for(l = forms; l != NULL; l = listNext(l)) {
    String form = listItem(l);

    /* detect duplicate query */
    if(hashGet(done, form)) {
      continue;
    }
    
    records = listAppendList(records, dbGetEntries(form));
    hashPut(done, strRegister(form), (Pointer)TRUE);
    if(0 == (++i % 1000)) {
      cdgPrintf(CDG_PROGRESS, "%d forms...\n", i);
    }
  }
  cdgPrintf(CDG_INFO, "INFO: loading %d items\n", listLength(records));

  return dbLoadEntries(records);
}


/* ------------------------------------------------------------------------- 
 * Initialize
 */

void dbInitialize(void) {
  done = hashNew( 500, 0.8,
                  hashStringHashFunction,
                  hashStringEqualFunction );
}

/* ------------------------------------------------------------------------- 
 * finalize 
 * close the database, delete \b done hash, remove temporary file
 */

void dbFinalize(void) {
  dbClose();
  if (tmpFilename) {
    remove(tmpFilename);
  }
  hashDelete(done);
}


/* ------------------------------------------------------------------------- 
 * open the cdg file.
 * opens the cdg file, which the database is indexing
 * you have to ensure that it is the right one, otherwise you will get crap
 * @param filename the name of the cdg file
 * @returns TRUE if success, else FALSE
 */

Boolean dbOpenCdgFile(String filename) {
  cdgstream = fopen(filename, "r");
  if(!cdgstream) {
    database = NULL;
  }
  return (cdgstream != NULL);
}

/* ------------------------------------------------------------------------- 
 * checks if there is a database handle available. 
 * @returns TRUE if there is a database opened, else FALSE
 */

Boolean dbAvailable(void) {
  return (database != NULL);
}
/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */
