/*
 * Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: constraintnet.h,v 1.61 2003/09/30 13:53:46 micha Exp $
 */

/* ---------------------------------------------------------------------------
 * @addtogroup Constraintnet
 * @{
 */

#ifndef CONSTRAINTNET_H
#define CONSTRAINTNET_H

/* ---------------------------------------------------------------------- */
#include <blah.h>
#include "input.h"
#include "scorematrix.h"
#include "lexemgraph.h"
#include "scache.h"

/* ---------------------------------------------------------------------- 
 * The types of a constraintnet edge
 */
typedef enum {
  cnEdgesOn,
  cnEdgesOff,
  cnEdgesFew,
  cnEdgesAll
} CnEdgesType;

/* ---------------------------------------------------------------------- 
 * The states of a constraintnet
 */
typedef enum {
   NSCreated,			       /**< net just created */
   NSUnInit,			       /**< node value initialized */
   NSUnSel,			       /**< unary selection function applied */
   NSUnFil,			       /**< node consistency */
   NSBinInit,			       /**< edges initialized */
   NSBinSel,			       /**< binary selection function applied */
   NSBinFil			       /**< edge consistency */
} NetState;

/* ----------------------------------------------------------------------
 * The constraint net.
 */
typedef struct {
   String id;			       /**< A unique identifier for the net.
					 Each net created by the system
					 is labeled as net <i>n</i>, where <i>n</i>
					 is the current value of cnCounter.  */
   NetState state;		       /**< State of the net */
   LexemGraph lexemgraph;	       /**< Points to the enriched word graph used
					 in constructing the net. */
   Vector values;		       /**< Vector of LevelValues */
   Vector nodes;		       /**< Vector of ConstraintNode */
   Vector edges;		       /**< Vector of ConstraintEdge */
   Agenda searchagenda;		       /**< Agenda for searching, it is used
					 by netsearch() */
   int totalNumberOfValues;	       /**< Holds the total number of LVs in the
					 constraint net. It should always be equal
					 to vectorSize(net->values). */
   List parses;		               /**< Contains all structures of type Parse
					 found by any solution method. */
   ScoreCache cache;		       /**< Holds the cache used to hold the results
					 of binary constraint evaluations. (This
					 cache only exists if scUseCache was set
					 when creating the net.) */
   int evalUnary;		       /**< Counting unary evaluations */
   int statUnary;		       /**< Counting unary statistics */
   int evalBinary;		       /**< Counting binary evaluations */
   Vector lvTotals;                    /**< Records how many LVs the net contained
					 at each successive step in its history.
					 This is only used by incrementalcompletion. */
}  ConstraintNetStruct;

typedef ConstraintNetStruct *ConstraintNet; /**< Pointer to the ConstraintNetStruct */


#ifndef SWIG
/* ----------------------------------------------------------------------
 * Models a node in a constraint net
 */
typedef struct {
   ConstraintNet net;		       /**< corresponding constraint net */
   Level level;		       /**< points to the level for which 
					 a constraint node was built. */
   GraphemNode gn;		       /**< corresponding grapheme node */
   Vector values;		       /**< holds shallow copies of all LVs
					 that may be used to bind this constraint
					 node. */
   int totalNumberOfValues;	       /**< number of level values */
   int noValidValues;		       /**< number of not yet deleted LVs */
   Vector *localPrevSupport;	       /**< correspond to the structure of the
					 same name in the arc consistency
					 algorithm */
   Vector *localNextSupport;	       /**< correspond to the structure of the
					 same name in the arc consistency
					 algorithm */
} ConstraintNodeStruct;

/**
 * Pointer to the ConstraintNodeStruct
 */
typedef ConstraintNodeStruct *ConstraintNode;


/* ----------------------------------------------------------------------
 * Models an edge between the two constraint nodes
 * \b start and \b stop. These two fields are shallow copies of
 * the nodes in the corresponding constraint net.
 */
typedef struct ConstraintEdgeStruct {
   ConstraintNode start;	       /**< start node */
   ConstraintNode stop;		       /**< end node */
   struct ConstraintEdgeStruct *reverse;	/**< points to the inverse
						  of the edge itself. */
   ScoreMatrix scores;		       /**< holds all binary scores calculated
					 for pairs of LVs from the two nodes */
   Boolean *m;			       /**< flag whether <x,a> is already processed */
   Vector *prevSupport;		       /**< used by module arcconsistency */
   Vector *nextSupport;		       /**< used by module arcconsistency */
   Vector *s;			       /**< used by module arcconsistency */
   int *counter;		       /**< counts support */
   Boolean *isMarked;		       /**< flag for marking, obsolete? */
} ConstraintEdgeStruct;

/**
 * Pointer to the ConstraintEdgeStruct
 */
typedef ConstraintEdgeStruct *ConstraintEdge;

/* ----------------------------------------------------------------------
 * This serves merely merely to pair a constraint node and an LV.
 */
typedef struct {
   ConstraintNode node;		       /**< constraint node */
   LevelValue value;		       /**< level value */
} NodeBindingStruct;

/**
 * Pointer to the NodeBindingStruct
 */
typedef NodeBindingStruct *NodeBinding;
#endif

/* ----------------------------------------------------------------------
 * Holds information about a constraint violated by a
 * solution.
 */
typedef struct {
   Constraint constraint;	       /**< violated constraint */
   Number penalty;		       /**< holds the penalty of this particular
					 instance of \b constraint (recall that a
					 constraint declaration specifies an entire
					 set of constraints, possibly with variable
					 penalties). */
   int nodeBindingIndex1;	       /**< holds the position of the LV causing
					 the conflict, as calculated by
					 lvIndex(). */
   LevelValue lv1;		       /**< first node binding */
   int nodeBindingIndex2;	       /**< holds the position of the LV causing
					 the conflict, as calculated by
					 lvIndex(). */
   LevelValue lv2;		       /**< second node binding, maybe empty */
} ConstraintViolationStruct;

/**
 * Pointer to the ConstraintViolationStruct
 */
typedef ConstraintViolationStruct *ConstraintViolation;

/* ---------------------------------------------------------------------- */

#ifndef SWIG
extern int cnCounter; /* FIXME: shouldn't this be static inside constraintnet.c */
#endif

#ifdef SWIG
#ifdef OLD_SWIG
%readwrite;
#else
%mutable;
#endif
#endif
/* pointer to most recently created constraint net */
extern ConstraintNet cnMostRecentlyCreatedNet;
#ifdef SWIG
#ifdef OLD_SWIG
%readonly;
#else
%immutable;
#endif
#endif

/* this replaces both cnSuppressEdgesFlag and cnFewEdgesFlag */
extern CnEdgesType cnEdgesFlag;

/* If TRUE, the deleted levelvalues are printed too. */
extern Boolean cnShowDeletedFlag;

/* If TRUE, construct NONSEPC levelvalues in constraint nets. */
extern Boolean cnUseNonSpec;

/*
 * Fraction of all levelvalues in one domain that are deleted based
 * only on the unary score. A factor of 1.0 means keep all values.
 */
extern Number cnUnaryPruningFraction;

/*
 * initial sortation of constraintnodes in constraintnets
 * (relevant in netsearch)
 * 0 no sorting of constraintnodes
 * 1 sort nodes by level->prio
 * 2 sort nodes by increasing domain size (i. e. small domains first)
 */
extern int cnSortNodesMethod;

/* ---------------------------------------------------------------------- */

/*
 * Looks for a constraint net with id ID and return it or NULL if
 * it can't find the net in inputCurrentGrammar.
 */
extern ConstraintNet cnFindNet(String id);

#ifndef SWIG
extern ConstraintNet cnBuildInit();
extern ConstraintNet cnBuildFinal(ConstraintNet net);
extern ConstraintNet cnBuild(Lattice lat);
extern void cnDelete(ConstraintNet net);
extern void cnDeleteNode(ConstraintNode cn);
extern void cnDeleteEdge(ConstraintEdge cn);
extern void cnDeleteBinding(NodeBinding nb);
extern void cnPrint(long unsigned int mode, ConstraintNet net);
extern void cnPrintNode(long unsigned int mode, ConstraintNode cn);
extern void cnPrintEdge(long unsigned int mode, ConstraintEdge e);
extern void cnBuildLevelValues(ConstraintNode node, Level level, GraphemNode modifier, GraphemNode modifiee);
extern Boolean cnBuildIter(ConstraintNet net, GraphemNode gn);
extern Boolean cnBuildNodes(ConstraintNet net);
extern void cnBuildEdges(ConstraintNet net);
extern void cnBuildLv(ConstraintNode node, List modifiers,
                      Level level, String label, List modifiees); 
extern Boolean cnRenew(ConstraintNet net);
extern Boolean cnIsStartNode(ConstraintNode n);
extern Boolean cnIsEndNode(ConstraintNode n);
extern Boolean cnNodeFollowsDirectly(ConstraintNode a, ConstraintNode b);
extern Boolean cnCompareViolation(ConstraintViolation a,
				  ConstraintViolation b); /* TODO: never used */
extern Boolean cnConnectedByArc(ConstraintNode a, ConstraintNode b);
extern void cnPrintInfo(ConstraintNet net);
extern void cnUnaryPruning(ConstraintNode node);
extern int cnOptimizeNet(ConstraintNet net);
extern int cnOptimizeNode(ConstraintNet net, ConstraintNode node);
extern void cnSortNodes(ConstraintNet net);
extern void cnSortLVs(ConstraintNet net);
extern void cnPrintParses(ConstraintNet net);
extern ConstraintNode cnFindNode(ConstraintNet net, LevelValue lv);

extern ConstraintViolation cvNew(Constraint c, LevelValue lva, LevelValue lvb);
extern void cvDelete(ConstraintViolation cv);
extern void cvAnalyse(ConstraintViolation cv, Vector context);
extern void cvPrint(unsigned long mode, ConstraintViolation cv);
extern Boolean cvContains(List conflicts, ConstraintViolation cv);
extern ConstraintViolation cvClone(ConstraintViolation cv);
extern Boolean cvCompare(ConstraintViolation a, ConstraintViolation b);
extern Boolean cvCompareNatural(ConstraintViolation a, ConstraintViolation b);
extern void cnPrintActiveLVs(ConstraintNet net);
extern void cnUndeleteAllLVs(ConstraintNet net);
extern void cnDeleteAllLVs(ConstraintNet net);
extern Boolean cnBuildUpdateArcs(ConstraintNet net, List listArcs);
extern GraphemNode cnGetGraphemNodeFromArc(ConstraintNet net, Arc arc);

extern void cnInitialize();
extern void cnCallback(String name, float *var);
#endif

extern Lattice cnGetLattice(ConstraintNet cn);

/* ---------------------------------------------------------------------- */
#endif /* don't insert anything after this #endif */
/** @} */
