/* 
 * Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Michael Schulz (see also AUTHORS and THANKS)
 * Birth: 12 Juli 1998
 *
 * $Id: hook.c,v 1.35 2003/09/04 07:50:21 micha Exp $
 */

/* -------------------------------------------------------------------------
 * @addtogroup HookCore HookCore - C Part in the core system
 * @ingroup Hook
 * @{
 *
 * The Hook module connects the core of the library to its outside world by a callback
 * mechanism.
 * Hook functions are to be supplied by the application programmer and will
 * be called at strategic points in the application. For instance, a
 * commonly used hook is \ref HOOK_PRINTF, that handles the output channels (see below).
 *
 * So hooking the \ref HOOK_PRINTF callback allows you to redefine and redirect the channels.
 * The following callbacks are defined:
 * - \ref HOOK_CNBUILDNODES
 * - \ref HOOK_EVAL
 * - \ref HOOK_NSSEARCH
 * - \ref HOOK_PRINTF
 * - \ref HOOK_FLUSH
 * - \ref HOOK_GLSINTERACTION
 * - \ref HOOK_GETS
 * - \ref HOOK_PROGRESS
 * - \ref HOOK_PARTIALRESULT
 * - \ref HOOK_ICINTERACTION
 * - \ref HOOK_RESET
 *
 * As a part of that the output channels are handled here. The following output
 * channels exist:
 * - \ref CDG_HINT
 * - \ref CDG_INFO
 * - \ref CDG_WARNING
 * - \ref CDG_PROLOG
 * - \ref CDG_EVAL
 * - \ref CDG_SEARCHRESULT
 * - \ref CDG_PROFILE
 * - \ref CDG_HOOK
 * - \ref CDG_ERROR
 * - \ref CDG_DEBUG
 * - \ref CDG_DEFAULT
 * - \ref CDG_PROGRESS
 * - \ref CDG_XML
 *
 * Note, that the term \e hook and \e callback are used mostly as synonyms. 
 * (Actually the term hook should be replaced by callback for clarity.)
 * 
 * As oposed to other modules some of the functions in this
 * module are not prefixed by \c hk as the \ref CodingStyle dictates. The reason
 * for that lies in the relative importance of this module and some of its 
 * functions to the overall system, i.e. cdgPrintf().
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <configure.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <blah.h>
#include "cdg.h"
#include "hook.h"
#include "gls.h"
#include "set.h"
#include "write.h"

/* -- VARIABLES --------------------------------------------------------- */

/* -------------------------------------------------------------------------
 * This Vector holds all defined hooks. It is initialized by the function 
 * hkInitialize(). After that, the symbolic constants can be used
 * to execute any of the hooks. 
 */
Vector hkHooks = NULL;

/* -------------------------------------------------------------------------
 * bitwise encoding of active output channels.
 *
 * This variable controls the currently active output channels.
 * All output is performed by the function cdgPrintf(), whose first argument 
 * is a bitwise encoding denoting the output channels (e.g. \ref CDG_INFO, \ref CDG_ERROR, 
 * or \ref CDG_DEBUG output) to which a message is to be routed.
 *
 * All messages tagged with a bit that is not set in \c hkVerbosity
 * are suppressed.
 */
unsigned long int hkVerbosity = CDG_DEFAULT | CDG_ERROR | CDG_INFO | CDG_WARNING | CDG_SEARCHRESULT;

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * validate access to variables in this module.
 *
 * This function is installed using setRegister() in the hkInitialize()
 * function of this module.
 * It is monitoring the access to the variable \c xml being
 * registered in hkInitialize(). 
 */
Boolean hkValidate(String name, String value, Boolean *var)
{
  if (!var) {
    return FALSE;
  }
  
  if (strcmp(name, "xml") == 0) {
    if ((strcmp(value, "on") == 0 && *var == TRUE) || 
	(strcmp(value, "off") == 0 && *var == FALSE)) {
      return TRUE;
    } 
  }
  
  return FALSE;
}

/* ----------------------------------------------------------------------
 * monitor access to variables in this module.
 *
 * This function is installed using setRegister() in the hkInitialize()
 * function of this module. When a variable is changed hkCallback() is
 * called handling appropriate sideeffects depending on the new variable
 * value. Right now only the variable \c xml is handled here.
 *
 * Switching on \c xml writes an appropriate xml header to the CDG_XML
 * channel. Switching off \c xml finishes all open xml tags.
 * See the write module for more information on the xml output channel.
 */
void hkCallback(String name, Boolean *var) 
{
  if (strcmp(name, "xml") == 0) {
    if (*var) {
      writeXmlHeader();
    } else {
      writeXmlEndAll();
    }
  }
}

/* ----------------------------------------------------------------------
 * initialization of the hook module.
 * This function must be called before the \c CDG library is functional.
 */
void hkInitialize(void)
{
   Hook hook;

   hkHooks = vectorNew(20);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "printf";
   hook->active = TRUE;
   hook->function = (HookFunction *)vprintf;

   hook->no = HOOK_PRINTF;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "flush";
   hook->active = TRUE;
   hook->function = NULL;
   hook->no = HOOK_FLUSH;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "buildnodes";
   hook->active = FALSE;
   hook->function = NULL;
   hook->no = HOOK_CNBUILDNODES;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "eval";
   hook->active = FALSE;
   hook->function = NULL;
   hook->no = HOOK_EVAL;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "netsearch";
   hook->active = FALSE;
   hook->function = NULL;
   hook->no = HOOK_NSSEARCH;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "reset";
   hook->active = TRUE;
   hook->function = NULL;
   hook->no = HOOK_RESET;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "glsInteraction";
   hook->active = TRUE;
   hook->function = glsInteraction;
   hook->no = HOOK_GLSINTERACTION;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "gets";
   hook->active = TRUE;
   hook->function = NULL;
   hook->no = HOOK_GETS;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "progress";
   hook->active = TRUE;
   hook->function = NULL;
   hook->no = HOOK_PROGRESS;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "partialresult";
   hook->active = TRUE;
   hook->function = NULL;
   hook->no = HOOK_PARTIALRESULT;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

   hook = (HookStruct *)memMalloc(sizeof(HookStruct));
   hook->count = 0;
   hook->name = "IC interaction";
   hook->active = TRUE;
   hook->function = NULL;
   hook->no = HOOK_ICINTERACTION;
   hook->cmd = NULL;
   vectorSetElementAt(hkHooks, hook, hook->no);

  setRegister("verbosity", SET_UNSLONG, &hkVerbosity, NULL, NULL, NULL,
                           "on", CDG_DEFAULT | CDG_ERROR | CDG_HINT | CDG_INFO | CDG_WARNING | CDG_SEARCHRESULT, 
			   "off", CDG_WARNING,
			   NULL);
  setRegister("hint",         SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_HINT,         NULL);
  setRegister("info",         SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_INFO,         NULL);
  setRegister("debug",        SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_DEBUG,        NULL);
  setRegister("error",        SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_ERROR,        NULL);
  setRegister("warning",      SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_WARNING,      NULL);
  setRegister("prolog",       SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_PROLOG,       NULL);
  setRegister("searchresult", SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_SEARCHRESULT, NULL);
  setRegister("eval",         SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_EVAL,         NULL);
  setRegister("profile",      SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_PROFILE,      NULL);
  setRegister("progress",     SET_BIT, &hkVerbosity, NULL, NULL, NULL, CDG_PROGRESS,     NULL);
  setRegister("xml",          SET_BIT, &hkVerbosity, NULL, hkValidate, hkCallback, CDG_XML, NULL);

}

/* ----------------------------------------------------------------------
 * finalization of this module.
 * After this call the \c CDG library is nearly shut down. That's why
 * this finalizer is to be called at last after calling the finalizer of
 * all other modules.
 */
void hkFinalize(void)
{
  int i;
  for (i=0; i<vectorSize(hkHooks); i++) {
    memFree(vectorElementAt(hkHooks, i));
  }
  vectorDelete(hkHooks);
}

/* ----------------------------------------------------------------------
 * execute a defined callback.
 * This is the standard way to call a hook. 
 * If the corresponding HookStruct::active is set, its HookStruct::count is
 * increased and its HookStruct::function is called handing over a pointer
 * to the HookStruct we are executing and the entire va_list passed to this function.
 * For example, \c cdgExecHook(HOOK_PARTIALRESULT, parse) is executing
 * \c HookStruct::function(hook, parse).
 * \param hookNo the identifier of the hook to be called, 
 *        e.g. HOOK_PROGRESS
 */
void cdgExecHook(int hookNo, ...)
{
#define DEBUG_CDGEXECHOOK 0
   va_list ap;
   Hook hook;

   if (hookNo >= vectorSize(hkHooks)) {
      cdgPrintf(CDG_ERROR, "ERROR: unknown hook `%d' called\n", hookNo);
      return;
   }

   hook = (Hook) vectorElementAt(hkHooks, hookNo);
#if DEBUG_CDGEXECHOOK
   printf("DEBUG: hook      = %p\n", hook);
   printf("       name      = %s\n", hook->name);
   printf("       active    = %d\n", hook->active);
   printf("       count     = %ld\n", hook->count);
   printf("       function  = %p\n", hook->function);
#endif
   if(hook->active) {
      hook->count++;
      if(hook->function) {
	 va_start(ap, hookNo);
	 (hook->function)(hook, ap);
	 va_end(ap);
      } else {
/*        printf("ERROR: hook function of `%s' not set\n", hook->name);*/
      }
   }
#if DEBUG_CDGEXECHOOK
   printf("DEBUG: done hook\n");
#endif
}


/* ----------------------------------------------------------------------
 * get the numerical identifier of a named hook.
 * \param name the name of a known hook
 * \returns the numerical identifier on success or -1 of no hook of that
 * name exists.
 */
int hkFindNoOfHook(String name)
{
   int index;
   Hook hook;
   Boolean found;

   for(index=0, found=FALSE; 
       index < vectorSize(hkHooks) && !found; 
       index++) {
      
      hook = (Hook)vectorElementAt(hkHooks, index);
      found = (strcmp(hook->name, name) == 0);
   }

   return (found?index-1:-1);
}

/* ----------------------------------------------------------------------
 * default get string callback.
 * This function gets one newline-terminated chunk of console input from
 * the user. Up to \c size characters are transferred into
 * \c buffer. For this purpose, \ref HOOK_GETS is executed. It this 
 * hook has no associated function, \c fgets() is called instead with 
 * an additional argument of \c stdin. This provides a transparent
 * way of reading input from a user. (Note that the normal interaction
 * with the shell of the cdgp parser uses a different and non-transparent
 * input routine.)
 */
void cdgGetString(String buffer, int size) {

   Hook hook;

   hook = vectorElementAt(hkHooks, HOOK_GETS);
   hook->count++;
   if(hook->function) {
     (hook->function)(hook, buffer, size);
   }
   else {
     fgets(buffer, size, stdin);
   }
}

/* ----------------------------------------------------------------------
 * output channel router.
 * This function takes care of routing messages to appropriate output
 * channels.
 * The first argument \c mode defines bitwise the channels this message 
 * is to be routed to. The rest of the arguments is passed to the 
 * \ref HOOK_PRINTF in printf mannor (so see the manual pages for printf).
 * We use \ref HOOK_PROGRESS exclusive here when the
 * \ref hkVerbosity contains the \ref CDG_PROGRESS flag and only
 * \ref HOOK_PRINTF otherwise. If we are producing xml output 
 * and the \ref CDG_WARNING or \ref CDG_ERROR channel is used, then
 * the message doubled according to the cdg-logfile.dtd by
 * writeXmlVaNotification().
 */
void cdgPrintf(int mode, String format, ...)
{
  va_list ap;
  Hook hook;

  if (hkVerbosity & mode) {
    va_start(ap, format);
    
    /* special handling of progress messages */
    if (mode & CDG_PROGRESS) {
      hook = vectorElementAt(hkHooks, HOOK_PROGRESS);
      hook->count++;

      /* check wether the progress hook is filled */
      if (hook->function != NULL) {
	(hook->function)(hook, format, ap);
      } 
      
      /* ... else fall back to normal cdgPrintf'ing */
      else {
	hook = vectorElementAt(hkHooks, HOOK_PRINTF);
	(hook->function) (format, ap);
      }

      /* flush the progress output of the cdg system */
      cdgFlush();

    } else {
      /* some verbosity generates XML messages */
      if (mode & CDG_WARNING) {
	writeXmlVaNotification("warning", format, ap);
      } else if (mode & CDG_ERROR) {
	writeXmlVaNotification("error", format, ap);
      } 

      hook = vectorElementAt(hkHooks, HOOK_PRINTF);
      hook->count++;
      (hook->function) (format, ap);
    }

    va_end(ap);
  }
}

/* ----------------------------------------------------------------------
 * default flush callback.
 * This is the default callback for \ref HOOK_FLUSH. If no hook is
 * installed, i.e. HookStruct::function is \ref NULL, then \c fflush
 * is called. In addition writeXmlFlush() is called to sync the 
 * \ref CDG_XML output channel.
 */
void cdgFlush(void)
{
   Hook hook;

   hook = vectorElementAt(hkHooks, HOOK_FLUSH);
   hook->count++;
   if(hook->function)
      (hook->function)();
   else {
      fflush(stdout);
   }

   writeXmlFlush();
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */

