/*
 * File: levelvalue.c
 * Author: Ingo Schroeder
 * Purpose: maintainance of level values
 * Birth: 24/4/97
 *
 * $Id: levelvalue.c,v 1.63 2003/09/29 08:37:17 foth Exp $
 */

/* ---------------------------------------------------------------------- */
#include <configure.h>

#include <stdio.h>
#include <string.h>
#include <blah.h>
#include "cdg.h"
#include "input.h"
#include "eval.h"
#include "statistics.h"
#include "levelvalue.h"
#include "lexemgraph.h"
#include "increment.h"
#include "hook.h"
#include "constraintnet.h"

/* ---------------------------------------------------------------------- */
/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * checks whether two level value are mutually compatible
 */
Boolean lvCompatible(LexemGraph lg, LevelValue a, LevelValue b)
{
   LexemNode atop = a->modifiee;
   List atops = a->modifiees;
   List abots = a->modifiers;

   LexemNode btop = b->modifiee;
   List btops = b->modifiees;
   List bbots = b->modifiers;

#define LVCOMPATIBLE 0

   if (!lgCompatibleSets(lg, abots, bbots)) {
#if LVCOMPATIBLE
     cdgPrintf(CDG_INFO, "*** lgcs abots bbots\n");
#endif
     return (FALSE);
   }

   if (atop != NULL && atop != NONSPEC) {
     if (!lgCompatibleSets(lg, atops, bbots)) {
#if LVCOMPATIBLE
       cdgPrintf(CDG_INFO, "*** lgcs atops bbots\n");
#endif
       return (FALSE);
     }

      if (btop != NULL && btop != NONSPEC) {
	 if (!lgCompatibleSets(lg, abots, btops)) {
#if LVCOMPATIBLE
	   cdgPrintf(CDG_INFO, "*** lgcs abots btops\n");
#endif
	   return (FALSE);
	 }
	 if (!lgCompatibleSets(lg, atops, btops)) {
#if LVCOMPATIBLE
	   cdgPrintf(CDG_INFO, "*** lgcs atops btops\n");
#endif
	   return (FALSE);
	 }
      }
   } else {
      if (btop != NULL && btop != NONSPEC &&
	  !lgCompatibleSets(lg, btops, abots)) {
#if LVCOMPATIBLE
	cdgPrintf(CDG_INFO, "*** lgcs btops abots\n");
#endif
	return (FALSE);
      }
   }
   return (TRUE);
}

/* ----------------------------------------------------------------------
 * builds level value for a constraint node, level and lexem nodes
 */
LevelValue lvNew(List modifiers, Level level, String label, List modifiees)
{
  LevelValue lv;

  lv = (LevelValue) memMalloc(sizeof (LevelValueStruct));
  lv->constraints = NULL;
  lv->no = -1;
  lv->score = 0.0;
  lv->limit = 1.0;
  lv->isDeleted = FALSE;
  lv->indexWRTNet = 0;
  lv->level = level;
  lv->label = strRegister(label);

  /* check for special modifiees */
  if (modifiees == NULL) {
    lv->modifiees = NULL;
    lv->modifiee = NULL;
  } else if (modifiees == NONSPEC) {
    lv->modifiees = NONSPEC;
    lv->modifiee = NONSPEC;
  } else {
    lv->modifiees = listClone(modifiees);
    lv->modifiee = (LexemNode) listItem(modifiees);
  }

  /* modifiers are never special */
  lv->modifier = listItem(modifiers);
  lv->modifiers = listClone(modifiers);

  return lv;
}

/* ----------------------------------------------------------------------
 * reset the values of an already existing levelvalue
 * - all list-containers are owned by the levelvalue 
 * - old modifiers- and modifiees-lists are reused
 */
LevelValue lvSet(LevelValue lv, List modifiers, Level level, String label,
		 List modifiees)
{
  lv->no = -1;
  lv->score = 0.0;
  lv->limit = 1.0;
  lv->isDeleted = FALSE;
  lv->indexWRTNet = 0;
  lv->level = level;
  lv->label = label;

  /* check for special modifiees */
  if (modifiees != NULL && modifiees != NONSPEC) {
    lv->modifiees = listSetList(lv->modifiees, modifiees);
    lv->modifiee = (LexemNode) listItem(modifiees);
  } else {
    lv->modifiees = NULL;
    lv->modifiee = (LexemNode) modifiees;
  }

  /* modifiers are never special */
  lv->modifier = (LexemNode) listItem(modifiers);
  lv->modifiers = listSetList(lv->modifiers, modifiers);

  /* old violations are deleted */
  listFree(lv->constraints);
  lv->constraints = NULL;

  return lv;
}

/* ----------------------------------------------------------------------
 * delete a levelvalue and all owned containers
 */
void lvDelete(LevelValue lv) 
{
  if (!lv) {
    return;
  }

  listFree(lv->modifiers); 
  if (spec(lv->modifiees)) {
    listFree(lv->modifiees);
  }
  listFree(lv->constraints);
  memFree(lv);
}

/* ----------------------------------------------------------------------
 * prints level value
 */
void lvPrint(long unsigned int mode, LevelValue lv, BitVector isDeleted)
{
   List l;
   LexemNode ln;
   Boolean first;
   int i, j, k;

   cdgPrintf(mode, "%0.2f (%02d-%02d) ", 
      lv->score,
      lv->modifier->arc->from,
      lv->modifier->arc->to);
   if (spec(lv->modifiee)) {
      cdgPrintf(mode, "(%02d-%02d) ",
	 lv->modifiee->arc->from,
	 lv->modifiee->arc->to);
   } else {
      cdgPrintf(mode, "        ");
   }
   cdgPrintf(mode, "%10s: ", lv->level->id);

   first = TRUE;
   for (l = lv->modifiers; l != NULL; l = listNext(l)) {
      ln = (LexemNode)(listItem(l));
      
      /* don't show deleted lexeme nodes */
      if (isDeleted && bvElementAt(isDeleted, ln->no))
	 continue;

      if (!first)
	 cdgPrintf(mode, "/");
      else 
	 first = FALSE;

      cdgPrintf(mode, "%s", ln->lexem->description);
   }

   /* take care to make all --LABEL--> arrows the same length */
   cdgPrintf(mode, "--");

   i = inputCurrentGrammar->maxLabelLength - strlen(lv->label);
   j = i / 2;
   for(k = 0; k < j; k++) cdgPrintf(mode, "-");

   cdgPrintf(mode, "%s", lv->label);

   if(i % 2 == 1) j++;
   for(k = 0; k < j; k++) cdgPrintf(mode, "-");

   cdgPrintf(mode, "-->");


   if (lv->modifiee == NONSPEC)
     cdgPrintf(mode, "NONSPEC");
   else if (lv->modifiee == NULL)
     cdgPrintf(mode, "NIL");
   else {

     first = TRUE;
     for (l = lv->modifiees; l != NULL; l = listNext(l)) {
       ln = (LexemNode)(listItem(l));
      
       /* don't show deleted lexeme nodes */
       if (isDeleted && bvElementAt(isDeleted, ln->no))
	 continue;

       if (!first)
	 cdgPrintf(mode, "/");
       else 
	 first = FALSE;

       cdgPrintf(mode, "%s", ln->lexem->description);
     }
   }
   
}

/* ----------------------------------------------------------------------
 * return the Direction of an LV
 */
Direction getDir(LevelValue x)
{

   if (x->modifiee == NULL)
      return Nil;
   else if (x->modifiee == NONSPEC)
      return Right;
   else if (x->modifiee->arc->from < x->modifier->arc->from)
      return Left;
   else
      return Right;
}


/* ----------------------------------------------------------------------
 * return the Connexion of two LVs
 */
Connexion getCon(LevelValue a, LevelValue b)
{

   GraphemNode
      atop = NULL,
      btop = NULL;

   GraphemNode
      abot = a->modifier->grapheme,
      bbot = b->modifier->grapheme;

   if (a->modifiee == NONSPEC)
     atop = NONSPEC;
   
   if(b->modifiee == NONSPEC)
     btop = NONSPEC;
      
   if (a->modifiee != NULL && a->modifiee != NONSPEC)
      atop = a->modifiee->grapheme;
   if (b->modifiee != NULL && b->modifiee != NONSPEC)
      btop = b->modifiee->grapheme;

   if (atop == btop && abot == bbot && atop != NONSPEC)
      return Identical;

   if (atop == bbot && abot == btop)
      return Inverse;

   if (atop == bbot)
      return Under;

   if (abot == btop)
      return Over;

   if (atop == btop && atop != NONSPEC)
      return Top;

   if (abot == bbot)
      return Bottom;

   return None;

}


/* ----------------------------------------------------------------------
 * Does this Signature apply to this LV?
 */
Boolean matchUnarySignature(LevelValue lv, Signature x)
{

   return (matchLevel(x->l1,lv->level) &&
      subsumesDirection(x->d1, getDir(lv)));

}

/* ----------------------------------------------------------------------
 * Does this Signature apply to these LVs?
 */
Boolean matchBinarySignature(LevelValue lva, LevelValue lvb, Signature x)
{

   return (matchLevel(x->l1, lva->level) &&
           matchLevel(x->l2, lvb->level) &&
      subsumesConnexion(x->conn, getCon(lva, lvb)) &&
      subsumesDirection(x->d1, getDir(lva)) &&
      subsumesDirection(x->d2, getDir(lvb)));

}

/* ----------------------------------------------------------------------
 * Define total ordering on LVs:
    compare by scores, and secondarily by their indexWRTNet
 */
Boolean lvCompare(LevelValue a, LevelValue b)
{
  if(a->score > b->score) return TRUE;
  if(a->score < b->score) return FALSE;
  return a->indexWRTNet < b->indexWRTNet;
}

/* ----------------------------------------------------------------------
 * like lvCompare(), but sorts only by limits.
 */
Boolean lvCompareByLimit(LevelValue a, LevelValue b)
{
   return (a->limit > b->limit);
}

/* ----------------------------------------------------------------------
 * like lvCompare(), but sorts by level & by time.
 */
Boolean lvCompareByNaturalOrder(LevelValue a, LevelValue b)
{
  if(a->modifier->arc->from < b->modifier->arc->from) {
    return TRUE;
  }
  if(a->modifier->arc->from > b->modifier->arc->from) {
    return FALSE;
  }
  if(a->level->no < b->level->no) {
    return TRUE;
  }

  return FALSE;
  
}

/* ----------------------------------------------------------------------
 * clone an LV
 */
LevelValue lvClone(LevelValue lv) {
  LevelValue result;

  if (!lv)
    return NULL;

  result = lvNew(lv->modifiers,
		 lv->level, lv->label, lv->modifiees);

  result->no = lv->no;
  result->modifier = lv->modifier;
  result->modifiee = lv->modifiee;
  result->indexWRTNet = lv->indexWRTNet;
  result->score = lv->score;
  result->limit = lv->limit;

  return result;
}

/* ----------------------------------------------------------------------
   Is the LV physically compatible with the lexemes allowed in the
   Vector?
  
   This uses a Vector of Boolean as produced by lgRequireLexeme().
 */
Boolean lvVectorCompatible(LevelValue lv, BitVector taboo)
{
  List l;
  Boolean found = FALSE;
  LexemNode ln;

  /* one modifiee must be allowed */
  if (spec(lv->modifiee)) {
    for (l = lv->modifiees; l; l = listNext(l)) {
      ln = listItem(l);
      if (!bvElementAt(taboo, ln->no)) {
	found = TRUE;
	break;
      }
    }
    if (!found) {
      return FALSE;
    }
  }

  /* one modifier must be allowed */
  for (l = lv->modifiers; l; l = listNext(l)) {
    ln = listItem(l);
    if (!bvElementAt(taboo, ln->no)) {
      return TRUE;
    }
  }

  return FALSE;

}

/* ----------------------------------------------------------------------
   Check that an LV is internally consistent. This is a void function
   because the answer "no" would be a fatal internal error anyway.
 */
void lvCheck(LevelValue lv) {

  /* unsuitable combination of list and node? */
  if (spec(lv->modifiee) && !listContains(lv->modifiees, lv->modifiee)) {
    cdgPrintf(CDG_ERROR, "PANIC: unsuitable modifiee `%s' for #%d\n",
	      lv->modifiee->lexem->description, lv->indexWRTNet);
    abort();
  }
  if (!listContains(lv->modifiers, lv->modifier)) {
    cdgPrintf(CDG_ERROR, "PANIC: unsuitable modifier `%s' for #%d\n",
	      lv->modifier->lexem->description, lv->indexWRTNet);
    abort();
  }
}


/* ----------------------------------------------------------------------
   This function computes the `logical' number of an LV in the set of LVs
   that constitute a dependency analysis: it is the position it will have if
   the LVs are sorted by time index first, and by level second. Some
   indices may not be assigned in two cases:

   (a) if there are lexeme nodes that span more than one time interval,
   lvIndex() will leave out those that correspond to positions within the
   long word.

   (b) if individual levels have their useFlag reset, the indices
   corresponding to LVs of these levels will also be skipped.

   This allows the function to determine exactly what position an LV holds,
   without any other context, but can be wasteful on space. The structure
   Analysis from the module frobbing uses this indexing method; the
   structure Parse uses a compact method that always gives consecutive index
   numbers. */
int lvIndex(LevelValue lv) {
  return
    inputCurrentGrammar->noOfLevels * lv->modifier->arc->from + lv->level->no;
}

/* ----------------------------------------------------------------------
 * Returns FALSE if all the modifiers or all the modifiees are deleted
 */
Boolean lvTestAllDeleted(LevelValue lv, BitVector isDeleted)
{
  List l;
  Boolean modifiers_ok = FALSE;

  for (l = lv->modifiers; l; l = listNext(l)) {
    LexemNode ln = (LexemNode)(listItem(l));
    if (bvElementAt(isDeleted, ln->no) == FALSE)
      modifiers_ok = TRUE;
  }

  if (modifiers_ok == FALSE)
    return FALSE;
  
  if (lv->modifiee == NULL)
    return TRUE;
  
  if (lv->modifiee == NONSPEC)
    return TRUE;

  for (l = lv->modifiees; l; l = listNext(l)) {
    LexemNode ln = (LexemNode)(listItem(l));
    if (bvElementAt(isDeleted, ln->no) == FALSE)
      return TRUE;
  }
  return FALSE;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
