/*
 * File: lexemgraph.c
 * Author: Ingo Schroeder
 * Purpose: maintainance of lexem graphs
 * Birth: 4/3/97
 *
 * $Id: lexemgraph.c,v 1.131 2004/01/13 13:26:10 foth Exp $
 */

/* ---------------------------------------------------------------------- */
#include <configure.h>

#include <ctype.h>
#include <stdio.h>
#include <blah.h>
#include <string.h>
#include <limits.h>
#include "hook.h"
#include "input.h"
#include "lexemgraph.h"
#include "eval.h"
#include "increment.h"
#include "set.h"
#include "tagger.h"
#include "parse.h"
#include "chunker.h"

/* -- VARIABLES --------------------------------------------------------- */
Boolean lgCompactLVs=TRUE;

extern Boolean inputSpuriousUppercase(LexemGraph lg, Arc arc);

/* ----------------------------------------------------------------------
 * (re-)computes the distance matrix
 */
void lgComputeDistances(LexemGraph lg)
{
   GraphemNode gn, gnNext;
   int noOfNodes = vectorSize(lg->graphemnodes);
   int i, j, k, ji,ik,jk,ij;

   /* arrayNew initializes the cells with NULL */
   if (!lg->distance) {
      lg->distance = arrayNew(noOfNodes, noOfNodes, 0);
   } else {
      arraySetAllElements(lg->distance, (Pointer) 0);
   }

   for (i = 0; i < noOfNodes; i++) {
      gn = (GraphemNode) vectorElementAt(lg->graphemnodes, i);
      /* If this gn has been deleted from the problem,
         leave its distances all at 0. This is necessary because
         the distance info does double duty as compatibility info. */
      if(lgAreDeletedNodes(lg, gn->lexemes)) {
        continue;
      }
      for (j = 0; j < noOfNodes; j++) {
	 gnNext = (GraphemNode) vectorElementAt(lg->graphemnodes, j);
         if(lgAreDeletedNodes(lg, gnNext->lexemes)) {
           continue;
         }
	 if (gn->arc->to == gnNext->arc->from) {
	    arraySetElement(lg->distance, (Pointer) 1, i, j);
	 }
      }
   }
   /* Warshall's algorithm for transitive closure */
   for (i = 0; i < noOfNodes; i++) {
      for (j = 0; j < noOfNodes; j++) {
	 ji = (int)arrayGetElement(lg->distance, j, i);

	 if (ji > 0) {
	    for (k = 0; k < noOfNodes; k++) {
	       ik = (int)arrayGetElement(lg->distance, i, k);
	       jk = (int)arrayGetElement(lg->distance, j, k);

	       if (ik > 0 && (jk == 0 || ji + ik < jk)) {
		  arraySetElement(lg->distance, (Pointer) (ji + ik), j, k);
	       }
	    }
	 }
      }
   }

   for (i = 0; i < noOfNodes; i++) {
      for (j = 0; j < noOfNodes; j++) {
	 ij = (int)arrayGetElement(lg->distance, i, j);
	 ji = (int)arrayGetElement(lg->distance, j, i);

	 if (ij == 0 && ji > 0) {
	    arraySetElement(lg->distance, (Pointer) (-ji), i, j);
	 }
      }
   }
}


/* ----------------------------------------------------------------------
 * computes noOfPathsFromStart
 */
long long computeNoOfPathsFromStart(LexemGraph lg, GraphemNode gn,
                                    long long sofar, long long maximal)
{
   long long i, sum;
   GraphemNode prevNode;
   long long noOfPathsFromStart;

   noOfPathsFromStart = lg->noOfPathsFromStart[gn->no];

   if (noOfPathsFromStart>= 0) {
      return noOfPathsFromStart;
   }

   if (!gn->live) {
     return lg->noOfPathsFromStart[gn->no] = 0;
   }

   if (sofar > maximal) {
      cdgPrintf(CDG_ERROR, "ERROR: lexem lattice contains cycles\n");
      abort();
   }

   if (gn->arc->from == lg->min) {
     return lg->noOfPathsFromStart[gn->no] = 1;
   }
   sum = 0;
   for (i = 0; i < maximal; i++) {
      prevNode = (GraphemNode) vectorElementAt(lg->graphemnodes, i);

      if (!prevNode->live
	  || prevNode->arc->to != gn->arc->from)
	 continue;

      sum += computeNoOfPathsFromStart(lg, prevNode, sofar + 1, maximal);

      if(sum < 0) {
        cdgPrintf(CDG_ERROR, "PANIC: integer overflow in computeNoOfPathsFromStart()!\n");
        abort();
      }
   }
   return lg->noOfPathsFromStart[gn->no] = sum;
}

/* ----------------------------------------------------------------------
 * computes noOfPathsToEnd
 */
long long computeNoOfPathsToEnd(LexemGraph lg, GraphemNode gn,
                                long long sofar, long long maximal)
{
   long long i, sum;
   GraphemNode nextNode;
   long long noOfPathsToEnd;

   noOfPathsToEnd = lg->noOfPathsToEnd[gn->no];

   if (noOfPathsToEnd >= 0)
      return noOfPathsToEnd;

   if (!gn->live) {
     return lg->noOfPathsToEnd[gn->no] = 0;
   }

   if (sofar > maximal) {
      cdgPrintf(CDG_ERROR, "ERROR: lexem lattice contains cycles\n");
      abort();
   }

   if (gn->arc->to == lg->max) {
     return lg->noOfPathsToEnd[gn->no] = 1;
   }
   
   sum = 0;
   for (i = 0; i < maximal; i++) {
      nextNode = (GraphemNode) vectorElementAt(lg->graphemnodes, i);

      if (!nextNode->live
	  || nextNode->arc->from != gn->arc->to) 
	 continue;

      sum += computeNoOfPathsToEnd(lg, nextNode, sofar + 1, maximal);
      if(sum < 0) {
        cdgPrintf(CDG_ERROR, "PANIC: integer overflow in computeNoOfPathsFromStart()!\n");
        abort();
      }
   }
   return lg->noOfPathsToEnd[gn->no] = sum;
}

/* ----------------------------------------------------------------------
 * computes # of paths possible in the graph, given the current state
   of deletions.
 */
void lgComputeNoOfPaths(LexemGraph lg)
{
   int i;
   LexemNode ln;
   GraphemNode gn;
   List l;
   int noOfNodes = vectorSize(lg->graphemnodes);
   
   /* append cells needed by iSearch */   
   for(i = bvSize(lg->isDeletedNode); i < vectorSize(lg->nodes); i++) {
     bvSetElementAt(lg->isDeletedNode, FALSE, i);
   }

   /* recompute ambiguities of each word */   
   for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
     gn = vectorElementAt(lg->graphemnodes, i);
     gn->live = FALSE;
     for(l = gn->lexemes; l != NULL; l = listNext(l)) {
       ln = listItem(l);
       if(!bvElementAt(lg->isDeletedNode, ln->no)) {
	 gn->live = TRUE;
         break;
       }
     }
   }

   /* maybe allocate the arrays */
   if(!lg->noOfPathsFromStart) {
     lg->noOfPathsFromStart = memMalloc(sizeof (long long) * noOfNodes);
   }
   if(!lg->noOfPathsToEnd) {
     lg->noOfPathsToEnd = memMalloc(sizeof (long long) * noOfNodes);
   }
   
   /* no paths from start/to end yet */
   for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
     lg->noOfPathsFromStart[i] = -1;
     lg->noOfPathsToEnd[i] = -1;
   }

   /* compute # of paths for all nodes */
   for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
      gn = (GraphemNode) vectorElementAt(lg->graphemnodes, i);

      if (lg->noOfPathsFromStart[gn->no] < 0) {
        computeNoOfPathsFromStart(lg, gn, 0, vectorSize(lg->graphemnodes));
      }
      if (lg->noOfPathsToEnd[gn->no] < 0) {
        computeNoOfPathsToEnd(lg, gn, 0, vectorSize(lg->graphemnodes));
      }
   }

   /* count # of paths in lg */
   lg->noOfPaths = 0;
   for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
      gn = (GraphemNode) vectorElementAt(lg->graphemnodes, i);

      if (!gn->live
	  || gn->arc->from != lg->min)
	 continue;

      lg->noOfPaths += lg->noOfPathsToEnd[gn->no];
   }
   
}

/* ----------------------------------------------------------------------
 * partitions a set of lexeme nodes into equivalence classes
 */
List lgPartitions(GraphemNode gn, BitString features)
{

/*

  result = [];

  FOR each lexeme l:
   IF l fits into one of the known classes,
    insert l there;
   ELSE
    create new class [lexem];
    insert the new class into result;
   FI
  ROF

  return result.
    
*/

   List result, p;
   List q;
   List class;
   LexemNode ln;
   LexiconItem lexeme1 = NULL, lexeme2 = NULL;

   result = NULL;

   /* partitioning the empty set
      yields the set with one class: { {} } */
   if(!gn) {
     return listPrependItem(NULL, NULL);
   } else if (gn == NONSPEC) {
     return listPrependItem(NULL, NONSPEC);
   }
#ifdef DEBUGPARTITION
   cdgPrintf(CDG_DEBUG,
	     "DEBUG: about to classify %d lexemes\n",
	     listLength(gn->lexemes));

   for (p = gn->lexemes; p != NULL; p = listNext(p)) {
      cdgPrintf(CDG_DEBUG, "%s ", ((LexemNode) listItem(p))->lexem->description);
   }
   cdgPrintf(CDG_DEBUG, "\n");
#endif

   /* partitioning a normal set yields sets of equivalent lexemes:
      { { das_ART_nom,das_ART_acc }, { das_PREL_nom,das_PREL_acc} } */
   for (p = gn->lexemes; p != NULL; p = listNext(p)) {
      ln = (LexemNode) listItem(p);
      lexeme1 = ln->lexem;

#ifdef DEBUGPARTITION
      cdgPrintf(CDG_DEBUG, "DEBUG: classifying lexeme %s\n", lexeme1->description);
#endif

      /* compare to the first item in each class */

      for (q = result; q != NULL; q = listNext(q)) {
	 class = (List) listItem(q);
	 lexeme2 = ((LexemNode) listItem(class))->lexem;

#ifdef DEBUGPARTITION
	 cdgPrintf(CDG_DEBUG, "DEBUG: comparing to lexeme %s... ", lexeme2->description);
#endif

	 /* insert into correct class */
	 if (lgCompactLVs &&
             !inputCompareLeByAtts(lexeme1, lexeme2, features)) {
#ifdef DEBUGPARTITION
	    cdgPrintf(CDG_DEBUG, "equivalent.\n");
	    cdgPrintf(CDG_DEBUG, "DEBUG: adding lexeme to this class.\n");
#endif
	    class = listAppendItem(class, ln);
	    break;
	 } else {
#ifdef DEBUGPARTITION
	    cdgPrintf(CDG_DEBUG, "differs.\n");
#endif
	 }
      }

      /* open new equivalence class */

      if (q == NULL) {

#ifdef DEBUGPARTITION
	 cdgPrintf(CDG_DEBUG, "DEBUG: creating new equivalence class.\n");
#endif

	 class = listPrependItem(NULL, ln);
	 result = listAppendItem(result, class);

      }
#ifdef DEBUGPARTITION
      cdgPrintf(CDG_DEBUG, "DEBUG: partition now looks like this:  ");
      for (q = result; q != NULL; q = listNext(q))
	 cdgPrintf(CDG_DEBUG, "%d ", listLength(listItem(q)));
      cdgPrintf(CDG_DEBUG, "\n");
#endif

   }

#ifdef DEBUGPARTITION
   cdgPrintf(CDG_DEBUG, "DEBUG: %d lexemes partitioned into %d classes.\n\n",
      listLength(gn->lexemes), listLength(result));
#endif


   return (result);

}

 
/* ----------------------------------------------------------------------
 * initializes the lexemgraph
 */
LexemGraph lgNewInit()
{
  LexemGraph lg;

  lg = (LexemGraph) memMalloc(sizeof(LexemGraphStruct));
  lg->lattice = NULL;
  lg->graphemnodes = vectorNew(20);
  lg->max = 0;
  lg->min = INT_MAX;
  lg->distance = NULL;
  lg->isDeletedNode = bvNew(20);
  lg->noOfPathsFromStart = NULL;
  lg->noOfPathsToEnd = NULL;
  lg->nodes = vectorNew(20);
  lg->tags = NULL;
  lg->chunks = NULL;
  return lg;
}


/* ----------------------------------------------------------------------
   Insert lexeme nodes into LG that correspond to ARC.
*/
Boolean lgNewIter(LexemGraph lg, Arc arc)
{
  GraphemNode gn;
  LexemNode ln;
  LexiconItem le;
  List j, entries = NULL;

  /* Maybe undo capitalisation introduced by orthographic convention.

     If the written word is uppercase, but that uppercase-ness is suspect
     because it is at the start of a phrase and might be mere orthographic
     convention, we have to decide which version we use for lexicon lookup.

     If our lexicon contains items for the lower-case version but none for
     the upper-case versions, we use only those; if it contains only
     items for the upper-case version, we use those; and if it contains
     neither, we allow both and hope that there is a lexical template which
     will catch this word.

     We do not use the obvious solution - look up both versions whenever a
     word is spurious - because it has the following defect: If a sentence
     starts with `Der', some naive lexical template could introduce a noun
     reading, and if POS tagging allows, it might actually survive even
     though it is exceedingly unlikely. Since we do not want this, we
     effectively force the reading to be `der'.

     Moral: If you really need to have open-class items in your lexicon
     that are near-homonymous with closed-class items, you can bloody well
     write proper lexicon items for them and not templates. */

  if(inputSpuriousUppercase(lg, arc)) {
    String low = strCopy(arc->word);
    Boolean have_upper, have_lower;

    low[0] = tolower(low[0]);
    have_upper = inputLexiconQuery(arc->word);
    have_lower = inputLexiconQuery(low);

    if(have_upper && !have_lower) {
      entries = listAppendList(entries, inputLexiconGet(arc->word));
    }
    else if(have_lower && !have_upper) {
      entries = listAppendList(entries, inputLexiconGet(low));
      arc->word = strCopy(low);
    } else {
      entries = listAppendList(entries, inputLexiconGet(low));
      entries = listAppendList(entries, inputLexiconGet(arc->word));
    }
    memFree(low);
  }

  /* Much the same goes for words in ALL UPPER CAPS, except that those can
     occur anywhere in a sentence, not only at the start, and we have to
     check three different spellings instea of two. */
  else if(inputALLUPPER(arc->word)) {
        String low = strCopy(arc->word);
        String startUpper = strCopy(arc->word);
        Boolean have_upper, have_lower, have_startUpper;
        int i;
    for (i=0;  i < strlen((unsigned char *)arc->word); i++){
        low[i] = tolower(low[i]);
    }
    for (i=1;  i < strlen((unsigned char *)arc->word); i++){
        startUpper[i] = tolower(low[i]);
    }

    have_lower = inputLexiconQuery(low);
    have_startUpper = inputLexiconQuery(startUpper);
    have_upper = inputLexiconQuery(arc->word);

    if( have_lower ){
         entries = listAppendList(entries, inputLexiconGet(low));
    }
    if( have_startUpper ){
         entries = listAppendList(entries, inputLexiconGet(startUpper));
    }
    if( have_upper ){
         entries = listAppendList(entries, inputLexiconGet(arc->word));
    }
    if(!have_lower && !have_startUpper && !have_upper){
         entries = listAppendList(entries, inputLexiconGet(low));
         entries = listAppendList(entries, inputLexiconGet(startUpper));
         entries = listAppendList(entries, inputLexiconGet(arc->word));
    }

    if(have_startUpper && !have_upper && !have_lower) {
      arc->word = strCopy(startUpper);
    }
    else if(have_lower && !have_upper && !have_startUpper) {
      arc->word = strCopy(low);
    }
    else if(!have_upper) {
      arc->word = strCopy(startUpper);
    }
 
    memFree(startUpper);
    memFree(low);
  }

  else {
    entries = inputLexiconGet(arc->word);
  }
  
  if (!entries) {
    cdgPrintf(CDG_WARNING, "WARNING: no lexical entry for `%s'.\n", arc->word);
    return FALSE;
  }

  lg->min = min(lg->min, arc->from);
  lg->max = max(lg->max, arc->to);

  gn = (GraphemNode) memMalloc(sizeof (GraphemNodeStruct));
  gn->no = vectorAddElement(lg->graphemnodes, gn);
  gn->lexemgraph = lg;
  gn->arc = arc;
  gn->lexemes = NULL;
  gn->chunk = NULL;

  /* force consistent order among homonyms */
  j = listSort(entries, leCompareByName);
  listFree(entries);
  entries = j;

  for (j = entries; j; j = listNext(j)) {
    le = (LexiconItem) listItem(j);
    ln = (LexemNode) memMalloc(sizeof (LexemNodeStruct));
    ln->no = vectorAddElement(lg->nodes, ln);
    ln->lexemgraph = lg;
    ln->arc = gn->arc;
    ln->lexem = le;
    ln->tagscore = 1.0;
    ln->limit = 1.0;
    bvSetElementAt(lg->isDeletedNode, FALSE, ln->no);

    /* link the nodes to each other */
    ln->grapheme = gn;
    gn->lexemes = listAppendItem(gn->lexemes, ln);
  }
  listFree(entries);

  return TRUE;
}

/* ----------------------------------------------------------------------
 * does the final computations for the lexemgraph
 */
Boolean lgNewFinal(LexemGraph lg)
{
  int noOfNodes;

  /* compute noOfPathsFromStart and noOfPathsToEnd */
  noOfNodes = vectorSize(lg->graphemnodes);
  lgComputeNoOfPaths(lg);

  /* no path? */
  if (lg->noOfPaths == 0) {
    cdgPrintf(CDG_WARNING, "WARNING: no path possible, invalid graph\n");
    return FALSE;
  }
  bvSetAllElements(lg->isDeletedNode, FALSE);
  lgComputeDistances(lg);

  /* build tags cheat sheet */
  {
    int i;
    AnnoEntry ae = findAnnotation(lg->lattice);
    List l, m;

    if (ae) {
      lg->tags = vectorNew(lg->max);
      for (i = 0; i < lg->max; i++) {
	vectorAddElement(lg->tags, strRegister(""));
      }

      for (l = ae->annos; l != NULL; l = listNext(l)) {
	Annotation a = listItem(l);

	for (m = a->specs; m != NULL; m = listNext(m)) {
	  Specification s = listItem(m);

	  if (s->type == STTag && strRegister("cat") == s->kind) {
	    /* CAUTION: this does not work reliable on wordgraphs */
	    vectorSetElementAt(lg->tags, s->name, a->from);
	  }
	}
      }
    } else {
      lg->tags = NULL;
    }
  }

#ifdef DEBUGLGNEW
  lgPrint(CDG_INFO, lg);
#endif

  return TRUE;
}

/* ----------------------------------------------------------------------
 * lgNew, with functionality split in different functions
 */
LexemGraph lgNew(Lattice lat)
{
  LexemGraph lg = lgNewInit();
  List l;

  lg->lattice = lat;
  for (l = lat->arcs; l; l = listNext(l)) {
    lgNewIter(lg, (Arc) listItem(l));
  }

  if (!lgNewFinal(lg)) {
    lgDelete(lg);
    lg = NULL;
  }

  return lg;
}

/* ----------------------------------------------------------------------
 * print lexem graph
 */
void lgPrint(long unsigned int mode, LexemGraph lg)
{
   Lattice lat = lg->lattice;
   int i;
   LexemNode ln;
   Arc a;
   Boolean deleted;
   
   cdgPrintf(mode, "\n%s : \n", lat->id);
   for (i = 0; i < vectorSize(lg->nodes); i++) {
      ln = (LexemNode) vectorElementAt(lg->nodes, i);
      deleted = bvElementAt(lg->isDeletedNode, ln->no);
      a = ln->arc;

      cdgPrintf(mode, "%%%d:%s%s(%d-%d)%s   %4.3e\n",
		ln->no,
		deleted ? "[" : " ",
		ln->lexem->description,
		a->from,
		a->to,
		deleted ? "]" : " ",
                ln->tagscore);
   }

   if (lg->chunks) {
    cdgPrintf(mode, "chunks:\n");
    chunkerPrintChunks(mode, lg->chunks);
   }
}

/* ----------------------------------------------------------------------
 * returns TRUE if node a directly follows node b
 */
Boolean lgNodeFollowsDirectly(GraphemNode a, GraphemNode b)
{
   if(a != NULL && b != NULL && b != NONSPEC) {
      if (a == NONSPEC)
	 return (b->arc->to == b->lexemgraph->max);
      else
	 return (b->arc->to == a->arc->from);
   } else
      return (FALSE);
}

/* ----------------------------------------------------------------------
 * returns TRUE if node is a start node
 */
Boolean lgIsStartNode(GraphemNode n)
{
   if(n == NULL || n == NONSPEC)
      return (FALSE);
   else
      return (n->arc->from == n->lexemgraph->min);
}

/* ----------------------------------------------------------------------
 * returns TRUE if node is an end node
 * TODO: should nodes on segment boundaries of incremental parsed input
 *       be considered to be endnodes
 */
Boolean lgIsEndNode(GraphemNode n)
{
   if(n == NULL || n == NONSPEC)
      return (FALSE);
   else
      return (n->arc->to == n->lexemgraph->max);
}

/* ----------------------------------------------------------------------
 * returns a distance measure for two lexem nodes
 */
int lgDistanceOfNodes(LexemGraph lg, LexemNode a, LexemNode b)
{
   int maxDistance,i,maxDim, distance;
   
   if (a == NULL || b == NULL || lg == NULL) {
      cdgPrintf(CDG_ERROR, "ERROR: lgDistanceOfNodes: argument is NULL\n");
      abort();
   }
   if (a == NONSPEC)
   {
      if (b == NONSPEC)
	 return 0;

      maxDim = arrayGetDimension(lg->distance, 1);
      maxDistance = 0;
      for (i = 0; i < maxDim; i++) {
	 distance = (int)arrayGetElement(lg->distance, b->grapheme->no, i);
	 if (maxDistance < distance) {
	    maxDistance = distance;
	 }
      }
      return -maxDistance-1;
   }

   if (b == NONSPEC)
   {
      maxDim = arrayGetDimension(lg->distance, 1);
      maxDistance = 0;
      for (i = 0; i < maxDim; i++) {
	 distance = (int)arrayGetElement(lg->distance, a->grapheme->no, i);
	 if (maxDistance < distance) {
	    maxDistance = distance;
	 }
      }
      return maxDistance+1;
   }

   return ((int)arrayGetElement(lg->distance, a->grapheme->no, b->grapheme->no));
}

/* ----------------------------------------------------------------------
 * may these words modify each other?
 */
Boolean lgMayModify(LexemGraph lg, GraphemNode down, GraphemNode up)
{
   LexemNode lna, lnb;
   if(!spec(up)) {
     return TRUE;
   }
   lna = listItem(down->lexemes);
   lnb = listItem(up->lexemes);
   
   return (lgDistanceOfNodes(lg, lna, lnb) != 0);
}

/* ----------------------------------------------------------------------
 * returns TRUE if lexem nodes a and b exist on one path
 */
Boolean lgCompatibleNodes(LexemGraph lg, LexemNode a, LexemNode b)
{
   if (lg == NULL || a == NULL || b == NULL) {
      cdgPrintf(CDG_ERROR, "ERROR: lgCompatibleNodes: argument is NULL\n");
      abort();
   }
   return (a == b || lgDistanceOfNodes(lg, a, b) != 0);
}

/* ----------------------------------------------------------------------
 * does existence of these lexemes exclude that lexeme?
 */
Boolean lgForbiddenBy(LexemGraph lg, LexemNode ln, List lexemes)
{
  if (!spec(lexemes))
    return FALSE;

  if (!lgCompatibleNodes(lg, ln, (LexemNode) listItem(lexemes))
      && !listContains(lexemes, ln)) {
    return (TRUE);
  } else {
    return (FALSE);
  }
}


/* ----------------------------------------------------------------------
 * are these sets of lexemes compatible, i.e. either unrelated
 * or intersecting?
 */

Boolean lgCompatibleSets(LexemGraph lg, List a, List b)
{

   LexemNode lna, lnb;
   List p;

   /* empty sets are compatible */

   if (a == NULL || b == NULL)
      return (TRUE);

   lna = (LexemNode) listItem(a);
   lnb = (LexemNode) listItem(b);

   /* unrelated sets are compatible */

   if (lgCompatibleNodes(lg, lna, lnb))
      return (TRUE);

   /* if we get here, the only chance for compatibility
      is for the lists to intersect. This is only possible
      if their elements are simultaneous. */

   if(lna->arc->from != lnb->arc->from ||
      lna->arc->to != lnb->arc->to)
     return FALSE;
   
   /* OK, the sets are simultaneous. Do they actually intersect? */

   for (p = a; p != NULL; p = listNext(p)) {
      if (listContains(b, listItem(p)))
	 return (TRUE);
   }

   /* mutually exclusive sets are not compatible */

   return (FALSE);

}

/* ----------------------------------------------------------------------
 * if the lexem node has been deleted TRUE is returned, if not FALSE
 */
Boolean lgIsDeletedNode(LexemGraph lg, LexemNode ln)
{
   return (bvElementAt(lg->isDeletedNode, ln->no));
}


/* ----------------------------------------------------------------------
 * if the lexem nodes have been deleted TRUE is returned, if not FALSE
 */
Boolean lgAreDeletedNodes(LexemGraph lg, List lexemes)
{
  List l;

  for (l = lexemes; l != NULL; l = listNext(l)) {
    if (!lgIsDeletedNode(lg, (LexemNode) listItem(l)))
      return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
 * if all of these lexeme nodes can be deleted without the lexem graph 
 * being torn apart, TRUE is returned, if not FALSE
 * 
 * !!!This function assumes that lexemes is a set of variants!!!
 */

Boolean lgAreDeletableNodes(LexemGraph lg, List lexemes)
{
   int affectedPaths = 0;
   List p;
   LexemNode ln;
   
   /* The lexem graph can only tear if we are trying to delete all of
      the remaining nodes of one grapheme node. So we first check this
      condition. */
   GraphemNode gn = ((LexemNode)listItem(lexemes))->grapheme;
   int noRemaining, noToDelete;

   /* the number of undeleted lexeme nodes for the grapheme node */   
   noRemaining = 0;
   for (p = gn->lexemes; p != NULL; p = listNext(p)) {
     ln = listItem(p);
     if(!bvElementAt(lg->isDeletedNode, ln->no)) {
       noRemaining++;
     }
   }
   if(noRemaining == 0) {
     return TRUE;
   }
   
   /* the number of lexemes that would be deleted */
   noToDelete = 0;
   for(p = lexemes; p != NULL; p = listNext(p)) {
     ln = listItem(p);
     if(!bvElementAt(lg->isDeletedNode, ln->no)) {
       noToDelete++;
     }
   }

   if(noToDelete < noRemaining) {
     return TRUE;
   }
   
   /* Now we know that the user wants do delete an entire grapheme
      node. This is allowed iff not all paths in the _grapheme_ graph
      go through gn. */

   affectedPaths =

     /* # of paths leading here */     
     lg->noOfPathsFromStart[gn->no]
     /* times # of paths leaving here */     
     * lg->noOfPathsToEnd[gn->no];

   /* this can't happen */   
   if(affectedPaths > lg->noOfPaths) {
     cdgPrintf(CDG_ERROR, "PANIC: node #%d supports more paths than exist!\n", gn->no);
     abort();
   }

   return (affectedPaths != lg->noOfPaths);

}

/* ----------------------------------------------------------------------
 * deletes a node from the lexeme graph itself.
 */
void lgDeleteNode(LexemGraph lg, LexemNode ln)
{

  int oldNo = lg->noOfPaths;

  /* already deleted, ok */
   if (bvElementAt(lg->isDeletedNode, ln->no))
      return;

   bvSetElementAt(lg->isDeletedNode, TRUE, ln->no);
   lgComputeNoOfPaths(lg);

   if (lg->noOfPaths == 0 && oldNo > 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgDeleteNode: deletion of node `%s' destroys last path\n",
	 ln->lexem->description);
   }

   /* have to do this because two nodes may become more distant
      by the deletion of the shortest connecting path */
   lgComputeDistances(lg);
}

/* ----------------------------------------------------------------------
 * delete a list of lexeme nodes
 */

void lgDeleteNodes(LexemGraph lg, List nodes)
{
  /* This is essentially a wrapper for several calls to
     lgDeleteNode(). But we don't actually want to call lgDeleteNode()
     several times because each invocation will execute a
     computeNoOfPaths(), which could waste a lot of time. */
   List p;
   int oldNo = lg->noOfPaths;
   LexemNode ln = NULL;
   
   /* So we inline the relevant code into a loop here... */
   for (p = nodes; p != NULL; p = listNext(p)) {
     ln = listItem(p);
     bvSetElementAt(lg->isDeletedNode, TRUE, ln->no);
   }

   /* ...and then do one call to computeNoOfPaths(). */   
   lgComputeNoOfPaths(lg);
   if(oldNo > 0 && lg->noOfPaths == 0) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgDeleteNodes: deletion of node `%s' destroys last path\n",
	 ln->lexem->description);
   }

   /* have to do this because two nodes may become more distant
      by the deletion of the shortest connecting path */
   lgComputeDistances(lg);
}


/* ----------------------------------------------------------------------
 * do the lexemes span the same time interval?
 */
Boolean lgSimultaneous(LexemNode a, LexemNode b)
{

   if (a == NULL && b != NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgSimultaneous: first argument must not be NULL\n");
      cdgPrintf(CDG_WARNING,
	 "WARNING: the other argument was `%s'\n", b->lexem->word);
   }
   if (b == NULL && a != NULL) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgSimultaneous: second argument must not be NULL\n");
      cdgPrintf(CDG_WARNING,
	 "WARNING: the other argument was `%s'\n", a->lexem->word);
   }
   if (a == NULL && b == NULL)
      cdgPrintf(CDG_WARNING,
	 "WARNING: lgSimultaneous: both arguments must not be NULL\n");

   if(a == NONSPEC || b == NONSPEC)
      return(TRUE);
   else if(a == NULL || b == NULL)
      return (FALSE);
   else
      return (a->arc->from == b->arc->from
	      && a->arc->to == b->arc->to);
}

/* ----------------------------------------------------------------------
 * is this lexeme a member of the this set?
 *
 * !!!This function assumes that lexemes is a set of variants!!!
 */

Boolean lgMember(LexemNode ln, List lexemes)
{
/* NIL binding is not an element of anything */
   if (ln == NULL)
      return (FALSE);

/* empty set has no member */
   if (lexemes == NULL)
      return (FALSE);

/* if ln belongs to another timespan, it's not a member */
   if (!lgSimultaneous(((LexemNode) listItem(lexemes)), ln))
      return (FALSE);

/* find it out the hard way */
   return (listContains(lexemes, ln));

}

/* ----------------------------------------------------------------------
 * is this lexeme set a subset of the that set?
 *
 * !!!This function assumes that a and b are sets of variants!!!
 */

Boolean lgSubset(List a, List b)
{

   List l;
   LexemNode ln;

/* empty set is always a subset */
   if (a == NULL)
      return (TRUE);

/* empty set has no proper subset */
   if (b == NULL)
      return (FALSE);

/* if a belongs to a different timespan, it's not a subset */
   if (!lgSimultaneous(((LexemNode) listItem(a)), (LexemNode) listItem(b)))
      return (FALSE);

/* find it out the hard way */
   for (l = a; l != NULL; l = listNext(l)) {
      ln = (LexemNode) listItem(l);
      if (!lgMember(ln, b))
	 return (FALSE);
   }

   return (TRUE);

}

/* ----------------------------------------------------------------------
 * Do two lexeme lists intersect?
 *
 * !!!This function assumes that a and b are sets of variants!!!
 */

Boolean lgIntersectingSets(List a, List b)
{

   List l;
   LexemNode ln;

   if (a == NULL && b != NULL) { return (FALSE); }

   if (a != NULL && b == NULL) { return (FALSE); }
   
   if (a == NULL && b == NULL) { return (TRUE); }

/* if a belongs to a different timespan, no intersection */
   if (!lgSimultaneous(((LexemNode) listItem(a)), (LexemNode) listItem(b))) {
      return (FALSE);
   }
   
/* find it out the hard way */
   for (l = a; l != NULL; l = listNext(l)) {
      ln = (LexemNode) listItem(l);
      if (listContains(b, ln)) {
	 return (TRUE);
      }
   }

   return (FALSE);

}

/* ----------------------------------------------------------------------
 * Takes a set of lexeme nodes, and extends it to a complete path
 * through the graph, composed of undeleted LexemNodes. 
 * Returns NULL if this is impossible,
 */
List lgMakePath(LexemGraph lg, List nodes) {

  LexemNode lna, lnb;
  List result = NULL;
  List l, m;
  int i, nrBoundPoints = 0;
  Boolean okay;
  
  /* We do this by simply appending arbitrary non-contradictory nodes
     until we have bound all time points. Note that for this approach
     to be correct, there must not be any undeleted dangling nodes in
     the graph. This condition must have ensured by cnOptimizeNet(). */
     
  /* is the list of nodes self-contradictory? */
  for(l = nodes; l != NULL; l = listNext(l)) {
    lna = listItem(l);
    nrBoundPoints += (lna->arc->to - lna->arc->from);
    for(m = listNext(l); m != NULL; m = listNext(m)) {
      lnb = listItem(m);
      if(!lgCompatibleNodes(lg, lna, lnb)) {
	return NULL;
      }
    }
  }
  
  /* choose arbitrary nodes until all time points are bound */  
  result = listClone(nodes);
  for(i = 0; i < vectorSize(lg->nodes); i++) {
    okay = TRUE;
    lna = vectorElementAt(lg->nodes, i);
    if(lgIsDeletedNode(lg, lna)) {
      continue;
    }
    for(l = result; l != NULL; l = listNext(l)) {
      lnb = listItem(l);
      if(lna == lnb ||
	 !lgCompatibleNodes(lg, lna, lnb)) {
	okay = FALSE;
	break;
      }
    }
    if(okay) {
      result = listAppendItem(result, lna);
      nrBoundPoints += (lna->arc->to - lna->arc->from);
    }
  }

  if(nrBoundPoints != lg->max - lg->min) {
    listFree(result);
    result = NULL;
  }
 
  return result;
}

/* ----------------------------------------------------------------------
 * deletes lexemgraph
 */
void lgDelete( LexemGraph lg )
{
  int i;
  GraphemNode gn;
  LexemNode ln;

  if (!lg) {
    return;
  }
  
  for ( i = 0; i < vectorSize( lg->graphemnodes ); i++ ) {
    gn = vectorElementAt(lg->graphemnodes, i);
    listFree(gn->lexemes);
    memFree(gn);
  }
  vectorDelete( lg->graphemnodes );

  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln = (LexemNode) vectorElementAt(lg->nodes, i);
    freeLexiconItem(ln->lexem);
    memFree( ln );
  }
  vectorDelete(lg->nodes);
  
  if(lg->distance != NULL) {
    arrayDelete(lg->distance);
  }
  
  memFree(lg->noOfPathsFromStart);
  memFree(lg->noOfPathsToEnd);
  bvDelete(lg->isDeletedNode);

  if(lg->tags) {
    vectorDelete(lg->tags);
  }

  listForeachFree(lg->chunks, chunkerChunkDelete);

  memFree( lg );
}


/* ----------------------------------------------------------------------
 * returns maximal ambiguity per time point
 */
int lgWidth (LexemGraph lg) {

  int result = 0;
  int i, j;
  int *ambiguity = memMalloc(sizeof(int)*lg->max);
  LexemNode ln;

  for(i = 0; i < lg->max; i++) {
    ambiguity[i] = 0;
  }

  for(i = 0; i < vectorSize(lg->nodes); i++) {
    ln = vectorElementAt(lg->nodes, i);
    for(j = ln->arc->from; j < ln->arc->to; j++) {
      ambiguity[j]++;
      if(ambiguity[j] > result) {
	result = ambiguity[j];
      }
    }
  }

  memFree(ambiguity);
  return result;
  
}

/* ----------------------------------------------------------------------
 * prints out a lexeme node
 */
void lgPrintNode(unsigned long mode, LexemNode ln) {

  Boolean dead = bvElementAt(ln->lexemgraph->isDeletedNode, ln->no);
  cdgPrintf(mode, "%s`%s'(%d-%d)%s",
            dead ? "[" : "",
	    ln->lexem->description,
	    ln->arc->from,
	    ln->arc->to,
            dead ? "]" : "");
}


/* ----------------------------------------------------------------------
   Do these lexeme nodes overlap?
  
   Returns TRUE if the two lexeme nodes have at least one time point in
   common.
   
   This is subtly different from the more common question, "Can the two
   nodes coexist on one path?": two nodes can be compatible although they
   overlap if they are identical. Conversely, a and b may be incompatible
   even if they do not overlap if there is no path between them.
 */
Boolean lgOverlap(LexemNode a, LexemNode b) {

  return((a->arc->to > b->arc->from && a->arc->from <= b->arc->from)
	 ||
	 (b->arc->to > a->arc->from && b->arc->from <= a->arc->from));
}

/* ----------------------------------------------------------------------
   Takes a Vector of Boolean, and sets all cells that correspond to the
   numbers of nodes incompatible with ln. This function can be used
   in combination with lvVectorCompatible() to decide whether an LV
   is compatible with a set of other LVs.
 */
void lgRequireLexeme(LexemGraph lg, BitVector v, LexemNode ln) {

  int i;
  LexemNode ln2;

  if (!spec(ln)) 
    return;

  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln2 = vectorElementAt(lg->nodes, i);
    if (!lgCompatibleNodes(lg, ln, ln2)) {
      bvSetElementAt(v, TRUE, ln2->no);
    }
  }
}

/* ----------------------------------------------------------------------
   Like lgRequireLexeme() but with a list of lexeme nodes.
 */
void lgRequireLexemes(LexemGraph lg, BitVector v, List which) {

  int i;
  LexemNode ln1, ln2;

  if (!spec(which))
    return;

  ln1 = listItem(which);

  if (!spec(ln1))
    return;

  for (i = 0; i < vectorSize(lg->nodes); i++) {
    ln2 = vectorElementAt(lg->nodes, i);
    if (!lgCompatibleNodes(lg, ln1, ln2) && !listContains(which, ln2)) {
      bvSetElementAt(v, TRUE, ln2->no);
    }
  }
}

/* ----------------------------------------------------------------------
 * updates the partial lexemgraph with the incoming arcs
 */
Boolean lgUpdateArcs(LexemGraph lg, Lattice lat, List listArcs)
{
  List m;

  for (m = listArcs; m; m = listNext(m)) {
    lgNewIter(lg, (Arc) listItem(m));
  }
  if (lg->distance != NULL) {
    arrayDelete(lg->distance);
    lg->distance = NULL;
  }
  
  lg->lattice = lat;
  return lgNewFinal(lg);
}

/* ----------------------------------------------------------------------
 * Copy the tag scores of one lexeme graph to another.
 * This is better than calling taggerTag() again
 * because that takes time and may fail.
 */
void lgCopyTagScores(LexemGraph destination, LexemGraph source)
{
  int i;
  List l;

  void backlink(Chunk chunk) {
    List l;

    for (l = chunk->nodes; l; l = listNext(l)) {
      GraphemNode gn = listItem(l);

      gn->chunk = chunk;
    }
    for (l = chunk->subChunks; l; l = listNext(l)) {
      backlink(listItem(l));
    }
  }

  for (i = 0; i < vectorSize(destination->nodes) && 
      i < vectorSize(source->nodes); i++) {
    LexemNode ln1 = vectorElementAt(destination->nodes, i);
    LexemNode ln2 = vectorElementAt(source->nodes, i);

    ln1->tagscore = ln2->tagscore;
  }

  /* Also copy the chunk information for much the same reason */
  for (l = source->chunks; l; l = listNext(l)) {
    Chunk chunk = listItem(l);

    chunk = chunkerCloneChunk(chunk);
    destination->chunks = listAppendItem(destination->chunks, chunk);
    chunkerReplaceGraphemes(chunk, destination);
    backlink(chunk);
  }

  /* copy tags cheat sheet */
  destination->tags = NULL;
  if(source->tags) {
    destination->tags = vectorClone(source->tags);
  }
  
}


/* ----------------------------------------------------------------------
 * Check that the LexiconItem is identical to the lexem field of one
 * of the lexeme nodes in the list
 */
Boolean lgLexemeInLexemNodeList(LexiconItem le, List list)
{
  List m;
  for (m = list; m != NULL; m = listNext(m)) {
    LexemNode ln = (LexemNode) listItem(m);
    if (ln->lexem == le)
      return TRUE;
  }
  return FALSE;
}

/* ----------------------------------------------------------------------
   Select the path in DST whose parts most closely match SRC.
 */
Boolean lgCopySelection(LexemGraph destination, LexemGraph source) {

  int i, j, k, record;
  LexemNode ln1, ln2, best;
  
  if(destination == source) {
    return TRUE;
  }

  bvSetAllElements(destination->isDeletedNode, TRUE);
  for(i = 0; i < bvSize(source->isDeletedNode); i++) {
    if(bvElementAt(source->isDeletedNode, i)) {
      continue;
    }

    /* OK, find the node in DST that corresponds to this one. */
    ln1 = vectorElementAt(source->nodes, i);
    best = NULL;
    record = INT_MAX;
    for(j = 0; j < vectorSize(destination->nodes); j++) {
      ln2 = vectorElementAt(destination->nodes, j);
      if(ln2->arc->from != ln1->arc->from ||
         ln2->arc->to != ln1->arc->to) {
        continue;
      }
      k = leCompare(inputCurrentGrammar, ln1->lexem, ln2->lexem, FALSE);
      if(k < record) {
        record = k;
        best = ln2;
      }
    }
    if(!best) {
      return FALSE;
    }
    bvSetElementAt(destination->isDeletedNode, FALSE, best->no);
  }
  
  return TRUE;
}

/* ----------------------------------------------------------------------
 *
 */
void lgInitialize()
{
  setRegister("compactlevelvalues", SET_BOOL, &lgCompactLVs, NULL, NULL, NULL, NULL);
}

/* ----------------------------------------------------------------------
   What categories can this node represent? (Needed while tagging.)
*/
List lgQueryCat(LexemGraph lg, Arc arc) {

  List items = inputLexiconGet(arc->word);
  List l;
  List result = NULL;

  if(inputSpuriousUppercase(lg, arc)) {
    String s = strCopy(arc->word);
    s[0] = tolower(s[0]);
    items = listAppendList(items, inputLexiconGet(s));
    memFree(s);
  }
  
  for(l = items; l != NULL; l = listNext(l)) {
    LexiconItem le = listItem(l);
    Value v = le->values[taggerCategoryIndex];
    if (v->type != VTString) {
      /* This should definitely not happen,
         but it will already generate errors elsewhere,
         so we just ignore it here. */
      continue;
    }
    result = listAddUniqueItem(result, strRegister(v->data.string));
    freeLexiconItem(le);
  }
  listFree(items);
  return result;
}

/* ----------------------------------------------------------------------
 * Does LG contain at least one instance of FORM? 
 */
Boolean lgContains(LexemGraph lg, String form) {
  int i;
  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElementAt(lg->graphemnodes, i);
    if(form == gn->arc->word) {
      return TRUE;
    }
    if(inputSpuriousUppercase(lg, gn->arc) &&
       !strcasecmp(form, gn->arc->word)) {
      return TRUE;
    }

  }
  return FALSE;
}
   
/* ----------------------------------------------------------------------
   Returns the most probable path, as defined by tagging scores.
*/
List lgMostProbablePath(LexemGraph lg) {
  List result = NULL;
  int i;
  List l;
  
  if(latticeBranches(lg->lattice)) {
    cdgPrintf(CDG_WARNING,
              "ERROR: lgSelectMostProbablePath() called no branching lattice!\n");
    return NULL;
  }
  
  for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElementAt(lg->graphemnodes, i);
    Number record = 0.0;
    LexemNode best = NULL;
    for(l = gn->lexemes; l != NULL; l = listNext(l)) {
      LexemNode ln = listItem(l);
      if(ln->tagscore > record) {
        record = ln->tagscore;
        best = ln;
      }
    }
    result = listPrependItem(result, best);
  }
  return result;
}

/* ----------------------------------------------------------------------
   Clone a grapheme node.

   The field ->lexemes is not set; the caller has to do that. (The two-way
   links between grapheme nodes and lexeme nodes can be set easier when all
   nodes are known.)
*/
GraphemNode gnClone(GraphemNode gn, Lattice lat) {

  GraphemNode result = 
  result = (GraphemNode) memMalloc(sizeof (GraphemNodeStruct));
  result->no = gn->no;
  result->lexemgraph = gn->lexemgraph;
  result->arc = inputFindArc(lat, gn->arc->no);
  result->lexemes = NULL;

  result->chunk = NULL;
  
  return result;
}


/* ----------------------------------------------------------------------
   Clone a lexeme graph.

   This performs a totally deep copy; even the underlying lattice, lexicon
   items etc. are cloned.
*/
LexemGraph lgClone(LexemGraph lg) {

  int i;
  LexemGraph result = memMalloc(sizeof(LexemGraphStruct));

  result->lattice = inputCloneLattice(lg->lattice);
  result->graphemnodes = vectorNew(20);
  for(i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElementAt(lg->graphemnodes, i);
    GraphemNode new = gnClone(gn, result->lattice);
    vectorAddElement(result->graphemnodes, new);
  }
  
  result->nodes = vectorNew(20);
  for(i = 0; i < vectorSize(lg->nodes); i++) {
    LexemNode ln = vectorElementAt(lg->nodes, i);
    LexemNode new = memMalloc(sizeof (LexemNodeStruct));
    new->no = ln->no;
    new->lexemgraph = result;
    new->arc = inputFindArc(result->lattice, ln->arc->no);
    new->lexem = inputCloneLI(ln->lexem);
    new->limit = ln->limit;
    new->grapheme = vectorElementAt(result->graphemnodes, ln->grapheme->no);
    new->grapheme->lexemes = listAppendItem(new->grapheme->lexemes, new);
    vectorAddElement(result->nodes, new);
  }

  result->max = lg->max;
  result->min = lg->min;

  result->isDeletedNode = bvClone(lg->isDeletedNode);

  result->distance = NULL;
  lgComputeDistances(result);

  result->noOfPathsToEnd = NULL;
  result->noOfPathsFromStart = NULL;
  lgComputeNoOfPaths(result);
  result->chunks = NULL;
  lgCopyTagScores(result, lg);

  return result;
    
}



/* -- ENDOFFILE --------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
