/* 
 * Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Birth: 26/11/99
 *
 */

/* ---------------------------------------------------------------------------
 * @addtogroup Parse Parse - Everything about parses. 
 * \author Kilian Foth (see also AUTHORS and THANKS for more)
 *
 * $Id: parse.c,v 1.170 2004/01/15 09:53:47 foth Exp $
 *
 * (long description)
 *
 * @{ 
 */


/* -- INCLUDES ---------------------------------------------------------- */
#include <configure.h>
#include <my_time.h>

#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <blah.h>
#include <limits.h>
#include "parse.h"
#include "cdg.h"
#include "eval.h"
#include "hook.h"
#include "input.h"
#include "set.h"
#include "tagger.h"

/* -- VARIABLES --------------------------------------------------------- */
Parse parseMostRecentlyCreatedParse = NULL;
Boolean parseVerifyPartially = FALSE;

/* -- MACROS ------------------------------------------------------------ */

/* -- TYPE DEFINITIONS -------------------------------------------------- */

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
   Return an unused, newly allocated parse id. 
 */
String parseMakeId()
{
  static int parseCounter = 0;
  return strPrintf("parse%d", parseCounter++);
}

/* ----------------------------------------------------------------------
   Return a new Parse structure with nothing set. Take the argument as the
   id of the new parse, or create an old-style parse0, parse1, ... id.
 */
Parse parseNew(String id) 
{
  Parse result = memMalloc(sizeof (ParseStruct));
  if (id == NULL) {
    result->id = parseMakeId(); /* no strRegister needed */
  } else {
    result->id = strRegister(id);
  }

  result->userName = cdgUser;
  result->latticeId = NULL;
  result->dateOfCreation = time(NULL);
  result->grammarFiles = listClone(inputCurrentGrammar->files);
  result->searchStrategy = NULL;
  result->comment = NULL;
  result->solutionNo = -1;
  result->noOfSolutions = -1;
  result->score = 0.0;
  result->optimal = FALSE;
  result->violatedConstraints = NULL;
  result->words = NULL;
  result->levels = NULL;
  result->labels = NULL;
  result->verticesStructure = NULL;
  result->verticesLabels = NULL;
  result->evalUnary = -1;
  result->evalBinary = -1;
  result->numberOfHits = -1;
  result->evaluationTime = -1;
  result->statisticComment = NULL;
  result->LVs = NULL;
  result->lg = NULL;
  result->violations = NULL;
  result->badness = NULL;
  result->map = NULL;
  result->rmap = NULL;
  
  return result;
}

/* ----------------------------------------------------------------------
   Delete a Parse.
 */
void parseDelete(Parse p)
{

  int i;

  cdgFreeString(p->userName);
  listFree(p->grammarFiles);
  cdgFreeString(p->searchStrategy);
  cdgFreeString(p->comment);
  listForeachFree(p->violatedConstraints, cdgFreeString);

  for (i = 0; i < vectorSize(p->words); i++) {
    deleteWord(vectorElementAt(p->words, i));
  }
  vectorDelete(p->words);

  listForeachFree(p->levels, cdgFreeString);
  listForeachFree(p->labels, cdgFreeString);

  if(p->verticesStructure) {
    vectorDelete(p->verticesStructure);
  }
  if(p->verticesLabels) {
    vectorDelete(p->verticesLabels);
  }
  cdgFreeString(p->statisticComment);

  if (!parseIsAbstract(p)) {
    parseStrip(p);
  }
  if (p==parseMostRecentlyCreatedParse) { parseMostRecentlyCreatedParse=NULL; }
  memFree(p);

}

/* ----------------------------------------------------------------------
  Compare two words by time index
*/
Boolean compareWord(Word a, Word b)
{
  return (a->from < b->from);
}

/* ----------------------------------------------------------------------
   Given a parse, the name of a level, and the index of a word,
   return the corresponding index into the field verticesStructure.
*/
int parseIndex(Parse p, String levelname, int wordno) {
  List l;
  int nrLevels = listLength(p->levels);
  int levelindex = 0;
  String s = NULL;

  levelname = strRegister(levelname);
  for(l = p->levels; l != NULL; l = listNext(l)) {
    s = listItem(l);
    if(s==levelname) {
      break;
    }
    levelindex++;
  }
  if(s != levelname) {
    cdgPrintf(CDG_ERROR, "ERROR: no level `%s' in parse %s\n", levelname, p->id);
    return -1;
  }
  if(wordno > vectorSize(p->words)) {
    cdgPrintf(CDG_ERROR, "ERROR: no %dth word in parse %s\n", wordno, p->id);
    return -1;
  }
  return nrLevels * wordno + levelindex;
}


/* ----------------------------------------------------------------------
   Compute p->map and p->rmap.
*/
Boolean parseComputeIndices(Parse p, LexemGraph lg) {

  int max = listLength(inputCurrentGrammar->levels) * lg->max;
  int i, vertexno, lvindex;
  Word w;
  List l;
  
  p->map = vectorNew(max);
  p->rmap = vectorNew(max);
  vectorSetAllElements(p->map, (Pointer)-1);
  vectorSetAllElements(p->rmap, (Pointer)-1);

  vertexno = 0;
  for (i = 0; i < vectorSize(p->words); i++) {
    w = vectorElementAt(p->words, i);
    for(l = p->levels; l != NULL; l = listNext(l)) {
      String levelname = listItem(l);
      Level level = findLevel(levelname);
      if(!level) {
        cdgPrintf(CDG_INFO, "ERROR: no level `%s' defined.\n", levelname);
        return FALSE;
      }
      lvindex = inputCurrentGrammar->noOfLevels * w->from + level->no;
      vectorSetElementAt(p->map, (Pointer)lvindex, vertexno);
      vectorSetElementAt(p->rmap, (Pointer)vertexno, lvindex);

      vertexno++;
    }
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
   Find the lv index corresponding to a vertex number
*/
int parseMap(Parse p, int vertexno) {
  if(vertexno == -1) {
    return -1;
  }
  if(!p->map) {
    return -1;
  }
  return (int)vectorElementAt(p->map, vertexno);
}

/* ----------------------------------------------------------------------
   Find the vertex number corresponding to an lv index 
*/
int parseRmap(Parse p, int lvindex) {
  if(lvindex == -1) {
    return -1;
  }
  if(!p->rmap) {

    return -1;
  }

  return (int)vectorElementAt(p->rmap, lvindex);
}

/* ----------------------------------------------------------------------
  Calculate the structure of a Parse, i.e. the fields words, levels
  etc., from the Vector of LVs.
*/
void parseComputeStructure(Parse p, Vector LVs, BitVector isDeleted)
{
  Vector map;
  LexemNode ln, ln2;
  LevelValue lv = NULL;
  List l, m;
  Word w;
  int i, j;
  int firstLevel = -1;
  List found = NULL;
  LexemGraph lg = NULL;

  /* grant this please */
  if (!LVs || vectorSize(LVs) == 0) {
    /* never reach */
    cdgPrintf(CDG_ERROR, "ERROR: no levelvalues given to parseComputeStructure\n");
    abort();
  }
  

  /* build words and levels */

  /* we're going to find all used words by selecting the correct modifier
   * from all LVs of the first level found. If the lexical path is
   * ambiguous, the user has already been warned. */

  /* search first value and level */
  for (i = 0; i < vectorSize(LVs); i++) {
    lv = vectorElementAt(LVs, 0);
    if (lv) {
      if(!lg) {
        lg = lv->modifier->lexemgraph;
      }
      firstLevel = lv->level->no;
      break;
    }
  }
  if (!lv) {
    cdgPrintf(CDG_ERROR, "ERROR: no levelvalues given to parseComputeStructure\n");
    abort();
  }

  map = vectorNew(lv->modifier->lexemgraph->max);
  p->levels = NULL;
  p->words = vectorNew(0);
  j = 0;
  for (i = 0; i < vectorSize(LVs); i++) {
    lv = vectorElementAt(LVs, i);
    if (!lv)
      continue;

    /* register level */
    p->levels = listAddUniqueItem(p->levels, lv->level->id);

    /* register label */
    p->labels = listAddUniqueItem(p->labels, lv->label);

    /* check for ambiguous LVs */
    for(l = lv->modifiers; l != NULL; l = listNext(l)) {
      ln = listItem(l);
      if(listContains(found, ln) ||
         bvElementAt(isDeleted, ln->no)) {
        continue;
      }
      for(m = listNext(l); m != NULL; m = listNext(m)) {
        ln2 = listItem(m);
        if(listContains(found, ln2) ||
           bvElementAt(isDeleted, ln2->no)) {
          continue;
        }

        found = listAddUniqueItem(found, ln);
        found = listAddUniqueItem(found, ln2);
        cdgPrintf(CDG_WARNING, "WARNING: ");
        lgPrintNode(CDG_WARNING, ln);
        cdgPrintf(CDG_WARNING, " chosen, but\n         ");
        lgPrintNode(CDG_WARNING, ln2);
        cdgPrintf(CDG_WARNING, " is undeleted as well!\n");
      }
    }
   
    /* register word */
    if (lv->level->no != firstLevel) {
      continue;
    }

    ln = NULL;
    for (l = lv->modifiers; l != NULL; l = listNext(l)) {
      ln = listItem(l);
      if (!bvElementAt(isDeleted, ln->no)) {
	break;
      }
    }

    w = newWord();
    w->from = ln->arc->from;
    w->to = ln->arc->to;
    w->word = ln->lexem->word;
    w->description = strRegister(ln->lexem->description);
    vectorAddElement(p->words, w);
  }
  listFree(found);
  vectorSort(p->words, compareWord);

  /* compute mapping from time index to word index */
  j = 0;
  for (i = 0; i < vectorSize(p->words); i++) {
    w = vectorElementAt(p->words, i);
    vectorSetElementAt(map, (Pointer)j++, w->from);
  }

  /* Note that parseComputeIndices() cannot fail here.
     It only returns FALSE if inputCurrentGrammar does
     not contain a level named in p, and that can't happen
     because we only just got the level's name from there. */
  (void)parseComputeIndices(p,lg);
  
  /* compute structure */
  p->verticesStructure = vectorNew(vectorSize(LVs));
  p->verticesLabels = vectorNew(vectorSize(LVs));

  for (i = 0; i < vectorSize(LVs); i++) {
    lv = vectorElementAt(LVs, i);
    if (!lv)
      continue;

    if (lv->modifiee == NULL) {
      j = -1;
    } else if (lv->modifiee == NONSPEC) {
      j = -2;
    } else {
      j = (int)vectorElementAt(map, lv->modifiee->arc->from);
    }
    vectorAddElement(p->verticesStructure, (void *)j);
    vectorAddElement(p->verticesLabels, lv->label);
  }

  vectorDelete(map);

}


/* ----------------------------------------------------------------------
 * Generate and return an LV that corresponds to the binding at position
 * where in the Parse.
 */
LevelValue parseFindLV(Parse p, int where)
{
  LevelValue lv = NULL;
  int nrLevels = listLength(p->levels);
  List modifiers = NULL;
  List modifiees = NULL;

  String levelId;
  Level level;
  String label;
  Word modifierWord;
  LexemNode modifier;
  Word modifieeWord;
  LexemNode modifiee;
  int j;
  String s = NULL;
  List l;
  
  if(0 == nrLevels) {
    return NULL;
  }

  levelId = strRegister(listNth(p->levels, (where % nrLevels) + 1));
  level = findLevel(levelId);
  label = strRegister(vectorElementAt(p->verticesLabels, where));
  modifierWord = vectorElementAt(p->words, where / nrLevels);
  modifier = parseFindLn(p, modifierWord, NULL);
  j = (int)vectorElementAt(p->verticesStructure, where);

  if (j == -1) {
    modifieeWord = NULL;
    modifiee = NULL;
  } else if (j == -2) {
    modifieeWord = NULL;
    modifiee = NONSPEC;
  } else {
    modifieeWord = vectorElementAt(p->words, j);
    modifiee = parseFindLn(p, modifieeWord, NULL);
  }

  if (!level) {
    cdgPrintf(CDG_ERROR, "ERROR: can't find level `%s'.\n", levelId);
    cdgPrintf(CDG_ERROR, "ERROR: can't decorate this parse.\n");
    return NULL;
  }

  /* check whether the level has this label */
  for(l = level->labels; l != NULL; l = listNext(l)) {
    s = listItem(l);
    if(s == label) {
      break;
    }
  }

  if (s != label) {
    cdgPrintf(CDG_WARNING,
              "WARNING: level `%s' has no label `%s', using `%s'.\n",
              levelId, label, s);
    label = s;
  }
  
  if (!modifier) {
    cdgPrintf(CDG_ERROR, "ERROR: can't find word `%s'.\n", modifierWord->word);
    cdgPrintf(CDG_ERROR, "ERROR: Can't decorate this parse.\n");
    return NULL;
  }

  if (modifieeWord && !modifiee) {
    cdgPrintf(CDG_ERROR, "ERROR: can't find word `%s'.\n", modifieeWord->word);
    cdgPrintf(CDG_ERROR, "ERROR: Can't decorate this parse.\n");
    return NULL;
  }

  
  /* Fabricate LV */
  modifiers = listPrependItem(NULL, modifier);
  if (modifiee != NULL && modifiee != NONSPEC) {
    modifiees = listPrependItem(NULL, modifiee);
  }
  lv = lvNew(modifiers, level, label, modifiees);
  listFree(modifiers);
  if (modifiees != NULL) {
    listFree(modifiees);
  }
  return lv;

}

/* ----------------------------------------------------------------------
   Check whether a Parse is abstract.
 */
Boolean parseIsAbstract(Parse p)
{
  return (p->lg == NULL || p->LVs == NULL || p->badness == NULL);
}

/* ----------------------------------------------------------------------
 * How well does an match le?
 */
int anCompare(Annotation an, LexiconItem le, Boolean verbose) {

  List l, path;
  Value v;
  String c = NULL;
  int result = 0;
  Hierarchy h = strlen(inputFeatureHierarchy) ?
    findHierarchy(inputFeatureHierarchy) :
    NULL;
  
  for (l = an->specs; l; l = listNext(l)) {
    Specification s = (Specification) listItem(l);
    if (s->type != STTag) {
      continue;
    }
    v = le->values[inputRegisterPath(path=scanPath(s->kind), inputCurrentGrammar)];
    listFree(path);
    if(v == ErrorValue) {
      if(verbose) {
        cdgPrintf(CDG_WARNING, "WARNING: feature `%s' does not exist\n", s->kind);
      }
        result++;
    } else if(!inputCompareFeatures(s->name, (c = sprintValue(v)), h)) {
      if(verbose) {
        cdgPrintf(CDG_WARNING, "WARNING: feature `%s' has value `%s', should be `%s'\n",
                  s->kind, c, s->name);
      }
      result++;
    }
  }
  return result;
}

/* ----------------------------------------------------------------------
   Translate Annotation to LexiconItem.

   Among the currently loaded LexiconItem structures, the one that matches 
   A best is selected and returned. If no entry even has the same word
   form, NULL may be returned.
*/
LexiconItem anno2le(Annotation an) {
  List l, entries;
  LexiconItem result = NULL;
  int i, record = INT_MAX;
  String lower = strCopy(an->word);
  
#ifdef DEBUGANNO2LE
  cdgPrintf(CDG_INFO, "DEBUG: translating `%s'\n", an->word);
#endif
  
  /* Pick the lexical entry that best matches the specified tags */
  entries = inputLexiconGet(an->word);

  /* allow lower case form of a word */
  lower[0] = tolower(lower[0]);
  entries = listAppendList(entries, inputLexiconGet(lower));

  for(l = entries; l != NULL; l = listNext(l)) {
    LexiconItem le = listItem(l);
    i = anCompare(an, le, FALSE);
#ifdef DEBUGANNO2LE
    cdgPrintf(CDG_INFO, "DEBUG:  `%s' has distance %d\n", le->description, i);
#endif
    if(i < record) {
      record = i;
      result = le;
    }
  }

  /* if there is no word in the lexicon that satisfies the annotation,
     use the one that matches best, but warn the user about the differences */
  if(record && result) {
    (void)anCompare(an, result, TRUE);
    cdgPrintf(CDG_WARNING, "WARNING: `%s' differs in %d feature%s.\n",
                           result->description,
                           record,
                           record == 1 ? "" : "s");
  }

  if(result) {
#ifdef DEBUGANNO2LE
    cdgPrintf(CDG_INFO, "DEBUG: selected `%s'\n", result->description);
#endif
    result = inputCloneLI(result);
  }

  listForeachFree(entries, freeLexiconItem);
  return result;
  
}

/* ----------------------------------------------------------------------
   Translate Annotation to LexemNode.

   Among the nodes of LG, the one that matches A best is selected and
   returned.

   An arbitrary number of feature mismatches are allowed if necessary, but
   if no node even has the same word form, NULL may be returned.
*/
LexemNode anno2ln(Annotation an, LexemGraph lg) {
  LexemNode result = NULL;
  int i, record = INT_MAX;
  
#ifdef DEBUGANNO2LE
  cdgPrintf(CDG_INFO, "DEBUG: translating `%s'\n", an->word);
#endif

  for(i = 0; i < vectorSize(lg->nodes); i++) {
    LexemNode ln = vectorElementAt(lg->nodes, i);
    int j;
    if(ln->arc->from != an->from || ln->arc->to != an->to) {
      continue;
    }
    
    j = anCompare(an, ln->lexem, FALSE);
#ifdef DEBUGANNO2LE
    cdgPrintf(CDG_INFO, "DEBUG:  `%s' has distance %d\n", ln->lexem->description, j);
#endif
    if(j < record) {
      record = j;
      result = ln;
    }
  }

  /* If a result is found but that satisfies the annotation,
     use the one that matches best,
     but warn the user about the differences */
  if(record && result) {
    (void)anCompare(an, result->lexem, TRUE);
    cdgPrintf(CDG_WARNING, "WARNING: %s: `%s' differs in %d feature%s.\n",
              lg->lattice->id,
              result->lexem->description,
              record,
              record == 1 ? "" : "s");
  }

  if(result) {
#ifdef DEBUGANNO2LE
    cdgPrintf(CDG_INFO, "DEBUG: selected `%s'\n", result->description);
#endif
  }

  return result;
  
}

/* ----------------------------------------------------------------------
   Find the lexeme node that corresponds to w best.

   If AE is not NULL, it is used to decide which of several homonyms was
   meant.
*/
LexemNode parseFindLn(Parse p, Word w, AnnoEntry ae) {

  Annotation an = NULL;
  List l;
  
  /* We cannot use parseIsAbstract() here,
     because this function is used by parseDecorate() itself,
     when the parse is still semi-decorated. */
  if(!p->lg) {
    return NULL;
  }
  
  /* If we have an AnnoEntry, find the relevant Annotation in it */
  if(ae) {
    for(l = ae->annos; l != NULL; l = listNext(l)) {
      Annotation a = listItem(l);
      if(a->from == w->from && a->to == w->to) {
        an = a;
        break;
      }
    }
  }

  /* If no annotation is available, just return a vaguely similar
     lexeme node */
  if(!an) {
    LexemNode result = NULL;
    int i;
    for(i = 0; i < vectorSize(p->lg->nodes); i++) {
      LexemNode ln = vectorElementAt(p->lg->nodes, i);
      if(ln->arc->from != w->from || ln->arc->to != w->to) {
        continue;
      }
      result = ln;
      if(ln->lexem->description == w->description) {
        return ln;
      }
    }
    if(!result) {
      cdgPrintf(CDG_DEBUG, "DEBUG: lexeme `%s' not found!\n", w->description);
    }
    return result;
  }

  /* If we have an annotation, use it to select the desired variant.  */
  return anno2ln(an, p->lg);
  
}

/* ----------------------------------------------------------------------
   Decorate a Parse with the optional elements required for eval'ing.
   The parse is not actually evaluated by this function.

   A clone of the lexeme graph passed is inserted into P, and fitting LVs
   are allocated and installed.
   
   Which lexical entry is used for each of the words depends on the third
   argument. It AE is not NULL, the lexeme from LG that corresponds most
   closely to the specifications in AE is used. Otherwise that lexeme that
   has the same name as the field w->description is used. If no such lexeme
   exists, one is chosen at random.
   
*/
Boolean parseDecorate(Parse p, LexemGraph lg, AnnoEntry ae)
{
  LevelValue lv = NULL;
  int i;
  LexemNode ln;
  Word w;

  /* clone a private lexeme graph all of our own */
  p->lg = lgClone(lg);
  lgCopyTagScores(p->lg, lg);
  bvSetAllElements(p->lg->isDeletedNode, TRUE);
  for(i = 0; i < vectorSize(p->words); i++) {
    w = vectorElementAt(p->words, i);
    if(!(ln = parseFindLn(p, w, ae))) {
      parseStrip(p);
      return FALSE;
    }
    bvSetElementAt(p->lg->isDeletedNode, FALSE, ln->no);
    w->description = ln->lexem->description;
  }
    
  if(!parseComputeIndices(p, lg)) {
    parseStrip(p);
    return FALSE;
  }
  
  /* allocate vector of LVs */
  if (p->LVs) {
    vectorDelete(p->LVs);
  }

  /* Note well: although we allocate p->LVs to the same size
     as p->verticesStructure, it may actually become larger later!
     p->verticesStructure is as large as the number of constraint edges.
     p->LVs may be larger if there are `long' words in the parse.
     p->LVs is always indexed by lvIndex(). */
  p->LVs = vectorNew(vectorSize(p->verticesStructure));

  /* fill it with LVs owned by nobody else but us */
  for (i = 0; i < vectorSize(p->verticesStructure); i++) {
    lv = parseFindLV(p, i);
    if(!lv) {
      parseStrip(p);
      return FALSE;
    }
    vectorSetElementAt(p->LVs, lv, lvIndex(lv));
  }

  p->violations = NULL;
  p->badness = bestBadness();
  return TRUE;
  
}

/* ----------------------------------------------------------------------
 * Make a decorated Parse abstract again.
 */
void parseStrip(Parse p)
{
  int i;
  LevelValue lv;

  if (p->LVs) {
    for (i = 0; i < vectorSize(p->LVs); i++) {
      lv = vectorElementAt(p->LVs, i);
      if(lv) {
        lvDelete(lv);
      }
    }
    vectorDelete(p->LVs);
    p->LVs = NULL;
  }

  if (p->violations) {
    listForeachFree(p->violations, cvDelete);
    p->violations = NULL;
  }

  if (p->lg) {
    freeLattice(p->lg->lattice);
    lgDelete(p->lg);
    p->lg = NULL;
  }
  if(p->badness) {
    bDelete(p->badness);
  }
  if(p->map) {
    vectorDelete(p->map);
  }
  if(p->rmap) {
    vectorDelete(p->rmap);
  }
}

/* ----------------------------------------------------------------------
 *   Compute the score of a Parse.
 */
void parseEval(Parse p)
{

  ConstraintViolation cv;
  List l;
  int i, j;
  LevelValue lva, lvb;

  /* must have the LVs to work! */
  if (parseIsAbstract(p)) {
    cdgPrintf(CDG_WARNING, "WARNING: no levelvalues in parse `%s'\n", p->id);
    return;
  }

  /* retract previous info */
  if (p->violations != NULL) {
    listForeachFree(p->violations, cvDelete);
    p->violations = NULL;
  }
  if (p->violatedConstraints != NULL) {
    listForeachFree(p->violatedConstraints, cdgFreeString);
    p->violatedConstraints = NULL;
  }

  p->score = 1.0;
  p->badness->no = 0;
  p->badness->hard = 0;
  p->badness->soft = 1.0;
  
  /* seek unary violations */
  for (i = 0; i < vectorSize(p->LVs); i++) {
    lva = vectorElementAt(p->LVs, i);
    if (lva)
      p->score *= evalUnary(lva, NULL, p->LVs, FALSE, p->badness, &p->violations);
  }

  /* seek binary violations */
  for (i = 0; i < vectorSize(p->LVs); i++) {
    lva = vectorElementAt(p->LVs, i);
    if (!lva)
      continue;
    for (j = i + 1; j < vectorSize(p->LVs); j++) {
      lvb = vectorElementAt(p->LVs, j);
      if (lvb)
	p->score *= evalBinary(lva, lvb, NULL, p->LVs, FALSE, p->badness, &p->violations);
    }
  }

  /* sort the list of violations */
  l = listSort(p->violations, cvCompare); 
  listFree(p->violations); 
  p->violations = l;

  /* compute constraint list */
  for (l = p->violations; l != NULL; l = listNext(l)) {
    cv = listItem(l);
    p->violatedConstraints =
      listAddUniqueItem(p->violatedConstraints, cv->constraint->id);
  }
 
}

/* ----------------------------------------------------------------------
 * Recompute conflicts and score after a change 
 * at the points indicated by l.
 */
void parseReEval(Parse p, List changes)
{
  ConstraintViolation cv;
  List l, m, newViolations = NULL;
  int i, j;
  LevelValue lva, lvb;
  List indices;
  
  /* must have the LVs to work! */
  if (!p->LVs) {
    return;
  }

  /* build list of lv indices */
  indices = NULL;
  for(l = changes; l != NULL; l = listNext(l)) {
    int i = (int)listItem(l);
    indices = listPrependItem(indices, vectorElementAt(p->map, i));
  }
  listFree(changes);

  /* retract conflicts that may have disappeared */
  for (l = p->violations; l != NULL; l = listNext(l)) {
    cv = listItem(l);

    if (!cv->constraint->is_context_sensitive &&
        !listContains(indices, (Pointer)cv->nodeBindingIndex1) &&
        (cv->nodeBindingIndex2 == -1
         || !listContains(indices, (Pointer)cv->nodeBindingIndex2))
        ) {
      newViolations = listAppendItem(newViolations, cv);

    } else {
      bSubtract(p->badness, cv->penalty);

      cvDelete(cv);
    }
  }
  listFree(p->violations);
  p->violations = newViolations;
  newViolations = NULL;

  for (l = indices; l != NULL; l = listNext(l)) {
    i = (int)listItem(l);

    lva = vectorElementAt(p->LVs, i);

    if (!lva) {
      continue;
    }

    /* seek new unary violations */
    (void)evalUnary(lva, NULL, NULL, FALSE, p->badness, &newViolations);

    /* seek binary violations */

    /* (A) binary conflicts between changed and unchanged LVs */
    for (j = 0; j < vectorSize(p->LVs); j++) {
      if (listContains(indices, (Pointer) j)) {
	continue;
      }

      lvb = vectorElementAt(p->LVs, j);
      if (lvb)
	(void)evalBinary(lva, lvb, NULL, NULL, FALSE, p->badness, &newViolations);
    }

    /* (B) binary conflicts between two changed LVs */
    for (m = listNext(l); m != NULL; m = listNext(m)) {

      j = (int)listItem(m);
      lvb = vectorElementAt(p->LVs, j);
      if (lvb)
	(void)evalBinary(lva, lvb, NULL, NULL, FALSE, p->badness, &newViolations);
    }
  }

  /* find context-sensitive conflicts */
  (void)evalInContext(p->LVs, p->LVs, p->badness, &newViolations);

  /* insert sorted the new violations into the remaining violations */
  for (l = newViolations; l; l = listNext(l)) {
    p->violations = listInsertSorted(p->violations, listItem(l), cvCompare);
  }
  listFree(newViolations);

  /* compute score */
  p->score = 1.0;
  for (l = p->violations; l != NULL; l = listNext(l)) {
    p->score *= ((ConstraintViolation) listItem(l))->penalty;
  }

  /* compute constraint list */
  listFree(p->violatedConstraints);
  p->violatedConstraints = NULL;
  for (l = p->violations; l != NULL; l = listNext(l)) {
    cv = listItem(l);
    p->violatedConstraints =
      listAddUniqueItem(p->violatedConstraints, cv->constraint->id);
  }
  listFree(indices);

}


/* ----------------------------------------------------------------------
 * Change one label in a Parse.
 */
void parseSwapLabel(Parse p, int where, String label)
{
  List l;
  LevelValue new, old;
  String oldLabel = vectorElementAt(p->verticesLabels, where);

  label = strRegister(label);

  /* swap label */
  vectorSetElementAt(p->verticesLabels, label, where);
  p->labels = listAddUniqueItem(p->labels, label);
  if (!listContains(p->labels, oldLabel)) {
    p->labels = listDeleteItem(p->labels, oldLabel);
  }

  if (!parseIsAbstract(p)) {

    /* change LV to suit */
    old = vectorElementAt(p->LVs, (int)vectorElementAt(p->map,where));
    new = parseFindLV(p, where);
    if (old == new) {
      return;
    }
    lvDelete(old);
    vectorSetElementAt(p->LVs, new, lvIndex(new));

    /* sanity check */
    if (!listContains(new->level->labels, label)) {
      cdgPrintf(CDG_WARNING,
		"WARNING: label `%s' not declared for level `%s'!\n",
		label, new->level->id);
    }

    /* re-eval grammar */
    l = listAppendItem(NULL, (Pointer) where);
    parseReEval(p, l);
  }
}

/* ----------------------------------------------------------------------
 * Change the governor of one edge in a Parse.
 */
void parseShiftEdge(Parse p, int where, int to)
{
  List l;
  LevelValue old,new;
  
  /* shift edge */
  vectorSetElementAt(p->verticesStructure, (void *)to, where);

  if (!parseIsAbstract(p)) {

    /* change LV to suit */

    old = vectorElementAt(p->LVs, (int)vectorElementAt(p->map, where));
    new = parseFindLV(p, where);
    if(old == new) {
      return;
    }
    lvDelete(old);
    vectorSetElementAt(p->LVs, new, lvIndex(new));

    /* re-eval grammar */
    l = listAppendItem(NULL, (Pointer) where);
    parseReEval(p, l);
  }
}

/* ----------------------------------------------------------------------
 * Exchanges a lexical reading in a Parse.
 */
void parseSwapWord(Parse p, int where, String id)
{
  LexemNode oldln = NULL, newln;
  LevelValue new, old;
  List changed;
  int nrLevels = listLength(p->levels);
  int i;
  Word w = vectorElementAt(p->words, where);
  String oldDescription = w->description;
  
  if (!parseIsAbstract(p)) {
    oldln = parseFindLn(p, w, NULL);
  }

  /* swap id */
  w->description = strRegister(id);
  
  if (!parseIsAbstract(p)) {

    newln = parseFindLn(p, w, NULL);
    
    if (!newln) {
      cdgPrintf(CDG_ERROR, "ERROR: lexeme node `%s' doesn't exist.\n", id);
      w->description = strRegister(oldDescription);
      return;
    }

    /* prevent insidious bug that would happen in the next line. */
    if(newln == oldln) {
      return;
    }
    
    /* select node */
    bvSetElementAt(p->lg->isDeletedNode, FALSE, newln->no);
    lgDeleteNode(p->lg, oldln);

    /* maybe change reading */
    w->word = newln->lexem->word;
    
    /* update LVs */
    changed = NULL;
    for (i = 0; i < vectorSize(p->verticesStructure); i++) {
      if (i / nrLevels == where ||
          (int)vectorElementAt(p->verticesStructure, i) == where) {
        new = parseFindLV(p, i);
        if(!new) {
          cdgPrintf(CDG_ERROR, "Error: parseFindLV() failed.\n");
          return;
        }
	old = vectorElementAt(p->LVs, lvIndex(new));
	if (old != new) {
	  vectorSetElementAt(p->LVs, new, lvIndex(new));
	  changed = listPrependItem(changed, (Pointer) i);
	  lvDelete(old);
	}
      }
    }
    
    /* re-eval */
    parseReEval(p, changed);
  }
}

/* ----------------------------------------------------------------------
 * Manipulate a Parse so that all mirror edges are unarily optimal.
 */
void parseMirror(Parse p)
{
  Badness best;
  List l, m;
  String label;
  int i, j, k;
  int nrLevels = listLength(p->levels);
  int nrWords = vectorSize(p->words);
  List troublespots = NULL;

  /* refuse to work without LVs */
  if(parseIsAbstract(p)) {
    return;
  }

  /* find instances of hard violations */
  for(l = p->violations; l != NULL; l = listNext(l)) {
    ConstraintViolation cv = listItem(l);
    if(!cv->penalty) {
      troublespots =
        listAddUniqueItem(troublespots, (Pointer)cv->nodeBindingIndex1);
      if(cv->nodeBindingIndex2 != -1) {
        troublespots =
          listAddUniqueItem(troublespots, (Pointer)cv->nodeBindingIndex2);
      }
    }
  }
  
  /* iterate over levels */
  for (i = 0, l = p->levels; l != NULL; l = listNext(l), i++) {
    String s = listItem(l);

    label = NULL;
    for (m = inputCurrentGrammar->levels; m != NULL; m = listNext(m)) {
      Level level = listItem(m);
      if (s == level->id && (label = getLevelProperty(level, "mirror")))
	break;
    }
    /* find mirror levels */
    if (!label) {
      continue;
    }
    
    best = bClone(p->badness);
    
    /* iterate over words */
    for (j = 0; j < nrWords; j++) {
      int index = nrLevels * j + i;
      LevelValue lv = vectorElementAt(p->LVs, (int)vectorElementAt(p->map, index));
      int record = (int)vectorElementAt(p->verticesStructure, index);

      if (!lv) /* never reach */
	continue;

      /* only search if this LV causes hard conflicts */
      if(!listContains(troublespots, (Pointer)index)) {
        continue;
      }
      
      bCopy(best, p->badness);

      /* iterate over attachment points */
      for(k = -1; k < nrWords; k++) {
        if(k == j) {
          continue;
        }
        parseShiftEdge(p, index, k);

        if(bCompare(p->badness, best)) {
          bCopy(best, p->badness);
          record = k;
        }
      }
      parseShiftEdge(p, index, record);
    }
  }
  listFree(troublespots);
}

/* ----------------------------------------------------------------------
   Computes the height of level id in Parse p.
 */
int parseHeight(Parse p, String levelname)
{
  int wordNo, height, treeHeight = 0;

  /* get the maximum height over all words */
  for (wordNo = 0; wordNo < vectorSize(p->words); wordNo++) {
    height = parseWordHeight(p, levelname, wordNo);

    /* cycle? */
    if (height > vectorSize(p->verticesStructure)) {
      treeHeight = -1;
      break;
    }
    treeHeight = max(treeHeight, height);
  }

  return treeHeight;
}

/* ------------------------------------------------------------------------- 
 * @brief get all child nodes of a word on a level tree
 *
 * @param parse the current parse that we examin
 * @param wordNo the word index in the word vector of the current parse
 * @param level a Level 
 * @return a list of word indices or NULL if there are no modifiers.
 *
 * @note you become the owner of the returned @ref List, 
 * so deallocate (@ref ListFree) it after you've consumed the result.
 */
List parseGetModifiers(Parse parse, Level level, int wordIndex)
{
  int i;
  List modifiers = NULL;
  int nrLevels = listLength(parse->levels);

  for (i = 0; i < vectorSize(parse->words); i ++) {
     int modifieeIndex = (int) 
       vectorElementAt(parse->verticesStructure, nrLevels * i + level->no);
    if (wordIndex == modifieeIndex) {
      modifiers = listAppendItem(modifiers, (Pointer)i);
    }
  }

  return modifiers;
}

/* ----------------------------------------------------------------------
   Computes the depth of a word on a level in Parse p, that is 
   the maximum number of hops to the leaf nodes
 */
int parseWordDepth(Parse parse, Level level, int wordIndex)
{
  int maxDepth = 0;
  List modifiers = parseGetModifiers(parse, level, wordIndex);
  
  if (listLength(modifiers) != 0) {
    List l;
    int depth;
  
    for (l = modifiers; l; l = listNext(l)) {
      int modifierIndex = (int) listItem(l);
      depth = parseWordDepth(parse, level, modifierIndex) + 1;
      if (depth > maxDepth)
	maxDepth = depth;
    }
  }

  return maxDepth;
}

/* ----------------------------------------------------------------------
   Computes the height of a word on a level in Parse p, that is 
   the number of hops to the root
 */
int parseWordHeight(Parse p, String levelname, int wordNo)
{
  int k, height = 0;
  int max = vectorSize(p->verticesStructure);


  /* find the longest path from a leaf to the root */
  while(wordNo != -1 && height < max) {
    k = parseIndex(p,levelname,wordNo);
    if(k == -1) {
      return 0;
    }
    wordNo = (int)vectorElementAt(p->verticesStructure, k);
    height++;
  }

  /* cycle? */
  if (height == max) {
    height = INT_MAX;
  }

  return height;
}

/* ----------------------------------------------------------------------
  Print a parse on stdout.
*/
void parsePrint(Parse parse)
{
  int i, j;
  String label, levelId;
  Word word;
  int nrLevels = listLength(parse->levels);
  Level level;
  ConstraintViolation cv;
  Vector violations;

  cdgPrintf(CDG_INFO, "\n"
	    "   parse id    : %s\n"
	    "   lattice id  : %s\n"
	    "   username    : %s\n"
	    "   date        : %s"
	    "   strategy    : %s\n"
	    "   comment     : %s\n"
	    "   score       : %4.3e\n"
	    "   optimal     : %s\n\n   bindings:\n",
	    parse->id,
	    parse->latticeId,
	    parse->userName,
	    ctime(&parse->dateOfCreation),
	    parse->searchStrategy,
	    parse->comment ? parse->comment : "<undefined>",
	    parse->score, 
	    parse->optimal ? "yes" : "no");

  for (i = 0; i < vectorSize(parse->verticesStructure); i++) {
    levelId = listNth(parse->levels, (i % nrLevels) + 1);
    level = findLevel(levelId);
    if (!level || !level->showflag)
      continue;

    label = vectorElementAt(parse->verticesLabels, i);
    word = vectorElementAt(parse->words, i / nrLevels);

    cdgPrintf(CDG_INFO, "   %03d %10s: %s(%d-%d)--%s-->",
	      word->from *
              inputCurrentGrammar->noOfLevels + level->no,
              levelId, word->description, word->from, word->to, label);

    j = (int)vectorElementAt(parse->verticesStructure, i);
    if (j == -1) {
      cdgPrintf(CDG_INFO, "NIL\n");
    } else if (j == -2) {
      cdgPrintf(CDG_INFO, "NONSPEC\n");
    } else {
      word = vectorElementAt(parse->words, j);
      cdgPrintf(CDG_INFO, "%s(%d-%d)\n",
		word->description, word->from, word->to);
    }
  }
  if (parse->violations) {
    cdgPrintf(CDG_INFO, "\n   Violations:\n");
    violations = listToVector(parse->violations);
    vectorSort(violations, cvCompare);
    for (i = 0; i < vectorSize(violations); i++) {
      cv = vectorElementAt(violations, i);
      levelId = listNth(parse->levels, (cv->nodeBindingIndex1 % nrLevels) + 1);
      level = findLevel(levelId);
      if (!level || !level->showflag)
	continue;
      cvPrint(CDG_INFO, cv);
      cdgPrintf(CDG_INFO, "\n");
    }
    vectorDelete(violations);
  }

  cdgPrintf(CDG_INFO, "\n");
}

/* ----------------------------------------------------------------------
  Return a new Word structure with nothing set.
*/
Word newWord(void)
{

  Word result = memMalloc(sizeof (WordStruct));

  result->from = -1;
  result->to = -1;
  result->word = NULL;
  result->description = NULL;

  return result;
}

/* ----------------------------------------------------------------------
  Delete a WordStruct.
*/
void deleteWord(Word w)
{

  cdgFreeString(w->word);
  cdgFreeString(w->description);

  memFree(w);
}

/* ----------------------------------------------------------------------
   Compute the mapping of words from vector a onto words from vector b;
   the two are vectors of Word
*/
void parseComputeMap(Vector map, Vector a, Vector b)
{
  int i, j;
  int la = vectorSize(a);
  int lb = vectorSize(b);

  Vector d = vectorNew((la + 1) * (lb + 1));

  /* Levenshtein distance matrix */
  Vector dir = vectorNew((la + 1) * (lb + 1));

  /* directions matrix, used to determine the mapping */

#define pos(x,y) ((x)*(lb+1)+(y))

  vectorSetElementAt(d, (void *)0, pos(0, 0));
  for (i = 1; i <= la; i++)
    vectorSetElementAt(d, (void *)i, pos(i, 0));
  for (j = 1; j <= lb; j++)
    vectorSetElementAt(d, (void *)j, pos(0, j));
  for (i = 1; i <= la; i++)
    for (j = 1; j <= lb; j++) {
      Word wa = (Word) vectorElementAt(a, i - 1);
      Word wb = (Word) vectorElementAt(b, j - 1);
      int m1 = (int)vectorElementAt(d, pos(i - 1, j - 1)) +
	(strcmp(wa->word, wb->word) == 0 ? 0 : 1);
      int m2 = (int)vectorElementAt(d, pos(i, j - 1)) + 1;
      int m3 = (int)vectorElementAt(d, pos(i - 1, j)) + 1;

      if (m1 <= m2 && m1 <= m3) {
	vectorSetElementAt(d, (void *)m1, pos(i, j));
	vectorSetElementAt(dir, (void *)1, pos(i, j));
      }
      if (m2 <= m1 && m2 <= m3) {
	vectorSetElementAt(d, (void *)m2, pos(i, j));
	vectorSetElementAt(dir, (void *)2, pos(i, j));
      }
      if (m3 <= m1 && m3 <= m2) {
	vectorSetElementAt(d, (void *)m3, pos(i, j));
	vectorSetElementAt(dir, (void *)3, pos(i, j));
      }
    }

#if 0
  for (i = 0; i <= la; i++) {
    for (j = 0; j <= lb; j++)
      cdgPrintf(CDG_INFO, "%d ", (int)vectorElementAt(d, pos(i, j)));
    cdgPrintf(CDG_INFO, "\n");
  }
#endif

  for (i = 0; i < la; i++)
    vectorSetElementAt(map, (void *)-1, i);
  i = la;
  j = lb;
  while (i >= 1 && j >= 1) {
    Word wa = (Word) vectorElementAt(a, i - 1);
    Word wb = (Word) vectorElementAt(b, j - 1);

    if (strcasecmp(wa->word, wb->word) == 0)
      vectorSetElementAt(map, (void *)(j - 1), i - 1);
    switch ((int)vectorElementAt(dir, pos(i, j))) {
      case 1:
	i--;
	j--;
	break;
      case 2:
	j--;
	break;
      case 3:
	i--;
	break;
    }
  }
  for (i = la - 1; i >= 0 && (int)vectorElementAt(map, i) == -1;
       vectorSetElementAt(map, (void *)-2, i--)) ;

  vectorDelete(d);
  vectorDelete(dir);
}

/* ----------------------------------------------------------------------
 * return the lexical category of a Word as a String, or "-".
 * This function returns the syntactic category of the word at position
 * \c wordno in \c p as a String. It can fail for many reasons
 * (\c taggerCategoryIndex not set, parse not decorated, etc.). If it
 * fails, the String `-' is returned.
 */
String wordCat(Parse p, int wordno) {
  
  /*  if we don't know which attribute encodes the syntactic category,
      we can't say anything */
  if(-1 != taggerCategoryIndex) {
    if(!parseIsAbstract(p)) {
      int index = wordno * inputCurrentGrammar->noOfLevels;
      if(index < vectorSize(p->LVs)) {
        LevelValue lv = vectorElementAt(p->LVs, index);
        if(lv) {
          Value v = lv->modifier->lexem->values[taggerCategoryIndex];
          if(v->type == VTString) {
            return v->data.string;
          }
        }
      }
    }
  }
  
  /* give up */
  return strRegister("-");

}

/* ----------------------------------------------------------------------
   A faade for parseCompare(): simply answers yes or no quickly.
 */
Boolean parseEquals(Parse a, Parse b)
{
  ParseCmpStatStruct stat;

  return parseCompare(a, b, &stat, PC_SILENT, PC_STRICT) == 0;
}

/* ----------------------------------------------------------------------
  Compare two parses

  Takes an integer defining the needed verbosity behaviour, and a boolean
  defining the type of mapping; Returns the total number of errors
*/
int parseCompare(Parse ref, Parse test, ParseCmpStat stat, int verbosity, int strict)
{
  if (ref == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: NULL reference parse\n");
    return -1;
  } else if (test == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: NULL test parse\n");
    return -1;
  } else {
  Vector map;	     /* mapping of words in first to words in second parse */
  int nrwordsprefixref; /* number of words in prefix of reference */
 
  /* 
   * values of the mapping: 
   * -1  = word not in the second wordgraph 
   * -2  = out of the common time span 
   * >=0 = position in the second wordgraph 
   */
  List l;
  int i, j, k1, k2, errors, index;
  int node1, node2;	     /* absolute number of current word in the two wordgraphs */
  int level1, level2;	     /* number of the current level in the two parses */
  int nodegov1, nodegov2;    /* number of the current governor in the two parses */
  String strword;	     /* the current word */
  String strcat = NULL;      /* its category */
  String strnode1, strnode2; /* the current word, including lexical info, in both parses */
  String strlevel;	     /* the current level */
  String strwordgov1, strwordgov2;	/* the current governors in both parses */
  String strnodegov1, strnodegov2;	/* same, including lexical info */
  String strcatgov1, strcatgov2;	/* their lexical category */
  String strlabel1, strlabel2;	/* the current labels in both parses */
  Word word1, word2;	     /* the curent words in both parses */
  Word gov1, gov2;	     /* the current governors in both parses */
  Vector levels_in_grammar;  /* the level ids in the current grammar */
  Vector totaledgecorrect_level;     /* total number of correct edges per level      */
  Vector totaledgetest_level;   /* total number of edges per level in test      */
  Vector totaledgeref_level;    /* total number of edges per level in reference */

  int nrwords1  = vectorSize(ref->words);
  int nrwords2  = vectorSize(test->words);
  int nrlevels1 = listLength(ref->levels);
  int nrlevels2 = listLength(test->levels);
  int nrlevels  = listLength(inputCurrentGrammar->levels);

  if (verbosity == PC_SUMMARY || verbosity == PC_ALL) {
    cdgPrintf(CDG_INFO, "\n");
    cdgPrintf(CDG_INFO, "INFO: Compare parses %s and %s\n\n", ref->id,
	      test->id);
  }

  map = vectorNew(nrwords1);
  parseComputeMap(map, ref->words, test->words);

#if 0
  cdgPrintf(CDG_INFO, "%d %d %d %d\n", nrwords1, nrwords2, nrlevels1,
	    nrlevels2);

  for (i = 0; i < vectorSize(map); i++)
    cdgPrintf(CDG_INFO, "%d ", (int)vectorElementAt(map, i));
  cdgPrintf(CDG_INFO, "\n");

  for (i = 0; i < nrwords1; i++)
    cdgPrintf(CDG_INFO, "%s ", ((Word) vectorElementAt(ref->words, i))->word);
  cdgPrintf(CDG_INFO, "\n");

  for (i = 0; i < nrwords2; i++)
    cdgPrintf(CDG_INFO, "%s ", ((Word) vectorElementAt(test->words, i))->word);
  cdgPrintf(CDG_INFO, "\n");
#endif

  stat->errorextrawordref = 0;
  stat->errorextrawordtest = 0;
  stat->errorwordlex = 0;
  stat->errorgovword = 0;
  stat->errorgovlex = 0;
  stat->erroredgelabel = 0;
  stat->matchedgebase = 0;
  stat->matchedgegov = 0;
  stat->matchedgetot = 0;
  stat->totalextrawordref = 0;
  stat->totalextrawordtest = 0;
  stat->totalwordlex = 0;
  stat->totalgovword = 0;
  stat->totalgovlex = 0;
  stat->totaledgelabel = 0;
  stat->totaledgebase = 0;
  stat->totaledgegov = 0;
  stat->totaledgetot = 0;

  totaledgecorrect_level   = vectorNew(nrlevels);
  totaledgetest_level = vectorNew(nrlevels);
  totaledgeref_level  = vectorNew(nrlevels);
  vectorSetAllElements(totaledgecorrect_level,   (Pointer) 0);
  vectorSetAllElements(totaledgetest_level, (Pointer) 0);
  vectorSetAllElements(totaledgeref_level,  (Pointer) 0);

  stat->totaledgetest = 0;
  stat->totaledgeref = 0;

  levels_in_grammar = vectorNew(nrlevels);
  for (l=inputCurrentGrammar->levels, i=0; l; l=listNext(l), i++) {
    Level level = (Level) listItem(l);
    vectorSetElementAt(levels_in_grammar, (Pointer) level->id, i);
  }

  /* errors: word in test but not in reference */
  for (i = 0; i < nrwords2; i++) {
    stat->totalextrawordtest++;
    if (vectorIndexOf(map, (void *)i, 0) == -1) {
      Word w = (Word) vectorElementAt(test->words, i);

      if (verbosity == PC_ALL)
	cdgPrintf(CDG_INFO,
		  "ERROR: word %s found in test, but not in reference\n",
		  w->word);
      stat->errorextrawordtest++;
    }
  }

  /* errors: word in reference but not in test */
  for (i = 0; i < nrwords1; i++) {
    node1 = i;
    node2 = (int)vectorElementAt(map, node1);
    word1 = (Word) vectorElementAt(ref->words, node1);
    strword = word1->word;
    strnode1 = word1->description;
    
    if (strict == PC_STRICT) {
      stat->totalextrawordref++;
      if (node2 == -2) {
	if (verbosity == PC_ALL)
	  cdgPrintf(CDG_INFO,
		    "ERROR: word %s found in reference, but not in test\n",
		    strword);
	stat->errorextrawordref++;
	continue;
      }
    } else {
      if (node2 == -2)
	continue;
    }

    stat->totalextrawordref++;
    if (node2 == -1) {
      if (verbosity == PC_ALL)
	cdgPrintf(CDG_INFO,
		  "ERROR: word %s found in reference, but not in test\n",
		  strword);
      stat->errorextrawordref++;
      continue;
    }
  }

  /* compute nrwordsprefixref */
  if (strict == PC_STRICT) {
    nrwordsprefixref = nrwords1;
  } else {
    int mapind;
    for (mapind = vectorSize(map)-1; mapind >= 0 && (int)vectorElementAt(map, mapind) < 0; mapind--);
    nrwordsprefixref = mapind + 1;
  }

  /* do the comparisons */
  for (i = 0; i < nrwords1; i++) {
    node1 = i;
    node2 = (int)vectorElementAt(map, node1);

    if (node2 == -1 || node2 == -2) 
      continue;

    word1 = (Word) vectorElementAt(ref->words, node1);
    word2 = (Word) vectorElementAt(test->words, node2);
    strnode1 = word1->description;
    strnode2 = word2->description;
    strword  = word1->word;

    stat->totalwordlex++;
    if (strcmp(strnode1, strnode2) != 0) {
      if (verbosity == PC_ALL)
	cdgPrintf(CDG_INFO,
		  "ERROR: Word with different lexical reading: %s should be %s\n",
		  strnode1, strnode2);
      stat->errorwordlex++;
    }

    for (j = 0; j < nrlevels1; j++) {
      level1 = j;
      strlevel = strRegister(listNth(ref->levels, level1 + 1));
      level2 = listIndex(test->levels, strlevel) - 1;
      if (level2 == -1)
	continue;		       /* level present in the first, but not in 
				        * the second parse */

      /* index of level in the vector of levels in the grammar */
      index = vectorIndexOf(levels_in_grammar, strlevel, 0);
      if(-1 == index) {
        continue;
      }
      
      k1 = node1 * nrlevels1 + level1;
      k2 = node2 * nrlevels2 + level2;
      strcat = wordCat(ref, node1);

      nodegov1 = (int)vectorElementAt(ref->verticesStructure, k1);
      nodegov2 = (int)vectorElementAt(test->verticesStructure, k2);
      strlabel1 = vectorElementAt(ref->verticesLabels, k1);
      strlabel2 = vectorElementAt(test->verticesLabels, k2);
      if (nodegov1 == -1) {
	gov1 = NULL;
	strwordgov1 = strRegister("NIL");
	strcatgov1 = strRegister("-");
	strnodegov1 = strRegister("NIL");
      } else if (nodegov1 == -2) {
	gov1 = NONSPEC;
	strwordgov1 = strRegister("NONSPEC");
	strcatgov1 = strRegister("-");
	strnodegov1 = strRegister("NONSPEC");
      } else {
	gov1 = (Word) vectorElementAt(ref->words, nodegov1);
	strwordgov1 = gov1->word;
	strcatgov1 = wordCat(test, nodegov1);
	strnodegov1 = gov1->description;
      }
      if (nodegov2 == -1) {
	gov2 = NULL;
	strwordgov2 = strRegister("NIL");
        strcatgov2 = strRegister("-");
	strnodegov2 = strRegister("NIL");
      } else if (nodegov2 == -2) {
	gov2 = NONSPEC;
	strwordgov2 = strRegister("NONSPEC");
        strcatgov2 = strRegister("-");
	strnodegov2 = strRegister("NONSPEC");
      } else {
	gov2 = (Word) vectorElementAt(test->words, nodegov2);
	strwordgov2 = gov2->word;
        strcatgov2 = wordCat(ref, nodegov2);
	strnodegov2 = gov2->description;
      }

      stat->totalgovword++;
      if (nodegov1 != nodegov2) {
	if (verbosity == PC_ALL)
	  cdgPrintf(CDG_INFO,
		    "ERROR: Governor of `%s/%s' on level `%s' is `%s/%s', should be `%s/%s'\n",
		    strword, strcat, strlevel, strwordgov1, strcatgov1, strwordgov2, strcatgov2);
	stat->errorgovword++;
      } else {
	stat->totalgovlex++;
	if (strcmp(strnodegov1, strnodegov2) != 0) {
	  if (verbosity == PC_ALL)
	    cdgPrintf(CDG_INFO,
		      "ERROR: Governor's lexical reading of node `%s' on level `%s' is `%s', should be `%s'\n",
		      strword, strlevel, strnodegov1, strnodegov2);
	  stat->errorgovlex++;
	}
      }
      stat->totaledgelabel++;
      if (strcmp(strlabel1, strlabel2) != 0) {
	if (verbosity == PC_ALL)
	  cdgPrintf(CDG_INFO,
		    "ERROR: Label of edge above node `%s/%s' on level `%s' is `%s', should be `%s'\n",
		    strword, strcat, strlevel, strlabel1, strlabel2);
	stat->erroredgelabel++;
      }
      stat->totaledgebase++;
      if (strcmp(strnode1, strnode2) == 0 && strcmp(strlabel1, strlabel2) == 0) {
	stat->matchedgebase++;
      }
      stat->totaledgegov++;
      if (strcmp(strnode1, strnode2) == 0 &&
	  strcmp(strlabel1, strlabel2) == 0 &&
	  strcmp(strwordgov1, strwordgov2) == 0) {
	stat->matchedgegov++;
      }
      stat->totaledgetot++;
      if (strcmp(strnode1, strnode2) == 0 &&
	  strcmp(strlabel1, strlabel2) == 0 &&
	  strcmp(strnodegov1, strnodegov2) == 0) {
        int currentvalue = (int)vectorElementAt(totaledgecorrect_level, index);
	vectorSetElementAt(totaledgecorrect_level, (Pointer)(1 + currentvalue), index);
	stat->matchedgetot++;
      }
    }
  }

  for (j = 0; j < nrlevels2; j++) {
    level2 = j;
    strlevel = listNth(test->levels, level2 + 1);
    index = vectorIndexOf(levels_in_grammar, strlevel, 0);
    if (index != -1) {
      vectorSetElementAt(totaledgetest_level, (Pointer)nrwords2, index);
    }
    stat->totaledgetest += nrwords2;
  }

  for (j = 0; j < nrlevels1; j++) {
    level1 = j;
    strlevel = listNth(ref->levels, level1 + 1);
    index = vectorIndexOf(levels_in_grammar, strlevel, 0);
    if (index != -1) {
      vectorSetElementAt(totaledgeref_level, (Pointer)nrwordsprefixref, index);
    }
    stat->totaledgeref += nrwordsprefixref;
  }

  errors = stat->errorextrawordref + stat->errorextrawordtest +
    stat->errorwordlex + stat->errorgovword + stat->errorgovlex +
    stat->erroredgelabel;

  if (verbosity == PC_SUMMARY || verbosity == PC_ALL) {
    if (errors != 0) {
      cdgPrintf(CDG_INFO, "\nINFO: Total of %d errors found\n\n", errors);
    } else {
      cdgPrintf(CDG_INFO, "INFO: No errors found\n\n");
    }
    cdgPrintf(CDG_INFO, "INFO: Summary of errors/checks in parseCompare\n");
    cdgPrintf(CDG_INFO, "\n");
    cdgPrintf(CDG_INFO,
	      "INFO: Extra words in reference:                 errors  %5d, checks %5d\n",
	      stat->errorextrawordref, stat->totalextrawordref);
    cdgPrintf(CDG_INFO,
	      "INFO: Extra words in test:                      errors  %5d, checks %5d\n",
	      stat->errorextrawordtest, stat->totalextrawordtest);
    cdgPrintf(CDG_INFO,
	      "INFO: Words with different lexical reading      errors  %5d, checks %5d\n",
	      stat->errorwordlex, stat->totalwordlex);
    cdgPrintf(CDG_INFO,
	      "INFO: Governors with different words:           errors  %5d, checks %5d\n",
	      stat->errorgovword, stat->totalgovword);
    cdgPrintf(CDG_INFO,
	      "INFO: Governors with different lexical reading: errors  %5d, checks %5d\n",
	      stat->errorgovlex, stat->totalgovlex);
    cdgPrintf(CDG_INFO,
	      "INFO: Edges with different label:               errors  %5d, checks %5d\n",
	      stat->erroredgelabel, stat->totaledgelabel);
    cdgPrintf(CDG_INFO,
	      "INFO: Identical edges (governors not counting): matches %5d, checks %5d\n",
	      stat->matchedgebase, stat->totaledgebase);
    cdgPrintf(CDG_INFO,
	      "INFO: Identical edges (same governors words):   matches %5d, checks %5d\n",
	      stat->matchedgegov, stat->totaledgegov);
    cdgPrintf(CDG_INFO,
	      "INFO: Identical edges (same gov. lex. read.):   matches %5d, checks %5d\n",
	      stat->matchedgetot, stat->totaledgetot);
    cdgPrintf(CDG_INFO, "\n");
  }

  if (verbosity == PC_SUMMARY || verbosity == PC_MEASURES || verbosity == PC_ALL) {
    cdgPrintf(CDG_INFO, "INFO: Total\tAccuracy = ");
    if (stat->totaledgetest == 0) {
      stat->accuracy = -1;
      cdgPrintf(CDG_INFO, "%s\tRecall = ", "n/a");
    } else {
      stat->accuracy = (float)stat->matchedgetot / (float)stat->totaledgetest;
      cdgPrintf(CDG_INFO, "%7.5f\tRecall = ", stat->accuracy);
    }
    if (stat->totaledgeref == 0) {
      stat->recall = -1;
      cdgPrintf(CDG_INFO, "%s\tF-measure = ", "n/a");
    } else {
      stat->recall = (float)stat->matchedgetot / (float)stat->totaledgeref;
      cdgPrintf(CDG_INFO, "%7.5f\tF-measure = ", stat->recall);
    }
    if (stat->totaledgeref + stat->totaledgetest == 0) {
      stat->fmeasure = -1;
      cdgPrintf(CDG_INFO, "%s\n", "n/a");
    } else {
      stat->fmeasure = 2 * (float)stat->matchedgetot / ((float)stat->totaledgetest + (float)stat->totaledgeref);
      cdgPrintf(CDG_INFO, "%7.5f\n", stat->fmeasure);
    }
    
    for (i=0; i<nrlevels; i++) {
      String strLevel   = (String)vectorElementAt(levels_in_grammar, i);
      int totaledgecorrect   = (int)vectorElementAt(totaledgecorrect_level, i);
      int totaledgetest = (int)vectorElementAt(totaledgetest_level, i);
      int totaledgeref  = (int)vectorElementAt(totaledgeref_level, i);
      cdgPrintf(CDG_INFO, "INFO: %s\tAccuracy = ", strLevel);
      if (totaledgetest == 0) {
	cdgPrintf(CDG_INFO, "%s\tRecall = ", "n/a");
      } else {
	cdgPrintf(CDG_INFO, "%7.5f\tRecall = ", (float)totaledgecorrect/(float)totaledgetest);
      }
      if (totaledgeref == 0) {
	cdgPrintf(CDG_INFO, "%s\tF-measure = ", "n/a");
      } else {
	cdgPrintf(CDG_INFO, "%7.5f\tF-measure = ", (float)totaledgecorrect/(float)totaledgeref);
      }
      if (totaledgeref + totaledgetest == 0) {
        cdgPrintf(CDG_INFO, "%s\n", "n/a");
      } else {
        cdgPrintf(CDG_INFO, "%7.5f\n", 2 * (float)totaledgecorrect/((float)totaledgetest + (float)totaledgeref));
      }
    }
  }
/* 
 * F-measure = 2 a r / (a + r)
 * a = total_correct / total_test
 * r = total_correct / total_ref
 * 
 * ==> F-measure = 2 total_correct / (total_test + total_ref)
 */

  vectorDelete(map);
  vectorDelete(levels_in_grammar);
  vectorDelete(totaledgecorrect_level);
  vectorDelete(totaledgetest_level);
  vectorDelete(totaledgeref_level);
  return errors;
 }
}

/* ----------------------------------------------------------------------
 * FIXME: add comments here
 */
Parse parseClone(Parse p)
{
  Parse p_copy;
  List l;

  p_copy = memMalloc(sizeof (ParseStruct));
  if (p->userName == NULL)
    p_copy->userName = NULL;
  else
    p_copy->userName = strRegister(p->userName);
  if (p->latticeId == NULL)
    p_copy->latticeId = NULL;
  else
    p_copy->latticeId = strRegister(p->latticeId);
  p_copy->dateOfCreation = p->dateOfCreation;
  p_copy->grammarFiles = NULL;
  for (l = p->grammarFiles; l; l = listNext(l)) {
    String s = strRegister((String) listItem(l));

    p_copy->grammarFiles = listAppendItem(p_copy->grammarFiles, s);
  }
  if (p->searchStrategy == NULL)
    p_copy->searchStrategy = NULL;
  else
    p_copy->searchStrategy = strRegister(p->searchStrategy);
  if (p->comment == NULL)
    p_copy->comment = NULL;
  else
    p_copy->comment = strRegister(p->comment);
  p_copy->solutionNo = p->solutionNo;
  p_copy->noOfSolutions = p->noOfSolutions;
  p_copy->score = p->score;
  p_copy->optimal = p->optimal;
  p_copy->violatedConstraints = NULL;
  for (l = p->violatedConstraints; l; l = listNext(l)) {
    String s = strRegister((String) listItem(l));

    p_copy->violatedConstraints =
      listAppendItem(p_copy->violatedConstraints, s);
  }
  if (p->words == NULL)
    p_copy->words = NULL;
  else {
    int s = vectorSize(p->words);
    int i;

    p_copy->words = vectorNew(s);
    for (i = 0; i < s; i++) {
      Word w0 = (Word) vectorElementAt(p->words, i);
      Word w = newWord();

      w->from = w0->from;
      w->to = w0->to;
      w->word = strRegister(w0->word);
      w->description = strRegister(w0->description);
      vectorSetElementAt(p_copy->words, w, i);
    }
  }
  p_copy->levels = NULL;
  for (l = p->levels; l; l = listNext(l)) {
    String s = strRegister((String) listItem(l));

    p_copy->levels = listAppendItem(p_copy->levels, s);
  }
  p_copy->labels = NULL;
  for (l = p->labels; l; l = listNext(l)) {
    String s = strRegister((String) listItem(l));

    p_copy->labels = listAppendItem(p_copy->labels, s);
  }
  if (p->verticesStructure == NULL)
    p_copy->verticesStructure = NULL;
  else 
    p_copy->verticesStructure = vectorClone(p->verticesStructure);
  if (p->verticesLabels == NULL)
    p_copy->verticesLabels = NULL;
  else 
    p_copy->verticesLabels = vectorClone(p->verticesLabels);
  
  p_copy->evalUnary = p->evalUnary;
  p_copy->evalBinary = p->evalBinary;
  p_copy->numberOfHits = p->numberOfHits;
  p_copy->evaluationTime = p->evaluationTime;
  if (p->statisticComment == NULL)
    p_copy->statisticComment = NULL;
  else
    p_copy->statisticComment = strRegister(p->statisticComment);
  p_copy->LVs = NULL;
  p_copy->lg = NULL;
  p_copy->violations = NULL;
  if (!parseIsAbstract(p)) {
    if(parseDecorate(p_copy, p->lg, NULL)) {
      parseEval(p_copy);
    }
  }

  p_copy->map = NULL;
  if(p->map) {
    p_copy->map = vectorClone(p->map);
  }
  p_copy->rmap = NULL;
  if(p->rmap) {
    p_copy->rmap = vectorClone(p->rmap);
  }
  
  p_copy->id = parseMakeId();

  return p_copy;
}



/* ----------------------------------------------------------------------
   Return new Word corresponding to an.
   Try very hard to set w->cat and w->description to fitting values.
   In particular, the description will later be used for selecting the correct
   lexeme nodes if the parse is ever decorated.    
*/
Word anno2word(Annotation an) {

  Word w = newWord();

  w->from = an->from;
  w->to = an->to;
  w->word = an->word;
  w->description = strRegister(w->word);

  return w;
  
}
  
/* -------------------------------------------------------------------
 * parseFromAnno() -- converts an AnnotatedEntry to Parse
 */
Parse parseFromAnno(AnnoEntry ae)
{
  List l, m;
  Parse p = parseNew(ae->id);
  int nrlevels, nrlabels, nrwords, posword, poslevel, index;
#ifdef DEBUGPARSEFROMANNO
  cdgPrintf(CDG_WARNING, "translating annotation `%s'\n", ae->id);
#endif
  
  
  p->searchStrategy = strRegister("anno2parse");
  p->latticeId = strRegister(ae->lattice);
  p->comment = NULL;
  p->solutionNo = 1;
  p->noOfSolutions = 1;
  p->words = vectorNew(10);

  /* add levels and labels */
  for (m = ae->annos; m; m = listNext(m)) {
    Annotation an = (Annotation) listItem(m);

    for (l = an->specs; l; l = listNext(l)) {
      Specification s = (Specification) listItem(l);

      if (s->type == STDep) {
	p->levels = listAddUniqueItem(p->levels, s->kind);
	p->labels = listAddUniqueItem(p->labels, s->name);
      }
    }
  }

  /* add words */
  for (m = ae->annos; m; m = listNext(m)) {
    Word w = anno2word(listItem(m));
    if(!w) {
      parseDelete(p);
      cdgPrintf(CDG_WARNING, "WARNING: can't translate annotation `%s'\n", ae->id);
      return NULL;
    }
    vectorAddElement(p->words, w);
  }

  nrlevels = listLength(p->levels);
  nrlabels = listLength(p->labels);
  nrwords = vectorSize(p->words);

  p->verticesStructure = vectorNew(nrwords * nrlevels);
  vectorSetAllElements(p->verticesStructure, (Pointer)-1);
  p->verticesLabels = vectorNew(nrwords * nrlevels);
  vectorSetAllElements(p->verticesStructure, NULL);

  /* add verticesStructure and verticesLabels */
  for (m = ae->annos, posword = 1; m; m = listNext(m), posword++) {
    Annotation an = (Annotation) listItem(m);

    for (l = an->specs; l; l = listNext(l)) {
      Specification s = (Specification) listItem(l);

      if (s->type != STDep) {
	continue;
      }
      index = parseIndex(p, s->kind, (posword - 1));
      vectorSetElementAt(p->verticesLabels, s->name, index);
      if(s->position > listLength(ae->annos)) {
        cdgPrintf(CDG_ERROR, "ERROR: Annotation specifies illegal governor index %d\n", s->position);
        parseDelete(p);
        return NULL;
      }
      vectorSetElementAt(p->verticesStructure, (void *)(s->position - 1), index);

    }
  }

  /* Add trivial values for unspecified edges */
  poslevel = 0;
  for(l = p->levels; l != NULL; l = listNext(l)) {
    
    /* try to find some label for this level */
    String levelname = listItem(l);
    Level level = NULL;
    String label;
    int i;
    Boolean warned = FALSE;
    for(m = inputCurrentGrammar->levels; m != NULL; m = listNext(m)) {
      Level ld = listItem(m);
      if(!strcmp(levelname,ld->id)) {
        level = ld;
        break;
      }
    }
    label = strRegister(level ? listItem(level->labels) : "");

    /* insert this label (the subordination is already -1) */
    for(i = 0; i < nrwords; i++) {
      index = i * nrlevels + poslevel;
      if(vectorSize(p->verticesLabels) <= index ||
         !vectorElementAt(p->verticesLabels, index)) {
        vectorSetElementAt(p->verticesLabels, label, index);
        if(!warned) {
          cdgPrintf(CDG_WARNING, "WARNING: Annotation `%s' is incomplete.\n", ae->id);
        }
        warned = TRUE;
      }
    }
    poslevel++;
  }
  
  return p;
}


/* ----------------------------------------------------------------------
   Compute the mapping of Annotations from vector a onto Words from vector b
   for parseVerify().
*/
void parseVerifyComputeMap(Vector map, Vector a, Vector b)
{
  int i, j;
  int la = vectorSize(a);
  int lb = vectorSize(b);

  Vector d = vectorNew((la + 1) * (lb + 1));

  /* Levenshtein distance matrix */
  Vector dir = vectorNew((la + 1) * (lb + 1));

  /* directions matrix, used to determine the mapping */

#define pos(x,y) ((x)*(lb+1)+(y))

  vectorSetElementAt(d, (void *)0, pos(0, 0));
  for (i = 1; i <= la; i++)
    vectorSetElementAt(d, (void *)i, pos(i, 0));
  for (j = 1; j <= lb; j++)
    vectorSetElementAt(d, (void *)j, pos(0, j));
  for (i = 1; i <= la; i++)
    for (j = 1; j <= lb; j++) {
      String wa = ((Annotation)vectorElementAt(a, i - 1))->word;
      String wb = ((Word)vectorElementAt(b, j - 1))->word;
      int m1 = (int)vectorElementAt(d, pos(i - 1, j - 1)) +
	(strcasecmp(wa, wb) == 0 ? 0 : 1);
      int m2 = (int)vectorElementAt(d, pos(i, j - 1)) + 1;
      int m3 = (int)vectorElementAt(d, pos(i - 1, j)) + 1;

      if (m1 <= m2 && m1 <= m3) {
	vectorSetElementAt(d, (void *)m1, pos(i, j));
	vectorSetElementAt(dir, (void *)1, pos(i, j));
      }
      if (m2 <= m1 && m2 <= m3) {
	vectorSetElementAt(d, (void *)m2, pos(i, j));
	vectorSetElementAt(dir, (void *)2, pos(i, j));
      }
      if (m3 <= m1 && m3 <= m2) {
	vectorSetElementAt(d, (void *)m3, pos(i, j));
	vectorSetElementAt(dir, (void *)3, pos(i, j));
      }
    }

#if 0
  for (i = 0; i <= la; i++) {
    for (j = 0; j <= lb; j++)
      cdgPrintf(CDG_INFO, "%d ", (int)vectorElementAt(d, pos(i, j)));
    cdgPrintf(CDG_INFO, "\n");
  }
#endif

  for (i = 0; i < la; i++)
    vectorSetElementAt(map, (void *)-1, i);
  i = la;
  j = lb;
  while (i >= 1 && j >= 1) {
    String wa = ((Annotation)vectorElementAt(a, i - 1))->word;
    String wb = ((Word)vectorElementAt(b, j - 1))->word;

    if(i == 1) {
      /* allow the first letter in the sentence to differ in case
         (cf. inputSpuriousUppercase()) */
      if (strcasecmp(wa, wb) == 0 &&
          strcmp((char *)&wa[1],(char *)&wb[1]) == 0) {
        vectorSetElementAt(map, (void *)(j - 1), i - 1);
      }
    } else {
      if (strcmp(wa, wb) == 0) {
        vectorSetElementAt(map, (void *)(j - 1), i - 1);
      }
    }
    
    switch ((int)vectorElementAt(dir, pos(i, j))) {
      case 1:
	i--;
	j--;
	break;
      case 2:
	j--;
	break;
      case 3:
	i--;
	break;
    }
  }
  for (i = la - 1; i >= 0 && (int)vectorElementAt(map, i) == -1;
       vectorSetElementAt(map, (void *)-2, i--)) ;

  /* remove duplicate mappings from the map:
     an LV should not be matched against two annotations,
     rather the missing LV should count as an error. */
  for(i = 1; i < vectorSize(map); i++) {
    if(vectorElementAt(map, i - 1) == vectorElementAt(map, i) &&
       vectorElementAt(map, i) > 0) {
      vectorSetElementAt(map, (Pointer)-1, i - 1);
    }
  }

  vectorDelete(d);
  vectorDelete(dir);
}

/* ----------------------------------------------------------------------
 * construct a new parse verification struct
 */
ParseVerification parseVerificationNew(Parse p)
{
  ParseVerification pv = 
    (ParseVerification)memMalloc(sizeof(ParseVerificationStruct));

  pv->parse = NULL;
  pv->annoEntry = NULL;
  pv->tries =
    hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);
  pv->structHits = 
    hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);
  pv->labelledHits = 
    hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);
  pv->lexicalHits = 
    hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);
  pv->strictHits = 
    hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);

  pv->lexErrors = vectorNew(vectorSize(p->words));
  pv->labelErrors = vectorNew(vectorSize(p->verticesLabels));
  pv->structErrors= vectorNew(vectorSize(p->verticesStructure));

  vectorSetAllElements(pv->lexErrors, (Pointer)0);
  vectorSetAllElements(pv->labelErrors, (Pointer)0);
  vectorSetAllElements(pv->structErrors, (Pointer)0);

  return pv;
}

/* ----------------------------------------------------------------------
 * destruct a parse verification
 */
void parseVerificationDelete(ParseVerification pv)
{
  hashDelete(pv->tries);
  hashDelete(pv->structHits);
  hashDelete(pv->labelledHits);
  hashDelete(pv->lexicalHits);
  hashDelete(pv->strictHits);

  vectorDelete(pv->lexErrors);
  vectorDelete(pv->labelErrors);
  vectorDelete(pv->structErrors);
  
  memFree(pv);
}

/* ----------------------------------------------------------------------
 * Compare Parse to Annotation and print results

 This function is exceedingly complex, although we make some pretty radical
 simplifying assumptions. Consider all of the differences that could exist
 between a Parse and an Annotation:

 - the structure may be different
 - the labels in analogous places may differ
 - the lexical reading of a word may differ
 - the phonetical form of a word may be different
 - indeed, the number of number of words may be different!
   This leads directly to the more general case:
 - the Annotation may contain less information than the Parse
   (labels, levels, words...) or more (levels or words).


 Here's what we make out of this mess:

 (1) In general, the Annotation is authoritative. Everything specified
     there is considered correct. If the Parse says something different,
     that is an error. If the Parse says nothing about something that the
     Annotation specifies, that is also an error. If information from the Parse 
     is not specified in the Annotation, it is ignored.
     
     Thus, the measure computed here is always a measure of Recall, never
     Precision. This is usually a problem, since perfect recall can be
     achieved simply by specifying everything. However, a Parse can only
     specify one thing for each degree of freedom (with one exception,
     considered next).

     Then again, solution methods are perfectly capable of returning
     multiple Parses for a problem, so actually we would want to assess
     several Parses against one Annotation and retrieve a single measure of
     correctness. But this isn't possible with this function.

 (2) There is one exception to the rule that Parses must be unambiguous
     about everything: if all of its LVs are disjunctive in the modifier,
     the lexical reading at that point is ambiguous. This can only
     happen when two homonyms are identical in all features ever touched by
     the grammar, which should not happen (don't model features that you
     can't distinguish). It does happen massively when the syntax level,
     which should touch all features, is switched off.

     Strictly speaking, such Parses are placeholders for several similar
     solutions. But for purposes of parseVerify() we treat them as if they
     were unambiguous: if one of the lexical items matches the conditions
     specified by the annotation, the Parse is correct, otherwise it is an
     error.

 (3) If the sequence of words differs between Annotation and Parse, we
     match them up with the Levenshtein algorithm, but no word in the
     Annotation can correspond to more than one word in the Parse. The words
     that it considers as corresponding to each other are then compared. Words
     left over in the Parse are ignored. Words left over in the Annotation and
     not in the Parse count as errors. This means that if the Parse has two
     words in reverse order (recognizer error?), one of them will be considered
     missing.

 (4) For each word in the annotation, the corresponding LVs are checked
     whether they comply with it. The result is counted both globally and
     per-level. The following levels of compliance exist:

     (4) everything correct
     (3) everything correct except the lexical reading of the modifiee
     (2) label and structure correct
     (1) structure only correct
     
     Any LV that does not meet either of these definitions is an error
     under all measures. Recall that a missing LV is always
     considered an error. 
*/
ParseVerification parseVerify(Parse p, AnnoEntry ae) 
{
  List l, path;
  int i, j, k;
  Vector wordHits;
  int gov1, gov2;
  int ignoredErrors = 0;
  
  /* compute word mapping from annotation to parse */
  Vector wordmap = vectorNew(20);
  List annolevels = NULL, missinglevels = NULL, missingwords = NULL;
  Vector annos = listToVector(ae->annos);
  int nrannowords = vectorSize(annos);
  Array matrix;
  char *measures[] = { "wrong", "structural", "labelled", "lexical", "strict"};
  ParseVerification pv = parseVerificationNew(p);
  Hierarchy h = findHierarchy(inputFeatureHierarchy);
  
  /* initialize the parse verification */
  pv->parse = p;
  pv->annoEntry = ae;

  parseVerifyComputeMap(wordmap, annos, p->words);
  
  /* Pre-compute how well each word was recalled:
      2 - lexically
      1 - orthographically
      0 - not at all   */
  wordHits = vectorNew(nrannowords);
  for(i = 0; i < nrannowords; i++) {
    Annotation a = vectorElementAt(annos, i);
    LexiconItem le = NULL;
    j = (int)vectorElementAt(wordmap, i);
    if(j > -1) {

      Word w = vectorElementAt(p->words, j);
#ifdef DEBUGPARSEVERIFY
      cdgPrintf(CDG_DEBUG, "DEBUG: anno: `%s'  word: `%s'\n", a->word, w->word);
#endif
      /* perfect match -- score 2 */
      vectorSetElementAt(wordHits, (Pointer)2, i);

      /* find underlying lexical entry */
      if(!parseIsAbstract(p)) {
        Boolean found = FALSE;
        for(k = 0; k < vectorSize(p->lg->nodes); k++) {
          le = ((LexemNode)vectorElementAt(p->lg->nodes, k))->lexem;
          if(le->description == w->description) {
            found = TRUE;
            break;
          }
        }

        /* compare features */
        if(found) {
          for(l = a->specs; l != NULL; l = listNext(l)) {
            Specification s = listItem(l);
            if(s->type == STTag) {
              Value annos = le->values[inputRegisterPath(path=scanPath(s->kind), inputCurrentGrammar)];
              String f = sprintValue(annos);

              listFree(path);
              /* missing feature */
              if(annos == ErrorValue) {
                cdgPrintf(CDG_WARNING,
                          "WARNING: word `%s': feature %s = %s not present in parse.\n",
                          w->word, s->kind, s->name);
                /* lexical error -- score 1 */
                vectorSetElementAt(pv->lexErrors, (Pointer)1, j);
                vectorSetElementAt(wordHits, (Pointer)1, i);
                break;
              }

              /* wrong feature */
              if(!inputCompareFeatures(f, s->name, h)) {
                cdgPrintf(CDG_WARNING,
                          "WARNING: word `%s': feature %s = %s should be %s.\n",
                          w->word, s->kind, f, s->name);
                /* lexical error -- score 1 */
                vectorSetElementAt(pv->lexErrors, (Pointer)1, j);
                vectorSetElementAt(wordHits, (Pointer)1, i);
                break;
              }
            }
          }
        }
      }
    } else {
      /* word not present -- score 0 */
      vectorSetElementAt(wordHits, 0, i);
    }
  }

  /* find all levels in annotation */
  for(i = 0; i < nrannowords; i++) {
    Annotation a = vectorElementAt(annos, i);
    for(l = a->specs; l != NULL; l = listNext(l)) {
      Specification s = listItem(l);
      if(s->type != STDep) {
        continue;
      }
      annolevels = listAddUniqueItem(annolevels, s->kind);
    }
  }

#ifdef DEBUGPARSEVERIFY
  cdgPrintf(CDG_DEBUG, "DEBUG: word map: ");
  for(i = 0; i < vectorSize(wordmap); i++) {
    int j = (int)vectorElementAt(wordmap, i);
    cdgPrintf(CDG_DEBUG, "%d ", j);
  }
  cdgPrintf(CDG_DEBUG, "\n");

  
  cdgPrintf(CDG_DEBUG, "DEBUG: word accuracy: ");
  for(i = 0; i < vectorSize(wordHits); i++) {
    int j = (int)vectorElementAt(wordHits, i);
    cdgPrintf(CDG_DEBUG, "%d ", j);
  }
  cdgPrintf(CDG_DEBUG, "\n");
#endif
  
  /* initialize verification matrix */
  matrix = arrayNew(nrannowords, listLength(annolevels), 0);
  for(i = 0; i < nrannowords; i++) {
    for(j = 0; j < listLength(annolevels); j++) {
      arraySetElement(matrix, (Pointer)-1, i, j);
    }
  }
                    
  /* Now compare all edges in the annotation against the parse */
  for(i = 0; i < nrannowords; i++) {
    int posword = (int)vectorElementAt(wordmap, i);
    Annotation a = vectorElementAt(annos, i);
    for(l = a->specs; l != NULL; l = listNext(l)) {
      Specification s = listItem(l);
      if(s->type == STDep) {
        int poslevel = listIndex(p->levels, s->kind) - 1;
        int posannolevel = listIndex(annolevels, s->kind) - 1;
        int parseindex = posword * listLength(p->levels) + poslevel;
        
        /* missing word -- score 0 */
        if(posword < 0) {
          if(!listContains(missingwords, a)) {
            cdgPrintf(CDG_WARNING, "WARNING: word `%s' not present in parse.\n", a->word);
            missingwords = listPrependItem(missingwords, a);
          }
          arraySetElement(matrix, (Pointer)0, i, posannolevel);
          continue;
        }

        /* missing level -- score 0 */
        if(poslevel == -1) {
          if(!listContains(missinglevels, s->kind)) {
	    Level levelDecl = findLevel(s->kind);
	    if (!levelDecl) {
	      cdgPrintf(CDG_WARNING, "WARNING: level `%s' not present in parse.\n", s->kind);
	    }
            missinglevels = listPrependItem(missinglevels, s->kind);
          }
          arraySetElement(matrix, (Pointer)0, i, posannolevel);
          continue;
        }
        
        /* structural error -- score 0 */
	gov1 = (int)vectorElementAt(p->verticesStructure, parseindex);
	gov2 = s->position - 1;
        if(gov1 != gov2) {
	  String s1 = "NIL", s2 = "NIL";

          vectorSetElementAt(pv->structErrors, (Pointer)1, parseindex);
          
          if(gov1>=0 && gov1 <vectorSize(wordmap)) {
	    int index = (int)vectorElementAt(wordmap, gov1);
	    if(index >=0) {
	      s1 = ((Word)vectorElementAt(p->words, index))->word;
	    }
	  }
	  if(gov2>=0) {
	    s2 = ((Annotation)vectorElementAt(annos, gov2))->word;
	  }
	  
	  cdgPrintf(CDG_WARNING,
		    "WARNING: subordination %s: %s(%d)-->%s(%d) should be %s(%d)-->%s(%d)\n",
		    s->kind,
		    a->word,
		    i+1,
		    s1,
		    gov1 + 1,
		    a->word,
		    i+1,
		    s2,
		    gov2 + 1);
	  
	  if (parseVerifyPartially && gov1 == -1) {
	    ignoredErrors++;
	    arraySetElement(matrix, (Pointer)4, i, posannolevel);
	  } else {
	    arraySetElement(matrix, (Pointer)0, i, posannolevel);
	  }
	  continue;
        }

        /* label error -- score 1 */
        if(vectorElementAt(p->verticesLabels, parseindex) != s->name) {
          vectorSetElementAt(pv->labelErrors, (Pointer)1, parseindex);
          cdgPrintf(CDG_WARNING,
                    "WARNING: label %s: %s--`%s'--> should be `%s'\n",
                    s->kind,
                    a->word,
                    vectorElementAt(p->verticesLabels, parseindex),
                    s->name);
          arraySetElement(matrix, (Pointer)1, i, posannolevel);
          continue;
        }

        /* lexical error in the modifier -- score 2 */
        if((int)vectorElementAt(wordHits, i) < 2) {
          arraySetElement(matrix, (Pointer)2, i, posannolevel);
          continue;
        }
        
        /* lexical error only in the modifiee -- score 3 */
        j = s->position - 1;
        if(j > 0 && (int)vectorElementAt(wordHits, j) < 2) {
          arraySetElement(matrix, (Pointer)3, i, posannolevel);
          continue;
        }

        /* no error -- score 4 */
        arraySetElement(matrix, (Pointer)4, i, posannolevel);
      }
    }
  }

#ifdef DEBUGPARSEVERIFY
  cdgPrintf(CDG_DEBUG, "DEBUG: matrix of compliance:\n");
  for(i = 0; i < nrannowords; i++) {
    for(j = 0; j < listLength(annolevels); j++) {
      cdgPrintf(CDG_DEBUG, "%3d ", arrayGetElement(matrix, i, j));
    }
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif

  /* Give matrix of information */
  cdgPrintf(CDG_INFO, "                   ");
  for(i = 1; i < 5; i++) {
    cdgPrintf(CDG_INFO, "%-13s", measures[i]);
  }
  cdgPrintf(CDG_INFO, "\n\n");


  i = 0;
  {
    int totaltries = 0;
    int totalhits[5] = { 0, 0, 0, 0, 0 };
    Hashtable hitsHash[5] = { 0,
      pv->structHits, pv->labelledHits,
      pv->lexicalHits, pv->strictHits };
    int n;
    for(l = annolevels; l != NULL; l = listNext(l)) {
      String level = listItem(l);
      Level levelDcl = findLevel(level);
      float score;
      int tries = 0;
      int hits[5] = { 0, 0, 0, 0, 0 };
      
      if (!levelDcl || !levelDcl->useflag) {
        i++;
	continue;
      }
      
      cdgPrintf(CDG_INFO, "%10s:", level);
      for(j = 0; j < nrannowords; j++) {
        int k = (int)arrayGetElement(matrix, j, i);
        
        /* the annotation may omit specification of some edges */
        if(k == -1) {
          continue;
        }
        tries++;
        totaltries++;
        for(n = 1; n < 5; n++) {
          if(k >= n) {
            hits[n]++;
            totalhits[n]++;
          }
        }
      }
      hashPut(pv->tries, level, (Pointer)(tries+1));
      for(n = 1; n < 5; n++) {
        score = 100.0 * hits[n] / tries; /* can't be 0, or we wouldn't have found it */
        cdgPrintf(CDG_INFO, "        %6.2f", score);
	hashPut(hitsHash[n], level, (Pointer)(hits[n]+1));
      }
      i++;
      cdgPrintf(CDG_INFO, "\n");
    }

    hashPut(pv->tries, strRegister("overall"), (Pointer)(totaltries+1));
    if(totaltries) {
      cdgPrintf(CDG_INFO, "%10s:", "overall");
      for(n = 1; n < 5; n++) {
        float score = 100.0 * totalhits[n] / totaltries;
        cdgPrintf(CDG_INFO, "        %6.2f", score);
	hashPut(hitsHash[n], strRegister("overall"), (Pointer)(totalhits[n]+1));
      }
      cdgPrintf(CDG_INFO, "\n");
    }
    if (ignoredErrors) {
      cdgPrintf(CDG_WARNING, "\nWARNING: ignoring %d error(s) in %d tries\n",
	  ignoredErrors, totaltries);
    }
  }

  arrayDelete(matrix);
  listFree(annolevels);
  listFree(missinglevels);
  listFree(missingwords);
  vectorDelete(annos);
  vectorDelete(wordHits);
  vectorDelete(wordmap);

  return pv;
}

/* ----------------------------------------------------------------------
 * returns the parse from the global parse list, that has the requested id
 */
Parse parseFind(String id)
{
  return hashGet(cdgParses, id);
}

/* ----------------------------------------------------------------------
 * Checks whether any modifiers are underspecified.
 */
Boolean parseIsUnderspecified(Parse p)
{
  int i;

  for (i = 0; i < vectorSize(p->verticesStructure); i++) {
    if ((int)vectorElementAt(p->verticesStructure, i) == -2) {
      return TRUE;
    }
  }
  return FALSE;
}

/* ----------------------------------------------------------------------
   Returns the creation date as a String.  
*/
String parseDate(Parse p)
{
  String result = asctime(localtime(&(p->dateOfCreation)));

  result[strlen(result) - 1] = ' ';    /* instead of '\n' */
  return strRegister(result);
}

/* ----------------------------------------------------------------------
   Return LV of a decorated Parse.
*/
LevelValue parseGetLV(Parse p, int where)
{
  if (parseIsAbstract(p)) {
    return NULL;
  }

  return vectorElementAt(p->LVs, (int)vectorElementAt(p->map, where));
}

/* ----------------------------------------------------------------------
   Return lexeme nodes of a decorated Parse.
*/
List parseGetLexemNodes(Parse p)
{
  List result = NULL;
  int i;
  if (!parseIsAbstract(p)) {
    Vector nodes = p->lg->nodes;
    for (i = 0; i < vectorSize(nodes); i++) {
      LexemNode ln = vectorElementAt(nodes, i);
      result = listPrependItem(result, ln);
    }
  }
  
  return result;
}

/* ----------------------------------------------------------------------
 * add a parse uniquely to the global list
 */
void parseRegister(Parse p)
{
  if(parseFind(p->id)) {
    return;
  }
  hashPut(cdgParses, p->id, p);
  parseMostRecentlyCreatedParse = p;
}


/* ----------------------------------------------------------------------
 * builds a parse from a vector of LVs
 */
Parse parseFromLVs(ConstraintNet net, Vector lvs, String strategy)
{
  Parse p = parseNew(NULL);
  p->latticeId = net->lexemgraph->lattice->id;
  p->searchStrategy = strRegister(strategy);
  p->comment = NULL;
  p->solutionNo = 1;
  p->noOfSolutions = 1;
  p->optimal = FALSE;
  p->evalUnary = net->evalUnary;
  p->evalBinary = net->evalBinary;
  p->numberOfHits = 0;
  p->evaluationTime = 0;
  p->statisticComment = NULL;

  vectorSort(lvs, lvCompareByNaturalOrder);
  parseComputeStructure(p, lvs, net->lexemgraph->isDeletedNode);
  parseDecorate(p, net->lexemgraph, NULL);
  parseEval(p);
  
  return p;
}

/* ----------------------------------------------------------------------
   Returns minimal timepoint covered by the Parse
*/
int parseMinFrom(Parse p) {

  int i, result = 0;
  for(i = 0; i < vectorSize(p->words); i++) {
    result = min(result, ((Word)vectorElementAt(p->words, i))->from);
  }
  
  return result;  
    
}

/* ----------------------------------------------------------------------
   Returns maximal timepoint covered by the Parse
*/
int parseMaxTo(Parse p) {

  int i, result = 0;
  for(i = 0; i < vectorSize(p->words); i++) {
    result = max(result, ((Word)vectorElementAt(p->words, i))->to);
  }
  
  return result;  
    
}

/* ----------------------------------------------------------------------
   Turn Parse into annotation.
*/
AnnoEntry parse2annotation(Parse p) {

  LexemNode ln = NULL;
  LevelValue lv;
  List l,m;
  int i,j;
  Word w;
  AnnoEntry result;
  Annotation a;
  Specification s;
  int nrLevels = listLength(p->levels);
  
  result = (AnnoEntry)memMalloc( sizeof(AnnoEntryStruct) );
  result->id = strRegister(p->latticeId);
  if(parseIsAbstract(p)) {
    result->lattice = p->latticeId;
  } else {
    result->lattice = p->lg->lattice->id;
  }
  result->annos = NULL;
  result->filename = NULL;
  result->lineNo = 0;

  /* iterate over words */
  for(i = 0; i < vectorSize(p->words); i++) {
    List desired_features;
    Boolean found;
    
    w = vectorElementAt(p->words, i);
    
    a = newAnnotation(w->from, w->to, strRegister(w->word));
    result->annos = listAppendItem(result->annos, a);

    if(!parseIsAbstract(p)) {
      ln = parseFindLn(p, w, NULL);
      
      /* include the distinctive features,
         plus all that the user explicitly ordered */
      desired_features = listClone(ln->lexem->distinctive_features);
      for(l = inputAnnoCats; l != NULL; l = listNext(l)) {
        found = FALSE;
        for(m = desired_features; m != NULL; m = listNext(m)) {
          if(comparePath(listItem(l), listItem(m))) {
            found = TRUE;
            break;
          }
        }
        if(!found) {
          desired_features = listPrependItem(desired_features, listItem(l));
        }
      }
      
      for(l = desired_features; l != NULL; l = listNext(l)) {
        Value v;
        List path;
        int k = inputLookupPath((List)listItem(l), inputCurrentGrammar);
        if(k < 0) {
          cdgPrintf(CDG_WARNING, "WARNING: can't lookup path `");
          printPath(CDG_WARNING, listItem(l));
          cdgPrintf(CDG_WARNING, "', ignoring\n");
          continue;
        }
        v = ln->lexem->values[k];
        path = listNth(inputCurrentGrammar->allAttributes, k+1);
        if(v == ErrorValue) {
          continue;
        }
        
        switch(v->type) {
        case VTString:
        case VTNumber:
        case VTList:
          s = newSpecification(STTag,
                               sprintPath(path),
                               sprintValue(v),
                               0);
          inputSpecificationAdd(a, s);
          break;
        default:
          break;
        }
      }
      listFree(desired_features);
    }
    
    /* record structure */    
    if(parseIsAbstract(p)) {
      for(j = 0; j < nrLevels; j++) {
        int index = i * nrLevels + j;
        s =  newSpecification(STDep,
                              listNth(p->levels, j + 1),
                              vectorElementAt(p->verticesLabels, index),
                              (int)vectorElementAt(p->verticesStructure, index) + 1);
        inputSpecificationAdd(a, s);
      }
    } else {
      for(j = 0; j < vectorSize(p->verticesStructure); j++) {
        lv = parseFindLV(p, j);
        if(!listContains(lv->modifiers, ln)) {
          lvDelete(lv);
          continue;
        }
        if(lv->modifiee == NONSPEC) {
          cdgPrintf(CDG_WARNING,
                    "WARNING: no write syntax defined for NONSPEC binding!\n");
        }
        s = newSpecification(STDep,
                             lv->level->id,
                             lv->label,
                             (int)vectorElementAt(p->verticesStructure, j) + 1);
        inputSpecificationAdd(a, s);
        lvDelete(lv);
      }
    }
  }
  

  return result;
  
}

/* ----------------------------------------------------------------------
   Calculate the label that would yield the best score at position WHERE
   if nothing else may be changed.

   If it isn't possible to improve the current score, the current label is
   returned even if other labels would yield the same score.
*/
String parseOptimizeLabel(Parse p, String levelname, int wordno) {
  Parse p2;
  List l;
  String result = NULL;
  Badness best;
  int where = parseIndex(p, levelname, wordno);
  Level level = findLevel(levelname);

  String current = vectorElementAt(p->verticesLabels, where);
  
  if(!level) {
    return NULL;
  }

  if(parseIsAbstract(p)) {
    return NULL;
  }
  
  p2 = parseClone(p);
  best = worstBadness();
  for(l = level->labels; l != NULL; l = listNext(l)) {
    String label = listItem(l);
    parseSwapLabel(p2, where, label);
    if(bCompare(p2->badness, best)) {
      result = label;
      bCopy(best, p2->badness);
    }
  }

  if(!bCompare(best, p->badness)) {
    result = current;
  }

  bDelete(best);
  parseDelete(p2);
  
  return result;
  
}


/* ----------------------------------------------------------------------
   Calculate the lexical variant that would yield the best score at position
   WHERE if nothing else may be changed.

   If it isn't possible to improve the current score, the current word is
   returned even if other words would yield the same score.
*/
String parseOptimizeWord(Parse p, int wordno) {
  Parse p2;
  LexemNode result = NULL;
  LexemNode current;
  Badness best;
  int i;
  Word w = vectorElementAt(p->words, wordno);

  if(parseIsAbstract(p)) {
    return NULL;
  }

  current = parseFindLn(p, w, NULL);
  
  p2 = parseClone(p);
  best = worstBadness();
  for(i = 0; i < vectorSize(p->lg->nodes); i++) {
    LexemNode ln = vectorElementAt(p->lg->nodes, i);
    if(ln->arc->from != wordno) {
      continue;
    }
    parseSwapWord(p2, wordno, ln->lexem->description);
    if(bCompare(p2->badness, best)) {
      result = ln;
      bCopy(best, p2->badness);
    }
  }

  if(!bCompare(best, p->badness)) {
    result = current;
  }

  bDelete(best);
  parseDelete(p2);

  return result ? result->lexem->description : NULL;
  
}

/* ----------------------------------------------------------------------
 * Initialize parse module
 */
void parseInitialize(void) {
  setRegister("verifypartially", SET_BOOL, &parseVerifyPartially, NULL, NULL,
	      NULL, NULL);
}

/* ----------------------------------------------------------------------
 * Finalize the parse module
 */
void parseFinalize(void) {
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
