/* Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without
 * modifications, as long as this notice is preserved.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Jochen Hagenstroem (see also AUTHORS and THANKS for more)
 * Purpose: interface CDG to an external POS tagger
 * Birth: 2001/02/07
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <unistd.h>
#include <stdio.h>
#include <strings.h>
#include <string.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <signal.h>
#include <blah.h>
#include "hook.h"
#include "set.h"
#include "lexemgraph.h"
#include "tagger.h"

/* -- VARIABLES ------------------------------------------------------------ */

/* exported globals */
int taggerCategoryIndex=-1;



/* only internally visible */
static String taggerCommand=NULL;
static String taggerCategoryPath=NULL;
static Boolean taggerIgnoreImpossible = TRUE;
static Boolean taggerNormalizeToOne = TRUE;
static Boolean is_up = TRUE;


/* -- TYPE DEFINITIONS ----------------------------------------------------- */
typedef struct {
   Number score;
   String tag;
} TagScoreStruct;
typedef TagScoreStruct *TagScore;


/* -- FUNCTIONS ------------------------------------------------------------ */

/* -------------------------------------------------------------------------
   Set taggerCategoryIndex.

   Looks up the feature index of the feature specified as
   `taggerCategoryPath' and stores it into taggerCategoryIndex.
*/
Boolean taggerLookupCategoryIndex()
{
  List path;

  taggerCategoryIndex=-1;
  if (!taggerCategoryPath || !*taggerCategoryPath)
    { cdgPrintf(CDG_WARNING, "WARNING: no category path\n"); return FALSE; }
  path=scanPath(taggerCategoryPath);
  taggerCategoryIndex=inputLookupPath(path, inputCurrentGrammar);
  inputFreePath(path);

  return taggerCategoryIndex>=0;
}

/* -------------------------------------------------------------------------
   Decorate a lexeme graph with part-of-speech scores.
 
   Calls the specified external tagger and assigns the returned tagscores
   to the field ->tagscores in each lexemnode of LG.
 */
int taggerTag(LexemGraph lg) {

  int gi; 
  int gmax;
  List l, m;
  int result = 0;
  String inputname = strPrintf("%s", tmpnam(0));
  String outputname = strPrintf("%s", tmpnam(0));
  String call = strPrintf("%s < %s > %s",
                          taggerCommand, inputname, outputname);
  FILE *input;
  FILE *output;
  char buffer [1024];
  
  if(!is_up) {
    return -1;
  }

  /* check if wordgraph is linear, only then tagging can be done */
  if (latticeBranches(lg->lattice)) {
    cdgPrintf(CDG_ERROR, "ERROR: can't tag branching lattices\n");
    return -1;
  }

  /* check for valid category index */
  if (taggerCategoryIndex < 0 && !taggerLookupCategoryIndex()) {
    cdgPrintf(CDG_ERROR, "ERROR: can't lookup category index\n");
    return -1;
  }
  
  /* iterate through words of the lattice 
     and write them to temporary file */
  if(NULL == (input = fopen(inputname, "w"))) {
    cdgPrintf(CDG_ERROR, "ERROR: Can't write tagger input file!\n");
    return -1;
  }
  for (l = lg->lattice->arcs; l; l = listNext(l)) {
    Arc a = (Arc) listItem(l);
    char b[1024]; 
    char *p;
    
    /* Words with embedded spaces confuse most taggers.
       Temporarily replace ' ' with an unprintable control character. */
    strncpy(b,a->word,1024);
    while(NULL != (p = strchr(b,' '))) {
      *p = 1;
    }
    
    /* pass the current word to the tagger */
    fprintf(input, "%s\n", b);
  }
  fclose(input);
  
  if(system(call)) {
    cdgPrintf(CDG_ERROR, "ERROR: Can't call tagger!\n");
    unlink(inputname);
    unlink(outputname);
    return -1;
  }
  
  /* set all tag scores to zero */
  for (gi = 0; gi < vectorSize(lg->nodes); gi++) {
    LexemNode ln = (LexemNode) vectorElementAt(lg->nodes, gi);
    ln->tagscore = 0.0;
  }

  gmax = vectorSize(lg->graphemnodes);
  gi = 0;
  if(NULL == (output = fopen(outputname, "r"))) {
    cdgPrintf(CDG_ERROR, "ERROR: Can't read tagger output file!\n");
    unlink(inputname);
    unlink(outputname);
    return -1;
  }

  /* go through tagger response and set tag score accordingly */
  while(fgets(buffer, 1024, output)) {

    GraphemNode gn = (GraphemNode) vectorElementAt(lg->graphemnodes, gi);
    String w = strtok(buffer, " \t\n"); 
    String t; 
    List lexCats, tagCats = NULL;
    Number total = 0.0;
    Number best = 0.0;
    char b[1024];
    char *p;

    if (!gn) {
      cdgPrintf(CDG_ERROR, "ERROR: invalid graphem node\n");
      result = -1;
      break;
    }
    if (!w) {
      cdgPrintf(CDG_ERROR, "ERROR: can't find word\n");
      result = -1;
      break;
    }

    /* Undo space escaping that we might have done */
    strncpy(b,w,1024);
    while(NULL != (p = strchr(b,1))) {
      *p = ' ';
    }

    if (strcmp(b, gn->arc->word)) {
      cdgPrintf(CDG_ERROR, "ERROR: word mismatch %s %s\n", b, gn->arc->word);
      result = -1;
      break;
    }

    lexCats = lgQueryCat(lg, gn->arc);

    /* loop through all tokens in a response line */
    for (t = strRegister(strtok(NULL, " \t\n"));
         t != strRegister("");
         t = strRegister(strtok(NULL, " \t\n"))) {
      String sc = strtok(NULL, " \t\n");
      Number score = 1.0;
      TagScore ts;
      
      /* read one tagscore checking for well-formedness of tagger output */
      if (sc && 1 != sscanf(sc, "%lf", &score)) {
	cdgPrintf(CDG_WARNING, "WARNING: can't read score from %s\n", sc);
        continue;
      }
     
      cdgPrintf(CDG_DEBUG, "DEBUG: %s  %s  %4.3e\n", b, t, score);
      if(!listContains(lexCats, t)) {
        if(taggerIgnoreImpossible) {
          cdgPrintf(CDG_DEBUG, "WARNING: category `%s' not in lexicon\n", t);
          continue;
        }
      }
      best = max(best, score);
      total += score;
      ts = memMalloc(sizeof(TagScoreStruct));
      ts->score = score;
      ts->tag = t;
      tagCats = listPrependItem(tagCats, ts);
    }
    listFree(lexCats);

    /* Normalize away the probabilities spent on impossible categories */
    if(taggerIgnoreImpossible) {
      for(l = tagCats; l != NULL; l = listNext(l)) {
        TagScore ts = listItem(l);
        ts->score /= total;
      }
      best /= total;
    }

    /* Convert probabilities to scores */
    if(taggerNormalizeToOne) {
      for(l = tagCats; l != NULL; l = listNext(l)) {
        TagScore ts = listItem(l);
        ts->score /= best;
      }
    }

    /* Now apply the scores, if there are any. */
    if(tagCats) {
      for(l = tagCats; l != NULL; l = listNext(l)) {
        TagScore ts = listItem(l);
        
        /* loop through all lexemnodes of a grapheme
           and write the tagscore to it */
        for (m = gn->lexemes; m; m = listNext(m)) {
          LexemNode ln = (LexemNode) listItem(m);
          Value v = ln->lexem->values[taggerCategoryIndex];
          
          if (v->type != VTString) {
            cdgPrintf(CDG_ERROR,
                      "ERROR: not a VTString at the tagger category index\n");
            continue;
          }
          
          /* only tag matching categories */
          if (!strcmp(v->data.string, ts->tag)) {
            ln->tagscore = ts->score;
          }
        }
	memFree(ts);
      }
      listFree(tagCats);
    } else {
      /* Otherwise just reset them to 1. */
      cdgPrintf(CDG_WARNING,
                "WARNING: none of the suggested tags "
                "for `%s' is in your lexicon.\n",
                b);
      for (m = gn->lexemes; m; m = listNext(m)) {
        LexemNode ln = (LexemNode) listItem(m);
        ln->tagscore = 1.0;
      }
    }
    gi++;
  }

  unlink(inputname);
  unlink(outputname);
  
  return result;
}

/* -------------------------------------------------------------------------
 * taggerInitialize
 */
void taggerInitialize()
{
  taggerCategoryPath=strRegister("cat");
  taggerLookupCategoryIndex();

  is_up = FALSE;
  
  taggerCommand = strRegister("");
  setRegister("taggerCommand", SET_STRING, &taggerCommand,
	      NULL, NULL, NULL,
	      NULL);
  /* callback function on set to get pathindex  */
  setRegister("taggerCategoryPath", SET_STRING, &taggerCategoryPath,
	      NULL, NULL, (VoidFunction *)taggerLookupCategoryIndex,
	      NULL);

  setRegister("taggerNormalizeToOne", SET_BOOL, &taggerNormalizeToOne, NULL, NULL, NULL, NULL);
  setRegister("taggerIgnoreImpossible", SET_BOOL, &taggerIgnoreImpossible, NULL, NULL, NULL, NULL);

}

/* ----------------------------------------------------------------------
   Return tagger status.

   If the tagger is in use, i.e. `newnet' commands will cause tagging to be
   done, TRUE is returned, otherwise FALSE.
*/
Boolean taggerUp(void) {
  return is_up;
}

/* ----------------------------------------------------------------------
   Toggle tagger status.
*/
int taggerStartStop(void) {
  is_up = !is_up;
  cdgPrintf(CDG_INFO, "INFO: tagger st%sed\n", is_up ? "art" : "opp");
  return 0;
}


/* ------------------------------------------------------------------- */
/* ------------------------------------------------------------------- */

