/* 
 * Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: transform.c,v 1.22 2003/12/19 14:16:21 foth Exp $ 
 */
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include "cdg.h"
#include "hook.h"
#include "timer.h"
#include "transform.h"

/* How many items should the local search agenda hold? */
int localSize = 3000;

/* How many of the Patches generated in a step should actually be tried out? */
int beamWidth = INT_MAX;

/* counter to generate unique problem ids */
int problem_counter = 0;

/* ----------------------------------------------------------------------
   Mark an Analysis as covering only part of an utterance.
*/
void aRestrict(Analysis a, int from, int to) {
  a->subStart = max(from, a->P->minTimepoint);
  a->subStop = min(to, a->P->maxTimepoint);

  /* This has almost certainly changed the score of the structure. */
  aJudge(a);
}

/* ----------------------------------------------------------------------
   Mark an Analysis as covering the entire utterance.
*/
void aUnrestrict(Analysis a) {
  a->subStart = a->P->minTimepoint;
  a->subStop = a->P->maxTimepoint;
  aJudge(a);
}

/* ---------------------------------------------------------------------
   Print the currently selected lexeme nodes.
*/
void printSelectedPath(unsigned long mode, Analysis a) {

  int i;

  for(i = 0; i < vectorSize(a->P->lg->nodes); i++) {
    if(bvElementAt(a->isUnselectedLexeme, i)) {
      continue;
    }
    lgPrintNode(mode, vectorElementAt(a->P->lg->nodes, i));
    cdgPrintf(mode, "\n");
  }
  cdgPrintf(mode, "\n");
}

/* ----------------------------------------------------------------------
   Return the lexeme node from the list that is selected in a, or NULL.
 */
LexemNode selectedLexeme(Analysis a, List lexemes) {

  LexemNode result = NULL;
  List l;

  if(!a) {
    return NULL;
  }

  for (l = lexemes; l; l = listNext(l)) {
    result = (LexemNode) listItem(l);
    if (!bvElementAt(a->isUnselectedLexeme, result->no)) {
      return result;
    }
  }

  return NULL;

}

/* ----------------------------------------------------------------------
   Return the lexeme node that is selected in a at time point WHERE.

   The return value is never NULL. If an existing time point is unbound in
   am Analysis, that is an internal error.
*/
LexemNode selectedAt(Analysis a, int where) {

  int i;
  LexemNode ln;

  /* Column #n of the vector lexemes contains a full listing
     of all lexemes that span timepoint #n.
     Therefore we need only look at that row. */
  for (i = 0; i < a->P->lgWidth; i++) {
    ln = vectorElementAt(a->P->lexemes, a->P->lgWidth * where + i);
    if (ln && !bvElementAt(a->isUnselectedLexeme, ln->no)) {
      return ln;
    }
  }

  /* This can't happen */
  printSelectedPath(CDG_ERROR, a);
  cdgPrintf(CDG_ERROR, "PANIC: unbound time interval %d\n", where);
  abort();

  return NULL;

}


/* Functions dealing with LevelValues
   ================================== 

  This module takes great care not to clone LVs gratuitously, since great
  numbers of LVs are modified during the repair process, and much more
  during lookahead to see which repair step should be taken.

  Instead it uses the precomputed LVs in the ConstraintNet wherever
  possible. This means that wherever an LV with indexWRTNet x
  appears in this module, it is not only equal but _identical_ to the
  corresponding LV in the constraint net. 

  Conversely, any LV that carries the indexWRTNet -1 is guaranteed to
  have the score 0.0 (otherwise it would have been found in the
  ConstraintNet). Such impossible LVs can only be introduced in two
  ways: either as a side effect of selecting another path through the
  lexeme graph, or by manipulating individual features of an LV during
  manual frobbing. In either case a totally new LV is cloned from the
  previously selected LV and changed as requested. The LVs contained in
  the ConstraintNodes of the net passed as a parameter to frobbing()
  are never changed at all, except to lower their limits. */


/* ----------------------------------------------------------------------
   Translates the level use flags into our own array for fast access.
*/
int gatherLevels(Problem P) {

  Level level;
  List l;
  
  P->nrLevels = inputCurrentGrammar->noOfLevels;
  P->levelActive = bvNew(P->nrLevels);

  P->nrActiveLevels = 0;
  for(l = inputCurrentGrammar->levels; l != NULL; l = listNext(l)) {
    level = listItem(l);
    if(level->useflag) {
      P->nrActiveLevels++;
      bvSetElementAt(P->levelActive, TRUE, level->no);
    } else {
      bvSetElementAt(P->levelActive, FALSE, level->no);
    }
  }
  return P->nrActiveLevels;
}

/* ----------------------------------------------------------------------
   Should we bind slot x? Returns FALSE if its level is inactive,
   or the timepoint is not used by the lexeme graph, or
   if we are currently solving only a subproblem.
*/
Boolean slotActive(Analysis a, int x) {
  if(x / a->P->nrLevels < a->subStart) {
    return FALSE;
  }
  if(x / a->P->nrLevels >= a->subStop) {
    return FALSE;
  }
  if(!bvElementAt(a->P->levelActive, x % a->P->nrLevels)) {
    return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
   Do lva and lvb cause the binary conflict cv?

   This function checks whether the two LVs violate the constraint recorded
   in CV. This checks both instantiation variants, so the order of the LVs
   is irrelevant. If the constraint has a variable penalty, TRUE is
   returned only if lva and lvb make the constraint fail with the same
   penalty as recorded in the conflict.

*/
Boolean lvsCauseConflict(Analysis a, LevelValue lva, LevelValue lvb,
                         ConstraintViolation cv) {

  if (matchBinarySignature(lva, lvb, cv->constraint->sig) &&
      !evalBinaryConstraint(cv->constraint, NULL, a->slots, lva, lvb) &&
      cv->constraint->penalty == cv->penalty) {
    return TRUE;
  }

  else if (matchBinarySignature(lvb, lva, cv->constraint->sig) &&
           !evalBinaryConstraint(cv->constraint, NULL, a->slots, lvb, lva) &&
           cv->constraint->penalty == cv->penalty) {
    return TRUE;
  }

  return FALSE;

}

/* ----------------------------------------------------------------------
   Is this LV valid in a?
 */
Boolean isValid(Analysis a, LevelValue lv) {

  if (spec(lv->modifiee) &&
      !listContains(lv->modifiees, selectedAt(a, lv->modifiee->arc->from))) {
    return FALSE;
  }

  if (!listContains(lv->modifiers, selectedAt(a, lv->modifier->arc->from))) {
    return FALSE;
  }

  return TRUE;

}

/* ----------------------------------------------------------------------
   Insert SOURCE int A.

   Does not respect the path criterion; if isValid(SOURCE, A) does not
   hold, A may become path inconsistent afterwards. You must use
   pLexicalize() before this function.
*/
void selectLV(Analysis a, LevelValue source) {

  int where = lvIndex(source);
  LevelValue target = vectorElementAt(a->slots, where);

  /* add to history */
  if (!listContains(a->P->history, (Pointer) where)) {
    a->P->history = listAppendItem(a->P->history, (Pointer) where);
  }

  /* self-replacement? */
  if (source == target) {
    return;
  }

  /* delete temporaries */
  if (target && target->indexWRTNet == -1) {
    lvDelete(target);
  }

  /* insert */
  vectorSetElementAt(a->slots, source, where);

}

/* ----------------------------------------------------------------------
   Return TRUE if LV would bind slot WHICH if it were selected.
 */
Boolean lvBindsSlot(Problem P, LevelValue lv, int where) {

  /* must bind same level */
  if (lv->level->no != where % P->nrLevels) {
    return FALSE;
  }

  /* must cover timespan */
  if (lv->modifier->arc->from > where / P->nrLevels) {
    return FALSE;
  }

  if (lv->modifier->arc->to <= where / P->nrLevels) {
    return FALSE;
  }

  return TRUE;

}

/* ----------------------------------------------------------------------
   Return the LV that binds slot WHERE in A.

   Note that the slot WHERE itself may be empty; in that case
   an LV in an earlier slot must exist that binds WHERE. 
 */
LevelValue slotFiller(Analysis a, int where) {

  LevelValue result = NULL;
  int i = where;

  while (i >= 0 && !result) {
    result = vectorElementAt(a->slots, i);
    i -= a->P->nrLevels;
  }

  /* if the Analysis is incomplete (under construction), then the
     latest previous LV may not actually bind that slot. In that case
     no LV binds it. */
  if(result && !lvBindsSlot(a->P, result, where)) {
    result = NULL;
  }

  return result;

}

/* ----------------------------------------------------------------------
   Propagate a limit to an LV.

   This can only happen for two reasons: either LV binds a set of lexeme
   nodes (as modifiers or as modifiees) that all have a common limit not
   higher than F, or there is, for a certain set of domains, no combined
   assignment including LV with a limit greater than F.

   Accordingly, the only calls to downgrade() are during the routines
   updateLimit() and dSearch().
*/
void downgrade(Problem P, LevelValue lv, Number f) {

  /* downgrade */
  lv->limit = f;

  /* downgrading an LV past the global record means deleting */
  if (significantlyGreater(P->globalMinimum, f)
      && !lv->isDeleted) {
    lv->isDeleted = TRUE;
    P->nrUndeletedLVs -= (lv->modifier->arc->to - lv->modifier->arc->from);
  }
}


/* ----------------------------------------------------------------------
   Recompute the limit of an LV based on the limits of the lexemes it binds.
*/
void updateLimit(Problem P, LevelValue lv) {

  Number f;
  List l;
  LexemNode lnb;

  /* deleted? */
  if(lv->isDeleted) {
    downgrade(P, lv, 0.0);
  }
  
  /* modifiers limited? */
  f = 0.0;
  for (l = lv->modifiers; l; l = listNext(l)) {
    lnb = listItem(l);
    if(lnb->limit > f) {
      f = lnb->limit;
    }
  }
  if(f < lv->limit) {
    downgrade(P, lv, f);
  }

  /* modifiees limited? */
  if (lv->modifiee != NULL && lv->modifiee != NONSPEC) { 

    f = 0.0;
    for (l = lv->modifiees; l; l = listNext(l)) {
      lnb = listItem(l);
      if (lnb->limit > f) {
        f = lnb->limit;
      }
    }
    if (f < lv->limit)
      downgrade(P, lv, f);
  }
}


/* ----------------------------------------------------------------------
   Return the limit of an LV.

   This function first updates the field ->limit of its argument, then
   returns that. This is the first of the two possibilities for a limit to
   be reduced. The decision to test the limits of an LV's lexemes in this
   routine is somewhat arbitrary but it guarantees that lexically motivated
   limits are propagated rather often.
*/
Number getLimit(Problem P, LevelValue lv) {
  updateLimit(P, lv);
  return lv->limit;

}
/* ----------------------------------------------------------------------
   Prints level value, does NOT ignore unselected lexeme nodes.

   In contrast to lvPrint(), this function even displays unselected lexeme
   nodes, enclosing them in square brackets. However, if the total space
   taken by such a list would probably exceed WIDTH, the ambiguity is
   displayed within braces instead.
 */
void displayLV(Problem P, LevelValue lv, int width) {

  List l;
  LexemNode ln;
  Boolean first;
  Number limit;
  int space;

  limit = getLimit(P,lv);

  if (lv->isDeleted) {
    cdgPrintf(CDG_INFO, "[ ");
  }

  if (lv->indexWRTNet == -1) {
    cdgPrintf(CDG_INFO, "  <invalid> ");
  } else {
    cdgPrintf(CDG_INFO, "#%d  ", lv->indexWRTNet);
  }

  if (lv->modifiers) {

    /* print all modifiers, unless that would take up huge amounts of space */
    space = listLength(lv->modifiers) * strlen(lv->modifier->lexem->description);
    if (space > width) {
      ln = selectedLexeme(P->current,lv->modifiers);
      if (!ln) {
        ln = lv->modifier;
      }
      cdgPrintf(CDG_INFO, "%s%s{%d}%s",
                lgIsDeletedNode(P->lg, ln) ? "[" : "",
                ln->lexem->description,
                listLength(lv->modifiers),
                lgIsDeletedNode(P->lg, ln) ? "]" : "");
    } else {

      first = TRUE;
      for (l = lv->modifiers; l; l = listNext(l)) {
        ln = (LexemNode) (listItem(l));

        if (!first)
          cdgPrintf(CDG_INFO, "/");
        else
          first = FALSE;

        if (lgIsDeletedNode(P->lg, ln))
          cdgPrintf(CDG_INFO, "[");

        cdgPrintf(CDG_INFO, "%s", ln->lexem->description);

        if (lgIsDeletedNode(P->lg, ln))
          cdgPrintf(CDG_INFO, "]");
      }
    }

  } else {
    cdgPrintf(CDG_INFO, "%s", lv->modifier->lexem->description);
  }

  cdgPrintf(CDG_INFO, "(%d-%d)/%s--%s-->",
            lv->modifier->arc->from,
            lv->modifier->arc->to, lv->level->id, lv->label);
  if (!lv->modifiee) {
    cdgPrintf(CDG_INFO, "NIL");
  } else if(lv->modifiee == NONSPEC) {
    cdgPrintf(CDG_INFO, "NONSPEC");
  } else {

    /* print all modifiees, unless that would take up huge amounts of space */
    space = listLength(lv->modifiees) * strlen(lv->modifiee->lexem->description);
    if (space > width) {
      ln = selectedLexeme(P->current,lv->modifiees);
      if (!ln) {
        ln = lv->modifiee;
      }
      cdgPrintf(CDG_INFO, "%s%s{%d}%s",
                lgIsDeletedNode(P->lg, ln) ? "[" : "",
                ln->lexem->description,
                listLength(lv->modifiees),
                lgIsDeletedNode(P->lg, ln) ? "]" : "");
    } else {

      first = TRUE;
      for (l = lv->modifiees; l; l = listNext(l)) {
        ln = (LexemNode) (listItem(l));

        if (!first)
          cdgPrintf(CDG_INFO, "/");
        else
          first = FALSE;

        if (lgIsDeletedNode(P->lg, ln))
          cdgPrintf(CDG_INFO, "[");

        cdgPrintf(CDG_INFO, "%s", ln->lexem->description);

        if (lgIsDeletedNode(P->lg, ln))
          cdgPrintf(CDG_INFO, "]");
      }
    }

    cdgPrintf(CDG_INFO, "(%d-%d)",
              lv->modifiee->arc->from, lv->modifiee->arc->to);
  }

  cdgPrintf(CDG_INFO, " [%4.3e]", lv->score);

  if (limit < lv->score && !lv->isDeleted) {
    cdgPrintf(CDG_INFO, " limit %4.3e", limit);
  }

  if (lv->isDeleted) {
    cdgPrintf(CDG_INFO, " ]");
  }

  cdgPrintf(CDG_INFO, "\n");

}

/* -----------------------------------------------------------------------
   Given LV, find an LV in a->P that has the same properties.

   Searches through the corresponding domains of a->P for an LV that

    - has the same label as LV
    - has the same structure as LV
    - is currently lexically valid.

   If one is found, it is returned, otherwise NULL is returned.

   This function is used to avoid unnecessary LV cloning. It judges LVs by
   comparing their lexeme nodes for identity, so it will only work
   correctly if P and lv are ultimately derived from the same lexeme graph. 
   It will fail if P derives from one constraint net and lv is taken from
   another, or from a Parse structure -- use translateLV() for that.
   
 */
LevelValue findExistingLV(Analysis a, LevelValue lv) {

  List l;
  LevelValue model;

  /* if the LV is physically inconsistent,
     don't even bother searching the domain */
  if (lv->modifiee != NULL && lv->modifiee != NONSPEC &&
      !lgCompatibleNodes(a->P->lg, lv->modifier, lv->modifiee)) {
#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG: findExistingLV() fails immediately\n");
#endif
    return NULL;
  }

  /* otherwise, inspect each LV in the domain closely */  
  for (l = a->P->domains[lvIndex(lv)]; l; l = listNext(l)) {
    model = (LevelValue) listItem(l);
#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG:  trying ");
    displayLV(a->P, model, 0);
#endif
    if (strcmp(model->label, lv->label)) {
#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG: wrong label\n");
#endif
      continue;
    }
    if (model->modifier->arc->from != lv->modifier->arc->from ||
        model->modifier->arc->to != lv->modifier->arc->to) {
#ifdef DEBUGFINDMODEL
      cdgPrintf(CDG_DEBUG, "DEBUG: wrong modifier\n");
#endif
      continue;
    }
      if (lv->modifiee == NONSPEC && model->modifiee != NONSPEC) {
#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG: not NONSPEC\n");
#endif
      continue;
      }
      if (lv->modifiee == NULL && model->modifiee != NULL) {
#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG: not NIL\n");
#endif
      continue;
      }
    if (lv->modifiee && lv->modifiee != NONSPEC &&
        (model->modifiee == NONSPEC || !model->modifiee)) {
#ifdef DEBUGFINDMODEL
      cdgPrintf(CDG_DEBUG, "DEBUG: not specified\n");
#endif
      continue;
      }
    if(lv->modifiee && model->modifiee &&
       (model->modifiee->arc->from != lv->modifiee->arc->from ||
       model->modifiee->arc->to != lv->modifiee->arc->to)) {
#ifdef DEBUGFINDMODEL
      cdgPrintf(CDG_DEBUG, "DEBUG: wrong modifiee\n");
#endif
      continue;
    }
    if (!isValid(a, model)) {
#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG: not lexically valid\n");
#endif
      continue;
     }

#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG: returning #%d\n", model->indexWRTNet);
#endif
    return model;
  }

#ifdef DEBUGFINDMODEL
    cdgPrintf(CDG_DEBUG, "DEBUG: returning NULL\n");
#endif
  return NULL;

}


/* ----------------------------------------------------------------------
   Find the lexeme node in a P's lexeme graph that corresponds to LN.
*/
LexemNode translateLn(Problem P, LexemNode ln) {
  int i;

  if(!ln) {
    return NULL;
  }
  
  for(i = 0; i < vectorSize(P->lg->nodes); i++) {
    LexemNode lnNew = vectorElementAt(P->lg->nodes, i);
    if(lnNew->arc->from != ln->arc->from ||
       lnNew->arc->to != ln->arc->to ||
       lnNew->lexem->description != ln->lexem->description) {
      continue;
    }

    return lnNew;
  }

  return NULL;

}

/* ----------------------------------------------------------------------
   Like findExistingLV(), but for `foreign' LVs, that is, LVs that
   do not belong to the same net and therefore have their own lexical
   entries. This means that it does not suffice to call listContains() &
   friends, which makes this function slower than findExistingLV. Indeed,
   this function may have to construct a totally new LV with index -1.
 */
LevelValue translateLV(Problem P, LevelValue lv) {

  List l, m;
  LevelValue model;
  LexiconItem le;
  int from, to;
  Boolean found;
#ifdef DEBUGTRANSLATELV
  cdgPrintf(CDG_INFO, "\nDEBUG: translating: ");
  displayLV(P, lv, 0);
#endif
  
  /* otherwise, inspect each LV in the domain closely */  
  for (l = P->domains[lvIndex(lv)]; l; l = listNext(l)) {
    model = (LevelValue) listItem(l);

#ifdef DEBUGTRANSLATELV
  cdgPrintf(CDG_INFO, "  comparing to: ");
  displayLV(P, model, 0);
#endif
  if (strcmp(model->label, lv->label)) {
#ifdef DEBUGTRANSLATELV
  cdgPrintf(CDG_INFO, "   ...wrong label.\n");
#endif
    continue;
  }
      
    le = lv->modifier->lexem;
    from = lv->modifier->arc->from;
    to = lv->modifier->arc->to;
    found = FALSE;
    for(m = model->modifiers; m != NULL; m = listNext(m)) {
      LexemNode ln = listItem(m);
      if(ln->lexem->description == le->description &&
         ln->arc->from == from &&
         ln->arc->to == to) {
        found = TRUE;
        break;
      }
    }
    if(!found) {
#ifdef DEBUGTRANSLATELV
  cdgPrintf(CDG_INFO, "   ...wrong modifier.\n");
#endif
      continue;
    }

    if (lv->modifiee != NONSPEC && model->modifiee == NONSPEC)
      continue; /* FIXME: Is this correct, Kilian */
    if (lv->modifiee == NONSPEC && model->modifiee != NONSPEC)
      continue;
    if (lv->modifiee == NULL && model->modifiee != NULL) {
#ifdef DEBUGTRANSLATELV
  cdgPrintf(CDG_INFO, "   ...wrong modifiee.\n");
#endif
      continue;
    }

    if (lv->modifiee != NULL && lv->modifiee != NONSPEC) {
      le = lv->modifiee->lexem;
      from = lv->modifiee->arc->from;
      to = lv->modifiee->arc->to;
      found = FALSE;
      for(m = model->modifiees; m != NULL; m = listNext(m)) {
        LexemNode ln = listItem(m);
        if(ln->lexem->description == le->description &&
           ln->arc->from == from &&
           ln->arc->to == to) {
          found = TRUE;
          break;
        }
      }
      if(!found) {
#ifdef DEBUGTRANSLATELV
        cdgPrintf(CDG_INFO, "   ...wrong modifiee.\n");
#endif
        continue;
      }
    }
    if(found) {
#ifdef DEBUGTRANSLATELV
        cdgPrintf(CDG_INFO, "   ...OK.\n");
#endif
      return model;
    }
  }

  /* The LV specified by the annotation is invalid, i.e. it violates a hard
     unary constraint. This can happen if the grammar is bad or has changed
     since the analysis was made. Ever helpful, we build a new LV in this
     case. */
  {
    LexemNode ln1 = translateLn(P, lv->modifier);
    LexemNode ln2 = translateLn(P, lv->modifiee);
    List modifiers = listPrependItem(NULL, ln1);
    List modifiees = NULL;
    LevelValue result;
    if(!ln1 || (lv->modifiee && !ln2)) {
      return NULL;
    }
    modifiers = listPrependItem(NULL, ln1);
    if(ln2) {
      modifiees = listPrependItem(NULL, ln2);
    }
    result = lvNew(modifiers, lv->level, lv->label, modifiees);
    result->indexWRTNet = -1;
    return result;
  }
}

/* ----------------------------------------------------------------------
   Construct an LV that can bind slot where in A.

   This is the exact opposite of a lexeme swap: rather than change the
   Analysis to suit an LV we would like to try, we invent an LV to suit the
   Analysis. This way of ensuring the path criterion is only used to build
   the initial Analysis.
 */
LevelValue inventLV(Analysis a, int slot) {

  LevelValue original, result;
  int from = slot / a->P->nrLevels;
  int levelno = slot % a->P->nrLevels;
  Level level;
  LexemNode modifier;
  List modifiers;
  String label;
  
  /* If there is at least one LV in the corresponding domain,
     we can steal and copy it. */
  if(a->P->domains[slot]) {
    result = lvClone(listItem(a->P->domains[slot]));
  

    result->indexWRTNet = -1;
    result->no = -1;
    listFree(result->modifiers);
    result->modifier = selectedAt(a, from);
    result->modifiers = listPrependItem(NULL, result->modifier);

    if (result->modifiee != NULL && result->modifiee != NONSPEC) {
      listFree(result->modifiees);
      result->modifiee = selectedAt(a, result->modifiee->arc->from);
      result->modifiees = listPrependItem(NULL, result->modifiee);
    }

    if ((original = findExistingLV(a, result)) != NULL) {
      lvDelete(result);
      result = original;
    }
  }
  /* Otherwise we have to work really hard to invent one out of thin air. */
  else {
    level = listNth(inputCurrentGrammar->levels, 1 + levelno);
    label = listItem(level->labels);
    modifier = selectedAt(a, from);
    modifiers = listPrependItem(NULL, modifier);
    result = lvNew(modifiers, level, label, NULL);
  }

  return result;

}

/* ----------------------------------------------------------------------
   Does LV violate something in A?
*/
Boolean causesConflict(Analysis a, LevelValue lv) {

  List l;
  ConstraintViolation cv;
  int index = lvIndex(lv);

  for (l = a->conflicts; l; l = listNext(l)) {
    cv = listItem(l);
    if (index == cv->nodeBindingIndex1)
      return TRUE;
    if (index == cv->nodeBindingIndex2)
      return TRUE;
  }

  return FALSE;

}

/* ----------------------------------------------------------------------
   Should LV be displayed in a solution?

   The reasoning here is that most slots in an analysis are useless
   (such as the OBL slots of function words) and do not contribute to
   the understanding of a solution, but merely take up space.
   Therefore we want to suppress displaying them.

   An lv is considered boring if it modifies NIL, and all other
   bindings of the same lexeme node also modify NIL. That way, if
   there are the LVs

   #200 = OBL1:ich(2,3)------>NIL  and
   #300 = OBL1:versichern(0,4)----->wir(5,6)

   then #200 will be boring and #300 will be interesting.

   However, FALSE is always returned if LV causes at least one conflict,
   so that conflicts will not refer to invisible LVs.
*/
Boolean isBoring(Analysis a, LevelValue lv) {

  List l;
  int where = lvIndex(lv);
  LevelValue lvb;

  if (causesConflict(a, lv)) {
    return FALSE;
  }

  for (l = a->P->domains[where]; l; l = listNext(l)) {
    lvb = listItem(l);
    if (lvb->modifiee != NULL && lvb->modifiee != NONSPEC &&
        lvb->modifier->arc->from == lv->modifier->arc->from &&
        lvb->modifier->arc->to == lv->modifier->arc->to) {
      return FALSE;
    }
  }

  return TRUE;

}


/* ----------------------------------------------------------------------
   Fix self-modifying LVs.

   When swapping in a lexeme that is so long that it overlaps both the
   modifier and the modifiee of an LV, an LV may end up with a lexeme
   node that modifies itself.

   Since this is unlikely to be helpful, we point those LVs to NIL
   immediately. Perhaps it would be smarter to point them wherever is
   best at the moment, but that is a problem for another day.  

   Note that since this LV is invalid, it cannot be the member of any
   domain. So we don't have to check its index to avoid changing the
   original LVs.
*/
void breakCircle(LevelValue lv) {


  if (lv->modifiee != NULL && lv->modifiee != NONSPEC && lgOverlap(lv->modifier, lv->modifiee)) {
#ifdef DEBUGLVBREAKCIRCLE
    cdgPrintf(CDG_WARNING,
              "WARNING: #%d modifies self at slot @%d\n",
              lv->indexWRTNet, lvIndex(lv));
#endif
    lv->modifiee = NULL;
    listFree(lv->modifiees);
    lv->modifiees = NULL;
  }
}

/* ----------------------------------------------------------------------
   Modify Analysis a so that the LV at slot #where points to the
   lexeme at time point direction. If no such LV exists, a temporary
   LV is allocated for the purpose.
*/
void aRedirect(Analysis a, int where, int direction) {
  LevelValue lv = (LevelValue) vectorElementAt(a->slots, where);
  LevelValue original = NULL;
  LevelValue tmpLV;

  while (!lv) {
    where--;
    lv = (LevelValue) vectorElementAt(a->slots, where);
  }
  tmpLV = lvClone(lv);
  if(lv->indexWRTNet == -1) {
    lvDelete(lv);
  }
  
  /* re-orientate */
  if (direction < 0) {
    tmpLV->modifiee = NULL;
    listFree(tmpLV->modifiees);
    tmpLV->modifiees = NULL;
  } else if(direction >= a->P->maxTimepoint) {
    tmpLV->modifiee = NONSPEC;
    listFree(tmpLV->modifiees);
    tmpLV->modifiees = NULL;
  } else {
    tmpLV->modifiee = selectedAt(a, direction);
    listFree(tmpLV->modifiees);
    tmpLV->modifiees = listPrependItem(NULL, tmpLV->modifiee);
  }

  /* update info */
  tmpLV->indexWRTNet = -1;
  tmpLV->no = -1;

  /* If the swap produced an LV that already exists,
     take this one and delete the clone. */
  if ((original = findExistingLV(a, tmpLV)) != NULL) {
    lvDelete(tmpLV);
    tmpLV = original;
  }

  /* Otherwise we retain the new LV.
     (It will be deallocated the next time this domain is changed.) */

  /* Finally, substitute the LV into the Analysis. */
  vectorSetElementAt(a->slots, tmpLV, where);

}

/* ----------------------------------------------------------------------
   Modify Analysis a so that the LV at slot #where carries the passed
   label. If no such LV exists, a temporary LV is allocated for the
   purpose.
*/
Boolean aRelabel(Analysis a, int where, String label) {
  LevelValue lv = (LevelValue) vectorElementAt(a->slots, where);
  LevelValue original = NULL;
  LevelValue tmpLV;

  while (!lv) {
    where--;
    lv = (LevelValue) vectorElementAt(a->slots, where);
  }

  /* sanity check */
  if(!listContains(lv->level->labels, label)) {
    cdgPrintf(CDG_WARNING, "WARNING: Level `%s' has no label `%s'.\n",
              lv->level->id,
              label);
    return FALSE;
  }
  
  tmpLV = lvClone(lv);

  /* swap */
  tmpLV->label = label;

  /* update info */
  tmpLV->indexWRTNet = -1;
  tmpLV->no = -1;

  /* If the swap produced an LV that already exists,
     take this one and delete the clone. */
  if ((original = findExistingLV(a, tmpLV)) != NULL) {
    lvDelete(tmpLV);
    tmpLV = original;
  }
  /* Otherwise we retain the new LV.
     (It will be deallocated the next time this domain is changed.) */

  /* Finally, substitute the LV into the Analysis. */
  vectorSetElementAt(a->slots, tmpLV, where);
  return TRUE;
  
}


/* Functions dealing with Analyses
   =============================== */

/* --------------------------------------------------------------------
   Calculate all conflicts in a analysis.

   Sets the fields ->b and ->conflicts. If these fields were non-empty, the
   contents are deallocated.
*/
void aJudge(Analysis a) {

  LevelValue lva, lvb;
  int i, j;
  List l;

  if(!a) {
    return;
  }

  /* remove old conflict set */
  if (a->conflicts) {
    listForeachFree(a->conflicts, cvDelete);
    a->conflicts = NULL;
  }
  /* start out with the perfect badness */
  a->b->no = 0;
  a->b->hard = 0;
  a->b->soft = 1.0;

  lock_tree(a->P->maxTimepoint);
  
  /* seek unary conflicts */
  for (i = 0; i < vectorSize(a->slots); i++) {
    if(!slotActive(a, i)) continue;
    lva = (LevelValue) vectorElementAt(a->slots, i);
    if (!lva) continue;
    (void)evalUnary(lva, a->P->net, a->slots, FALSE, a->b, &a->conflicts);

    /* seek binary conflicts */
    for (j = i + 1; j < vectorSize(a->slots); j++) {
      if(!slotActive(a, j)) continue;
      lvb = (LevelValue) vectorElementAt(a->slots, j);
      if (!lvb)
        continue;

      (void)evalBinary(lva, lvb, a->P->net, a->slots, FALSE, a->b, &a->conflicts);
    }
  }

  unlock_tree();

  /* sort conflicts */
  if(a->conflicts) {
    l = a->conflicts;
    a->conflicts = listSort(a->conflicts, cvCompare);
    listFree(l);
  }
}

/* ----------------------------------------------------------------------
   Re-compute conflicts after applying a patch.

   After applying a Patch to an Analysis, it is much faster not to
   evaluate the grammar on the entire new Analysis, but simply to
   re-evaluate it merely on the changed bits. This is slightly
   complicated by the fact that the LVs listed in the Patch are not
   the only ones that might have changed. If the Patch also changed
   the path though the lexeme graph, other LVs might have changed to
   follow suit.

   Basically this procedure first inspects every LV in a to see if it
   has changed. Then all conflicts caused by the changed LVs are
   erased, because those are the ones that may have been resolved.
   Finally the grammar is evaluated on the changed LVs.
*/
void aRejudge(Analysis a, Patch p) {

  ConstraintViolation cv;
  LevelValue lv, lvb;
  LexemNode ln;
  BitVector changedLV = bvNew(a->P->nrSlots);
  BitVector changedInterval = bvNew(a->P->maxTimepoint);

  List l, m;
  List lvList;
  List old = NULL, new = NULL;
  int i,j;

  Vector partialslots = vectorClone(a->slots);
  
  /* 1. Note down all LVs that were changed by the patch. */
  bvSetAllElements(changedLV, FALSE);

  /* explicitly changed LVs */
  for (i = 0; i < vectorSize(p->selection); i++) {
    lv = vectorElementAt(p->selection, i);
    if (!lv)
      continue;
    bvSetElementAt(changedLV, TRUE, lvIndex(lv));
  }

  /* implicitly changed LVs:
     to find those, we note down all changed time intervals */
  if (p->newLexemes) {

    /* make vector of changed time intervals */
    bvSetAllElements(changedInterval, FALSE);
    for (l = p->newLexemes; l; l = listNext(l)) {
      ln = listItem(l);
      bvSetElements(changedInterval, TRUE, ln->arc->from, ln->arc->to);
    }

    /* Mark those LVs that modify one of the new lexemes... */
    for (i = 0; i < vectorSize(a->slots); i++) {
      lv = vectorElementAt(a->slots, i);
      if (!lv)
        continue;
      if (!(lv->modifiee != NULL && lv->modifiee != NONSPEC))
        continue;
      for (j = lv->modifiee->arc->from; j < lv->modifiee->arc->to; j++) {
        if (bvElementAt(changedInterval, j)) {
          bvSetElementAt(changedLV, TRUE, i);
          break;
        }
      }
    }

    /* ...or are modified by one of them */
    for (i = 0; i < vectorSize(a->slots); i++) {
      lv = vectorElementAt(a->slots, i);
      if (!lv)
        continue;
      for (j = lv->modifier->arc->from; j < lv->modifier->arc->to; j++) {
        if (bvElementAt(changedInterval,j)) {
          bvSetElementAt(changedLV, TRUE, i);
          break;
        }
      }
    }
  }

  /* 2. Remove conflicts that may have disappeared. */
  for (l = a->conflicts; l; l = listNext(l)) {
    cv = (ConstraintViolation) listItem(l);

    /* caused by a changed lv? */
    if (bvElementAt(changedLV,lvIndex(cv->lv1)) ||
        (cv->lv2 && bvElementAt(changedLV,lvIndex(cv->lv2))) ||

        /* context-sensitive conflict? This can disappear
           whenever anything change at all. */
        cv->constraint->is_context_sensitive ||
        /* caused by an lv that is unchanged but has disappeared?
           (This can only happen during initialization.) */
        vectorElementAt(a->slots, lvIndex(cv->lv1)) == NULL ||
        (cv->lv2 && vectorElementAt(a->slots, lvIndex(cv->lv2)) == NULL)) {

      /* remove */

      /* Since we are going to remove this conflict,
         we must remove its penalty as well. */
      bSubtract(a->b, cv->penalty);

      /* de-allocate */
      cvDelete(cv);

    } else {

      /* retain */
      old = listAppendItem(old, cv);
    }

  }

  /* de-allocate old list, don't free items:
     each item was either freed above, or is needed in the new list. */
  listFree(a->conflicts);

  /* 3. Calculate newly introduced conflicts. */

  /* make list of new LVs */
  lvList = NULL;
  for (i = 0; i < vectorSize(a->slots); i++) {
    if(!slotActive(a, i)) {
      vectorSetElementAt(partialslots, NULL, i);
      continue;
    }      
    lv = vectorElementAt(a->slots, i);
    if (!lv)
      continue;
    if (bvElementAt(changedLV,i)) {
      lvList = listAppendItem(lvList, lv);
    }
  }

  lock_tree(a->P->maxTimepoint);
  
  /* seek new unary conflicts */
  for (l = lvList; l; l = listNext(l)) {
    lv = listItem(l);
    (void)evalUnary(lv, a->P->net, NULL, FALSE, a->b, &new);

    /* seek new binary conflicts between changed & unchanged LVs */
    for (i = 0; i < vectorSize(a->slots); i++) {
      if(!slotActive(a, i)) continue;
      lvb = (LevelValue) vectorElementAt(a->slots, i);
      if (!lvb)
        continue;

      /* don't evaluate against another member of lvList */
      if (bvElementAt(changedLV,i))
        continue;

      (void)evalBinary(lv, lvb, a->P->net, NULL, FALSE, a->b, &new);
    }

    /* seek new binary conflicts between changed LVs */
    for (m = listNext(l); m; m = listNext(m)) {
      lvb = listItem(m);
      (void)evalBinary(lv, lvb, a->P->net, NULL, FALSE, a->b, &new);
    }
  }

  /* find context-sensitive conflicts */
  evalInContext(partialslots, a->slots, a->b, &new);
  vectorDelete(partialslots);
  
  unlock_tree();

  /* sort new conflicts */
  if(new) {
    l =  new;
    new = listSort(new, cvCompare);
    listFree(l);
  }

  /* ensure that new conflicts precede old conflicts */
  a->conflicts = listAppendList(new, old);

  listFree(lvList);

  bvDelete(changedLV);
  bvDelete(changedInterval);
  
}

/* ----------------------------------------------------------------------
   Re-compute conflicts after manipulating a single LV.

   This presupposes that the only change since the last aJudge() call was
   to insert an LV at slot and this did not change the path through the
   lexeme graph.
*/
void rejudgeLV(Analysis a, int where) {

  ConstraintViolation cv;
  LevelValue lv, lvb;

  List l;
  List old = NULL, new = NULL;
  int i;

  /* remove old conflicts */
  for (l = a->conflicts; l; l = listNext(l)) {
    cv = (ConstraintViolation) listItem(l);

    /* caused by lv? */
    if (cv->nodeBindingIndex1 == where ||
        (cv->lv2 && cv->nodeBindingIndex2 == where) ||
        /* context-sensitive? */
        cv->constraint->is_context_sensitive) {
     
      /* remove */

      /* Since we are going to remove this conflict,
         we must remove its penalty as well. */
      bSubtract(a->b, cv->penalty);

      /* de-allocate */
      cvDelete(cv);

    } else {

      /* retain */
      old = listAppendItem(old, cv);
    }

  }

  /* de-allocate old list, don't free items:
     each item was either freed above, or is needed in the new list. */
  listFree(a->conflicts);

  lock_tree(a->P->maxTimepoint);

  lv = vectorElementAt(a->slots, where);
  (void)evalUnary(lv, a->P->net, NULL, FALSE, a->b, &new);

  /* seek new binary conflicts between lv and other LVs */
  for (i = 0; i < vectorSize(a->slots); i++) {

    lvb = (LevelValue) vectorElementAt(a->slots, i);
    if (!lvb)
      continue;

    /* don't evaluate against self */
    if (lv == lvb)
      continue;

    (void)evalBinary(lv, lvb, a->P->net, NULL, FALSE, a->b, &new);
  }

  evalInContext(a->slots, a->slots, a->b, &new);
  unlock_tree();
  
  /* sort new conflicts */
  if(new) {
    l =  new;
    new = listSort(new, cvCompare);
    listFree(l);
  }
  
  /* ensure that new conflicts precede old conflicts */
  a->conflicts = listAppendList(new, old);

}

/* ----------------------------------------------------------------------
   Allocate an Analysis. 

   This function allocates a new Analysis for the Problem P. It does not
   allocate all fields of the structure, since the first Analysis in a
   Problem will be completed by net2problem(), and further structures are
   produced by aClone().

*/
Analysis aNew(Problem P) {

  Analysis result = memMalloc(sizeof (AnalysisStruct));
  result->P = P;
  result->slots = vectorNew(P->nrSlots);
  result->isUnselectedLexeme = bvNew(P->nrLexemes);
  result->b = bestBadness();
  result->conflicts = NULL;
  result->subStart = P->minTimepoint;
  result->subStop = P->maxTimepoint;
  
  vectorSetAllElements(result->slots, NULL);
  bvSetAllElements(result->isUnselectedLexeme, TRUE);

  return result;
}

/* ----------------------------------------------------------------------
   Clone an Analysis.

   This function creates a deep copy of A, except for the constituting LVs,
   which are usually simply copied rather than cloned. LVs are only cloned
   if they are not owned by the underlying constraint net (i.e., have
   indexWRTNet == -1).
*/
Analysis aClone(Analysis a) {
  int i;
  LevelValue lv;
  List l;
  ConstraintViolation cv;

  Analysis result = memMalloc(sizeof (AnalysisStruct));
  result->P = a->P;

  /* mostly shallow copy of LVs */
  result->slots = vectorClone(a->slots);
  for (i = 0; i < vectorSize(result->slots); i++) {
    lv = vectorElementAt(a->slots, i);
    if(!lv) {
      continue;
    }
    if (lv->indexWRTNet == -1) {
      lv = lvClone(lv);
      vectorSetElementAt(result->slots, lv, i);
    }
  }

  /* deep copy of everything else */
  result->isUnselectedLexeme = bvClone(a->isUnselectedLexeme);
  result->conflicts = NULL;
  for (l = a->conflicts; l; l = listNext(l)) {
    cv = cvClone((ConstraintViolation) listItem(l));
    result->conflicts = listAppendItem(result->conflicts, cv);
  }
  result->b = bClone(a->b);
  result->subStart = a->subStart;
  result->subStop = a->subStop;
  
  
  return result;

}

/* --------------------------------------------------------------------
   De-allocate a analysis.

   Deletes the dynamic parts except the enclosing Problem and the
   original LVs.
*/
void aDelete(Analysis a) {
  int i;
  LevelValue lv;
  if (a) {
    if (a->isUnselectedLexeme) {
      bvDelete(a->isUnselectedLexeme);
    }

    if (a->slots) {
      for (i = 0; i < vectorSize(a->slots); i++) {
        lv = vectorElementAt(a->slots, i);
        if (lv && lv->indexWRTNet == -1) {
          lvDelete(vectorElementAt(a->slots, i));
        }
      }
      vectorDelete(a->slots);
    }

    listForeachFree(a->conflicts, cvDelete);
    if(a->b) {
      bDelete(a->b);
    }

    memFree(a);

  }
}

/* ----------------------------------------------------------------------
   Return TRUE if A has no hard violations.
*/
Boolean aValid(Analysis a) {

  return (a->b->hard == 0);

}

/* ----------------------------------------------------------------------
   Re-calculate the soft score of an Analysis by multiplying the
   scores of all soft conflicts.

   This is to prevent rounding errors after repeated division and
   multiplication during the repair process.
*/
void aCalibrate(Analysis a) {

  List l;
  ConstraintViolation cv;
  Number score = 1;

  for (l = a->conflicts; l; l = listNext(l)) {
    cv = (ConstraintViolation) listItem(l);
    if (cv->penalty)
      score *= cv->penalty;
  }

  if (score != a->b->soft) {

#ifdef DEBUGACALIBRATE
    cdgPrintf(CDG_DEBUG, "adjusted by %4.3e)\n", score - a->b->soft);
#endif

    a->b->soft = score;

  }
}

/* ----------------------------------------------------------------------
   Check that an LV is consistent internally _and_ with the
   containing Analysis.
*/
void checkValidity(Analysis a, LevelValue lv) {

  if (isValid(a, lv)) {
    return;
  }

  if (lv->modifiee != NULL && lv->modifiee != NONSPEC && !selectedLexeme(a, lv->modifiees)) {
    cdgPrintf(CDG_ERROR, "\nPANIC: modifiees of #%d are all unselected:\n",
              lv->indexWRTNet);
    displayLV(a->P, lv, 999);
    printSelectedPath(CDG_ERROR, a);
    abort();
  }

  if (!selectedLexeme(a, lv->modifiers)) {
    cdgPrintf(CDG_ERROR, "\nPANIC: modifiers of #%d are all unselected:\n",
              lv->indexWRTNet);
    displayLV(a->P, lv, 999);
    printSelectedPath(CDG_ERROR, a);
    abort();
  }
}




/* ----------------------------------------------------------------------
   Show which LVs are not the unary best ones.

   Displays all domains that do not have their optimal element selected
   (i.e. the one with the highest limit).

*/
void showUnary(Analysis a) {

  int i;
  LevelValue lv, optimal;
  Number x, y;

  for (i = 0; i < vectorSize(a->slots); i++) {
    lv = vectorElementAt(a->slots, i);
    if (!lv)
      continue;

    optimal = listItem(a->P->domains[i]);

    x = getLimit(a->P,lv);
    y = getLimit(a->P,optimal);

    if (x < y) {
      cdgPrintf(CDG_INFO, "@%d: #%d limit %4.3e, #%d limit %4.3e\n",
                i, lv->indexWRTNet, x, optimal->indexWRTNet, y);
    }
  }

  cdgPrintf(CDG_INFO, "\n\n");

}

/* ----------------------------------------------------------------------
   Report on current distance to the annotation.

   Compares P->current and P->annotation and displays the number of LVs
   differing between them, as well as the total number of LVs in the
   annotation. Two LVs are considered equal if they have the same
   indexWRTNet.

*/
void reportCorrectness(unsigned long int mode, Problem P) {

  int i;
  LevelValue lva,lvb;
  int links = 0, correct = 0;

  if(!P->annotation) {
    return;
  }
  
  for(i = 0; i < P->nrSlots; i++) {
    lvb = vectorElementAt(P->annotation->slots,i);
    if(!lvb) {
      continue;
    }
    links++;
    lva = vectorElementAt(P->current->slots,i);
    if(lva != NULL &&
       lva->indexWRTNet == lvb->indexWRTNet) {
      correct++;
    }
  }
}

/* ----------------------------------------------------------------------
   Selects list of lexemes and switches all LVs correspondingly.

   Substitutes all lexeme nodes in WHICH into A. All previously selected
   LVs are changed just enough so that they comply with the new path
   selected. Naturally, this may make part of the structure and the labels
   of the Analysis totally meaningless if the new path differs
   substantially from the old one.

*/
void selectPath(Analysis a, List which) {
  LevelValue oldLV, newLV, original;
  LexemNode ln, lnold, lnnew;
  List l;
  int index, i, j;
  BitVector changed;

  /* It is admissible to call selectPath() vacuously. */
  if (!which) {
    return;
  }

  changed = bvNew(a->P->maxTimepoint);
  bvSetAllElements(changed, FALSE);

#ifdef DEBUGSELECTPATH
  cdgPrintf(CDG_DEBUG, "old path:\n");
  for (i = 0; i < a->P->nrLexemes; i++) {
    ln = vectorElementAt(a->P->lg->nodes, i);
    if (bvElementAt(a->isUnselectedLexeme, i))
      continue;
    cdgPrintf(CDG_DEBUG, "%%%d: ", ln->no);
    lgPrintNode(CDG_DEBUG, ln);
    cdgPrintf(CDG_DEBUG, "\n");
  }
  cdgPrintf(CDG_DEBUG, "\nselecting path:\n");
  for (l = which; l; l = listNext(l)) {
    ln = listItem(l);
    cdgPrintf(CDG_DEBUG, "%%%d: ", ln->no);
    lgPrintNode(CDG_DEBUG, ln);
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif

  for (l = which; l; l = listNext(l)) {
    lnnew = listItem(l);
    if (!bvElementAt(a->isUnselectedLexeme, lnnew->no)) {
      continue;
    }
    for (i = lnnew->arc->from; i < lnnew->arc->to; i++) {
#ifdef DEBUGSELECTPATH
      cdgPrintf(CDG_DEBUG,
                "  interval %d has changed in word `%s'.\n",
                i, lnnew->lexem->description);
#endif
      bvSetElementAt(changed, TRUE, i);
    }
  }

  /* unselect all lexemes that cover changed timespans */
  for (i = 0; i < a->P->nrLexemes; i++) {
    lnold = vectorElementAt(a->P->lg->nodes, i);
    if (!lnold) {
      continue;
    }
    for (j = lnold->arc->from; j < lnold->arc->to; j++) {
      if (bvElementAt(changed,j)) {
        if(!bvElementAt(a->isUnselectedLexeme, lnold->no)) {
          bvSetElementAt(a->isUnselectedLexeme, TRUE, lnold->no);
#ifdef DEBUGSELECTPATH
          cdgPrintf(CDG_DEBUG, "  unselecting ");
          lgPrintNode(CDG_DEBUG, lnold);
          cdgPrintf(CDG_DEBUG, " because of changed time span %d.\n", j);
#endif
          break;
        }
      }
    }
  }

  /* record the selection */
  for (l = which; l; l = listNext(l)) {
    lnnew = listItem(l);
    bvSetElementAt(a->isUnselectedLexeme, FALSE, lnnew->no);
  }

#ifdef DEBUGSELECTPATH
  cdgPrintf(CDG_DEBUG, "new path:\n");
  for (i = 0; i < a->P->nrLexemes; i++) {
    ln = vectorElementAt(a->P->lg->nodes, i);
    if (bvElementAt(a->isUnselectedLexeme, i))
      continue;
    cdgPrintf(CDG_DEBUG, "%%%d: ", ln->no);
    lgPrintNode(CDG_DEBUG, ln);
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif

  /* exchange all LVs that bind unselected lexemes. These are:

  [A] LVs that use one of them as a modifiee. They are simply
  switched to the simultaneously selected lexeme. If that results
  in an LV that already exists, no additional LV must be allocated.

  To find them, we must iterate over all LVs. */
  for (index = 0; index < vectorSize(a->slots); index++) {
    oldLV = (LevelValue) vectorElementAt(a->slots, index);

    /* hit? */
    if (!oldLV ||
        !(oldLV->modifiee != NULL && oldLV->modifiee != NONSPEC) ||
        selectedLexeme(a, oldLV->modifiees)) {
      continue;
    }

#ifdef DEBUGSELECTPATH
    cdgPrintf(CDG_DEBUG, "[A] have to replace #%d in @%d\n", oldLV->indexWRTNet, index);
#endif

    /* clone the LV (perhaps only temporarily) */
    newLV = lvClone(oldLV);

    /* realize change */
    ln = selectedAt(a, oldLV->modifiee->arc->from);
    newLV->modifiee = ln;
    listFree(newLV->modifiees);
    newLV->modifiees = listAppendItem(NULL, ln);
    newLV->indexWRTNet = -1;
    newLV->no = -1;

    /* If the swap produced an LV that already exists,
       take this one and delete the clone. */
    if ((original = findExistingLV(a, newLV)) != NULL) {
      lvDelete(newLV);
      newLV = original;
    } else {

      /* A manipulated LV may even be self-modifying. */
      breakCircle(newLV);
    }

    /* Finally, substitute the new LV into the Analysis. */
    vectorSetElementAt(a->slots, newLV, index);
    /* This overwrites the old content of the slot,
       so there is no need to remove the old LV. */
#ifdef DEBUGSELECTPATH
    cdgPrintf(CDG_DEBUG, "replaced with #%d.\n", newLV->indexWRTNet);
#endif

    /* de-allocate temporary */
    if (oldLV->indexWRTNet == -1) {
      lvDelete(oldLV);
    }
  }

  /* [B] Those LVs that use an unselected lexeme as a modifier.
     Note that an LV may fall under both category [A] and [B]. */
  for (index = 0; index < vectorSize(a->slots); index++) {
    oldLV = (LevelValue) vectorElementAt(a->slots, index);

    /* hit? */
    if (!oldLV || selectedLexeme(a, oldLV->modifiers)) {
      continue;
    }

    /* since case [B] may replace an LV from domain @X
       with an lV that goes into domain @Y,
       we must take care to remove the old LV from its slot. */
    vectorSetElementAt(a->slots, NULL, index);

#ifdef DEBUGSELECTPATH
    cdgPrintf(CDG_DEBUG, "[B] have to replace #%d in @%d\n", oldLV->indexWRTNet, index);
#endif


    /* Beware of holes in the Analysis!

    If we replace the lexeme

    DATUM(0-3)

    by the lexemes

    der(0-1) erste(1-2) April(2-3),

    then the old LV bound three domains. We cannot simply replace
    this LV with one that binds der(0-1) because that would leave
    two domains vacant. Instead we inspect all of these domains and
    make sure that they all are bound. */
    i = oldLV->modifier->arc->from;
    while (i < oldLV->modifier->arc->to) {

      /* clone the LV (perhaps only temporarily) */
      newLV = lvClone(oldLV);

      /* realize change */
      ln = selectedAt(a, i);
      newLV->modifier = ln;
      listFree(newLV->modifiers);
      newLV->modifiers = listAppendItem(NULL, ln);
      newLV->indexWRTNet = -1;
      newLV->no = -1;

      /* If the swap produced an LV that already exists, take this one and delete the clone. */
      if ((original = findExistingLV(a, newLV)) != NULL) {
        lvDelete(newLV);
        newLV = original;
      } else {

        /* A manipulated LV may even be self-modifying. */
        breakCircle(newLV);
      }

      /* Finally, substitute the LV into the Analysis. */
      vectorSetElementAt(a->slots, newLV, lvIndex(newLV));
#ifdef DEBUGSELECTPATH
      cdgPrintf(CDG_DEBUG, "replaced with #%d.\n", newLV->indexWRTNet);
#endif

      /* step over this interval */
      i = ln->arc->to;
    }

    /* de-allocate temporary */
    if (oldLV->indexWRTNet == -1) {
      lvDelete(oldLV);
    }
  }
  bvDelete(changed);

}



/* Functions dealing with Patches
   ============================== */

/* ----------------------------------------------------------------------
   Return new Patch.

   Returns a new WIDTH-fold Patch and initializes the internal control
   variables. The result contains no LV yet.

*/
Patch pNew(Analysis a, int width, int flags) {

  Patch result = memMalloc(sizeof (PatchStruct));

  result->a = a;
  result->local = bestBadness();
  result->limit = 1.0;
  result->selection = vectorNew(width);
  result->index = 0;
  result->nrNewLVs = 0;
  result->predicted = NULL;
  result->newLexemes = NULL;
  result->bound = bvNew(a->P->nrSlots);

  /* normally, disallow deleted lexeme nodes implicitly */
  result->forbiddenLexemes = bvClone(a->P->lg->isDeletedNode);

  /* upon request, allow them all */
  if(flags & OD_DELETED) {
    bvSetAllElements(result->forbiddenLexemes, FALSE);
  }

  bvSetAllElements(result->bound, FALSE);

  return result;
}

/* ----------------------------------------------------------------------
   Clone a Patch.

   Only the PatchStruct itself, the Badness and the Vectors and Lists are
   actually allocated anew, not their contents.
*/
Patch pClone(Patch p) {

  Patch result;
  
  if(!p) {
    return NULL;
  }
  
  result = pNew(p->a, vectorSize(p->selection), 0);

  result->index = p->index;
  result-> nrNewLVs = p->nrNewLVs;
  result->limit = p->limit;

  /* deep copy */  
  vectorCopy(result->selection, p->selection);
  result->newLexemes = listClone(p->newLexemes);

  bvCopy(result->bound, p->bound);
  bvCopy(result->forbiddenLexemes, p->forbiddenLexemes);

  if (p->predicted)
    result->predicted = bClone(p->predicted);

  if (p->local)
    bCopy(result->local, p->local);
  else
    result->local = NULL;

  return result;

}

/* ----------------------------------------------------------------------
   Delete Patch.
*/
void pDelete(Patch p) {

  if(!p) {
    return;
  }

  vectorDelete(p->selection);
  if (p->local)
    bDelete(p->local);
  if (p->predicted)
    bDelete(p->predicted);
  listFree(p->newLexemes);
  bvDelete(p->forbiddenLexemes);
  bvDelete(p->bound);
  memFree(p);

}

/* ----------------------------------------------------------------------
   Build a Patch that binds one more LV than the original one.

   Usually, Patches that cannot help in solving the problem are not built,
   and NULL is returned instead. A Patch is considered unhelpful if it
   contains a deleted LV, or has a lower limit than an existing solution,
   or has score 0.

   This is normally a Good Thing, but in some places we expressly want to
   build a Patch even though we know it won't help us, so the pruning can
   be suppressed by setting OD_DELETED.

   In no case will pNext() ever construct a Patch with physically
   incompatible LVs.
*/
Patch pNext(Patch old, LevelValue lv, int flags) {

  LevelValue current;
  int i, j = 0;
  Badness b;
  Boolean prune = ((flags & OD_DELETED) == 0);
  Boolean debug = (flags & OD_DEBUG) != 0;
  Patch new;

  /* reject deleted LVs */  
  if (prune && lv->isDeleted) {
    if(debug) {
      cdgPrintf(CDG_DEBUG, "DEBUG: #%d is deleted.\n", lv->indexWRTNet);
    }
    return NULL;
  }

  /* new LV must satisfy the search flags */
  current = slotFiller(old->a,lvIndex(lv));

  if((flags & OD_ORIENT) == 0) {
    if(current) {
      int new = -1;
      int old = -1;
      if(current->modifiee) {
        old = current->modifiee->arc->from;
      }
      if(lv->modifiee) {
        new = lv->modifiee->arc->from;
      }
      if(old != new) {
        if(debug) {
          cdgPrintf(CDG_DEBUG,
                    "DEBUG: #%d would change the structure\n",
                    lv->indexWRTNet);
        }
        return NULL;
      }
    }
  }
  if((flags & OD_LABEL) == 0) {
    if(current) {
      if(current->label != lv->label) {
        if(debug) {
          cdgPrintf(CDG_DEBUG,
                    "DEBUG: #%d would change the label\n",
                    lv->indexWRTNet);
        }
        return NULL;
      }
    }
  }
  if((flags & OD_LEXICAL) == 0) {
    if(!isValid(old->a, lv)) {
      if(debug) {
        cdgPrintf(CDG_DEBUG,
                  "DEBUG: #%d would change the lexemes\n",
                  lv->indexWRTNet);
      }
      return NULL;
    }
  }  

  /* ...and the path criterion */
  if (!lvVectorCompatible(lv, old->forbiddenLexemes)) {
    if(debug) {
      cdgPrintf(CDG_DEBUG,
                "DEBUG: #%d is lexically incompatible.\n",
                lv->indexWRTNet);
    }
    return NULL;
  }

      
  /* ...and the grammaticality constraints */
  b = bestBadness();
  for (i = 0; i < vectorSize(old->selection); i++) {
    LevelValue oldlv = (LevelValue) vectorElementAt(old->selection, i);
    if (!oldlv)
      continue;
    if (prune &&
        evalBinary(lv, oldlv, old->a->P->net, NULL, FALSE, b, NULL) == 0.0) {
      if(debug) {
        cdgPrintf(CDG_DEBUG, "DEBUG: pruning Patch with score 0.0\n");
      }
      bDelete(b);
      return NULL;
    }
  }


  /* build new Patch */    
  new = pClone(old);

  /* extend by one LV */  
  vectorSetElementAt(new->selection, lv, new->index);
  new->index++;

  /* note down all domains bound by this LV */
  for (i = lv->modifier->arc->from; i < lv->modifier->arc->to; i++) {
    j = old->a->P->nrLevels * i + lv->level->no;
    bvSetElementAt(new->bound, TRUE, j);
  }

  /* update Badness */  
  bAddBadness(new->local, b);
  if (lv->score != 1.0) {
    bAdd(new->local, lv->score);
  }
  /* This is incorrect. A score of f < 1.0 for the new LV may have
     been caused by several unary faults, so it might have to be
     represented by (2/0/f) rather than (1/0/f). But we ignore this
     for the time being. */

  /* update limits */  
  if (new->local->soft < new->limit) {
    new->limit = new->local->soft;
  }
  if (lv->limit < new->limit) {
    new->limit = lv->limit;
  }

  /* reject Patches with limit that is too low */  
  if (prune) {

    if (significantlyGreater(old->a->P->globalMinimum, new->limit)) {
      if(debug) {
        cdgPrintf(CDG_DEBUG, "DEBUG: limit %4.3e too low\n", new->limit);
      }
      pDelete(new);
      bDelete(b);
      return NULL;
    }
  }

  /* update compatibility Vector */
  lgRequireLexemes(old->a->P->lg, new->forbiddenLexemes, lv->modifiees);
  lgRequireLexemes(old->a->P->lg, new->forbiddenLexemes, lv->modifiers);

  bDelete(b);
  return new;

}

/* ----------------------------------------------------------------------
   Compare Patches, first by limit, then by local Badness.
*/
Boolean pCompare(Patch a, Patch b) {

  if (a->limit > b->limit)
    return TRUE;
  if (a->limit < b->limit)
    return FALSE;

  return (bCompare(a->local, b->local));

}

/* ----------------------------------------------------------------------
   Print exactly what effect P has on A
*/
void pDebug(Analysis a, Patch p) {
  List l;
  Analysis newA = aClone(a);
  apply(newA, pClone(p));
  for(l = newA->conflicts; l != NULL; l = listNext(l)) {
    ConstraintViolation new = listItem(l);
    if(!cvContains(a->conflicts, new)) {
      cdgPrintf(CDG_DEBUG, "DEBUG: introduces ");
      cvPrint(CDG_DEBUG, new);
      cdgPrintf(CDG_DEBUG, "\n");

    }
  }
  for(l = a->conflicts; l != NULL; l = listNext(l)) {
    ConstraintViolation old = listItem(l);
    if(!cvContains(newA->conflicts, old)) {
      cdgPrintf(CDG_DEBUG, "DEBUG: removes ");
      cvPrint(CDG_DEBUG, old);
      cdgPrintf(CDG_DEBUG, "\n");
    }
  }

  aDelete(newA);

}

/* ----------------------------------------------------------------------
   Display a Patch.
  
   This function displays a Patch using cdgPrintf().

   If BRIEF is set, self-substitutions are omitted from the output for
   better readability.
*/
void pPrint(unsigned long mode, Patch p, Boolean brief) {

  int i;
  LevelValue lv;
  List l;
  LevelValue original;
  Boolean empty = TRUE;

  /* show new LVs */  
  for (i = 0; i < vectorSize(p->selection); i++) {
    lv = (LevelValue) vectorElementAt(p->selection, i);
    if (!lv)
      continue;

    /* only show those LVs that actually are new */
    original = vectorElementAt(p->a->slots, lvIndex(lv));
    if (brief && original && original->indexWRTNet == lv->indexWRTNet)
      continue;

    cdgPrintf(mode, "#%d ", lv->indexWRTNet);
    empty = FALSE;
  }

  if (empty) {
    cdgPrintf(mode, "[no change]{%d} ", vectorSize(p->selection));
  }

  cdgPrintf(mode, "/ ");

  /* show new lexeme nodes */  
  for (l = p->newLexemes; l; l = listNext(l)) {
    lgPrintNode(mode, listItem(l));
    cdgPrintf(mode, " ");
  }

  /* show local score */  
  cdgPrintf(mode, "%4.3e ", p->local->soft);
  if (p->local->soft > p->limit) {
    cdgPrintf(mode, "limit %4.3e  ", p->limit);
  }

  /* show predicted Badness */  
  if (p->predicted) {
    bPrint(mode, p->predicted);
    cdgPrintf(mode, " predicted");
  }
  cdgPrintf(mode, "\n");

}

/* ----------------------------------------------------------------------
   Apply a Patch to an Analysis. P must already lexicalized.
   
   After apply(),
   - A contains all Lvs from P
   - all LVs from P are valid in A
   - A is path consistent 

   Note that this deletes the Patch when done, so a Patch can only be
   applied once.

   It is not meaningful to apply a Patch to any other Analysis than the
   one it was built for. A different but equivalent Analysis is fine,
   though -- that is in fact what pJudge() does.

*/
void apply(Analysis a, Patch p) {

  LevelValue source;
  int i;

  /* swap in all new lexemes */
  selectPath(a, p->newLexemes);

  /* insert all new LVs */
  for (i = 0; i < vectorSize(p->selection); i++) {
    source = vectorElementAt(p->selection, i);
    if (!source) {
      continue;
    }
    selectLV(a, source);
  }

  /* rejudge in all the necessary places */ 
  aRejudge(a, p);
  aCalibrate(a);

  /* cleanup */
  pDelete(p);

}

/* ----------------------------------------------------------------------
   Predict accurately the Badness that applying P would yield.

   To this end, the Patch is temporarily applied to its Analysis and the
   resulting Badness is noted.
*/
void pJudge(Patch p) {

  Analysis newA = aClone(p->a);

  apply(newA, pClone(p));
  if(p->predicted) {
    bDelete(p->predicted);
  }
  p->predicted = bClone(newA->b);

  aDelete(newA);

}

/* -----------------------------------------------------------------------
   Given a tuple of (compatible) LVs and an Analysis A, compute the set of
   lexemes that have to be substituted into A so it has a complete and
   valid path that supports all the LVs in P.

   Uses the vector p->forbiddenLexemes to decide which lexemes are
   incompatible with one of the LVs. The result is put into p->newLexemes.

   Returns FALSE if no path can be built, TRUE otherwise. Usually, this
   function normally does not select deleted lexeme nodes. Therefore, if a
   Patch was built with OD_DELETED, the flag must also be given to
   pLexicalize(). This is the only flag that pLexicalize() reads.

   If a Patch was not built with OD_DELETED, this function can only fail
   because the lexeme graph has become altogether disconnected, e.g. there
   is no path at all that spans the entire graph.
   
   It is a critical error if the lexicalisation fails (this should be
   changed to something more sensible). As it is, the caller is guaranteed
   to get a valid path if one exists, or an assertion otherwise. No
   provision is made to create a path that is as `nearly' valid as
   possible.
   
*/
Boolean pLexicalize(Patch p, int flags, Boolean debugme) {
  int i, j, k;
  LexemNode ln, orig, found;
  Boolean exactfit;
  Analysis a = p->a;
  LexemGraph lg = a->P->lg;
  
  /* stores the hypothetically selected lexeme nodes per time point */
  Vector bound=vectorNew(a->P->maxTimepoint);
  
  /* As of now, all time points are bound. */
  int coverage = a->P->nrTimepoints;

  if(debugme) {
    cdgPrintf(CDG_DEBUG, "DEBUG: pLexicalize() called. These are the new LVs:\n");
  }
  for(i = 0; i < vectorSize(p->selection); i++) {
    LevelValue lv = vectorElementAt(p->selection, i);
    if(lv) {
      if(debugme) {
        displayLV(a->P, lv, 99999);
      }
 
    }
  }

  /* if the lexeme graph has become disconnected,
     pLexicalize() cannot possibly succeed. */
  if(lg->noOfPaths == 0) {
    if(debugme) {
      cdgPrintf(CDG_DEBUG, "DEBUG: the graph is disconnected. Giving up.\n\n");
    }
    return FALSE;
  }

  /* Otherwise it must succeed. If it doesn't,
     that is a programming error. */
  if(debugme) {
    {
      LevelValue lv1, lv2;
      for(i = 0; i < vectorSize(p->selection); i++) {
        lv1 = vectorElementAt(p->selection, i);
        for(j = 0; j < vectorSize(p->selection); j++) {
          lv2 = vectorElementAt(p->selection, j);
          if(!lvCompatible(p->a->P->lg, lv1, lv2)) {
            cdgPrintf(CDG_DEBUG, "PANIC: pNext is about to build an invalid Patch!\n");
            abort();
          } else {
            cdgPrintf(CDG_DEBUG, "#%d - #%d: compatible.\n", lv1->indexWRTNet, lv2->indexWRTNet);
          }
        }
      }
    }
  }

  if(debugme) {
    cdgPrintf(CDG_DEBUG,
              "checking %d time points.\n",
              a->P->nrTimepoints);
    cdgPrintf(CDG_DEBUG, "Old path:\n");
    printSelectedPath(CDG_DEBUG, a);
    cdgPrintf(CDG_DEBUG, "Here is the entire lexeme graph:\n");
    lgPrint(CDG_DEBUG, lg);

  }

  i = a->P->minTimepoint;
  while(i < a->P->maxTimepoint) {
    ln = selectedAt(a, i);
    for(j = ln->arc->from; j < ln->arc->to; j++) {
      vectorSetElementAt(bound, (Pointer)ln, j);
    }
    i = ln->arc->to;
  }

  /* Look for points that must be unbound. */
  i = a->P->minTimepoint;
  while (i < a->P->maxTimepoint) {
    ln = vectorElementAt(bound,i);
    
    if (bvElementAt(p->forbiddenLexemes, ln->no)) {
      coverage -= (ln->arc->to - ln->arc->from);
      for(j = ln->arc->from; j < ln->arc->to; j++) {
        vectorSetElementAt(bound, NULL, j);
      }
      if(debugme) {
        cdgPrintf(CDG_DEBUG, "unbinding ");
        lgPrintNode(CDG_DEBUG, ln);
        cdgPrintf(CDG_DEBUG, ", %d points bound\n", coverage);
        if(coverage > a->P->nrTimepoints) {
          cdgPrintf(CDG_DEBUG, "Something's not right!!\n");
          abort();
        }
      }
    }

    /* step to next (possibly) unbound time point */
    i = ln->arc->to;
  }

  while(coverage < a->P->nrTimepoints) {
    if(debugme) {
      cdgPrintf(CDG_DEBUG, "%d points bound, still patching...\n", coverage);
    }
  
    /* Find replacements. */
    i = a->P->minTimepoint;
    while(i < a->P->maxTimepoint) {
      if(debugme) {
        cdgPrintf(CDG_DEBUG, "checking point %d...\n", i);
      }
      if(vectorElementAt(bound,i)) {
        i = ((LexemNode)vectorElementAt(bound,i))->arc->to;
        continue;
      }
      /* Find a new lexeme binding point #i. */
      orig = selectedAt(a, i);
      if(debugme) {
        cdgPrintf(CDG_DEBUG, "Trying to replace ");
        lgPrintNode(CDG_DEBUG, orig);
        cdgPrintf(CDG_DEBUG, " at point %d\n", i);
      }
      found = NULL;
      exactfit = FALSE;
      for (j = 0; j < a->P->lgWidth; j++) {
        ln = vectorElementAt(a->P->lexemes, a->P->lgWidth * i + j);
        if(!ln) {
          continue;
        }
        if(debugme) {
          cdgPrintf(CDG_DEBUG, "Trying ");
          lgPrintNode(CDG_DEBUG, ln);
          cdgPrintf(CDG_DEBUG, "... ");
        }
        if(!(flags & OD_DELETED) &&
           bvElementAt(lg->isDeletedNode, ln->no)) {
          if(debugme) {
            cdgPrintf(CDG_DEBUG, "deleted.\n");
          }
          continue;
        }
        if(bvElementAt(p->forbiddenLexemes, ln->no)) {
          if(debugme) {
            cdgPrintf(CDG_DEBUG, "forbidden.\n");
          }
          continue;
        }
        /* Try very hard to make a good choice. We want to prefer nodes
           that bind exactly the same time slot as before, because then
           we don't have to make subsidiary swaps. Secondarily, we want
           nodes with high limits because those lead to better solutions.
        */
        if(exactfit && (orig->arc->from != ln->arc->from ||
                        orig->arc->from != ln->arc->from)) {
          if(debugme) {
            cdgPrintf(CDG_DEBUG, "not an exact fit.\n");
          }
          continue;
        }
        if(found && found->limit > ln->limit) {
          if(debugme) {
            cdgPrintf(CDG_DEBUG, "got a better one already.\n");
          }
          continue;
        }
        if(debugme) {
          cdgPrintf(CDG_DEBUG, "OK.\n");
        }
        found = ln;
        exactfit = (orig->arc->from == ln->arc->from &&
                    orig->arc->from == ln->arc->from);
      
      }

      if (!found) {
        if(debugme) {
          abort();
        } else {
          cdgPrintf(CDG_DEBUG, "PANIC: couldn't lexicalize Patch, transcript follows...\n");
          pLexicalize(p,flags, TRUE);
        }
      }

      /* OK, we'll take it. */
      lgRequireLexeme(lg, p->forbiddenLexemes, found);
      
      /* clear all lexemes that overlap it... */
      j = found->arc->from;
      while(j < found->arc->to) {
        if(NULL != (ln = vectorElementAt(bound,j))) {
          coverage -= (ln->arc->to - ln->arc->from);
          for(k = ln->arc->from; k < ln->arc->to; k++) {
            vectorSetElementAt(bound, NULL, k);
          }
          j = ln->arc->to;
          if(debugme) {
            cdgPrintf(CDG_DEBUG, "unbinding ");
            lgPrintNode(CDG_DEBUG, ln);
            cdgPrintf(CDG_DEBUG, ", %d points bound\n", coverage);
            if(coverage > a->P->nrTimepoints) {
              cdgPrintf(CDG_DEBUG, "Something's not right!!\n");
              exit(1);
            }
          }
        } else {
          j++;
        }
      }
      
      /* ...and insert it. */
      for(j = found->arc->from; j < found->arc->to; j++) {
        vectorSetElementAt(bound, found, j);
        coverage++;
      }
      if(debugme) {
        cdgPrintf(CDG_DEBUG, "binding "); 
        lgPrintNode(CDG_DEBUG, found);
        cdgPrintf(CDG_DEBUG, ", %d points bound\n", coverage);
        if(coverage > a->P->nrTimepoints) {
          cdgPrintf(CDG_DEBUG, "Something's not right!!\n");
          abort();
        }
      }
      i = found->arc->to;
    }
  }

  /* translate array into lexeme List */
  i = a->P->minTimepoint;
  while(i < a->P->maxTimepoint) {
    ln = vectorElementAt(bound,i);
    if(ln != selectedAt(a, i)) {
      p->newLexemes = listPrependItem(p->newLexemes, ln);
    }
    if(!ln) {
      cdgPrintf(CDG_DEBUG, "PANIC: pLexicalize failed to bind %d!\n", i);
      abort();
    }
    i = ln->arc->to;
  }
  vectorDelete(bound);

  return TRUE;
}

/* ----------------------------------------------------------------------
   Does P imply a lexical change when applied to A?
*/
Boolean pIsLexical(Analysis a, Patch p) {

  int i;
  LevelValue lv;

  for(i = 0; i < vectorSize(p->selection); i++) {
    lv = vectorElementAt(p->selection, i);
    if(!lv) {
      continue;
    }
    if(!isValid(a, lv)) {
      return TRUE;
    }
  }
  return FALSE;

}

/* ----------------------------------------------------------------------
   If P were applied to A, would A then repeat a conflict from
   a->P->removedConflicts?
*/

Boolean detectCircle(Analysis a, Patch p, int flags) {

  Boolean result = FALSE;
  int i;
  ConstraintViolation cv = NULL;
  Analysis newA;
  int arity;

  /* these LVs are actually checked */
  LevelValue lv1 = NULL, lv2 = NULL;
  List l;

  /* We may have to construct an Analysis with a different path
     selected than a to see whether conflicts would be re-introduced
     or not. But that is only necessary if p introduces any new
     lexemes. */
  newA = aClone(a);

  /* preview what the Analysis would look like
     after swapping in the new lexemes */
  selectPath(newA, p->newLexemes);

  /* and the new LVs */
  for (i = 0; i < vectorSize(p->selection); i++) {
    LevelValue lv = vectorElementAt(p->selection, i);
    if(lv) {
      selectLV(newA, lv);
    }
  }
  
  /* check all recently removed conflicts */
  for (l = a->P->removedConflicts; l; l = listNext(l)) {
    cv = (ConstraintViolation) listItem(l);
    arity = listLength(cv->constraint->vars);
    
    /* find the corresponding LV. */
    lv1 = slotFiller(newA, cv->nodeBindingIndex1);
    
    /* check it */
    if (arity == 1 &&
        matchUnarySignature(lv1, cv->constraint->sig) &&
        !evalUnaryConstraint(cv->constraint, NULL, newA->slots, lv1) &&
        cv->constraint->penalty == cv->penalty) {
      result = TRUE;
      break;
    }
    
    /* binary conflict */
    else {

      /* Let lv1 and lv2 be the LVs that may or may not re-introduce cv. */
      lv2 = slotFiller(newA, cv->nodeBindingIndex2);

      if (lv2 && lvsCauseConflict(newA, lv1, lv2, cv)) {
        result = TRUE;
        break;
      }
    }
  }

  if (result && (flags & OD_DEBUG)) {
    pPrint(CDG_DEBUG, p, TRUE);
    cdgPrintf(CDG_DEBUG, " would re-violate ");
    cvPrint(CDG_DEBUG, cv);
    cdgPrintf(CDG_DEBUG, "\n");

  }
  
  aDelete(newA);
  return result;

}

/* Functions dealing with domains
   ============================== */

/* Domains are not a data type in their own. Each domain in a problem
   is simply a List of all LVs that may bind a specific level/time
   point pair. This means that a domain for time point 0 may contain
   LVs with the modifier wer(0,1) and also werden(0,2). Conversely,
   shallow copies of the same LV appear in several domains if its
   modifier spans more than one time interval. */


/* ----------------------------------------------------------------------
   Count remaining LVs in Problem.

   This function computes the field P->nrUndeletedLVs. Note that usually
   this field is maintained by the frobbing routines that may delete LVs. 
   This function must be called after a foreign routine has deleted LVs.

*/
void determineAmbiguity(Problem P) {
  int i;
  List l;
  
  P->nrUndeletedLVs = 0;
  for(i = 0; i < P->nrSlots; i++) {
    for(l = P->domains[i]; l; l = listNext(l)) {
      LevelValue lv = listItem(l);
      if(lv->level->useflag && !lv->isDeleted) {
        P->nrUndeletedLVs +=
          (lv->modifier->arc->to - lv->modifier->arc->from);
      }
    }
  }
}



/* ----------------------------------------------------------------------
   Print all LVs in a domain.
*/
void dPrint(unsigned long mode, Problem P, int where) {

  List m;
  LevelValue lv;
  Number limit = 0.0;
  Number f;

  List l = P->domains[where];

  if (l) {
    lv = (LevelValue) listItem(l);
    cdgPrintf(mode, "@%d: %d/%s\n", where, where / P->nrLevels, lv->level->id);
  } else {
    cdgPrintf(mode, "Empty domain.\n");
  }

  for (m = l; m; m = listNext(m)) {
    displayLV(P, listItem(m), 40);
    f = getLimit(P,listItem(m));
    if (f > limit)
      limit = f;
  }

  cdgPrintf(mode, "\n limit %4.3e\n", limit);
}

/* ----------------------------------------------------------------------
   Re-calculate limits for all LVs in a domain.

   This function propagates limits bidirectionally between LVs and lexeme
   nodes. It is a generalization of cnOptimizeNode(): the former function
   can only delete structures if they are absolutely impossible. This
   function can also detect implausible structures:

   The limit of an LV cannot exceed the limits of all its modifiers (or all
   of its modifiees).

   The limit of a lexeme node cannot exceed the limit of all LVs that bind
   it as a modifier. (Note that a lexeme node can appear in an Analysis
   without ever being a modifiee, so this second condition is not
   symmetrical.)

   Theoretically, the limit of a lexeme node cannot exceed the limits of
   all other lexemes that are needed to connect it to the start and end of
   the word graph. However, because this condition is expensive to compute,
   and is also approximated by the first two conditions, it is not
   enforced.

*/
void dCheck(Problem P, int which) {

  List l, m;
  Number *localLimit;
  LevelValue lv;
  LexemNode ln;
  int i, no, from;
  Number f;
  Boolean change;
  
#ifdef DEBUGDCHECK
    cdgPrintf(CDG_DEBUG, "\nDEBUG: dCheck(@%d)\n", which);
#endif
  /* empty domain ? */
  if (!P->domains[which])
    return;
 
  localLimit=memMalloc(sizeof(Number)*P->nrLexemes);
  from = which / P->nrLevels;

  /* localLimits start out at 0 */
  for (i = 0; i < P->nrLexemes; i++) {
    localLimit[i] = 0.0;
  }

  /* propagate limit of LVs to lexeme nodes */
  for (l = P->domains[which]; l; l = listNext(l)) {
    lv = listItem(l);
    f = getLimit(P,lv);

#ifdef DEBUGDCHECK
    cdgPrintf(CDG_DEBUG, "DEBUG: #%d has limit %4.3e\n", lv->indexWRTNet, f);
#endif
    if (significantlyGreater(P->globalMinimum, f) && !lv->isDeleted) {
      lv->isDeleted = TRUE;
      P->nrUndeletedLVs -= (lv->modifier->arc->to - lv->modifier->arc->from);
      continue;
    }

    for (m = lv->modifiers; m; m = listNext(m)) {
      no = ((LexemNode) listItem(m))->no;
      if (f > localLimit[no]) {
        localLimit[no] = f;
#ifdef DEBUGDCHECK
    cdgPrintf(CDG_DEBUG, "DEBUG: locallimit[%%%d] = %4.3e\n", no, f);
#endif
      }
    }
  }

  /* lower limits of lexeme nodes that have too little support */
  change = FALSE;
  for (i = 0; i < P->lgWidth; i++) {
    ln = vectorElementAt(P->lexemes, from * P->lgWidth + i);
    if(!ln) {
      break;
    }
    if (localLimit[ln->no] < ln->limit) {
      change = TRUE;
      ln->limit = localLimit[ln->no];
#ifdef DEBUGDCHECK
    cdgPrintf(CDG_DEBUG, "DEBUG: %%%d->limit  = %4.3e\n", ln->no, localLimit[ln->no]);
#endif
      if (significantlyGreater(P->globalMinimum, ln->limit)) {
#ifdef DEBUGDCHECK
        cdgPrintf(CDG_DEBUG, "DEBUG: deleting ");
        lgPrintNode(CDG_DEBUG, ln);
        cdgPrintf(CDG_DEBUG, "\n");
#endif
        lgDeleteNode(P->lg, ln);
      }
    }
  }

  /* propagate limits from lexeme nodes to LV, and re-sort domain */
  l = P->domains[which];
  P->domains[which] = NULL;
  for (m = l; m; m = listNext(m)) {
    updateLimit(P,listItem(m));
    P->domains[which] =
      listInsertSorted(P->domains[which], listItem(m), lvCompareByLimit);
  }
  listFree(l);
  memFree(localLimit);
}

/* ----------------------------------------------------------------------
   Optimize several domains after another.

   This function behaves like separate calls to dCheck() with all
   the integers in which as arguments, but is more efficient
   because the cleanup code only runs once.
*/
void dCheckAll(Problem P, List which) {

  List l;

  for(l = which; l != NULL; l = listNext(l)) {
    dCheck(P, (int)listItem(l));
  }

  listFree(which);
  
  if(P->net) {
    cnOptimizeNet(P->net);
  }
  /* Since this may have deleted LVs,
     we must recount the remaining ones. */
  determineAmbiguity(P);

  /* And since deleting lexeme nodes may change the distance between
     nodes, which may change the score of constraint violations,
     which may change the score of the current Analysis, we should
     re-judge it here. */
  aJudge(P->current);
  aJudge(P->best);
  aJudge(P->annotation);
}


/* ----------------------------------------------------------------------
   How many undeleted LVs does the domain WHERE contain?
*/
int dSize(Problem P, int where) {

  int result = 0;
  List l;

  for (l = P->domains[where]; l; l = listNext(l)) {
    if (!((LevelValue) listItem(l))->isDeleted) {
      result++;
    }
  }

  return result;

}

/* ----------------------------------------------------------------------
   Search specified domains for good combined assignments.

   The domains searched are the ones whose number is contained in WHICH.
   See below, however, about the OD_INT flag.
   
   This executes a local search among all LVs of the domains in question. 
   Care is taken not to form tuples that are physically incompatible, nor
   is the same LV ever added twice to the same tuple, even if WHICH
   contains duplicates.
   
   The result is a List of all possible Patches, ordered by pCompare().
   Only the fields ->local and ->limit are computed, not the field
   ->predicted (see optimizeDomains() for a way to do that).
   
   FLAGS specifies particular kinds of behaviour. The following flags are
   defined:

   OD_INT: the argument WHICH is not a Vector, but an integer
   specifying the one domain to be searched. Not pretty, but it saves
   us the allocation of a Vector for most calls.

   OD_PREDICT: choose the Patch with the best predicted Badness.

   OD_LIMIT: return the Patch with the best limit (default).

   OD_SINGLE: return only unary Patches even when multiple domains are
   specified.

   OD_ORIENT: allow Patches that change the direction of edges.

   OD_LEXICAL: allow Patches that change lexical items.

   OD_LABEL: allow Patches that change labels.

   OD_ALLOWALL: a synonym for specifying all of the three previous flags.
   This ensures that no Patch is rejected for one of its LVs.

   OD_DELETED: allow Patches that contain deleted LVs or lexemes.

   Note that OD_DELETED only re-allows LVs that are present, but were
   marked as deleted during the preceding parsing process. It can not
   magically revive LVs that were rejected when the constraint net was
   first built. Therefore, LVs that violate hard unary constraints are
   usually not found even with OD_DELETE.

   If you specify OD_DELETED to a function in the toolchain, you must also
   pass it to all later functions that operate on the resulting Patches, to
   avoid spooking the internal consistency checks.

   OD_PRUNE: delete LVs that do not appear in any of the possible Patches.

   OD_NARROW: ignore Patches with suboptimal limit.

   OD_INTERACTIVE: ask the user to select a Patch instead of computing the
   result autonomously.
    
   OD_DEBUG: give overwhelming amount of output.

   OD_CONTEXT: use context-sensitive constraints for evaluating a Patch.
   This will give wrong results unless WHICH actually contains the entire
   context.
   
   OD_IMMEDIATE: return immediately if a Patch is found that improves A
   beyond a->P->best->b.

   Note that these flags are used by various other functions that ultimately
   call or are called by dSearch(), such as optimizeDomains() and pNext(),
   and not all flags are used by all of them. For instance, OD_IMMEDIATE
   is read only by optimizeDomains() since it implies Patch previewing.

   The set of possible transformations is generated recursively, using an
   Agenda with size localSize. Therefore a simultaneous search in many
   domains will usually be incomplete.

   If the search was not truncated (and OD_PRUNE is set), the result of the
   search is used to propagate limits to LVs from all domains searched. 
   Each LV is assigned a limit that is not higher than the best limit of a
   Patch it appears in.
   
*/
List dSearch(Analysis a, Vector which, int flags) {

  Agenda ag;
  LevelValue lv;
  List l = NULL;
  List result = NULL;
  Number *localLimits=memMalloc(sizeof(Number)*a->P->nrLVs);
  Number bestLimit = 0.0;
  Patch p, newP;
  int i, j, index, width;
  double nrCombinations = 1;

  /* Do not trust the caller; several other invocation options
     would render pruning invalid, so test them explicitly. */
  Boolean doPruning =
    (flags & OD_PRUNE) &&
    (flags & OD_ALLOWALL) == OD_ALLOWALL;
  LevelValue current = NULL;
  Boolean verbose = (flags & OD_INTERACTIVE) != 0;
  Boolean debug = (flags & OD_DEBUG) != 0;

  Vector context = NULL;
  if(flags & OD_INT) {
    width = 1;
    nrCombinations = dSize(a->P, (int)which);
  } else {
    width = vectorSize(which);
    
    /* preview how much effort it's going to be */
    for (i = 0; i < vectorSize(which); i++) {
      
      /* this is the number of the domain */
      index = (int)vectorElementAt(which, i);
      l = listPrependItem(l, (Pointer)index);
      
      /* this is the current size of the domain,
         which gives the actual complexity of searching it now */
      nrCombinations *= dSize(a->P, index);
    }
  }

  if(debug) {
    cdgPrintf(CDG_DEBUG, "\nDEBUG: dSearch(");
    if(flags & OD_INT) {
      cdgPrintf(CDG_DEBUG, "@%d", (int)which);
    } else {
      for(i = 0; i < vectorSize(which); i++) {
        j = (int)vectorElementAt(which, i);
        cdgPrintf(CDG_DEBUG, "@%d", j);
        if(i < vectorSize(which)-1) {
          cdgPrintf(CDG_DEBUG, " ");
        }
      }
    }
    cdgPrintf(CDG_DEBUG, ")\n");
  }
  
  if (verbose) {
    cdgPrintf(CDG_INFO, "searching %.0f combination%s\n",
              nrCombinations, nrCombinations > 1 ? "s" : "");
  }

  if (doPruning) {

    dCheckAll(a->P, l);

    /* table of locally determined limits */
    for (i = 0; i < a->P->nrLVs; i++) {
      localLimits[i] = 0.0;
    }
  }

  /* conduct local search */
  ag = agNew(localSize, pDelete);
  agSetVerbosity(ag, verbose);
  cdgAgInsert(ag, 1.0, pNew(a, width, flags));

  while (!agIsEmpty(ag)) {

    /* read */
    p = (Patch) agRemoveBest(ag);

    if(debug) {
      cdgPrintf(CDG_DEBUG, "DEBUG: read ");
      pPrint(CDG_DEBUG, p, TRUE);
    }

    /* time limit reached? */
    if(timerExpired) {
      pDelete(p);
      continue;
    }

    /* interrupt? */
    if (cdgCtrlCTrapped) {
      pDelete(p);
      continue;
    }

    /* Should we delete Patches with suboptimal limits or not?

    This is one of the truly important parts of the module.
    Basically, we only want to investigate Patches with the optimal
    limit, because that is the idea behind having limits. That is
    the first line of the condition.

    Also, a Patch with a suboptimal limit may still be useful for
    assessing the limits of the LVs contained in it. So we only
    want to suppress Patches if we already know that that stage
    will be skipped later on -- because the agenda overflowed, or
    because the user has told us not to. That is the second line of
    the condition. */
    if (flags & OD_LIMIT &&
        p->limit < bestLimit &&
        (agIsTruncated(ag) || !doPruning)) {
      if(debug) {
        cdgPrintf(CDG_DEBUG, "delete because of insufficient limit.\n");
        pPrint(CDG_DEBUG, p, TRUE);
      }
      pDelete(p);
      continue;
    }

    /* Since in general word graphs, one lexeme can cover multiple
       intervals, one LV can bind multiple domains. It is therefore
       possible that we have already bound the next domain ordered. */

    /* find the next domain to bind */
    while (p->index < width) {
      /* next domain ordered */
      if(flags & OD_INT) {
        i = (int)which;
      } else {
        i = (int)vectorElementAt(which, p->index);
      }
      /* bound already? */
      if (!bvElementAt(p->bound, i)) {
        break;
      } else {
        /* skip it */
        vectorSetElementAt(p->selection, NULL, p->index++);
      }
    }

    /* retain */
    if (p->index == width) {

      if(debug) {
        cdgPrintf(CDG_DEBUG, "DEBUG: retained ");
        pPrint(CDG_DEBUG, p, TRUE);
      }

      /* apply context-sensitive constraints */
      if(flags & OD_CONTEXT) {
        context = vectorNew(a->P->nrSlots);

        for(i = 0; i < vectorSize(p->selection); i++) {
          LevelValue lv = vectorElementAt(p->selection, i);
          if(lv) {
            vectorSetElementAt(context, lv, lvIndex(lv));
          }
        }

        evalInContext(p->selection, context, p->local, NULL);
        vectorDelete(context);

        /* maybe update limit */
        if (p->local->soft < p->limit) {
          p->limit = p->local->soft;
        }
        if (p->local->hard) {
          p->limit = 0.0;
        }

      }
      
      result = listInsertSorted(result, p, pCompare);
      if (p->limit > bestLimit) {
        if (verbose) {
          cdgPrintf(CDG_INFO, "INFO: Patch with limit %4.3e found\n", p->limit);
        }
        bestLimit = p->limit;
      }
      continue;
    }

    /* extend */
    if(flags & OD_INT) {
      i = (int)which;
    } else {
      i = (int)vectorElementAt(which, p->index);
    }
    if(debug) {
      cdgPrintf(CDG_DEBUG, "DEBUG: considering @%d (length %d)\n",
                i, listLength(a->P->domains[i]));
    }

    current = slotFiller(a,i);
    for (l = a->P->domains[i]; l; l = listNext(l)) {
      lv = listItem(l);
      if(debug) {
        cdgPrintf(CDG_DEBUG, "DEBUG: Trying #%d... ", lv->indexWRTNet);
      }
      /* reject deleted LV */
      if((flags & OD_DELETED) == 0 && lv->isDeleted) {
        if(debug) {
          cdgPrintf(CDG_DEBUG, "is deleted\n", lv->indexWRTNet);
        }
        continue;
      } else {
        if(debug) {
          cdgPrintf(CDG_DEBUG, "\n");
        }
      }

      if ((newP = pNext(p, lv, flags)) != NULL) {

        /* is it an actual change, or merely a substitution X --> X? */
        if (lv != current) {
          newP->nrNewLVs++;
        }
        
        /* write */
        cdgAgInsert(ag, newP->limit, newP);
        
        if(debug) {
          cdgPrintf(CDG_DEBUG, "DEBUG: wrote ");
          pPrint(CDG_DEBUG, newP, TRUE);
        }
      } 
    }

    pDelete(p);

  }


  /* Now, propagate limits to all LVs in the searched domains. */  
  if (doPruning && !cdgCtrlCTrapped && !agIsTruncated(ag)) {

    if (verbose) {
      cdgPrintf(CDG_INFO, "%d of %.0f combination%s possible\n",
                listLength(result),
                nrCombinations, nrCombinations > 1 ? "s" : "");
    }

    /* update global limit */
    if (a->P->globalMaximum > bestLimit) {
      a->P->globalMaximum = bestLimit;
      cdgPrintf(CDG_INFO, "INFO: it won't get any better than %4.3e.\n",
                a->P->globalMaximum);
    }

    /* if an LV appears in a Patch with limit f,
       then its limit may be as high as that */
    for (l = result; l; l = listNext(l)) {
      p = (Patch) listItem(l);
      if(p->index < width) continue;
      for (i = 0; i < vectorSize(p->selection); i++) {
        lv = (LevelValue) vectorElementAt(p->selection, i);
        if (!lv)
          continue;
        index = lv->indexWRTNet;
        
        if (localLimits[index] < p->limit) {
          if(debug) {
            cdgPrintf(CDG_DEBUG,
                      "DEBUG: #%d can attain %4.3e\n",
                      index, p->limit);
          }
          localLimits[index] = p->limit;
        }
      }
    }

    /* degrade LVs that do not cooperate */
    for (i = 0; i < width; i++) {
      if(flags & OD_INT) {
        index = (int)which;
      } else {
        index = (int)vectorElementAt(which, i);
      }
      for (l = a->P->domains[index]; l; l = listNext(l)) {
        lv = (LevelValue) listItem(l);
        if (lv->isDeleted)
          continue;

        j = lv->indexWRTNet;

        if (localLimits[j] < lv->limit) {
          if(debug) {
            cdgPrintf(CDG_DEBUG,
                      "DEBUG: limit(#%d) = %4.3e\n",
                      lv->indexWRTNet, localLimits[j]);
          }
          downgrade(a->P, lv, localLimits[j]);
        }
      }
    }
  }

  if (verbose) {
    cdgCtrlCTrapped = FALSE;
  }

  agDelete(ag);
  memFree(localLimits);

  return result;

}

/* ----------------------------------------------------------------------
   Generate a Patch that swaps in LN and changes nothing else.
*/
Patch makeLexicalPatch(Analysis a, LexemNode ln) {

  Patch result = pNew(a, 0, 0);
  result->newLexemes = listPrependItem(NULL, ln);
  lgRequireLexeme(a->P->lg, result->forbiddenLexemes, ln);
  return result;

}

/* ----------------------------------------------------------------------
   Optimise n domains in the context of A.

   Performs a dSearch() on the specified domains and applies pJudge() to
   some or all of the resulting patches to decide which should be applied.

   The result is the Patch that is most suitable in the context of the
   current Problem. NULL may be returned if no Patch is possible. 

   If the returned Patch is not NULL, it has its field `predicted' set, and
   the value is an accurate prediction of the Badness of A if that Patch
   were applied to it. It has also been successfully been lexicalized and
   can immediately be applied to A.
   
   Various additional restrictions are obeyed. The variable beamWidth
   limits the number of patches that are considered at all. If a Patch
   fails the detectCircle() test, it is discarded. If OD_IMMEDIATE is set,
   the first Patch that improves on P->best->b stops further previewing. If
   OD_INTERACTIVE is set, the user is asked for a decision via stdout.
   
   Other flags works as in dSearch().
*/
Patch optimizeDomains(Analysis a, Vector which, int flags) {

  Badness predictedOptimum = worstBadness();
  List l, patches = NULL;
  Patch p, result;

  char buffer[BUFFERSIZE];
  Number bestLimit = 0.0;
  int bestPatch = 0;
  int i,no = 0;

  /* create only single-spot Patches */
  if(flags & OD_SINGLE &&
     !(flags & OD_INT)) {
    int i;
    for(i = 0; i < vectorSize(which); i++) {
      int where = (int)vectorElementAt(which, i);
      patches = listAppendList(patches, dSearch(a, (Pointer)where, flags | OD_INT));
    }
  }

  /* create Patches by combining LVs from the appropriate domains */
  else {
    patches = dSearch(a, which, flags);
  }

  /* consider all patches... */
  for (l = patches; l; l = listNext(l)) {
    no++;

    /* ...within beam width */
    if(no > beamWidth) {
      break;
    }

    /* react to user break */    
    if (cdgCtrlCTrapped) {
      cdgPrintf(CDG_INFO, "\nmaybe that analysis took a little long...\n");
      break;
    }

    p = (Patch) listItem(l);

    /* if an LV in the Patch was deleted by dSearch itself,
       disregard the entire Patch */
    for(i = 0; i < vectorSize(p->selection); i++) {
      LevelValue lv = vectorElementAt(p->selection, i);
      if(lv && lv->isDeleted && (flags & OD_DELETED) == 0) {
        i = -1;
        break;
      }
    }
    if(i == -1) {
      continue;
    }
    
    /* obey narrowSearch flag */
    if ((flags & OD_NARROW) && p->limit < bestLimit && p->nrNewLVs > 1) {
      continue;
    }

    /* pLexicalize() only fails if the lexeme graph has become
       disconnected. In this case all further calls to pLexicalize() will
       also fail, so we can immediately give up. */
    if(!pLexicalize(p, flags, FALSE)) {
      break;
    }
  
    if(flags & OD_DEBUG) {
      cdgPrintf(CDG_DEBUG, "judging Patch ");
      pPrint(CDG_DEBUG, p, TRUE);
    }

    /* do not consider Patches that would re-introduce conflicts */
    if (detectCircle(a, p, flags)) {
      continue;
    }

    /* update the best limit of patches we have found so far */    
    if (p->limit > bestLimit) {
      bestLimit = p->limit;
    }

    pJudge(p);
    if (flags & OD_INTERACTIVE) {
      cdgPrintf(CDG_INFO, "\n%d) ", no);
      pPrint(CDG_INFO, p, TRUE);
    }

    /* Now for the tricky part: which of the Patches should be
       returned? It depends on the method. In classical minConflict
       style, it is the Patch that will result in the best immediate
       improvement. */
    if (flags & OD_PREDICT) {

      /* the best patch is the one with the best predicted Badness. */
      if (bCompare(p->predicted, predictedOptimum)) {
        *predictedOptimum = *p->predicted;
        bestPatch = no;
      }
    } else {

      /* the best patch is the one with the best limit and the best
         predicted Badness.

         But locally non-optimal patches are allowed if they yield an
         immediate global improvement. */

      if ((p->limit == bestLimit && bCompare(p->predicted, predictedOptimum))
          ||
          bCompare(p->predicted, a->P->best->b)) {
        *predictedOptimum = *p->predicted;
        bestPatch = no;
      }
    }


    /* if we have found a method to improve the Analysis globally;
       don't even bother with further previewing */
    if ((flags & OD_IMMEDIATE) &&
        bCompare(p->predicted, a->P->best->b)) {
      break;
    }
  }

  /* Decide which patch to return */
  result = NULL;
  if (patches) {
    if (0 == (flags & OD_INTERACTIVE)) {
      no = bestPatch;
    } else if (bestPatch > 0) {

      /* get input */
      cdgPrintf(CDG_INFO, "patch (%d): ", bestPatch);
      cdgFlush();
      cdgGetString(buffer, BUFFERSIZE);

      /* remove trailing whitespace */
      while (isspace((int)buffer[strlen(buffer) - 1])) {
        buffer[strlen(buffer) - 1] = '\0';
      }

      no = -1;
      if (!strcmp(buffer, "")) {
        no = bestPatch;
      } else {
        no = 0;
        sscanf(buffer, "%d", &no);
      }
    } else {
      no = 0;
    }
  }

  /* return Patch */
  if (no > 0 && no <= listLength(patches)) {
    result = pClone(listNth(patches, no));
  }

  /* free */
  listForeachFree(patches, pDelete);
  bDelete(predictedOptimum);

  return result;

}


/* Functions dealing with Problems
   =============================== */

/* ----------------------------------------------------------------------
   Commit to the current value of domain WHERE.

   All other LVs in the domain are marked deleted.
*/
void freezeDomain(Analysis a, int where) {

  LevelValue theOne = slotFiller(a, where);
  List l;
  cdgPrintf(CDG_DEBUG, "DEBUG: Freezing @%d, retaining #%d, deleting %d values.\n",
            where, theOne->indexWRTNet, dSize(a->P,where));

  for(l = a->P->domains[where]; l != NULL; l = listNext(l)) {
    LevelValue lv = listItem(l);
    if(lv != theOne &&
       !lv->isDeleted) {
      lv->isDeleted = TRUE;
      a->P->nrUndeletedLVs -= (lv->modifier->arc->to - lv->modifier->arc->from);
    }
  }
}

/* --------------------------------------------------------------------
   Detect conflicts introduced by repair.

   A conflict CV is `recent' if there is a `best' Analysis
   and CV is not in it.
*/
Boolean isRecent(Problem P, ConstraintViolation cv) {

  if(!P->best) {
    return FALSE;
  }

  return !cvContains(P->best->conflicts, cv);

}

/* ----------------------------------------------------------------------
   Detect unremovable conflicts.

   This simply returns TRUE if CV is in P->unremovable.
*/
Boolean isUnremovable(Problem P, ConstraintViolation cv) {

  return cvContains(P->unremovable, cv);

}


/* ----------------------------------------------------------------------
   Sort all conflicts in A by recentness and score.

   This is to ensure that conflicts introduced by a repair step will end up
   in the front of the list, and will thus be attacked before other
   conflicts.
*/
void sortConflicts(Analysis a) {

  List old = NULL, new = NULL;
  List l;

  for (l = a->conflicts; l; l = listNext(l)) {
    if (isRecent(a->P, listItem(l))) {
      new = listInsertSorted(new, listItem(l), cvCompare);
    } else {
      old = listInsertSorted(old, listItem(l), cvCompare);
    }
  }
  listFree(a->conflicts);
  a->conflicts = listAppendList(new, old);

}

/* ----------------------------------------------------------------------
   List all conflicts in an Analysis.

   cvPrint() is used to display conflicts.
*/
void printConflicts(unsigned long mode, Analysis a) {

  int counter;
  List l;
  ConstraintViolation cv;

  if (a->conflicts)
    cdgPrintf(mode, "\n");
  counter = 0;

  /* show all conflicts */
  for (l = a->conflicts; l; l = listNext(l)) {
    cv = (ConstraintViolation) listItem(l);
    cdgPrintf(mode, "%3d. ", counter++);
    cvPrint(mode, cv);

    /* mark if this conflict as old or new */
    if (isRecent(a->P, cv)) {
      cdgPrintf(mode, "    (recently introduced)");
    } else if (isUnremovable(a->P, cv)) {
      cdgPrintf(mode, "    (obstinate)");
    }
    cdgPrintf(mode, "\n");

  }

}

/* ----------------------------------------------------------------------
   Reset the transformation history info for a Problem.

   Deallocates the fields P->removedConflicts and P->history as well as
   their elements.

*/
void resetProblem(Problem P) {

  listForeachFree(P->removedConflicts, cvDelete);
  P->removedConflicts = NULL;

  listFree(P->history);
  P->history = NULL;

  sortConflicts(P->current);

}

/* ----------------------------------------------------------------------
   Free a remembered Problem state.
*/
void psDelete(ProblemState ps) {
  if(!ps) {
    return;
  }
  memFree(ps->LVLimits);
  bvDelete(ps->deletedLVs);
  memFree(ps->LnLimits);
  bvDelete(ps->deletedLns);
  vectorDelete(ps->valuesPerNode);
  memFree(ps);
}

/* ----------------------------------------------------------------------
   Take a snapshot of the Problem at the current time.

   This only remembers the state of deletion of LVs, lexeme nodes, etc.
   The ProblemState will no longer work if you add or delete any
   nodes later.
*/
ProblemState psNew(Problem P) {
  ProblemState result;
  ConstraintNet net = P->net;
  LexemGraph lg;
  int i;

  if(!net) {
    return NULL;
  }
  
  lg = net->lexemgraph;
  result = memMalloc(sizeof(ProblemStateStruct));
  result->LVLimits = memMalloc(sizeof(Number) * vectorSize(net->values));
  result->deletedLVs = bvNew(vectorSize(net->values));
  result->LnLimits = memMalloc(sizeof(Number) * vectorSize(lg->nodes));
  result->deletedLns = bvNew(vectorSize(lg->nodes));
  result->valuesPerNode = vectorNew(vectorSize(net->nodes));
  
  /* remember state of LVs */
  for(i = 0; i < vectorSize(net->values); i++) {
    LevelValue lv = vectorElementAt(net->values, i);
    result->LVLimits[i] = lv->limit;
    bvSetElementAt(result->deletedLVs, lv->isDeleted, i);
  }
  
  /* remember state of lexeme nodes*/
  for(i = 0; i < vectorSize(lg->nodes); i++) {
    LexemNode ln = vectorElementAt(lg->nodes, i);
    result->LnLimits[i] = ln->limit;
    bvCopy(result->deletedLns, lg->isDeletedNode);
  }

  /* remember value count in ConstraintNodes */
  for(i = 0; i < vectorSize(net->nodes); i++) {
    ConstraintNode cn = vectorElementAt(net->nodes, i);
    vectorSetElementAt(result->valuesPerNode, (Pointer)cn->noValidValues, i);
  }
  return result;
}

/* ----------------------------------------------------------------------
   Record the current state of the Problem for posterity.
*/
void rememberProblem(Problem P) {
  if(P->savedState) {
    psDelete(P->savedState);
  }
  P->savedState = psNew(P);
}


/* ----------------------------------------------------------------------
   Reset Problem to the remembered state.

   This function restores P and its underlying
   constraint net to the state that it had at the time when
   rememberProblem() was last called on it. 
*/
void restoreProblem(Problem P, ProblemState ps) {

  int i;
  if(!ps) {
    return;
  }

  if(!P->net) {
    return;
  }
  
  /* restore state of LVs */
  for(i = 0; i < vectorSize(P->net->values); i++) {
    LevelValue lv = vectorElementAt(P->net->values, i);
    lv->limit = ps->LVLimits[i];
    lv->isDeleted = bvElementAt(ps->deletedLVs, i);
  }

  /* restore state of lexeme nodes */
  bvCopy(P->lg->isDeletedNode, ps->deletedLns);
  for(i = 0; i < vectorSize(P->lg->nodes); i++) {
    LexemNode ln = vectorElementAt(P->lg->nodes, i);
    ln->limit = ps->LnLimits[i];
  }
  lgComputeNoOfPaths(P->lg);
  lgComputeDistances(P->lg);

  /* restore value counts in ConstraintNodes */
  for(i = 0; i < vectorSize(P->net->nodes); i++) {
    ConstraintNode cn = vectorElementAt(P->net->nodes, i);
    cn->noValidValues = (int)vectorElementAt(ps->valuesPerNode, i);
  }
}

/* ----------------------------------------------------------------------
   When the pressure specified by the user turns out too great for the
   problem, halve it and reset the net, because all the deletions done
   so far cannot be trusted any more.

   However, all is not lost; the partial solution found during the
   high-pressure search is likely to be useful for further search. */
Boolean reducePressure(Problem P) {

  resetProblem(P);
  P->globalMaximum = 1.0;

  if(P->globalMinimum < 1e-8) {
    cdgPrintf(CDG_PROFILE, "INFO: cannot reduce pressure further...\n");
    cdgPrintf(CDG_PROFILE, "INFO: perhaps there *is* no solution?\n");
    return FALSE;
  }

  P->globalMinimum /= 10.0;
  cdgPrintf(CDG_INFO, "INFO: Pressure too great, resetting to %4.3e\n", P->globalMinimum);
  restoreProblem(P, P->savedState);

  /* By restoring the entire Problem, we have repaired the lexeme
     graph and reinstated its distance information. Since this information
     is used by the predicate distance(), each and every constraint formula
     has to be re-evaluated now, even in unary constraints. */
  aJudge(P->current);
  aJudge(P->best);

  return TRUE;
} 

/* ----------------------------------------------------------------------
   Revert to the best known location in the search space;
   that is, make P->current a clone of P->best.
*/
void backTrace(Problem P) {

  if (P->current) {
    aDelete(P->current);
    P->current = NULL;
  }
  if (P->best) {
    P->current = aClone(P->best);
  }

  resetProblem(P);

}

/* ----------------------------------------------------------------------
   De-allocate a Problem.
*/
void deleteProblem(Problem P) {

  int i;

  if(P->levelActive) {
    bvDelete(P->levelActive);
  }

  if (P->lexemes) {
    vectorDelete(P->lexemes);
    P->lexemes = NULL;
  }

  if (P->domains) {
    for (i = 0; i < P->nrSlots; i++)
      listFree(P->domains[i]);
    memFree(P->domains);
    P->domains = NULL;
  }

  if (P->best) {
    aDelete(P->best);
    P->best = NULL;
  }

  if (P->current) {
    aDelete(P->current);
    P->current = NULL;
  }
  if (P->annotation) {
    aDelete(P->annotation);
    P->annotation = NULL;
  }

  listForeachFree(P->removedConflicts, cvDelete);
  listFree(P->history);
  listForeachFree(P->unremovable, cvDelete);

  psDelete(P->savedState);
  
  if (P->profileTime) {
    timerFree(P->profileTime);
  }

  memFree(P);

}

/* ----------------------------------------------------------------------
   Transform Analysis to Parse.

   This function returns a new decorated Parse built from A.
   It calls parseComputeStructure() to do the real work.

*/
Parse analysis2parse(Analysis a) {
  int i;
  Parse p = parseNew(NULL);
  LevelValue lv;
  Vector LVs;

  /* general info */  
  p->latticeId = a->P->lg->lattice->id;
  p->searchStrategy = strRegister("frobbing");
  p->comment = NULL;
  p->solutionNo = 1;
  p->noOfSolutions = 1;
  p->score = a->b->hard ? 0.0 : a->b->soft;
  p->optimal = (!significantlyGreater(a->P->globalMaximum, a->b->soft));

  LVs = vectorNew(0);
  for(i = 0; i < vectorSize(a->slots); i++) {
    lv = vectorElementAt(a->slots, i);
    if(!lv) {
      continue;
    }
    vectorAddElement(LVs, lv);
  }
  parseComputeStructure(p, LVs, a->isUnselectedLexeme);
  /* can't fail here, the grammar is definitely present */
  (void)parseDecorate(p, a->P->lg, NULL);
  vectorDelete(LVs);

  /* conflicts */
  parseEval(p);

  return p;

}

/* ----------------------------------------------------------------------
   Call parseVerify()
*/
void aVerify(Analysis a) {

  AnnoEntry ae = findAnnotation(a->P->lg->lattice);
  Parse p;
  if(!ae) {
    return;
  }
  p = analysis2parse(a);
  parseVerificationDelete(parseVerify(p, ae));
  parseDelete(p);
}



/* ----------------------------------------------------------------------
   Register a new globally best Analysis.
   
   This function registers a new globally optimal score found for an
   Analysis. It deallocates P->best and replaces it with a clone of
   P->current. If the Analysis carries a positive score, this is assigned
   to P->globalMinimum. If a conflict found in P->unremovable does not also
   appear in P->current->conflicts, it is removed from P->unremovable
   (since it was obviously not unremovable).

*/
void newMaximum(Analysis a) {

  List l, m;
  ConstraintViolation cv;
  int i;
  Parse p;
  Number f = 1.0;
  Problem P = a->P;

  P->lastSolFound = timerElapsed(P->profileTime);
  if(aValid(a) &&
     0.0 == P->firstSoftFound &&
     a->subStart == P->minTimepoint &&
     a->subStop == P->maxTimepoint) {
    P->firstSoftFound = P->lastSolFound;
  }

  /* register new record for Badness */
  if (P->best) {
    aDelete(P->best);
  }
  P->best = aClone(a);
  
  /* note new record for score */
  if (aValid(P->best)) {

    P->globalMinimum = P->best->b->soft;
    l = NULL;
    for(i = 0; i < P->nrSlots; i++) {
      if(!slotActive(a,i)) {
        continue;
      }
      l = listPrependItem(l, (Pointer)i);
    }
    dCheckAll(P, l);
  }

  /* groom the Analysis a bit */
  aCalibrate(P->best);

  /* If an `unremovable' conflict was removed anyway,
     it was obviously not unremovable.
     So we retract it from the List.  */
  m = NULL;
  for (l = P->unremovable; l; l = listNext(l)) {
    cv = listItem(l);
    if (!cvContains(P->best->conflicts, cv)) {
      cdgPrintf(CDG_DEBUG,
                "DEBUG: Hey, `%s' was removed after all...\n",
                cv->constraint->id);
      cvDelete(cv);
      f = min(f, cv->penalty);
    }
    
    /* By the same token, retract all softer conflicts from this list as
       well.
       If we were wrong about not being able to repair a hard conflict, we
       could also have been wrong about not being able to repair a soft
       conflict. */
    else if (cv->penalty > f) {
      cdgPrintf(CDG_DEBUG,
                "DEBUG: I'll take another shot at `%s'.\n",
                cv->constraint->id);
      cvDelete(cv);
    }

    /* yup, still obstinate */
    else {
      m = listPrependItem(m, cv);
    }
  }
  
  listFree(P->unremovable);
  P->unremovable = m;

  resetProblem(P);

  /* hand the Parse to Tcl for displaying */
  p = analysis2parse(P->best);
  cdgPrintf(CDG_INFO, "*");
  cdgExecHook(HOOK_PARTIALRESULT, p);
  parseRegister(p);

}

/* ----------------------------------------------------------------------
   Show remaining ambiguity of P.

   Prints out the average remaining ambiguity in the
   Problem, i.e. the ratio of remaining LVs to domains.
*/
void printAmbiguity(unsigned long mode, Problem P) {

  if (P->nrUndeletedLVs == P->nrActiveSlots) {
    cdgPrintf(mode, " (unique)");
  } else {
    cdgPrintf(mode, " ambiguity %.1f", (float)P->nrUndeletedLVs / P->nrActiveSlots);
  }

}

/* --------------------------------------------------------------------
   Show what was done recently.

   Displays the conflicts recently removed from the Analysis and the
   domains that were changed for that purpose.
*/
void printHistory(unsigned long mode, Problem P) {

  List l;

  /* show removedConflicts */
  if (P->removedConflicts) {
    cdgPrintf(mode, "\nrecently removed:\n");
    for (l = P->removedConflicts; l; l = listNext(l)) {
      cvPrint(mode, listItem(l));
      cdgPrintf(mode, "\n");
    }
    cdgPrintf(mode, "\n");
  }

  /* show history */
  if (P->history) {
    for (l = P->history; l; l = listNext(l)) {
      cdgPrintf(mode, "@%d ", (int)listItem(l));
    }
    cdgPrintf(mode, "\n\n");
  }
}

/* ----------------------------------------------------------------------
   Print all LVs of an Analysis.

   Displays all LVs in A in their natural order.
*/
void aPrint(unsigned long int mode, Analysis a) {
  int i, j, k;
  
  /* iterate over all values */
  for (i = 0; i < vectorSize(a->slots); i++) {
    LevelValue lv;
    if(!slotActive(a,i)) {
      continue;
    }
    lv = (LevelValue) vectorElementAt(a->slots, i);

    if (!lv) {
      if (mode != CDG_INFO) {
        cdgPrintf(mode, "@%d: <implicitly bound domain>\n", i);
      }
      continue;
    }

    /* LVs aren't shown if they have no competitors and cause no conflicts. */
    if (mode == CDG_INFO && isBoring(a, lv)) {
      continue;
    }
    
    /* align number to 6 characters */
    j = lvIndex(lv);
    for (k = 1000; k > j; k /= 10) {
      if (k > 1 || j != 0) {
        cdgPrintf(mode, " ");
      }
    }
    cdgPrintf(mode, "@%d: ", i, lv->indexWRTNet);
    j = lv->indexWRTNet;
    if (j < 0) {
      cdgPrintf(mode, "    ");
    } else {
      for (k = 100000; k > j; k /= 10) {
        if (k > 1 || j != 0) {
          cdgPrintf(mode, " ");
        }
      }
    }
    cdgPrintf(mode, "#%d ", j);
    if(lv->isDeleted) {
      cdgPrintf(mode, "[");
    }
    lvPrint(mode, lv, a->isUnselectedLexeme);
    if(lv->isDeleted) {
      cdgPrintf(mode, "]");
    }
    cdgPrintf(mode, "\n");
  }
}

/* --------------------------------------------------------------------
   Print current state of the problem.

   Displays the current state of the Problem. In particular, prints the
   current Analysis as well as the recent transformation history.
*/
void printProblem(unsigned long mode, Problem P) {

  aCalibrate(P->current);
  cdgPrintf(mode, "\n\n\n");

  aPrint(mode, P->current);
  
  /* print conflicts */
  printConflicts(mode, P->current);

  /* print Badness */
  cdgPrintf(mode, "\n");
  bPrint(mode, P->current->b);

  /* special case: suboptimal Badness during a repair step */  
  if (P->best && bCompare(P->best->b, P->current->b)) {
    cdgPrintf(mode, ", chasing ");
    bPrint(mode, P->best->b);
  }

  /* special case: optimal Badness reached */  
  if (P->current->b->hard == 0 &&
      !significantlyGreater(P->globalMaximum, P->current->b->soft)) {
    cdgPrintf(mode, " (optimal)");
    resetProblem(P);
  }

  /* some more measures about the current Analysis */  
  printAmbiguity(mode, P);
  cdgPrintf(mode, "\n");

  /* print history */
  printHistory(mode, P);

}

/* ----------------------------------------------------------------------
   Check Analysis for internal consistency, and abort with warning if
   an error is found.

   Different levels of checking are necessary so that aCheck() can be used
   during the construction of Analyses, when some of the structures checked
   here are incomplete because they haven't been built yet.
*/
void aCheck(Analysis a, int level) {

  int i, j, k;
  LevelValue lv;
  LexemNode lna, lnb;
  Analysis b;
  LexemGraph lg = a->P->lg;

  BitVector bound=bvNew(a->P->maxTimepoint);
  BitVector slots=bvNew(a->P->nrSlots);

  bvSetAllElements(slots, FALSE);

  /* invalid LV in the Analysis? */
  for (i = 0; i < vectorSize(a->slots); i++) {
    if(!slotActive(a,i)) {
      continue;
    }
    lv = vectorElementAt(a->slots, i);
    if (!lv)
      continue;

#if 0
    /* zoning violation? */
    if(lv->modifiee &&
       (lv->modifiee->arc->from < a->P->subStart ||
        lv->modifiee->arc->from >= a->P->subStop)) {
      cdgPrintf(CDG_INFO, "PANIC: illegal LV @%d/#%d points to %d rather than [%d..%d]\n",
                i, lv->indexWRTNet, lv->modifiee->arc->from, a->P->subStart, a->P->subStop);
      abort();
    }
#endif
    
    lvCheck(lv);
    checkValidity(a, lv);

    /* LV residing in wrong domain? */
    if ((j = lvIndex(lv)) != i) {
      cdgPrintf(CDG_ERROR, "PANIC: LV at @%d has lvIndex() of %d:\n\n", i, j);
      displayLV(a->P, lv, 40);
      abort();
    }

    for (j = lv->modifier->arc->from; j < lv->modifier->arc->to; j++) {
      k = a->P->nrLevels * j + lv->level->no;
      bvSetElementAt(slots, TRUE, k);
    }

  }

  /* path criterion */
  bvSetAllElements(bound, FALSE);

  for (i = 0; i < a->P->nrLexemes; i++) {
    lna = (LexemNode) vectorElementAt(lg->nodes, i);

    if (bvElementAt(a->isUnselectedLexeme, lna->no))
      continue;

    bvSetElements(bound, TRUE, lna->arc->from, lna->arc->to);

    for (j = i + 1; j < a->P->nrLexemes; j++) {
      lnb = (LexemNode) vectorElementAt(lg->nodes, j);

      if (bvElementAt(a->isUnselectedLexeme, lnb->no))
        continue;

      if (lgOverlap(lna, lnb)) {
        cdgPrintf(CDG_ERROR, "PANIC: ");
        lgPrintNode(CDG_ERROR, lna);
        cdgPrintf(CDG_ERROR, " and ");
        lgPrintNode(CDG_ERROR, lnb);
        cdgPrintf(CDG_ERROR, " both selected at time point %d\n",
                  lna->arc->from);
        abort();
      }
    }
  }

  /* skip Badness check */  
  if (level >= 2) {

    /* wrong Badness calculated? */
    b = aClone(a);
    aJudge(b);
    if (bCompare(a->b, b->b) || bCompare(b->b, a->b)) {

      cdgPrintf(CDG_ERROR, "PANIC: misevaluated Analysis!\n");
      aPrint(CDG_ERROR, b);
      cdgPrintf(CDG_ERROR, "computed conflicts=\n");
      cdgPrintf(CDG_ERROR, "\n");
      bPrint(CDG_ERROR, a->b);
      printConflicts(CDG_ERROR, a);
      cdgPrintf(CDG_ERROR, "actual conflicts=\n");
      cdgPrintf(CDG_ERROR, "\n");
      bPrint(CDG_ERROR, b->b);
      printConflicts(CDG_ERROR, b);
      abort();
    }
    aDelete(b);

    /* skip completeness check */
    if (level >= 3) {

      /* path criterion */
      for (i = a->subStart; i < a->subStop; i++) {
        if (!bvElementAt(bound,i)) {
          cdgPrintf(CDG_ERROR, "PANIC: time point %d not bound\n", i);
          abort();
        }
      }
      /* All domains bound? */
      for (i = 0; i < a->P->nrSlots; i++) {
        if(!slotActive(a,i)) {
          continue;
        }
        if (!bvElementAt(slots,i)) {
          cdgPrintf(CDG_ERROR, "PANIC: domain @%d not bound\n", i);
          printProblem(CDG_ERROR, a->P);
          abort();
        }
      }
    }
  }
  bvDelete(bound);
  bvDelete(slots);
}


/* ----------------------------------------------------------------------
   Build the fields P->lgWidth and P->lexemes.
*/
void buildLexemeTable(Problem P) {

  LexemNode ln;
  LexemGraph lg = P->lg;
  int i, j, k;


  P->lgWidth = lgWidth(lg);
  P->nrLexemes = vectorSize(lg->nodes);

  if(P->lexemes) {
    vectorDelete(P->lexemes);
  }

  P->lexemes = vectorNew(P->maxTimepoint * P->lgWidth);
  vectorSetAllElements(P->lexemes, NULL);
  for (i = 0; i < P->nrLexemes; i++) {
    ln = (LexemNode) vectorElementAt(P->lg->nodes, i);
    for (j = ln->arc->from; j < ln->arc->to; j++) {
      k = j * P->lgWidth;
      while (vectorElementAt(P->lexemes, k) != NULL) {
        k++;
      }
      vectorSetElementAt(P->lexemes, ln, k);
    }
  }

#ifdef DEBUGBUILDLEXEMETABLE
  cdgPrintf(CDG_DEBUG, "Lexeme table: lgwidth = %d\n", P->lgWidth);
  for(i = 0; i < vectorSize(P->lexemes); i++) {
    ln = vectorElementAt(P->lexemes, i);
    if (!ln) continue;
    cdgPrintf(CDG_DEBUG, "Cell %d (timeslot %d item %d): ",
              i,
              i / P->lgWidth,
              i % P->lgWidth
              );
    lgPrintNode(CDG_DEBUG, ln);
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif
  
}




/* ----------------------------------------------------------------------
   Add LVs to an Analysis so that it binds slots 0 through slot.
*/
void aExtend(Analysis a, int slot) {

  Patch p, q;
  LevelValue lv;
  List l;

  /* bound already? */
  if (slotFiller(a, slot) != NULL) {
    return;
  }

  /* select another LV.

  This is not as easy as it looks. For one thing, we must select a set
  of LVs that are at least physically compatible, which is non-trivial. 
  But the whole process must also be fast, so we don't want to employ
  backtracking. So we simply get one LV from the domain, and if it is
  incompatible with the previous choices, we _make_ it compatible. This
  always leads to a complete and consistent Analysis because all LVs
  that are physically impossible have already been deleted by
  cnOptimizeNet(). */

  lv = NULL;
  for (l = a->P->domains[slot]; l; l = listNext(l)) {
    lv = (LevelValue) listItem(a->P->domains[slot]);
    if(lv->isDeleted) {
      continue;
    }
    if (isValid(a, lv)) {
      break;
    }
  }

  if (!isValid(a, lv)) {
    lv = inventLV(a, slot);
  }

  p = pNew(a, 1, OD_DELETED);
  q = pNext(p, lv, OD_ALLOWALL | OD_DELETED);
  pDelete(p);
  (void)pLexicalize(q,OD_DELETED,FALSE);
  apply(a, q);

#ifdef DEBUGAEXTEND
  displayLV(a->P,lv, 40);
  cdgPrintf(CDG_DEBUG, "#%d binds", lv->indexWRTNet);
  for (j = lv->modifier->arc->from; j < lv->modifier->arc->to; j++) {
    k = P->nrLevels * j + lv->level->no;
    cdgPrintf(CDG_DEBUG, " @%d", k);
  }
  cdgPrintf(CDG_DEBUG, "\n");

  for (j = 0; j < vectorSize(a->slots); j++) {
    if (j % P->nrLevels == 0) {
      cdgPrintf(CDG_DEBUG, "\n");
    }
    if (vectorElementAt(a->slots, j) != NULL) {
      cdgPrintf(CDG_DEBUG, "X");
    } else if (slotFiller(a,j)) {
      cdgPrintf(CDG_DEBUG, "I");
    } else {
      cdgPrintf(CDG_DEBUG, "0");
    }
  }
  cdgPrintf(CDG_DEBUG, "\n");
#endif

}


/* ----------------------------------------------------------------------
   Put an LV into all appropriate domains, and count it.

   This function registers lv in all domains of P where it is
   appropriate. It also updates P->nrLVs and
   P->nrUndeletedLVs.
*/
void addLVToProblem(Problem P, LevelValue lv) {

  int k, index;

  /* put this LV into whatever domains need it */
  for (k = lv->modifier->arc->from; k < lv->modifier->arc->to; k++) {
    index = P->nrLevels * k + lv->level->no;
    P->domains[index] = listInsertSorted(P->domains[index], lv, lvCompare);
  }

  /* make sure that
     (a) LV has a non-negative number
     (b) P remembers to allocate at least this much space when
         reasoning about its LVs
  */
  if(lv->indexWRTNet < 0) {
    lv->indexWRTNet = P->nrLVs;
  }
  P->nrLVs++;


  /* The field nrUndeletedLVs is special because it counts, not
     the LVs, but the slots that the remaining LVs may bind. That
     way we can just divide nrUndeletedLVs by the number of all
     slots and get the exact ambiguity of the Problem.  */      
  if (!lv->isDeleted) {
    P->nrUndeletedLVs += (lv->modifier->arc->to - lv->modifier->arc->from);
  }
}
/* ----------------------------------------------------------------------
   Transform Parse to Analysis.

   Builds an Analysis from a decorated Parse. Returns NULL if p does not
   analyse the entire utterance modeled by the constraint net.
*/
Analysis parse2analysis(Problem P, Parse p) {
  Analysis result;
  LevelValue lv, model;
  int i;
  BitVector bv;
  
  /* Parse must be decorated */
  if(!p->LVs) {
    return(NULL);
  }

  /* Parse must contain the correct number of LVs */
  if(vectorSize(p->LVs) != P->nrSlots) {
    return NULL;
  }
  
  /* copy lexical selection */
  result = aNew(P);  
  bv = bvClone(P->lg->isDeletedNode);
  if(!lgCopySelection(P->lg, p->lg)) {
    return NULL;
  }
  bvCopy(result->isUnselectedLexeme, P->lg->isDeletedNode);
  bvCopy(P->lg->isDeletedNode, bv);
  bvDelete(bv);

  /* copy LVs */
  for(i = 0; i < vectorSize(p->LVs); i++) {
    lv = vectorElementAt(p->LVs, i);
    if(!lv || !lv->level->useflag) {
      continue;
    }
    
    model = translateLV(P, lv);

    if(!model) {
      aDelete(result);
      cdgPrintf(CDG_WARNING, "WARNING: net contains no such LV:\n");
      displayLV(P, lv, 0);
      return NULL;
    }
    vectorSetElementAt(result->slots, model, lvIndex(model));
  }

  /* Again, the Parse may not provide all required LVs,
     so we make sure to extend the Analysis as needed. */
  for(i = 0; i < vectorSize(result->slots); i++) {
    if(slotActive(result,i) && !slotFiller(result, i)) {
      LevelValue lv = inventLV(result, i);
      vectorSetElementAt(result->slots, lv, i);
    }
  }
  return result;

}

/* ----------------------------------------------------------------------
   Transform AnnoEntry to Analysis.

   Converts AE to a parse, decorates it, and calls parse2analysis().

   Either of there steps may fail so that NULL is returned.
*/
Analysis anno2analysis(Problem P, AnnoEntry ae) {

  Parse p;
  Analysis a = NULL;

  if(!ae) {
    cdgPrintf(CDG_WARNING, "WARNING: No annotation available.\n");
    return NULL;
  }
  
  /* Luckily, this is a solved problem:
     we let the parse module do the work for us. */
  if((p = parseFromAnno(ae)) &&
     parseDecorate(p, P->lg, ae)) {
    a = parse2analysis(P, p);
  }

  if(p) {
    parseDelete(p);
  }
  if(a) {
    return a;
  } else {
    cdgPrintf(CDG_WARNING, "WARNING: Could not build reference analysis!\n");
    return NULL;
  }
}


/* ----------------------------------------------------------------------
   Build a Problem with an initial Analysis from a net.

   This function allocates a new Problem structure and initializes it. In
   particular, it creates an initial Analysis for the Problem as
   P->current. This Analysis is guaranteed to be complete and acoustically
   consistent, but it may be ungrammatical. (At least it will be
   constructed quickly.) If this fails, NULL is returned.

   If NET already contains a Parse, then the initial Analysis will be
   constructed to match it.

   The parameter PRESSURE can be used to exert temporal pressure
   on the processing. It is assigned to the globalMinimum of the
   resulting Problem.

*/
Problem net2problem(ConstraintNet net, Number pressure) {

  Analysis A;
  ConstraintNode n;
  LevelValue lv = NULL;
  LexemGraph lg = net->lexemgraph;
  List l;
  Problem P = memMalloc(sizeof (ProblemStruct));
  Number minimum = 0.0;
  
  int i, j;

  P->id = strPrintf("problem%d", problem_counter++);
  P->net = net;
  P->lg = net->lexemgraph;
  P->globalMinimum = 0.0;
  P->judgeCounter = 0;

  P->firstSoftFound = 0.0;
  P->lastSolFound = 0.0;
  
  if(0 == gatherLevels(P)) {
    cdgPrintf(CDG_WARNING, "No levels active -- nothing to transform.\n");
    return NULL;
  }

  /* translate external time pressure into internal controlling
     variables. At the moment this is done by using the parameter
     `pressure' as a non-zero initialization for globalMinimum, but
     other methods could be invented. */
  minimum = pressure;

  P->minTimepoint = lg->min;
  P->maxTimepoint = lg->max;
  P->nrTimepoints = lg->max - lg->min;
  P->nrSlots = P->maxTimepoint * P->nrLevels;
  P->nrActiveSlots = P->nrTimepoints * P->nrActiveLevels;
  P->globalMaximum = 1.0;
  P->unremovable = NULL;
  P->removedConflicts = NULL;
  P->history = NULL;
  P->lexemes = NULL;
  P->annotation=NULL;
  P->savedState = NULL;
  
  /* build table of lexeme nodes */
  cdgPrintf(CDG_INFO, "Building table of lexemes...");
  cdgFlush();

  buildLexemeTable(P);
  cdgPrintf(CDG_INFO, " done.\n");

  /* build table of domains */
  cdgPrintf(CDG_INFO, "Building domains...");
  cdgFlush();
  P->domains = (List *) memMalloc(sizeof (List) * P->nrSlots);
  for (i = 0; i < P->nrSlots; i++) {
    P->domains[i] = NULL;
  }

  /* populate domains with precomputed LVs */
  P->nrLVs = P->nrUndeletedLVs = 0;
  for (i = 0; i < vectorSize(P->net->nodes); i++) {
    n = (ConstraintNode) vectorElementAt(P->net->nodes, i);
    if(!n->level->useflag) {
      continue;
    }
    for (j = 0; j < vectorSize(n->values); j++) {
      lv = vectorElementAt(n->values, j);
      addLVToProblem(P, lv);
    }
  }

  P->current = NULL;
  P->best = NULL;

  /* If a domain remains empty, no solution is possible. */
  for (i = 0; i < P->nrSlots; i++) {
    if(!bvElementAt(P->levelActive, i % P->nrLevels)) {
      continue;
    }
    dCheck(P, i);
    if (P->domains[i] == NULL) {
      Level level = listNth(inputCurrentGrammar->levels, 1 + (i % P->nrLevels));
      LexemNode ln = vectorElementAt(P->lexemes, P->lgWidth * i / P->nrLevels);
      cdgPrintf(CDG_WARNING,
                "\nWARNING: empty domain `%s'/%s, problem has no solution.\n",
                ln->arc->word, level->id);
      return NULL;
    }
  }
  /* This is important because later on we rely on each domain having
     at least one value. */
  cdgPrintf(CDG_INFO, " done.\n");

  /* initial analysis */
  cdgPrintf(CDG_INFO, "Building initial analysis...");
  cdgFlush();

  A = aNew(P);

  P->current = A;
  P->best = NULL;

  vectorSetAllElements(A->slots, NULL);

  /* build an arbitrary path through the graph. */
  bvSetAllElements(A->isUnselectedLexeme, TRUE);

  /* Since the calls to dCheck() in the previous step may have deleted
     lexemes, we have to re-optimize the net so that lgMakePath will
     not select dangling lexemes. */
  if(cnOptimizeNet(net) < 0) {
    deleteProblem(P);
    return NULL;
  };

  l = lgMakePath(net->lexemgraph, NULL);
  /* This can only fail if there is no path though the lattice at all.
     Ideally, CDG should have refused to build a net for such a graph,
     or the lattice should have been rejected at load time. But just
     in case an invalid lexeme graph sneaks through here, test it
     anyway. */
  if (!l) {
    cdgPrintf(CDG_WARNING,
              "\nWARNING: no path through the lexeme graph possible, bailing out...\n");
    return NULL;
  }

  selectPath(A, l);
  listFree(l);
  /* There is no guarantee that this path supports an analysis. We
     know that each of the lexemes now selected can be bound at every
     level, but we might still encounter situations such as this one:

     -- A1 --   -- B1 --   -- C1 --
     /        \ /        \ /        \
     o         o          o          o
     \        / \        / \        /
     -- A2 --   -- B2 --   -- C2 --

     where the only permissible LVs are:

     A2-->B1
     A1-->B2
     B1-->NIL
     B2-->NIL
     C1-->B1
     C2-->B2

     If we arbitrarily select the path A1B1C1, we cannot build the
     initial Analysis out of existing LVs.

     On the other hand, we don't want to start swapping lexemes around
     while building the initial Analysis. For one, that is dreadfully
     expensive. Also, we could then get into trouble with our
     bookkeeping about which domains are already bound. Therefore we
     simply select the first valid LV from each unbound domain, and if
     there is none, we invent one out of thin air. */

  /* iterate over all slots */
  for (i = 0; i < P->nrSlots; i++) {
    if(!slotActive(P->current,i)) {
      continue;
    }
    aExtend(A, i);
  }

  aCheck(A, 3);

  cdgPrintf(CDG_INFO, " done.\n");

  /* compute conflicts */
  cdgPrintf(CDG_INFO, "Computing conflicts...");
  cdgFlush();
  aJudge(P->current);

  cdgPrintf(CDG_INFO, " done.\n");

  /* Now clone the completed Analysis */
  P->best = aClone(P->current);

  /* we *might* just have constructed a valid Analysis by sheer luck... */
  if (aValid(P->current) && P->current->b->soft > P->globalMinimum) {
    P->globalMinimum = P->current->b->soft;
  }

  /* build another Analysis that corresponds to an annotation we may
     have available */
  P->annotation = anno2analysis(P, findAnnotation(net->lexemgraph->lattice));

  /* Judge the annotation so it won't confuse the optimization
     with its flashy score of 1.0 */
  aJudge(P->annotation);
  

  /* now apply pressure to the problem */
  P->globalMinimum = minimum;

  
  P->profileTime = timerNew();
  
  resetProblem(P);
  rememberProblem(P);
  newMaximum(P->current);

  /* register problem */
  hashPut(cdgProblems, P->id, P);

  return P;

}

/* ----------------------------------------------------------------------
   Create a Problem without a constraint net, just from a decorated parse.

   A constraint net will be built, but it will only contain very few values
   and be unsuitable for real parsing; it is just useful for being able to
   use the transform module at all.
*/
Problem parse2problem(Parse p, Number pressure) {

  Analysis temp;
  Problem P;
  Number minimum = 0.0;
  int index = 0;
  int i, j;
  
  if(parseIsAbstract(p)) {
    cdgPrintf(CDG_ERROR, "ERROR: need a decorated parse.\n");
    return NULL;
  }
  
  P = memMalloc(sizeof (ProblemStruct));
  P->id = strPrintf("problem%d", problem_counter++);
  P->net = NULL;
  P->lg = p->lg;
  P->globalMinimum = 0.0;
  P->judgeCounter = 0;

  P->firstSoftFound = 0.0;
  P->lastSolFound = 0.0;
  
  
  if(0 == gatherLevels(P)) {
    cdgPrintf(CDG_WARNING, "No levels active -- nothing to transform.\n");
    return NULL;
  }

  /* translate external time pressure into internal controlling
     variables. At the moment this is done by using the parameter
     `pressure' as a non-zero initialization for globalMinimum, but
     other methods could be invented. */
  minimum = pressure;
 
  if(p->score > minimum) {
    minimum = p->score;
  }

  P->minTimepoint = P->lg->min;
  P->maxTimepoint = P->lg->max;
  P->nrTimepoints = P->lg->max - P->lg->min;
  P->nrSlots = P->maxTimepoint * P->nrLevels;
  P->nrActiveSlots = P->nrTimepoints * P->nrActiveLevels;
  P->globalMaximum = 1.0;
  P->unremovable = NULL;
  P->removedConflicts = NULL;
  P->history = NULL;
  P->lexemes = NULL;
  P->annotation=NULL;
  P->savedState = NULL;
  
  /* build table of lexeme nodes */
  cdgPrintf(CDG_INFO, "Building table of lexemes...");
  cdgFlush();

  buildLexemeTable(P);
  cdgPrintf(CDG_INFO, " done.\n");

  /* build table of domains */
  cdgPrintf(CDG_INFO, "Building domains...");
  cdgFlush();
  P->domains = (List *) memMalloc(sizeof (List) * P->nrSlots);
  for (i = 0; i < P->nrSlots; i++) {
    P->domains[i] = NULL;
  }

  cdgPrintf(CDG_INFO, " done.\n");

  P->best = NULL;
  temp = parse2analysis(P, p);
  if(!temp) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot transform `%s' to Analysis\n", p->id);
    return NULL;
  }

  /* populate domains */
  P->nrLVs = P->nrUndeletedLVs = 0;
  for(i = 0; i < P->nrSlots; i++) {
    LevelValue lv = slotFiller(temp, i);
    int from = lv->modifier->arc->from;

    /* add one LV for each modifier variant */
    for (j = 0; j < P->lgWidth; j++) {
      LexemNode ln = vectorElementAt(P->lexemes, P->lgWidth * from + j);
      LevelValue alt;
      if(ln) {

        alt = lvClone(lv);
        alt->modifier = ln;
        listFree(alt->modifiers);
        alt->modifiers = listPrependItem(NULL, ln);
        alt->indexWRTNet = index++;
        alt->score = evalUnary(alt, NULL, NULL, FALSE, NULL, NULL);
        alt->limit = alt->score;
        addLVToProblem(P, alt);
      }
    }
  }
  aDelete(temp);
  P->current = parse2analysis(P, p);
  
  /* compute conflicts */
  cdgPrintf(CDG_INFO, "Computing conflicts...");
  cdgFlush();
  aJudge(P->current);
  aCheck(P->current, 3);
  cdgPrintf(CDG_INFO, " done.\n");

  /* Now clone the completed Analysis */
  P->best = aClone(P->current);

  if (aValid(P->current) && P->current->b->soft > P->globalMinimum) {
    P->globalMinimum = P->current->b->soft;
  }

  /* build another Analysis that corresponds to an annotation we may
     have available */  
  P->annotation = anno2analysis(P, findAnnotation(P->lg->lattice));

  /* Judge the annotation so it won't confuse the optimization
     with its flashy score of 1.0 */
  aJudge(P->annotation);

  /* now apply pressure to the problem */
  P->globalMinimum = minimum;
 
  P->profileTime = timerNew();
  resetProblem(P);
  rememberProblem(P);
  newMaximum(P->current);

  /* register problem */
  hashPut(cdgProblems, P->id, P);
  
  return P;

}

/* ----------------------------------------------------------------------
   Register new Parse (if it is really new).

   Converts P->best to a Parse and adds it to the constraint net underlying
   P. This is only done if the net does not already contain an identical
   Parse.

*/
void addParse(Analysis a) {

  List l;
  Parse parse;
  Problem P = a->P;

  if(!a) {
    return;
  }

  /* only register solutions that aren't already known */  
  parse = analysis2parse(a);
  if(P->net) {
    for(l = P->net->parses; l != NULL; l = listNext(l)) {
      if(parseEquals(listItem(l), parse)) {
        parseDelete(parse);
        return;
      }
    }
    
    /* add to net */
    P->net->parses = 
      listPrependItem(P->net->parses, parse);
  }

  /* add to global list */
  parseRegister(parse);
  cdgPrintf(CDG_INFO,
            "INFO: parse %s added to the global parses list.\n",
            parse->id);

}


/* ----------------------------------------------------------------------
   Return a Vector of Badness by domain of A.
*/
Vector aAnalyze(Analysis a) {

  Vector byDomain=vectorNew(a->P->nrSlots);
  int i;
  List l;
  ConstraintViolation cv;
  Badness b;

  for (i = 0; i < a->P->nrSlots; i++) {
    vectorSetElementAt(byDomain, bestBadness(), i);
  }

  /* count all conflicts */
  for (l = a->conflicts; l; l = listNext(l)) {
    cv = listItem(l);

    b = vectorElementAt(byDomain,cv->nodeBindingIndex1);
    bAdd(b, cv->penalty);

    if (cv->lv2) {
      b = vectorElementAt(byDomain,cv->nodeBindingIndex2);
      bAdd(b, cv->penalty);
    }
  }
  return byDomain;

}

/* ----------------------------------------------------------------------
   Return a domain to change.

   This simply returns any domain that causes at least one conflict.
*/
int selectVariable(Problem P) {

  Vector byDomain=aAnalyze(P->current);
  int i, start;
  int result;

  /* just pick one that violates something */
  do {
    start = result = rand() % P->nrSlots;

    while (((Badness)vectorElementAt(byDomain,result))->no == 0 ||
           listContains(P->history, (Pointer) result)) {
      result = (result + 1) % P->nrSlots;
      if (result == start) {
        result = -1;
        break;
      }
    }
  } while(!slotActive(P->current,start));

  for (i = 0; i < P->nrSlots; i++) {
    bDelete(vectorElementAt(byDomain,i));
  }
  vectorDelete(byDomain);
  return result;

}
/* ----------------------------------------------------------------------
   Compare Analyses by Badness.
*/
Boolean aCompare(Analysis a, Analysis b) {
  return bCompare(a->b, b->b);
}

/* ----------------------------------------------------------------------
   Copy of genetic::cross().
*/
Analysis aCross(Analysis a, Analysis b) {

  Analysis result = aClone(a);
  int cut = 0;
  int i = 0;
  Patch p = pNew(result, a->P->nrSlots - cut + 1, 0);
  Patch q;

  if(!a->P->net) {
    return NULL;
  }
  
  do {
    cut = rand() % a->P->nrSlots;
  } while (!slotActive(a,i));
  
  for(i = cut; i < a->P->nrSlots; i++) {
    LevelValue lv;
    if(!slotActive(a,i)) {
      continue;
    }
    lv = vectorElementAt(b->slots, i);
    if(lv->indexWRTNet == -1) {
      lv = lvClone(lv);
    }
    q = pNext(p, lv, OD_ALLOWALL | OD_DELETED);
    pDelete(p);
    p = q;
  }
  (void)pLexicalize(p,OD_DELETED,FALSE);
  apply(result, p);

  /* mutate the result */
  p = pNew(a, 1, OD_DELETED);
  q = pNext(p,
            vectorElementAt(a->P->net->values,
                            rand() % vectorSize(a->P->net->values)),
            OD_ALLOWALL | OD_DELETED);
  if(q) {
    (void)pLexicalize(q,OD_DELETED,FALSE);
    apply(result, q);
  }
  pDelete(p);

  return result;
  
}

/* ----------------------------------------------------------------------
   Return LV with indexWRTNet INDEX in P, or NULL.
*/
LevelValue lvByNumber(Problem P, int no) {
  List l;
  int i;

  for(i = 0; i < P->nrSlots; i++) {
    for(l = P->domains[i]; l != NULL; l = listNext(l)) {
      LevelValue lv = listItem(l);
      if(lv->indexWRTNet == no) {
        return lv;
      }
    }    
  }
  return NULL;
}

/* ----------------------------------------------------------------------
  If there is already a Problem associated with NET, return it. If there
  isn't, return NULL.
*/
Problem findProblem(ConstraintNet net) {
  List l, m;

  for(l = m= hashListOfKeys(cdgProblems); l != NULL; l = listNext(l)) {
    Problem P = hashGet(cdgProblems, listItem(l));
    if(P->net == net) {
      return P;
    }
  }
  return NULL;
}


/* -- ENDOFFILE ------------------------------------------------------------ */


