/* Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: transform.h,v 1.12 2003/11/05 14:29:23 foth Exp $
 */

#ifndef TRANSFORM_H
#define TRANSFORM_H

/* ---------------------------------------------------------------------------
 @addtogroup Transform 
 @{

*/

/* -- INCLUDES ------------------------------------------------------------- */
#include <configure.h>
#include <limits.h>
#include "eval.h"
#include "lexemgraph.h"
#include "parse.h"
#include "timer.h"


/* Bit flags for dSearch() and relted functions.

   See dSearch() for full documentation. */
#define OD_INT         (1 << 0)    /* arg 1 is int, not Vector */
#define OD_PREDICT     (1 << 1)    /* choose Patch with best Badness */
#define OD_LIMIT       (1 << 2)    /* choose Patch with best limit  */
#define OD_SINGLE      (1 << 3)    /* do not search domains jointly */
#define OD_ORIENT      (1 << 4)    /* allow structural change */
#define OD_LEXICAL     (1 << 5)    /* allow lexical change */
#define OD_LABEL       (1 << 6)    /* allow label change */
#define OD_ALLOWALL    (OD_ORIENT | OD_LEXICAL | OD_LABEL)
#define OD_DELETED     (1 << 7)    /* allow combinations with score 0 */
#define OD_PRUNE       (1 << 8)    /* delete LVs that are proven wrong */
#define OD_NARROW      (1 << 9)    /* ignore Patches with suboptimal limit */
#define OD_INTERACTIVE (1 << 10)   /* give output & receive input via stdin */
#define OD_DEBUG       (1 << 11)   /* give overwhelming amount of output */
#define OD_CONTEXT     (1 << 12)   /* use context-sensitive constraints also */
#define OD_IMMEDIATE   (1 << 13)   /* break if new maximum found */



/* -- FORWARD DECLARATIONS ---------------------------------------------- */

/* Problem and Analysis are mutually dependent */
struct AnalysisStruct;

/* -- TYPE DEFINITIONS -------------------------------------------------- */

/* Assembled information about the state of a Problem.

   This structure is used to remember the state that a ConstraintNet was in
   before frobbing began to mess with it. This is used for backing out
   possibly wrong changes that were made by frobbing under pressure.
*/
struct ProblemStateStruct {
  Number *LVLimits;
  Number *LnLimits;
  BitVector deletedLVs;
  BitVector deletedLns;
  Vector valuesPerNode;
};
typedef struct ProblemStateStruct ProblemStateStruct, *ProblemState;


/* Instance of the parsing CSOP.

   A Problem is essentially a ConstraintNet decorated with some additional
   information, none of which is strictly necessary, but which is
   transformed into forms where it can be accessed more efficiently. For
   instance, the field LEXEMES contains no information that could not be
   read from P->net->lexemgraph->nodes, it just rearranges the nodes so that
   they can be accessed in constant rather than linear time.

   The Problem structure would be unnecessary if someone could be bothered
   to integrate its features into ConstraintNet. As it is, Problem
   structures exist as separate entities, but a Problem is usually based on
   an existing ConstraintNet. What's more, there should never be more than
   one Problem for a ConstraintNet. All existing Problem structures are
   kept in the global hash table cdgProblems. To find the Problem, if any,
   associated with a net, use findProblem().
   
   A Problem can be built without a ConstraintNet via parse2problem().
   However, it will then only contain those LVs that were already present
   in the Parse and be unsuitable for actual parsing unless you add more
   LVs later. 

   All the int fields are results of measurements about the net, cached to
   avoid recomputing. Likewise, the Vectors lexemes and domains contain
   shallow copies of structures from the ConstraintNet, but arranged in a
   way that makes access more efficient for a repair process.

   The Vector lexemes contains all lexeme nodes in the underlying lexeme
   graph. It is indexed by time point and by an index smaller than lgWidth:
   if i == t * lgWidth + n, then cell i contains the nth lexeme node that
   binds time point t. Each lexeme node occurs in every column of this
   vector that corresponds to a time point that it binds (where the Arc
   (from,to) `binds' the time points from, from+1, ... to-1). That is, a
   lexeme occurs several times in the Vector if it binds more than one time
   point. Conversely, some cells of the Vector may be NULL if not all time
   points in the graph are equally ambiguous.

   This organization serves to ensure that all lexemes that span a given
   time point can be retrieved quickly. (Note that although the arcs of a
   word graph are usually notated ordered by time, this property is not
   guaranteed by the module lexemgraph.)

   The array domains contains all LVs that were built for the
   ConstraintNet, rearranged to speed up access. Each cell of the array
   corresponds to a pair of Level and time point: LVs that bind time point
   t on level l are found at cell i == nrLevels * t + l->no. Note that
   shallow copies of an LV may be present in more than one domain, but one
   LV can only occupy one slot in an Analysis.

   The fields current and best hold the currently active Analysis and the
   best Analysis encountered during this run.

   The Lists removedConflicts and unremovable hold deep copies of
   ConstraintViolations that have recently been repaired or that have been
   encountered so often that they are considered unrepairable. The list
   history holds the indices of domains recently changed. Different repair
   strategies may make different use of these fields.

   The fields globalMinimum and globalMaximum give two correct estimates
   for the score of the best Analysis possible for this Problem. They can
   be set in different ways. If repair is applied to a net that already
   contains solutions, then their scores are obviously a lower limit for
   this number. Also, if the repair process finds a better solution, it
   updates the globalMinimum. The globalMaximum can be set to x when all
   LVs in one domain have a limit not better than x, or when the exhaustive
   search among several domains fails to find a better partial analysis
   than x.

*/
struct ProblemStruct {

  String id;               /* unique name */

  /* Core data */
  ConstraintNet net;       /* underlying SLCN */
  LexemGraph lg;           /* underlying lexeme graph */
  Vector lexemes;          /* for looking up lexemes quickly */
  List *domains;           /* array of all domains */

  /* Analyses */
  struct AnalysisStruct *best;       /* best Analysis found so far */
  struct AnalysisStruct *current;    /* current Analysis */
  struct AnalysisStruct *annotation; /* Analysis built from annotation */
  
  /* Statistics */
  int judgeCounter;        /* # of calls to pJudge() so far */
  Timer profileTime;       /* time when frobbing started */
  double firstSoftFound;   /* time when 1st sol. >0 was found */
  double lastSolFound;     /* time when final result was found */
  
  /* the following fields are more or less specific to frobbing,
     and should be moved there if possible */

  Number globalMaximum;    /* upper bound on score for the utterance */
  Number globalMinimum;    /* lower bound on score for the utterance*/
  List removedConflicts;   /* recently removed conflicts */
  List history;            /* recently changed domains */
  List unremovable;        /* conflicts that have been given up on */
  ProblemState savedState; /* state to return to if pruning deleted too much */

  /* the following fields are all calculated once from the
     underlying net and then cached to avoid recomputation. */
  
  int nrTimepoints;        /* # of time slots to be bound */
  int maxTimepoint;        /* min time slot to be bound */
  int minTimepoint;        /* max time slot to be bound */
  int nrLexemes;           /* cached # of nodes in lexem graph */
  int nrSlots;             /* nrLevels * nrTimepoints */
  int nrActiveSlots;       /* nrSlots - slots disabled by useflags */
  int lgWidth;             /* maximal ambiguity per timepoint */
  int nrLVs;               /* # of LVs in entire problem */
  int nrUndeletedLVs;      /* # of LVs not deleted */
  BitVector levelActive;   /* cache level use flags for fast access */
  int nrLevels;            /* Number of levels in the problem. */
  int nrActiveLevels;      /* # of levels with their uselevel flag set */
  
};
typedef struct ProblemStruct ProblemStruct, *Problem;

/* A single consistent, but not necessarily grammatical, parse.

   The main data structure manipulated by this module is the Analysis.
   An Analysis corresponds to a complete parse of the underlying
   utterance, that is, a binding of all time points of the underlying
   lattice on all levels of analysis. The parse is not guaranteed to
   be grammatical. In fact, an Analysis may violate any number of
   defined constraints, even those with score 0.0. This is to ensure
   that an Analysis can be built quickly for any lattice, so that we
   have something to repair.

   The central part of an Analysis is the Vector LVs. This vector
   contains all LVs that together form the dependency trees that make
   up the parse. Each LV only occupies one slot, whose number is
   calculated using the function lvIndex(). If an utterance contains
   lexeme hypotheses that extend over several timepoints, then some of 
   the slots will be empty. This is to ensure that if we want to test
   whether a certain LV is in an Analysis, we only have to search one
   slot rather than several. A simple example illustrates the layout
   of several LVs within the Vector: assume that we have the levels
   SYN and SEM, and the lattice

        -- der --   ------ Mann ---------   -- schlft --
       /         \ /                     \ /             \
      0           1          2            3               4
       \                    / \          / \             /
        -------- das -------   -- Lamm --   --- steht ---


    then the Vector could look like this:

              LVs[0]: SYN:der(0,1)---1--->Mann(1,3)
              LVs[1]: SEM:der(0,1)--DET-->Mann(1,3)
              LVs[2]: SYN:Mann(1,3)--1-->schlft(3,4)
              LVs[3]: SEM:Mann(1,3)--AG->schlft(3,4)
              LVs[4]: NULL
              LVs[5]: NULL
              LVs[6]: SYN:schlft(3,4)--MOD-->NIL
              LVs[7]: SEM:schlft(3,4)--PROP->NIL

    The LVs stored in this vector must make up a complete set of
    dependency trees at any time.

    The Vector isUnselectedLexeme notes which lexemes from the
    underlying lexeme graph are selected for analysis, much like the
    field isDeletedLexemNode of a NetSearchState.
    `isUnselectedLexeme' is admittedly an inelegant way of putting
    it, but we need an array of negative information (TRUE means not
    used) so we can pass it as an argument to lvPrint, and we don't
    want to call it `isDeletedLexeme' to avoid confusion with lexemes
    that are actually deleted in the graph. The Vector must select a
    set of lexemes that form a complete path through the graph at any
    time. All LVs in the Analysis must be valid under this path.

    The List conflicts notes all ConstraintViolations that the
    Analysis causes.

    If the fields subStart and subStop differ from the extreme time points
    of the enclosing Problem, then all functions changing and evaluating
    Analyses pretend that the world ends at these boundaries rather than
    the real ones.

    An Analysis owns only its Badness and the ConstraintViolation
    structures in conflicts.
 */
struct AnalysisStruct {
  Problem P;                      /* enclosing Problem */
  Vector slots;                   /* set of constituting LVs*/
  BitVector isUnselectedLexeme;   /* info about currently selected lexemes */
  List conflicts;                 /* list of ConstraintViolations */
  Badness b;                      /* generalized score */
  int subStart;                   /* start of current active subrange  */
  int subStop;                    /* end of current active subrange  */
};
typedef struct AnalysisStruct AnalysisStruct, *Analysis;



/* Collected info about a possible repair step.

   The canonical method to transform an Analysis is to construct a Patch
   for it, then apply that Patch. A Patch is essentially a list of LVs to
   be inserted into the Analysis. These LVs are held in the Vector
   selection.

   The List newLexemes holds a set of lexemes needed to make all these LVs
   valid. The Vector forbiddenLexemes marks all those lexemes that
   contradict one of the LVs chosen so far. No incompatible LVs are allowed 
   in the same Patch: any LV that relies on a set of `forbidden'
   lexemes cannot be added by pNext().

   The field newLexemes is rather expensive to calculate, since
   changing one lexeme in an Analysis may require more swaps on both
   sides. Therefore it is not calculated as a Patch is constructed.
   To actually decide which new set of lexemes is to be used for the
   Patch, the function pLexicalize() must be called.

   The two Badnesses local and predicted give estimates about how
   good (by the generalized measure) the partial Analysis represented
   by the Patch and the resulting complete Analysis will be. The
   Number limit gives the lowest boundary that can be calculated for
   the score of an Analysis that results from this Patch. Again,
   different repair strategies can use this information in different
   ways.

   The field nrNewLVs counts the number of LVs that are not
   already present in the current Analysis (in principle, it is
   legal to substitute an LV for itself, though care must be taken to
   ensure termination).

   The Vector bound holds the indices of all domains bound by the
   LVs in the Patch. Note that since an LV may bind several
   domains, this Vector may contain more elements than the selection.

   The field local holds the Badness of the partial
   Analysis constituted by the LVs.

   The field predicted holds an estimate of the Badness of
   the Analysis that will result from applying the Patch.

   The field limit holds an upper bound to the classical score of
   the Analysis that will result from applying the Patch.

   Patches are allocated by the function pNew(). The canonical method
   to construct a Patch that binds the LVs A and B is to call the
   functions pNext(B, pNext(A, pNew())) (modulo some more
   parameters). A Patch p is then applied to the Analysis a by
   calling the functions pLexicalize(a, p) and apply(a, p).

   Neither LevelValue nor Analysis structures are owned by a
   Patch, only its own Vectors and Lists.

*/
struct PatchStruct {
  Analysis a;                   /* Analysis that the patch applies to */
  Vector selection;             /* LVs to swap in */

  List newLexemes;              /* lexemes to be swapped in */
  int index;                    /* pointer into selection */
  int nrNewLVs;                 /* # of non-vacuous switches */
  BitVector forbiddenLexemes;   /* mark incompatible lexemes */
  BitVector bound;              /* mark domains bound by a Patch */

  Badness local;                /* mutual score of LVs in a Patch */
  Badness predicted;            /* Badness of resulting Analysis */
  Number limit;                 /* maximally attainable score */
};
typedef struct PatchStruct PatchStruct, *Patch;

/* How many items should the local search agenda hold? */
extern int localSize;

/* How many of the Patches generated in a step should actually be tried out? */
extern int beamWidth;

extern Analysis aClone(Analysis a);
extern Analysis aCross(Analysis a, Analysis b);
extern Boolean aCompare(Analysis a, Analysis b);
extern Boolean aRelabel(Analysis a, int where, String label);
extern Boolean aValid(Analysis a);
extern Boolean isUnremovable(Problem P, ConstraintViolation cv);
extern Boolean pLexicalize(Patch p, int flags, Boolean debugme);
extern Boolean reducePressure(Problem P);
extern Boolean slotActive(Analysis a, int x);
extern LevelValue slotFiller(Analysis a, int where);
extern LevelValue lvByNumber(Problem P, int i);
extern LexemNode selectedAt(Analysis a, int where);
extern LexemNode selectedLexeme(Analysis a, List lexemes);
extern List dSearch(Analysis a, Vector which, int flags);
extern Parse analysis2parse(Analysis a);
extern Patch optimizeDomains(Analysis a, Vector which, int flags);
extern Patch pNew(Analysis a, int width, int flags);
extern Patch pNext(Patch old, LevelValue lv, int flags);
extern Problem net2problem(ConstraintNet net, Number pressure);
extern Problem parse2problem(Parse parse, Number pressure);
extern ProblemState psNew(Problem P);
extern Vector aAnalyze(Analysis a);
extern int dSize(Problem P, int where);
extern int selectVariable(Problem P);
extern void aCalibrate(Analysis a);
extern void aCheck(Analysis a, int level);
extern void aDelete(Analysis a);
extern void aJudge(Analysis a);
extern void aPrint(unsigned long int mode, Analysis a);
extern void aRedirect(Analysis a, int where, int direction);
extern void aRestrict(Analysis a, int from, int to);
extern void aUnrestrict(Analysis a);
extern void aVerify(Analysis a);
extern void addParse(Analysis a);
extern void apply(Analysis a, Patch p);
extern void backTrace(Problem P);
extern void dPrint(unsigned long mode, Problem P, int where);
extern void deleteProblem(Problem P);
extern void determineAmbiguity(Problem P);
extern void displayLV(Problem P, LevelValue lv, int width);
extern void freezeDomain(Analysis a, int where);
extern void newMaximum(Analysis a);
extern void pDelete(Patch p);
extern void pJudge(Patch p);
extern void pPrint(unsigned long mode, Patch p, Boolean brief);
extern void printAmbiguity(unsigned long mode, Problem P);
extern void printConflicts(unsigned long mode, Analysis a);
extern void printHistory(unsigned long mode, Problem P);
extern void printProblem(unsigned long mode, Problem P);
extern void rejudgeLV(Analysis a, int where);
extern void rememberProblem(Problem P);
extern void reportCorrectness(unsigned long int mode, Problem P);
extern void resetProblem(Problem P);
extern void restoreProblem(Problem P, ProblemState ps);
extern void selectPath(Analysis a, List which);
extern void showUnary(Analysis a);
extern void sortConflicts(Analysis a);
extern Problem findProblem(ConstraintNet net);
extern void addLVToProblem(Problem P, LevelValue lv);
extern Patch makeLexicalPatch(Analysis a, LexemNode ln);

#endif /* don't insert anything after this #endif */
