/*
 * File: write.h
 * Author: Horia F. Pop
 * Purpose: header file for write.c
 * Birth: 19/11/99
 *
 * $Id: write.h,v 1.18 2003/08/27 16:19:14 tomas Exp $
 */

#ifndef WRITE_H
#define WRITE_H

/* -- TYPE DEFINITIONS -------------------------------------------------- */
/* -- INCLUDES ---------------------------------------------------------- */
#include "constraintnet.h"
#include "input.h"
#include "parse.h"
#include "stdarg.h"
#include "input.h"

/* -- MACROS ------------------------------------------------------------ */
/* -- VARIABLES --------------------------------------------------------- */
/* -- FUNCTIONS --------------------------------------------------------- */

#ifndef SWIG
extern void writeInitialize(void);
extern void writeFinalize(void);
#endif

/* -- XFIG FUNCTIONS ---------------------------------------------------- */

#ifndef SWIG

extern void writeXFigHeader(FILE *file);
/*
 * writes the XFig header
 *
 * arguments:
 * - file = the FILE pointer to the XFig file currently written
 *
 * - it is the responsibility of the caller to open the file before 
 *   using this function and to close it after that
 */

extern void writeXFigCircle(FILE *file, int x0, int y0, int r);
/*
 * writes a circle to an XFig file
 *
 * arguments:
 * - file      = the FILE pointer to the XFig file currently written
 * - x0 and y0 = the coordinates of the center of the circle
 * - r         = the radius of the circle
 *
 * - it is the responsibility of the caller to open the file before 
 *   using this function and to close it after that
 */

extern void writeXFigSolidCircle(FILE *file, int x0, int y0, int r);
/*
 * writes a solid (filled) circle to an XFig file
 *
 * arguments:
 * - file      = the FILE pointer to the XFig file currently written
 * - x0 and y0 = the coordinates of the center of the circle
 * - r         = the radius of the circle
 *
 * - it is the responsibility of the caller to open the file before 
 *   using this function and to close it after that
 */

extern void writeXFigLine(FILE *file, int x1, int y1, int x2, int y2);
/*
 * writes a line segment to an XFig file
 *
 * arguments:
 * - file      = the FILE pointer to the XFig file currently written
 * - x1 and y1 = the coordinates of one end of the line segment
 * - x2 and y2 = the coordinates of the other end of the line segment
 *
 * - it is the responsibility of the caller to open the file before 
 *   using this function and to close it after that
 */

extern void writeXFigText(FILE *file, int l, int x, int y, String s, int align);
/*
 * writes a text string to an XFig file
 *
 * arguments:
 * - file  = the FILE pointer to the XFig file currently written
 * - l     = the width of the box containing the text
 * - x, y  = the coordonates of the origin of the string:
 *           lower left corner  if align = 0
 *           lower center       if align = 1
 *           lower right corner if align = 2
 * - s     = the text to be written
 * - align = text alignment in the box:
 *           0 = left align
 *           1 = center 
 *           2 = right align
 * 
 * - it is the responsibility of the caller to open the file before 
 *   using this function and to close it after that
 */

extern void writeXFigSpline(FILE *file, int x1, int y1, int x2, int y2, int ho, int h);
/*
 * writes a spline curve to an XFig file
 *
 * arguments:
 * - file      = the FILE pointer to the XFig file currently written
 * - x1 and y1 = the coordinated of the first end
 * - x2 and y2 = the coordonates of the second end
 * - ho and h  = used to set three intermediary fixed points
 *              
 * - the X coordinates of the three intermediary points:
 *       x1 + d/8, x1 + d/2, x2 - d/8 (d = x2 - x1)
 * - the heights of the three points over (under) the line segment
 *       ho + h * 6/7, ho + h, ho + h * 6 / 7
 * 
 * - it is the responsibility of the caller to open the file before 
 *   using this function and to close it after that
 */

/* -- FUNCTIONS THAT WRITE REPRESENTATIONS OF CDG STRUCTURES ------------ */

extern void writeConstraintNet(ConstraintNet net, String filename);
/*
 * writes a table representation of a constraint net to a LaTeX file
 * 
 * arguments:
 * - net      = the constraint net whose representation is to be written
 * - filename = the name of the LaTeX file that will contain the table 
 *              representation of the net
 */

extern void writeLattice(Lattice l, Boolean withWeights, String filename);
/*
 * writes a graphic representation of a wordgraph to an XFig file
 *
 * arguments:
 * - l           = the wordgraph whose graphic representation is written
 * - withWeights = specifies whether the word arcs weights are printed
 * - filename    = the name of the XFig file that will contain the 
 *                 graphic representation of the wordgraph
 */

extern void writeParse(Parse parse, String main_level, String graph_filename, String info_filename);
/*
 * writes a graphic representation of a parse to an XFig file, and other 
 * relevant information into a LaTeX file
 *
 * arguments:
 * - parse          = the parse whose representation is to be written
 * - main_level     = the level for which the parse tree will be built
 * - graph_filename = the name of the XFig file that will contain the 
 *                    graphic representation
 * - info_filename  = the name of the LaTeX file that will contain the 
 *                    other relevant information
 */

extern void writeParses(List parses, String main_level, String graph_filename, String info_filename);
/*
 * writes all parses to different XFig and LaTeX files
 *
 * arguments:
 * - parses         = the list of parses whose representations are written
 * - main_level     = the level for which the parse trees will be built
 * - graph_filename = the root name of the XFig files that will contain 
 *                    the graphic representations
 * - info_filename  = the root name of the LaTeX files that will contain 
 *                    the other relevant information
 *
 * - the filenames effectively created will be of the form file-1.fig, 
 * file-2.fig, ..., and, respectively, file-1.tex, file-2.tex, ...
 */

/* ----------------------------------------------------------------------
 *
 * File naming conventions
 *
 * Since, for a certain wordgraph, we have functions that write more files,
 * it would be a good idea to name these files so that they could be easily
 * recognized. I suggest the following naming convention:
 * 
 * name-wrd.fig    = the XFig file with the wordgraph representation;
 * name-net.tex    = the LaTeX file with the constraint net information;
 * name-sol-n.fig  = the XFig file with the n-th solution representation;
 * name-inf-n.tex  = the LaTeX file with the n-th solution information,
 *
 * where:
 * 
 * - 'name' is the name of the wordgraph;
 * - 'n' is the solution number, starting from 1..
 * 
 * These are suggestions for the users. For now, the write functions do not 
 * use any kind of automatic file name generation, with the single exception 
 * that writeParses automatically appends '-n.fig' and '-n.tex' after the 
 * input filename, to facilitate the generation of more than one set of files.
 *
 * ---------------------------------------------------------------------- */

/* -- XML FUNCTIONS ----------------------------------------------------- */

/*
 * for further explanation of xml see http://www.w3c.com
 */

extern FILE *writeXmlFileSet(String fileName);
/*
 * set the xml output file; the variables xmlFile and xmlFileName are changed
 * accordingly; previous files are closed and guaranteed to be wellformed
 */

extern FILE *writeXmlFileGet(void);
/* returns the xmlFile handle */

extern void writeXmlFlush(void);
/* flushes the xmlFile */

extern void writeXmlIndent(void);
/*
 * print the correct ammount of indentation to the xml file output stream;
 * correct indentation is calulated in writeXmlStartTag and writeXmlEndTag
 */

extern void writeXmlStartTag(String name, String format, ...);
/*
 * write an xml start tag; the name of this opend tag is pushed on a stack
 * and poped when writeXmlEndTag is encountered accordingly
 */

extern void writeXmlEndTag(void);
/*
 * pop the last xml tag from the xml tag stack and write an xml end tag
 * for it
 */

extern void writeXmlEndAll(void);
/* pop all xml tags from the xml tag stack and write xml end tags */

extern void writeXmlEmptyTag(String name, String format, ...);
/* write an xml empty tag; see writeXmlStartTag */

extern void writeXmlProcessingInstruction(String target, String format, ...);
/* write an xml processing instruction, only needed in the header hear */

extern void writeXmlHeader(void);
/* write a propper header for the cdgp xml logfile format */

extern void writeXmlNotification(String type, String format, ...);
/* write an xml notification */

extern void writeXmlVaNotification(String type, String format, va_list ap);
/*
 * write an xml notification using a variable list 
 * see writeXmlNotification
 */

extern void writeXmlStatistics(int type, String method, String description, ...);
/* write an xml statistics, a set of numbers */

extern void writeXmlCommand(String commandLine);
/* write an xml command tag */

extern void writeXmlUser(String name, String mail);
/* write an xml user */

extern void writeXmlDate(time_t time);
/* write the given time in xml format */

extern void writeXmlFile(String fileName, int lineNo);
/* write information about some file to the xml output stream */

extern void writeXmlGrammar(List files);
/* write xml information about a used grammar */

extern void writeXmlReference(String name, String type, Pointer object);
/* write the xml representation of an object reference */

#define writeXmlNumber(name, value) _writeXmlNumber(name, (long double) value)
extern void _writeXmlNumber(String name, long double value);
/* write a Number representation to the xml output stream */

extern void writeXmlBoolean(String name, Boolean value);
/* write a Boolean representation to the xml output stream */

extern void writeXmlNetInfo(ConstraintNet net);
/* write ConstraintNet information to the xml output stream */

extern void writeXmlArc(Arc arc);
/* write a Arc representation to the xml output stream */

extern void writeXmlLattice(Lattice lattice);
/* write a Lattice representation to the xml output stream */

extern void writeXmlMethod(String method);
/* write a method tag */

extern void writeXmlParse(Parse parse);
/* write a Parse representation to xml */

extern void writeXmlAnnoEntry(AnnoEntry annoEntry);
/* write an AnnoEntry representation to xml */

extern void writeXmlParses(List parses);
/*
 * write a set of Parse representations to xml
 * see writeXmlParse
 */

extern void writeXmlParseComparison(Parse parse1, Parse parse2, ParseCmpStat stat);
/* write a parse comparison to xml */

extern void writeXmlParseVerification(ParseVerification pv);
/* write a parse verification to xml */

extern void writeXmlWord(Word word);
/*
 * write a Word representation to xml; this actualy produces a
 * lexem tag from a word
 */
#endif /* SWIG */

/* Writes an AnnoEntry to a the file FOO.cda
 *  in a format suitable for input to cdgp. 
 */
extern void writeAnnotation(AnnoEntry ae, String id);

/* -- PROLOG FUNCTIONS ----------------------------------------------------- */

void writePrologParse(FILE*, Parse);
void writePrologAE(FILE*,AnnoEntry);

#endif /* don't insert anything after this #endif */
