#!/usr/bin/env perl
# utils/check-annotations.pl

# Quickly finds all violations that the annotations produce
# under the current grammar.
# Requires cdgp in your PATH.

use warnings;
use strict;

my @edges;
my $threshold = 0.5;

our $grammar = shift;
if (defined $grammar) {
    $grammar =~ s/.cdg$//;
} else {
 $grammar = `pwd`;
 chomp $grammar;
 $grammar =~ s|^.+/||;
}

# remake the grammar
system "make $grammar.cdg &>/dev/null";

# sort like ls -v does
sub byVName {

  # don't change the names themselves
  my ($x,$y) = ($a,$b);

  while ($x ne "" && $y ne "") {

    # skip identical alphabetic parts
    while (substr($x,0,1) eq substr($y,0,1) &&
           $x !~ /^[0-9]/) {
      substr($x,0,1) = "";
      substr($y,0,1) = "";
    }

    # compare alphabetical parts normally
    if ($x !~ /^\d/ || $y !~ /^\d/) {
      return $x cmp $y;
    }

    # compare numeric parts numerically
    $x=~s/^(\d+)//;
    my $nx = $1;
    $y=~s/^(\d+)//;
    my $ny = $1;
    my $result = $nx <=> $ny;
    next unless $result;
    return $result;
  }
}

# extract all sentence ids
our @sentences;
if ($grammar eq 'negra') {
  @sentences =
    `(cd /data/corpus/negra/cdg; ls -v gold-s*cda) | sed s/.cda//g`;
} else {
  @sentences =
    sort byVName
      map {/\'?([^\']+)\'? *: *.+<->/;$1}
        `egrep ":.+<->" $grammar.cdg`;
}

open(OUT,">commands") or die "can't write to `commands': $!\n";
print OUT "set debug on\n";
print OUT "set taggerNormalizeToOne on\n";
print OUT "set taggerIgnoreImpossible on\n";
print OUT "load $grammar.cdg\n";
print OUT "tagger\n";
for(@sentences) {
  print OUT "anno2parse $_\n" if defined;
  print OUT "printparse $_\n" if defined;
}
print OUT "quit\n";
close(OUT);

# call cdgp
open(LOG, "$ENV{CDGHOME}/cdgp/cdgp -i commands|tee cdgp.log|");
open(OUT, ">new_errors") or die "Can't write `new_errors': $!\n";
my @words = ();
my $printme = 0;
my $info = '';
my $lattice = '';
while(<LOG>) {

  if(/anno2parse +(.+)/) {
    print OUT $info if $printme;
    print OUT "$lattice: @words" if $printme;
    $printme = 0;
    @words = ();
    $info = "\n\n\n\nSentence $1\n\n";
  }

  if(/lattice id *: *(.+)/) {
    $lattice = $1;
  }

  # remember edges to better display the conflict
  if(/^   (\d+)        (.+)/) {
    my ($index,$edge) = ($1,$2);
    $edge =~ s/\(\d+-\d+\)//g;
    $edges[$index]="$edge";
  }

  # remember words to display the sentence if necessary
  if(/ +SYN: (.+?)(_|\(\d+-\d+\))/) {
    push @words, $1;
  }

  # maybe produce output
  if(/^   Violations:/) {
  }

  # match `000 : 030 : 1.000e-01 : Vorfeld'
  if(/((\d+) : )?(\d+) : ([0-9.e+-]+) : (.+)$/) {
    my $id = $5;
    my $score = $4;
    my $lva;
    my $lvb;
    if($1) {
      ($lva,$lvb) = ($2,$3);
    } else {
      ($lva,$lvb) = ($3,'');
    }

    # maybe remember conflict
    if($score < $threshold) {
      $printme = 1;
      $info .= "$edges[$lva]\n";
      if($lvb) {
        $info .= "$edges[$lvb]\n";
      }

      my $c = "$id($score): \@$lva";
      if($lvb) {
        $c .= "-\@$lvb";
      }

      $info .= "$c\n\n";
    }
  }
}

print OUT $info if $printme;
print OUT "$lattice: @words" if $printme;

close(LOG);
close(OUT);
unlink 'commands';
if( -f 'known_errors') {
    system('diff new_errors known_errors');
} else {
    system "cat new_errors; echo";
}
