#!/usr/bin/env perl
# cdg/utils/merge-annos.pl

# Given two annotations, create a third that contains a superset
# of information.
#
# Use this tool in the following scenario:
#
# - you have a corpus of detailed annotations
# - your grammar leaves some features underspecified
#   for performance reasons
# - you want to make changes by editing trees in xcdg
#
# Say that the golden annotation for a word looks like this:
#
#   3 4 Aktien
#   'case' / 'gen'
#   'cat' / 'NN'
#   'gender' / 'fem'
#   'number' / 'pl'
#   'person' / 'third'
#   'SYN' -> 'GMOD' -> 2 /* ('Prozent') */
#   'SUB' -> '' -> 0
#   'REF' -> '' -> 0
#
# with all features carefully specified. When creating the
# corresponding edge in the actual grammar, it looks like this:
#
#   3 4 Aktien
#   'case' / 'bot'
#   'cat' / 'NN'
#   'gender' / 'fem'
#   'number' / 'bot'
#   'person' / 'third'
#   'SYN' -> 'GMOD' -> 2 /* ('Prozent') */
#   'SUB' -> '' -> 0
#   'REF' -> '' -> 0
#
# Now suppose that the annotator wants to change the label to read
# `APP'. If we simply hit `save' and copy the new annotation to the
# gold repository, the case and number information will be lost.
# Instead, call merge-annos.pl <new file name> <repository directory>,
# and the information will be carefully preserved.
#
# If one of the annotations has a more specific value for a feature
# than the other, that will be preserved. Otherwise the new annotation
# is considered correct. The result is a file that reads
#
#   3 4 Aktien
#   'case' / 'gen'
#   'cat' / 'NN'
#   'gender' / 'fem'
#   'number' / 'pl'
#   'person' / 'third'
#   'SYN' -> 'APP' -> 2 /* ('Prozent') */
#   'SUB' -> '' -> 0
#   'REF' -> '' -> 0
#
# as desired.

use warnings;
use strict;
BEGIN {
 if(!defined $ENV{CDGHOME}) {
   die "Please set CDGHOME.\n";
 }
 $ENV{CDGPERL} = "$ENV{CDGHOME}/bindings";
}
use lib "$ENV{CDGPERL}";
use lib "$ENV{CDGPERL}/.libs";
use CDG;
use CDG::Utils qw(:functions);

die "Usage: $0 <gold dir>\n" unless @ARGV == 1;

our $golddir = shift;

for my $anno (<*.cda>) {

# load new annotation
  commandEval 'set verbosity 0';
  commandEval "load $anno";
  my $annotations = InputStruct_annotations_get($CDG::inputCurrentGrammar);
  my $new;
  my $id;
  my %newedges;
  for (my $l = hashListOfKeys($annotations); $l; $l = listNext $l) {
    $new = hashGet($annotations, listItem($l));
    $id = AnnoEntryStruct_id_get $new;

    # Slurp it into perl space; this is necessary because it would
    # probably get overwritten, since we are about to load another
    # annotation with the same name.
    $new = AnnoEntry2hash $new;

    # keep edges in a hash instead of an array, in case the level
    # ordering is inconsistent
    for my $e (@{$$new{edges}}) {
      $newedges{$$e{level}}[$$e{child}] = $e;
    }
    last;
  }


  commandEval "reset\n";


# load corresponding gold annotation
  die "No such directory: $golddir" unless -d $golddir;
  my $goldname = "$golddir/$id.cda";
  die "No such file: $goldname" unless -r $goldname;
  commandEval "load $goldname\n";
  my $ann = InputStruct_annotations_get($CDG::inputCurrentGrammar);


# Why can't we just say hashGet($ann, $id) here?
  my $l = hashListOfKeys $ann;
  my $old = hashGet($ann, listItem $l);
  $old = AnnoEntry2hash $old;

  # keep edges in a hash instead of an array, in case the level
  # ordering is inconsistent
  my %oldedges;
  for my $e (@{$$old{edges}}) {
    $oldedges{$$e{level}}[$$e{child}] = $e;
  }


# merge annotations asymetrically

# compare words
  my $i = 0;
  for my $w1 (@{$$old{words}}) {
    my $w2 = $$new{words}[$i++];
    next if $$w1{word} eq 'NIL';

    if ($$w1{word} ne $$w2{word}) {
      die "Mismatched words: $$w1{word}/$$w2{word}\n";
    }

    for my $a (keys %$w1) {
      next if $a eq 'word';
      my $v1 = $$w1{$a};
      my $v2 = exists $$w2{$a} ? $$w2{$a} : '<undef>';

      if ($v1 ne $v2) {
#      print "Word `$$w1{word}': mismatched feature $a: $v1/$v2\n";

        # def beats undef
        if ($v2 eq '<undef>') {
          # do nothing
        }

        # acc beats nom_dat_acc
        elsif ($v2 eq 'bot' or $v2 =~ /$v1/) {
          # do nothing
        }

        # masc beats not_neut
        elsif ($v2 =~ /_/ && $v1 !~ /_/) {
          # do nothing
        }

        # otherwise acknowledge the new feature
        else {
          $$w1{$a} = $$w2{$a};
          print "Word `$$w1{word}': changing feature $a: $v1 --> $v2\n";
        }
      }
    }
  }

  for my $level (keys %oldedges) {
    for my $i (1..$#{$$old{words}}) {

      my $e1 = $oldedges{$level}[$i];
      my $e2 = $newedges{$level}[$i];

      if (!exists $$e1{label}) {
        die "ERROR: oldedges{$level}[$i] does not exist.\n";
      }
      if (!exists $$e2{label}) {
        die "ERROR: newedges{$level}[$i] does not exist.\n";
      }

      if ($$e1{label} ne $$e2{label}) {
        print "Changing label $$e1{label} --> $$e2{label}\n";
        $$e1{label} = $$e2{label};
      }

      if ($$e1{parent} ne $$e2{parent}) {
        print "Changing subordination $$e1{parent} --> $$e2{parent}\n";
        $$e1{parent} = $$e2{parent};
      }
    }
  }

  my $oldname = "$id.cda.orig";
  rename "$id.cda", "$id.cda.orig";
  open(OUT, ">$id.cda")
    or die "Can't write `>$id.cda': $!\n";
  print OUT format_anno $old;
  close(OUT);

  commandEval 'reset';

}
