# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllConstraints - manage all constraints.
##
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
## $Id: allconstraints.tcl,v 1.27 2003/10/28 13:07:56 ddreyer Exp $
## ----------------------------------------------------------------------------
class AllConstraints {
  inherit DataBrowser

  
  # public methods
  public method init_data {} 

  ## replaces getconstraint
  public method getCData {id}; ## \type TclString

  public method getTclData {row constraint}; ## \type TclNumber, Constraint

  public method getAllConstraintIds; ## \type TclList

  constructor {args} {}; ## \type TclList

  # private methods
  private method _init_data {}
  private method showbutton_action {}
  private method editbutton_action {}
  private method usebutton_action {}
  private method usegroupbutton_action {}
  private method uselevelbutton_action {}


  # private variables

  ## hash mapping constraint ids to Constraint structures
  private variable constraints; ## \type TclArray
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body AllConstraints::constructor {args} {
    set _idColumnIndex 0
    set table $itk_component(table)
    $table configure -cols 7


  itk_component add editbutton {
    button $itk_component(childsite).edit\
	-anchor w \
	-text "Edit" \
	-command [code $this editbutton_action]
  } { }

  itk_component add showbutton {
    button $itk_component(childsite).show\
	-anchor w \
	-text "Show" \
	-command [code $this showbutton_action]
  } { }

  itk_component add usebutton {
    button $itk_component(childsite).use\
	-anchor w \
	-text "Use" \
	-command [code $this usebutton_action]
  } { }

  itk_component add usegroupbutton {
    button $itk_component(childsite).usegroup\
	-anchor w \
	-text "Use Group" \
	-command [code $this usegroupbutton_action]
  } { }

  itk_component add uselevelbutton {
    button $itk_component(childsite).uselevel\
	-anchor w \
	-text "Use Level" \
	-command [code $this uselevelbutton_action]
  } { }


  # packing
  pack \
      $itk_component(showbutton) \
      $itk_component(editbutton) \
      $itk_component(usebutton) \
      $itk_component(usegroupbutton) \
      $itk_component(uselevelbutton) \
      -side left -padx 3 

  # get the initial data
  init_data
  
  # register the helpmessages
  .cdgmain help sethelpstr \
      $itk_component(showbutton) "Print a constraint in the shell" \
      $itk_component(editbutton) "Edit the source of the constraint" \
      $itk_component(usebutton) "(Don't) Use this constraint" \
      $itk_component(usegroupbutton) "(Don't) Use constraints from this group" \
      $itk_component(uselevelbutton) "(Don't) Use constraints from this level" \
      $itk_component(table) \
      "Click selects row, Ctrl-Click selects more rows, NW cell selects all"


  eval itk_initialize $args
}



## ----------------------------------------------------------------------------
## call _init_data() if necessary.
## ----------------------------------------------------------------------------
body AllConstraints::init_data {} {

    if {[InputStruct_xcdgconstraintsToDate_get [inputCurrentGrammar_get]]} {
       return
   }
    
  .cdgmain busy compute "Init Constraints" [code $this _init_data]

}

## ----------------------------------------------------------------------------
## get data from the cdg tool.
## ----------------------------------------------------------------------------
body AllConstraints::_init_data {} {

  set table $itk_component(table)
  $table configure -state normal
  set rmax [$table cget -rows]
  $table clear all
  $table delete rows 1 [$table cget -rows]
  $table configure -rows $rmax
  $table tag config default
  $table tag config colored -bg gray80 
  $table tag config ids -anchor w
  $table tag col ids 0
  $table width 0 17 1 10 2 12 3 7 4 7 5 5 6 3
  
  # prepare
  foreach id [array names constraints] {
    unset constraints($id)
  }

  # import data to Tcl
  set h [InputStruct_constraints_get [inputCurrentGrammar_get]]
  set m [hashListOfKeys $h]
  for {set l $m} { $l != "NULL" } { set l [listNext $l ] } {
    set id [listItem $l]
    set constraint [hashGet $h $id]
    set id [ConstraintStruct_id_get $constraint]
    set constraints($id) $constraint
  }
  listFree $m

  $table setCell row 0,0 { "" "File" "Group" "Level1" \
      "Level2" "Penalty" "active" }

  set r 1
  foreach id [lsort [array names constraints]] {
    refreshrow $r $constraints($id)
    incr r
    if {[expr $r % 100] == 0} {
      update 
    }
  }

  _setCount [hashSize [InputStruct_constraints_get [inputCurrentGrammar_get]]]
  setSelection

  InputStruct_xcdgconstraintsToDate_set [inputCurrentGrammar_get] 1

  $table configure -state disabled
}


## ----------------------------------------------------------------------------
## fill a row with the values from a specific constraint.
## ----------------------------------------------------------------------------
body AllConstraints::getTclData {row constraint} {
  set table $itk_component(table)
  # constraint name
  set constraint_id [ConstraintStruct_id_get $constraint]

  # base name of defining file  
  regexp "\[^/\]*$" [ConstraintStruct_filename_get $constraint] basename

  # section of constraint
  set group [SectionStruct_id_get [ConstraintStruct_section_get $constraint]]

  # names of related levels
  set i 1
  set level1 "---" 
  set level2 "---" 
  for {set l [ConstraintStruct_vars_get $constraint] } \
      { $l != "NULL" } \
      { set l [listNext $l ] } \
      {
	set varinfo [listItem $l]
	set level$i [VarInfoStruct_levelname_get $varinfo]
	incr i
      }

  # constraint penalty
  set penalty [ConstraintStruct_penalty_get $constraint]

  # variable penalties are just called "variable"...
  set term [ConstraintStruct_penaltyTerm_get $constraint]
  if { $term != "NULL" } { 
    set penalty "variable" 

    # ...but sometimes people just notate a constant in brackets.
    # Let's outsmart them!
    set type [TermStruct_type_get $term]
    if {$type == "10"} { # 10 corresponds to TTNumber
      set data [TermStruct_data_get $term]
      set penalty [TermStruct_data_number_get $data]
    }
  }

  # usage flag
  set active [ConstraintStruct_active_get $constraint]
  if { $active == 0 } { set active no }
  if { $active == 1 } { set active yes }
 
  $table setCell row $row,0 [list \
      $constraint_id $basename $group $level1 \
      $level2 $penalty $active ]

}


## ----------------------------------------------------------------------------
## toggle usage of groups of selected constraints .
## ----------------------------------------------------------------------------
body AllConstraints::usegroupbutton_action {} {
  # avoid flickering
  if {$_selection == ""} return

  # find all affected groups
  foreach id $_selection {
    set constraint $constraints($id)
    set group [ConstraintStruct_section_get $constraint]
    set groups($group) 1
  }

  # toggle all constraints within them 
  foreach id [array names constraints] {
    set constraint $constraints($id)
    set group [ConstraintStruct_section_get $constraint]
    if { [info exists groups($group)] } { 
      commandEval "useconstraint $id"
      refreshid $id 
    }
  }
  
}

## ----------------------------------------------------------------------------
## toggle usage of level of selected constraints.
## ----------------------------------------------------------------------------
body AllConstraints::uselevelbutton_action {} {
  # avoid flickering
  if {$_selection == ""} return

  # find all affected levels
  foreach id $_selection {
    set constraint $constraints($id)
    for { set l [ConstraintStruct_vars_get $constraint] } \
	{ $l != "NULL" } \
	{ set l [listNext $l] } \
	{
	  set varinfo [listItem $l]
	  set level [VarInfoStruct_levelname_get $varinfo]
	  set levels($level) 1
	}
  }

  # toggle all found levels
  foreach id [array names levels] {
    commandEval "uselevel $id"
  }

  # since this changes nothing in our table, no need to redraw.
  # But the levelpage needs refreshing.
  .cdgmain levels init_data

}

## ----------------------------------------------------------------------------
## toggle usage of selected constraints.
## ----------------------------------------------------------------------------
body AllConstraints::usebutton_action {} {
  # avoid flickering
  if {$_selection == ""} return

  foreach id $_selection { commandEval "useconstraint $id" }
  foreach id $_selection { refreshid $id }
}

## ----------------------------------------------------------------------------
## call an editor to view the constraint source
## ----------------------------------------------------------------------------
body AllConstraints::editbutton_action {} {
  # avoid flickering
  if {$_selection == ""} return

  # build list of filenames
  set target ""
  foreach id $_selection {

    # I suppose most editors grok "+<linenum> file" nowadays...
    set constraint $constraints($id)
    set linenum [ConstraintStruct_lineNo_get $constraint]
    set filename [ConstraintStruct_filename_get $constraint]
    append target "+$linenum $filename "
  }

  # build command line
  regsub -all "%f" "/bin/sh -c \"[.cdgmain cget -editor]\"" $target command

  # execute
  if {[catch {eval exec $command >/dev/null &} errMsg]} {
    ::cmd::Puts "ERROR: $errMsg"
  }
}

## ----------------------------------------------------------------------------
## display the constraint in the shell.
## ----------------------------------------------------------------------------
body AllConstraints::showbutton_action {} {
  foreach id $_selection { commandEval "constraint $id" }
}



## ----------------------------------------------------------------------------
## get pointer to a constraint_id, cache it in array constraints.
## ----------------------------------------------------------------------------
body AllConstraints::getCData {id} {
  if {![info exists constraints($id)]} {
    set found 0
    for {set l [cdgConstraints_get]} \
	{$l != "NULL"} \
	{ set l [listNext $l ] } {

	  set constraint [listItem $l]
	  set constraint_id [ConstraintStruct_id_get $constraint]
	  if {$constraint_id == $id} {
	    incr found
	    break
	  }
	}
    if {!$found} {
      return "NULL"
    }
    set constraints($id) $constraint
    _setCount [listLength [cdgConstraints_get]]
  }
  return $constraints($id)
}


## ----------------------------------------------------------------------------
## Return list of all constraints' names
## ----------------------------------------------------------------------------
body AllConstraints::getAllConstraintIds {} {
  init_data
  return [array names constraints]
}
