# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## AllWordgraphs - manage all grammar wordgraphs.
##
## \author Michael Daum (see also AUTHORS and THANKS for more)
## $Id: allwordgraphs.tcl,v 1.62 2003/10/28 13:07:56 ddreyer Exp $
## ----------------------------------------------------------------------------
class AllWordgraphs {
  inherit DataBrowser


    # public methods
    public method getNext {latticeId}; ## \type TclString
    public method getPrev {latticeId}; ## \type TclString
    public method getAllLatticeIds {}; ## \type TclList
    public method refreshid {id}; ## \type TclString
    
    # replaces getwordgraph
    public method getCData {id}; ## \type TclString

    public method getTclData {row id}; ## \type TclNumber, TclString 

   

    constructor {args} {}; ## \type TclList

    # protected methods
    public method init_data {}

    # private methods
    private method _init_data {}
    private method editbutton_action {}
    private method detailsbutton_action {}
    private method newbutton_action {}
    private method deletebutton_action {}
    private method annobutton_action {}
    private method icbutton_action {}
    private method interactivebutton_action {}
    private method details {}
    private method newnet {}


  # private variables

  ## hash mapping wordgraph ids to their C structure
  private variable wordgraph; ## \type TclArray

  ## list of ids sorted as in the table
  private variable allsorted

  ## list of selected ids will be stored in inherited
  ## protected variable DataBrowser::_selection 
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body AllWordgraphs::constructor {args} {
  set _idColumnIndex 0
  set table $itk_component(table)
  $table configure -cols 2 -colstretchmode last \
      -cellanchor w -selecttype cell

  itk_component add newbutton {
    button $itk_component(childsite).new \
	-anchor w \
	-text "New" \
	-command [code $this newbutton_action]
  } { }
  
  itk_component add editbutton {
    button $itk_component(childsite).edit \
	-anchor w \
	-text "Edit"\
	-command [code $this editbutton_action]
  } { }
  
  itk_component add detailsbutton {
    button $itk_component(childsite).details\
	-anchor w \
	-text "Details" \
	-command [code $this detailsbutton_action]
  } { }

  itk_component add deletebutton {
    button $itk_component(childsite).delete\
	-anchor w \
	-text "Delete" \
	-command [code $this deletebutton_action]
  } { }

  itk_component add newnetbutton {
    button $itk_component(childsite).newnet\
	-anchor w \
	-text "New Net" \
	-command [code $this newnet]
  } { }

  itk_component add annobutton {
    button $itk_component(childsite).annotation\
	-anchor w \
	-text "Annotations" \
	-command [code $this annobutton_action]
  } { }

  itk_component add icbutton {
    button $itk_component(childsite).ic\
	-anchor w \
	-text "IC" \
	-command [code $this icbutton_action]
  } { }

  itk_component add interactivebutton {
    button $itk_component(childsite).interactive\
	-anchor w \
	-text "Interactive" \
	-command [code $this interactivebutton_action]
  } { }

  itk_component add newnetDialog {
    NewnetDialog $itk_interior.newnetDialog \
	-master . \
	-title "Xcdg: Newnet" \
	-modality application
  } { }

  # packing
  pack $itk_component(newbutton) \
      $itk_component(detailsbutton) \
      $itk_component(newnetbutton) \
      $itk_component(annobutton) \
      $itk_component(icbutton) \
      $itk_component(interactivebutton) \
      -side left -padx 3 

  # (editbutton & deletebutton remain unpacked)

  # register the helpmessages
  .cdgmain help sethelpstr \
      $itk_component(newbutton) "Input a new wordgraph" \
      $itk_component(editbutton) "Edit the selected wordgraph" \
      $itk_component(deletebutton) "Delete the selected wordgraph" \
      $itk_component(detailsbutton) "Show details of the selected wordgraph" \
      $itk_component(newnetbutton) "Compute a new constraint net" \
      $itk_component(annobutton) "Find the corresponding parse in the input" \
      $itk_component(icbutton) "Apply incremental search to the utterance" \
      $itk_component(interactivebutton) "Parse user-supplied input." \
      $itk_component(entryfield) "Enter a wordgraph-id" \
      $itk_component(table) "Click to select"

  # get the initial data
  init_data

  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## refresh the displayed data for a specific ID
## ----------------------------------------------------------------------------
body AllWordgraphs::refreshid {id} {
  set data [getCData $id]
  if {$data == "NULL" } return

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set ids [$table getCell 0,0 $rmax,0]
  set r [lsearch -exact $ids $id]

  if {$r == -1} {
    set r [expr $rmax + 1]
  }
  $table configure -rows $r
  refreshrow $r $data
}


## ----------------------------------------------------------------------------
## fill a row with the values from a specific wordgraph.
## \param row row index to insert data
## \param ptr Pointer to lattice_id
## ----------------------------------------------------------------------------
body AllWordgraphs::getTclData {row ptr} {
    set table $itk_component(table)
    set id [LatticeStruct_id_get $ptr]
    $table setCell $row,0 [LatticeStruct_id_get $ptr]
  
    set sentence ""
    for {set m [LatticeStruct_arcs_get $ptr]}\
	{$m != "NULL"} \
	{set m [listNext $m]} {
	
	    set arc [listItem $m]
	    append sentence " [ArcStruct_word_get $arc]"
	}
    $table setCell $row,1 $sentence
}




## ----------------------------------------------------------------------------
## call _init_data if necessary
## ----------------------------------------------------------------------------
body AllWordgraphs::init_data {} {

  if {[InputStruct_xcdgwordgraphsToDate_get [inputCurrentGrammar_get]]} {
    return
  }

  .cdgmain busy compute "Init Word Graphs" [code $this _init_data]
}



## ----------------------------------------------------------------------------
## get the data from the cdg-tool
## ----------------------------------------------------------------------------
body AllWordgraphs::_init_data {} {

  set table $itk_component(table)
  $table configure -state normal
  set rmax [$table cget -rows]
  $table clear all
  $table delete rows 1 [$table cget -rows]
  $table configure -rows $rmax
  $table tag config default
  $table tag config colored -bg gray80 
  $table width 0 20
  $table setCell row 0,0 { "Id" "Words" }

  array unset wordgraph
  set h [InputStruct_lattices_get [inputCurrentGrammar_get]]
  set m [hashListOfKeys $h]
  for {set l $m}  { $l != "NULL" } { set l [listNext $l ] } {
    set lattice [hashGet $h [listItem $l]]
    set lattice_id [LatticeStruct_id_get $lattice]
    set wordgraph($lattice_id) $lattice
  }
  listFree $m

  set r 1
  set list [inputGetSortedLatticeNames [inputCurrentGrammar_get]]
  for {set l $list}  { $l != "NULL" } { set l [listNext $l ] } {
    set name [pointer2string [listItem $l]]
    lappend allsorted $name
    set lattice $wordgraph($name)
    refreshrow $r $lattice
    incr r
    if {[expr $r % 100] == 0} {
      update
    }
  }
  listFree $list;

  _setCount [hashSize [InputStruct_lattices_get [inputCurrentGrammar_get]]]
  setSelection

  InputStruct_xcdgwordgraphsToDate_set [inputCurrentGrammar_get] 1
  $table configure -state disabled
}



## ----------------------------------------------------------------------------
## actions to take place on pressing the editbutton
## ----------------------------------------------------------------------------
body AllWordgraphs::editbutton_action {} {
  details
  configure -detailstate normal
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the detailsbutton
## ----------------------------------------------------------------------------
body AllWordgraphs::detailsbutton_action {} {
  details
  configure -detailstate disabled
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the newbutton
## ----------------------------------------------------------------------------
body AllWordgraphs::newbutton_action {} {
  .cdgmain confirm "Under Construction\nPlease use `inputwordgraph' in the shell."
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the deletebutton
## ----------------------------------------------------------------------------
body AllWordgraphs::deletebutton_action {} {
  .cdgmain not_yet
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the annobutton
## ----------------------------------------------------------------------------
body AllWordgraphs::annobutton_action {} {

  if {$_selection == ""} {
    ::cmd::Puts "ERROR: Please select a wordgraph"
    return
  }

  foreach item $_selection {
    set lat [findLattice $item]
    if {$lat == "NULL"} {
      ::cmd::Puts "ERROR: no lattice `$item' found."
      return
    }
    for {set l [findAnnotations $lat]} {$l != "NULL" } {set l [listNext $l]} {
     set id [AnnoEntryStruct_id_get [listItem $l]]
      set parse [.cdgmain parses getCData $id]
      .cdgmain parses showparse $id
    }
  }
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the icbutton
## ----------------------------------------------------------------------------
body AllWordgraphs::icbutton_action {} {

  if {$_selection == ""} {
    ::cmd::Puts "ERROR: Please select a wordgraph"
    return
  }

  foreach id $_selection {
    ::cmd::IC $id
  }
}

## ----------------------------------------------------------------------------
## actions to take place on pressing the interactivebutton
## ----------------------------------------------------------------------------
body AllWordgraphs::interactivebutton_action {} {

  # if there already is an interactive window, do nothing
  foreach vis [itcl::find objects -isa VisParses] {
    if {[$vis cget -title] == "IC:interactive"} {
      ::cmd::Puts "ERROR: There already is an open interactive window."
      return
    }
  }

  # build a window to display results
  set interactiveWindow [VisParses ".#auto" \
                             -title "Xcdg - Interactive parsing" \
                             -iconname "Interactive" \
                             -interactive 1 ]
  
  # call IC in interactive mode
  ::cmd::IC interactive


}


## ----------------------------------------------------------------------------
## show the details-window
## ----------------------------------------------------------------------------
body AllWordgraphs::details {} {

  if {[info exist itk_component(details)] && \
	  [winfo exists $itk_component(details)] } {
    $itk_component(details) show [lindex $_selection 0]
  } else {
    itk_component add details {
      WordgraphTable $itk_interior.details [lindex $_selection 0]
    } {
      rename -state -detailstate detailState State
      keep -background -cursor
    }
  }
}

## ----------------------------------------------------------------------------
## actions taking place when the newnetbutton is pressed
## ----------------------------------------------------------------------------
body AllWordgraphs::newnet {} {

  if {"$_selection" != ""} {
    $itk_component(newnetDialog) activate $_selection
  }
}

## ----------------------------------------------------------------------------
## get pointer to a lattice_id, cache it in array wordgraph
## ----------------------------------------------------------------------------
body AllWordgraphs::getCData {id} {
  if {![info exists wordgraph($id)]} {
    set found 0
    set h [InputStruct_lattices_get [inputCurrentGrammar_get]]
    set lattice [hashGet $h [string2pointer $id]]
    if {$lattice == "NULL"} {
      return "NULL"
    }
    set wordgraph($id) $lattice
  }
  return $wordgraph($id)
}



## ----------------------------------------------------------------------------
## get the lattice next after $id.
## ----------------------------------------------------------------------------
body AllWordgraphs::getNext {latticeId} {

  init_data
  if {![info exists wordgraph($latticeId)]} {
    return ""
  }
  set ids [getAllLatticeIds]
  set index [lsearch -exact $ids $latticeId]
  set index [expr $index + 1]
  return [lindex $ids $index]

}

## ----------------------------------------------------------------------------
## Show the annotation for the lattice above $id.
## ----------------------------------------------------------------------------
body AllWordgraphs::getPrev {latticeId} {

  init_data
  if {![info exists wordgraph($latticeId)]} {
    return ""
  }
  set ids [getAllLatticeIds]
  set index [lsearch -exact $ids $latticeId]
  set index [expr $index - 1]
  return [lindex $ids $index]
}


## ----------------------------------------------------------------------------
## Return list of all lattices' names
## ----------------------------------------------------------------------------
body AllWordgraphs::getAllLatticeIds {} {
  init_data
  return $allsorted
}
