# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## DataBrowser - Abstract base class for tabbed panes containing 
## a top frame of buttons and a center table
##
##
## Rules for subclassing:
##
## - Override init_data in order to perform any additional initializations
##
## - Call init_data whenever the frontend has to be set back to an initial
##   state, e.g. after loading a new file 
##
## - Always call init_data at the end of derived constructor just before
##   itk_initialize 
##
## - Add and pack additional components in derived constructor 
##   use component childsite to insert buttons into top panel
##
## - All remaining methods possess default implementation, which may be
##   overridden 
##
## - Use variable _selection to store table row selections
##
## - Set variable _idColumnIndex to index number of column, that will contain 
##   selectable IDs (set at beginning of derived constructor!)
##
## - Override getCData to get a pointer to the underlying C-Structure
## - Override getTclData to get data corresponding to the pointer retrieved by getCData
##   and writing it into the table
##
## \author Dietmar Dreyer (see also AUTHORS and THANKS for more)
##
## $Id: databrowser.tcl,v 1.7 2003/10/28 13:11:27 ddreyer Exp $
## ----------------------------------------------------------------------------

class DataBrowser {
   inherit ::itk::Widget

    # public methods
    public method getSelection {}
    public method setSelection {args}; ## \type TclList

    ## abstract methods for retrieving the relevant C or Tcl data values for a given ID 
    public method getTclData {row id}; ## \type TclNumber, TclString
    public method getCData {id}; ## \type TclString

    public method refreshid {id}; ## \type TclString

    # protected methods
    protected method init_data {}; ## call at the end of derived constructor
    protected method _rowtag {row}; ## \type TclNumber
    protected method _return_action {}    
    protected method _browse_action {w}; ## \type TclWidget
    protected method _motion_action {w x y }; ## \type TclWidget,TclNumber,TclNumber
    protected method _setCount {n}; ## \type TclNumber

    ## abstract method called in refreshid
    protected method refreshrow {row item}; ## \type TclNumber, TclString

    # private methods
    
    private method _getMatchingRows {regex varnameListIDs varnameMapIDs}; ## \type TclString,TclString,TclString
    private method _evalSelectionPatterns {}; ## returns list of matching IDs 

    constructor {args} {};

    # protected data members

    ## string of current selected row ids
    protected variable _selection ""
    
    ## column index of table that contains selectable ids (used in _browse_action)
    protected variable _idColumnIndex ""
    
};

## ----------------------------------------------------------------------------
## A DataBrowser constructor.
## \param args arguments passed to itk_initialize
## ----------------------------------------------------------------------------
body DataBrowser::constructor {args} {
   set _idColumnIndex 1
   itk_component add childsite {
      frame $itk_interior.childsite \
		-borderwidth 0 \
		-relief flat \
   } { 
      keep -cursor -background
   }

   itk_component add entryfield {
      iwidgets::entryfield $itk_interior.entry \
		-textvariable [scope _selection] \
		-labelpos w \
		-textbackground gray80 \
		-labeltext "Selection :"
   } {
      keep -background -cursor -foreground 
      rename -labeltext -entrytext entryText  EntryText
      rename -textfont -entryfont entryFont Font
      rename -textbackground -entrybackground entryBackground Background
   }

   itk_component add table {
      MyTable $itk_interior.table \
		-exportselection 0 \
		-rows 20 \
		-cols 9 \
		-cellanchor c \
		-state disabled \
		-colstretchmode all \
		-colseparator " " \
		-rowstretchmode none \
		-selectmode extended \
		-selecttype row \
		-titlecols 1 \
		-titlerows 1 \
		-rowtagcommand [code $this _rowtag]
   } { 
      keep -background -cursor 
   }

   itk_component add countlabel {
      label $itk_interior.countlabel \
		-relief flat \
		-borderwidth 0
   } { 
      rename -text -count count Amount
      keep -background -cursor
   }

   # packing
   grid $itk_component(childsite) -row 0 -column 0 -sticky ew -padx 3
   grid $itk_component(countlabel) -row 0 -column 1 -sticky ew -padx 3
   grid $itk_component(entryfield) -row 1 -column 0 -columnspan 2 -sticky ew -padx 3 -pady 3
   grid $itk_component(table) -row 2 -column 0 -columnspan 2 -sticky news -padx 3
   grid rowconfigure $itk_interior 2 -weight 1
   grid columnconfigure $itk_interior 0 -weight 1

    # some bindings
    set table [$itk_component(table) component table]
    bind $table <Motion> "[code $this _motion_action %W %x %y]; break"
    set ef [$itk_component(entryfield) component entry]
    bind $ef <Return> [code $this _return_action]
    #bind $ef <KeyRelease> [code "$this _keypress_action; continue"]
    bind $table <ButtonRelease-1> [code $this _browse_action %W]

   # Default Table Configurations
   $table tag config default
   $table tag config colored -bg gray80

   eval itk_initialize $args
}


## ----------------------------------------------------------------------------
## browse slot.
## This method adjusts the selected file in AllFiles::_selection.
## \param w the widget bound to this slot.
## ----------------------------------------------------------------------------
body DataBrowser::_browse_action {w} {

   set ids ""
   set lastrow 0
   foreach index [$w curselection] {
      scan $index "%d,%d" row col
      if {$lastrow == $row } {
	 continue
      }
      set lastrow $row
      set id [$w get $row,$_idColumnIndex] 
      if {$id != ""} {
	 lappend ids $id
      }
   }
   eval setSelection $ids
}


## ----------------------------------------------------------------------------
## Returns a list of all row indices matching 'regex' in the ID column
## \param regex regular expression string  
## \param varnameListIDs in-out-parameter, name of local variable to append matching IDs
## \param varnameMapIDs in-out-parameter, name of local array-variable to insert matching IDs
##        in order to prohibit duplicate selection entries in entry field
## ----------------------------------------------------------------------------
body DataBrowser::_getMatchingRows {regex varnameListIDs varnameMapIDs} {
    upvar $varnameListIDs matchingIDs
    upvar $varnameMapIDs mapIDs
    set table $itk_component(table)

    set rmax [$table cget -rows]
    set ids [$table getCell 1,$_idColumnIndex $rmax,$_idColumnIndex]
  
    set rows ""
    foreach id $ids {
	if {[regexp -expanded -- "^$regex$" $id match] == 1} {
	    set row [lsearch $ids $match]
	
	    if {$row >= 0 && [info exists mapIDs($match)] == 0} {
		lappend rows $row
		lappend matchingIDs $match
		set mapIDs($match) 1
	    }
	}
    }
    return $rows
}


## ----------------------------------------------------------------------------
## actions to take place on pressing return in the entryfield.
## ----------------------------------------------------------------------------
body DataBrowser::_return_action {} {
    set _selection [_evalSelectionPatterns]
    set rows [eval setSelection $_selection]
   
   if {$rows != {} } {
      $itk_component(table) see "[lindex $rows end],0"
   }
}

## ----------------------------------------------------------------------------
## Evaluates input to the entry field as a list of regular expressions
## and matches them against all IDs in order to return a list of matching
## IDs 
## ----------------------------------------------------------------------------
body DataBrowser::_evalSelectionPatterns {} {
   set table $itk_component(table)

  
   set rows ""
   set matchingIDs ""
   set mapIDs(0) ""
  
   foreach pattern $_selection {
       set match [_getMatchingRows $pattern matchingIDs mapIDs]
      
       if { $match != "" } {
	   set rows [concat $rows $match] 
       }
     
       if {[llength $rows] > 0} {
	   foreach row $rows {
	       $table selection set $row,0
	   }
       }
   }

   if {$rows != {} } {
      $itk_component(table) see "[lindex $rows end],0"
   }
    
   return $matchingIDs

}


## ----------------------------------------------------------------------------
## return a list of selected ids
## ----------------------------------------------------------------------------
body DataBrowser::getSelection {} {
   return $_selection
}



## ----------------------------------------------------------------------------
## select one or more rows.
## This method sets the AllFiles::_selection variable.
## A previous selection is cleared;
## an empty selection removes all selections.
## \param args a list of all row ids in the files table
## ----------------------------------------------------------------------------
body DataBrowser::setSelection {args} {
   set table $itk_component(table)

   set rmax [$table cget -rows]
   set ids [$table getCell 0,$_idColumnIndex $rmax,$_idColumnIndex]
   $table selection clear 0,0 end
  
   set newselection ""
   set rows ""
   foreach id $args {
      set row [lsearch $ids $id]
      if {$row >= 0} {
	 $table selection set $row,0
	 lappend newselection $id
	 lappend rows $row
      }
   }
   set _selection $newselection

   return $rows
}

## ----------------------------------------------------------------------------
## colorize the table rows
## This method is a callback configured to the table in order to colorize the
## rows.
## ----------------------------------------------------------------------------
body DataBrowser::_rowtag {row} {
  if {[expr $row % 2]} {
    return "default"
  } else {
    return "colored"
  }
}


## ----------------------------------------------------------------------------
## display the count-label.
## This number should reflect the number of items selected
## \param n the number to be set
## ----------------------------------------------------------------------------
body DataBrowser::_setCount {n} {
   $itk_component(countlabel) configure -text "Count: $n"
}


## ----------------------------------------------------------------------------
## Default motion slot.
## \param w the widget where the motion was detected
## \param x the x coords of the mouse 
## \param y the y coords of the mouse
## ----------------------------------------------------------------------------
body DataBrowser::_motion_action {w x y} {
    focus $w
}


## ----------------------------------------------------------------------------
## refresh the displayed data for a specific ID
## ----------------------------------------------------------------------------
body DataBrowser::refreshid {id} {
  set data [getCData $id]
  if {$data == "NULL"} return

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set ids [$table getCell 0,0 $rmax,0]
  set r [lsearch -exact $ids $id]

  if {$r == -1} {
    init_data
  } else {
    refreshrow $r $data
  }
}

## ----------------------------------------------------------------------------
## fill a row with the values from a specific item
## ----------------------------------------------------------------------------
body DataBrowser::refreshrow {row item} {
  set table $itk_component(table)
  $table configure -state normal
  set rmax [$table cget -rows]
  if {$row >= $rmax} {
    $table configure -rows [expr $row + 1]
  }  
  getTclData $row $item
  $table configure -state disabled
}
