# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## CdgProcess - interface for a cdgp process.
## This class wraps arround a cdgp command spawned in the background and tries
## to provide all you need to handle a cdgp job. The class communicates with
## the cdgp job connecting to its STDIN and reads back all its output on STDOUT.
## \ingroup YadaScheduler
##
## \author Michael Daum 
##
## $Id: CdgProcess.tcl,v 1.11 2003/11/19 09:39:06 micha Exp $
## ----------------------------------------------------------------------------
class CdgProcess {
  inherit Process

  # variables ----------------------------------------------------------------

  ## the file where to store the xml output of the logfile
  public variable xmlFile	""
  
  ## if you like to call cdgp with an init file (-i commandline option), put
  ## that filename in here.
  public variable initFile	""

  ## binary copy usage.
  ## This flag indicates whether to use a private copy of the cdgp binary in a
  ## temp directory or not.
  public variable useCopy	0

  ## binary location.
  ## This variable stores the absolute path to the cdgp binary that we wrap arround.
  public variable cdgp		""

  constructor {args} {}; ## \type TclList
  destructor {}


  # methods ------------------------------------------------------------------
  public method start {{command ""}}; ## \type TclString
  public method stop {}

};

## ----------------------------------------------------------------------------
## constructor.
## This constructor builds a cdgp wrapper object. If you don't specify a
## cdgp binary yourself (with -cdgp "/here/it/is/cdgp"), we assume the binary
## to be in \c "$env(YADA_TMP)/cdgp-[pid]".
## ----------------------------------------------------------------------------
body CdgProcess::constructor {args} {
  global env

  eval configure $args

  # did we already provide a command or not
  if {$cdgp == ""} {
    if {$useCopy} {
      set cdgp "$env(YADA_TMP)/cdgp-[pid]"

      # making a copy of the original cdgp 
      exec cp $env(YADA_CDGP) $cdgp
    } else {
      set cdgp $env(YADA_CDGP)
    }
  }
}

## ----------------------------------------------------------------------------
## destructor.
## Before we leave we remove the private copy of the cdgp binary iff \c useCopy
## is \c true.
## ----------------------------------------------------------------------------
body CdgProcess::destructor {} {

  if {$useCopy} {
    # delete the binary which we made a copy of
    exec rm $cdgp
  }
}

## ----------------------------------------------------------------------------
## start the cdgp process.
## This forks the cdgp in the background, which is actually done in Process::start().
## \param command this lets you overwrite the command given in CdgProcess::cdgp.
## ----------------------------------------------------------------------------
body CdgProcess::start {{command ""}} {

  # prepair the command line
  if {$command == ""} {
    if {![file exists $cdgp]} {
      return -1;
    }
    set command "$cdgp"
  }
  
  if {$xmlFile != ""} {
    # strip off gz 
    regsub {^(.*)\.gz$} $xmlFile {\1} clearXmlFile
    append command " -x $clearXmlFile"
  } 

  if {$initFile == ""} {
    append command " -i none"
  } else {
    append command " -i $initFile"
  }

  # start the process
  set pid [chain $command]
  
  return $pid
}

## ----------------------------------------------------------------------------
## stop the cdgp process.
## This method stops a started cdgp process and gzips the xmllogfile. Afterwards
## Process::stop() is called to finalize everything. There  the job scheduled
## to be \a consumed.
## ----------------------------------------------------------------------------
body CdgProcess::stop {} {

  # compress xmlFile
  if {$xmlFile != ""} {
    regsub {^(.*)\.gz$} $xmlFile {\1} clearXmlFile
    if {[file exists $clearXmlFile]} {
      if {[catch {exec gzip -f $clearXmlFile} errMsg]} {
	print "ERROR: $errMsg\n$::errorInfo\n"
      }
    }
  }

  # Process::stop and consume
  chain
}
