# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## Process - interface for a forked process.
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## $Id: Process.tcl,v 1.13 2003/02/27 13:02:26 micha Exp $
## ----------------------------------------------------------------------------
class Process {

  # variables ----------------------------------------------------------------
  public variable logFile	""
  public variable printCommand  ""
  public variable sync		0
  public variable job		""
  public variable consumerCommand ""
  public variable showProgress	0
  public variable bufferSize	81920

  private variable _pid		-1
  private variable _pipe	""
  private variable _logFileId	-1
  
  # methods ------------------------------------------------------------------
  public method start {command}; ## \type TclString \virtual
  public method stop {}; ## \virtual
  public method tell {args}; ## \type TclList
  public method isAlive {}
  public method wait {}
  public method signal {signo}; ## \type TclNumber
  public method interrupt {}
  public method term {}
  public method kill {}
  public method nice {{prio 10}}; ## \type TclNumber
  public method getPid {}
  public method getJobState {}
  public method getPassedJobStates {}

  constructor {args} {}; ## \type TclList
  destructor {}

  private method _readPipe {}
  protected method _print {text}; ## \type TclString
  
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body Process::constructor {args} {
  set printCommand [code puts -nonewline]
  eval configure $args
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body Process::destructor {} {

  # terminate the process
  if {$_pid > 0} {
    if {[catch {exec kill $_pid} errMsg]} {
      _print "ERROR: $errMsg\n$::errorInfo\n"
    }
  }
  set _pid -1

  # close the command pipe
  if {$_pipe != ""} {
    fileevent $_pipe readable ""
    catch {close $_pipe}
    set _pipe ""
  }

  # close the logfile
  if {$_logFileId > 0} {
    catch {close $_logFileId}
    set _logFileId -1

    # compress logFile
    regsub {^(.*)\.gz$} $logFile {\1} clearLogFile
    if {[file exists $clearLogFile]} {
      if {[catch {exec gzip -f [glob $clearLogFile]} errMsg]} {
	_print "ERROR: $errMsg\n$::errorInfo\n"
      }
    }
  }

  # call the consumer
  if {$consumerCommand != "" && $job != ""} {

    # this process can only be consumed once
    set thisConsumerCommand "$consumerCommand"
    set consumerCommand ""
    if {[catch {$thisConsumerCommand $job} errMsg]} {
      _print "ERROR: $errMsg\n$::errorInfo\n"
    }
  }

  # unregister from machine
  if {$job != ""} {
    set machine ""
    catch {set machine [$job cget -machine]}
    if {$machine != ""} {
      $machine stopJob $job
    }
  }
}

## ----------------------------------------------------------------------------
## getPid
## ----------------------------------------------------------------------------
body Process::getPid {} {
  return $_pid
}

## ----------------------------------------------------------------------------
## start
## ----------------------------------------------------------------------------
body Process::start {command} {

  if {$logFile == ""} {
    set cmdPostfix "> /dev/null"
    set _logFileId -1
  } else {
    set cmdPostfix ""
    regsub {^(.*)\.gz$} $logFile {\1} clearLogFile
    set dirname [file dirname $clearLogFile]
    file mkdir $dirname
    set _logFileId [open $clearLogFile w]
  }

  if {[catch {set _pipe [open "|bash -c \"exec $command 2>&1 $cmdPostfix\"" r+]} errMsg]} {
    _print "ERROR: $errMsg\n$::errorInfo\n"
    stop
    return 0
  }

  set _pid [pid $_pipe]

  fconfigure $_pipe -blocking 0 -translation binary -buffersize $bufferSize
  fileevent $_pipe readable [code $this _readPipe ]

  return $_pid
}

## ----------------------------------------------------------------------------
## stop
## ----------------------------------------------------------------------------
body Process::stop {} {
  itcl::delete object $this
}

## ----------------------------------------------------------------------------
## _readPipe
## ----------------------------------------------------------------------------
body Process::_readPipe {} {

  # temporarily disable read pipe event
  fileevent $_pipe readable ""

  if {$_pipe == "" || [eof $_pipe]} {
    after idle [code $this stop]
    return
  } 

  set buffer [read $_pipe $bufferSize]

  if {$_logFileId > 0} {
    puts -nonewline $_logFileId "$buffer"
    flush $_logFileId
  }

  if {$showProgress} {
    _print $buffer
  }

  # reenable read pipe event
  fileevent $_pipe readable [code $this _readPipe]
}

## ----------------------------------------------------------------------------
## _print
## ----------------------------------------------------------------------------
body Process::_print {text} {

  if {$printCommand != ""} {
    $printCommand "$text"
  }
}


## ----------------------------------------------------------------------------
## wait
## ----------------------------------------------------------------------------
body Process::wait {} {

  if {$_pid > 0 && ![eof $_pipe]} {
    flush $_pipe
    vwait [scope _pid]
  }
}

## ----------------------------------------------------------------------------
## isAlive
## ----------------------------------------------------------------------------
body Process::isAlive {} {
  global tcl_platform

  switch $tcl_platform(os) {
    "SunOS" {
      set psCommand "ps -a"
    }
    "Linux" {
      set psCommand "ps -x"
    }
    "Windows NT" {
      set psCommand "ps"
    }
    "Windows 95" {
      set psCommand "ps"
    }
    "Windows 98" {
      set psCommand "ps"
    }
    default {
      error "unknown platfprm $tcl_platform(os)"
    }
  }

  #puts "ps='[eval exec $psCommand]'"

  if {[catch {set psInfo [eval exec $psCommand | grep $_pid | grep -v grep]}]} {
    return 0
  }

  if {$psInfo != ""} {
    return 1
  }

  return 0
}

## ----------------------------------------------------------------------------
## interrupt
## ----------------------------------------------------------------------------
body Process::interrupt {} {
  signal 2
}

## ----------------------------------------------------------------------------
## term
## ----------------------------------------------------------------------------
body Process::term {} {
  signal 15
}

## ----------------------------------------------------------------------------
## kill
## ----------------------------------------------------------------------------
body Process::kill {} {
  signal 9
}

## ----------------------------------------------------------------------------
## signal
## ----------------------------------------------------------------------------
body Process::signal {signo} {
  if {$_pid > 0} {
    if {[catch {exec kill -$signo $_pid} errMsg]} {
      _print "ERROR: $errMsg\n$::errorInfo\n"
    }
  }
}

## ----------------------------------------------------------------------------
## tell
## ----------------------------------------------------------------------------
body Process::tell {args} {

  if {[eof $_pipe]} {
    return
  }

  if {$args == ""} {
    return
  }

  foreach something $args {
    # puts "DEBUG: tell `$something'"
    puts $_pipe $something
  }

  if {$sync} {
    flush $_pipe
  }
}

## ----------------------------------------------------------------------------
## nice
## ----------------------------------------------------------------------------
body Process::nice {{prio 10}} {
  if {$_pid > 0} {
    if {[catch {exec renice $prio $_pid} errMsg]} {
      _print "ERROR: $errMsg\n$::errorInfo\n"
    }
  }
}

## ----------------------------------------------------------------------------
## getJobState
## ----------------------------------------------------------------------------
body Process::getJobState {} {
  if {$job == ""} {
    return "unknown"
  }
  return [$job getState]
}

## ----------------------------------------------------------------------------
## getPassedJobStates
## ----------------------------------------------------------------------------
body Process::getPassedJobStates {} {
  if {$job == ""} {
    return [list "unknown"]
  }
  return [$job getPassedStates]
}

## ----------------------------------------------------------------------------
## callback for Process::showProgress
## showProgress
## ----------------------------------------------------------------------------
configbody Process::showProgress {
  if {$_logFileId > 0} {
    flush $_logFileId
  }
}

