# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaExperiment - bundled information about a experiment.
## A YadaExperiment is the central object of the YADA application. It brings
## together all the parts that are needed to start YadaJob objects forking the parser
## process in the background and consuming the produced data.
##
## So a YadaExperiment 
## - has a unique name,
## - needs a YadaGrammar, 
## - needs a command to fork the parser
## - specifies a YadaMachine where all jobs are started on,
## - a script to be played to the parser itself and
## - a unique data directory where the results of the experiments are stored
##   (two files per wordgraph: the standard output and the xml output data)
##
## All YadaExperiment objects are accessed via the YadaExperiments configuration
## document. YadaGrammars and YadaMachines that are available are configured separately.
##
## \ingroup YadaConfiguration
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## $Id: YadaExperiment.tcl,v 1.4 2003/12/18 12:59:27 micha Exp $
## ----------------------------------------------------------------------------
class YadaExperiment {
  inherit YadaConfigItem

  # variables ----------------------------------------------------------------

  ## the name of the experiment
  public variable name ""

  ## the type of the experiment. 
  ## The type of the experiment specifies the parsing flavour used.
  ## Possible sensefull values by now are 
  ## - annotation (which is actually now parsing flavour but a way to compute
  ##   parses via anno2parse())
  ## - arcconsistency 
  ## - frobbing 
  ## - genetic
  ## - gls 
  ## - incremental completition
  ## - isearch 
  ## - netsearch 
  ## - pruning 
  ## - wfst
  public variable type ""

  ## path to cdgp binary file
  public variable command ""

  ## path where to store all computed data files
  public variable dataDir ""

  ## YadaGrammar to appy this experiment.
  public variable grammarName ""

  ## YadaMachine where to schedule resulting YadaJobs on.
  public variable machineName "localhost"

  ## CDGP script to feed into the STDIN of a CdgProcess.
  public variable script ""

  ## flag indicating whether this experiment is currently waiting for 
  ## a YadaJob to finish.
  public variable isActive 1

  ## flag indicating whether all data has been computed for this
  ## experiment, so no more YadaJobs are needed.
  public variable isComplete 0

  ## flag indicating whether all statistics in all data has been
  ## collected or not.
  public variable isUpToDate 0

  ## tcl command used to print a message. 
  ## \see print()
  public variable printCommand ""

  ## total scores collected in the experiments data files. 
  public variable totalScore 0

  ## labelled hits collected in the experiments data files. 
  public variable labelledHits 0

  ## strict hits collected in the experiments data files. 
  public variable strictHits 0

  ## lexical hits collected in the experiments data files. 
  public variable lexicalHits 0

  ## struct hits collected in the experiments data files. 
  public variable structHits 0

  ## total number of tries collected in the experiments data files. 
  public variable totalTries 0

  ## total time collected in the experiments data files. 
  public variable totalTime 0

  ## total solution time collected in the experiments data files. 
  public variable totalSolTime 0

  ## total soft time collected in the experiments data files. 
  public variable totalSoftTime 0

  # methods ------------------------------------------------------------------
  public method clone {args}; ## \type TclList
  public method check {}
  public method start {job}; ## \type YadaJob
  public method print {message}; ## \type TclString
  public method resetStatistics {}
  public method getXmlFileName {wgName}; ## \type TclString
  public method getLogFileName {wgName}; ## \type TclString
  public method getDocFileName {}

  constructor {args} {}; ## \type TclList

};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaExperiment::constructor {args} {

  eval configure $args

  # check obliatory options
  if {"$name" == ""} {
    error "ERROR: unnamed experiment"
  }
  if {"$type" == ""} {
    error "ERROR: experiment $name untyped"
  }
  if {"$grammarName" == ""} {
    error "ERROR: experiment $name w/o grammar"
  }
  if {"$command" == ""} {
    error "ERROR: experiment $name has no command"
  }
  
  # create default data store in
  # YADA_DATA/machineName/grammarName/type/name
  if {"$dataDir" == ""} {
    set dataDir [file join $::env(YADA_DATA) $machineName $grammarName $type $name]   
  }

  # security filter for dataDir
  if {$name != "<none>"} {
    set securityFilter {[\\\*\?\~\^\$\@\%\`\"\'\&\;\|\<\>\x00-\x1F]} ;# from twiki
    regsub -all $securityFilter $dataDir "" dataDir
  }

  # check if the named experiment already exists
  foreach experiment [itcl_info object -class YadaExperiment] {
    if {$experiment == $this} {
      continue
    }
    if {[$experiment cget -name] == $name} {
      error "ERROR: experiment $name already exists"
    }
    if {[$experiment cget -dataDir] == $dataDir} {
      error "ERROR: experiment $name uses same store"
    }
  }

  # create data store
  if {$name != "<none>" && ![file exists "$dataDir"]} {
    file mkdir "$dataDir"
    set isComplete 0
  }
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaExperiment::clone {args} {
  set newExperiment [YadaExperiment ::#auto \
    -name $name \
    -type $type \
    -command $command \
    -dataDir $dataDir \
    -grammarName $grammarName \
    -machineName $machineName \
    -script $script \
    -isActive $isActive
  ]

  eval $newExperiment configure $args

  return $newExperiment
}

## ----------------------------------------------------------------------------
## check
## ----------------------------------------------------------------------------
body YadaExperiment::check {} {

  if {$name == "<none>"} {
    return 1
  }
  
  if {$dataDir == "" || $grammarName == ""} {
    return 0
  }

  # check if the experiment is complete
  if {!$isComplete} {
    set grammar [.main getGrammar $grammarName]
    set wordgraphs [$grammar getSelection]
    set isComplete 1
    foreach wordgraph $wordgraphs {
      set file [getXmlFileName $wordgraph]
      if {![file exists $file]} {
	set isComplete 0
	break
      }
    }
  }

  return $isComplete
}

## ----------------------------------------------------------------------------
## start
## ----------------------------------------------------------------------------
body YadaExperiment::start {job} {
  set wordgraph [$job cget -wordgraphName]
  set grammar [.main getGrammar $grammarName]
  set process [$job cget -process]
  if {$process == ""} {
    error "ERROR: experiment cannot be applied to an unstarted job"
  }

  # replace makros in the script
  set tmpScript $script
  foreach file [$grammar getFiles] {
    regsub -all -- "%load" $tmpScript "load $file\n%load" tmpScript
  }
  regsub -all -- "%load" $tmpScript "" tmpScript
  regsub -all -- "%wordgraph" $tmpScript $wordgraph tmpScript
  regsub -all -- "%path" $tmpScript $dataDir tmpScript

  # start the script
  $process tell "$tmpScript\nquit\n"
}

## ----------------------------------------------------------------------------
## print
## ----------------------------------------------------------------------------
body YadaExperiment::print {message} {
  if {$printCommand != ""} {
    $printCommand $message
  }
}

## ----------------------------------------------------------------------------
## getXmlFileName
## ----------------------------------------------------------------------------
body YadaExperiment::getXmlFileName {wgName} {
  return [file join $dataDir ${wgName}.xml.gz]
}

## ----------------------------------------------------------------------------
## getLogFileName
## ----------------------------------------------------------------------------
body YadaExperiment::getLogFileName {wgName} {
  return [file join $dataDir ${wgName}.log.gz]
}

## ----------------------------------------------------------------------------
## getDocFileName
## ----------------------------------------------------------------------------
body YadaExperiment::getDocFileName {} {
  return [file join $dataDir yada.xml]
}

## ----------------------------------------------------------------------------
## resetStatistics
## ----------------------------------------------------------------------------
body YadaExperiment::resetStatistics {} {
  set totalScore 0
  set labelledHits 0
  set structHits 0
  set lexicalHits 0
  set strictHits 0
  set totalTries 0
  set totalTime 0
  set totalSolTime 0
  set totalSoftTime 0
  set noSolutions 0
  set isUpToDate 0
}

## ----------------------------------------------------------------------------
## callback for script.
## ----------------------------------------------------------------------------
configbody YadaExperiment::script {
  regsub -all {^\s+} $script "" script
  regsub -all {\n\s+} $script "\n" script
}

