# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaGrammar - bundled information about a CDG grammar
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaGrammar.tcl,v 1.20 2003/11/19 09:39:06 micha Exp $
## ----------------------------------------------------------------------------
class YadaGrammar {
  inherit YadaConfigItem

  # variables ----------------------------------------------------------------
  
  ## name of the grammar
  public variable name ""

  ## list of all files making up the grammar
  private variable _files {}

  ## list of selected wordgraph ids
  private variable _selection ""

  ## array mapping wordgraph ids to YadaWordgraph s
  private variable _wordgraphs; ## \type TclArray

  ## flag indicating whether the grammars _wordgraphs are up to date. If not
  ## we reconstruct this array in scan() or rescan().
  private variable _isUpToDate 0

  # methods ------------------------------------------------------------------

  constructor {args} {}; ## \type TclList
  destructor {}

  public method deleteWordgraphs {}
  public method getWordgraphs {}
  public method getWordgraph {wgName}; ## \type TclString
  public method setWordgraph {wordgraph}; ## \type YadaWordgraph
  public method setSelection {args}; ## \type TclList
  public method getSelection {}
  public method setFiles {args}; ## \type TclList
  public method getFiles {}
  public method scan {}
  public method rescan {}
  public method clone {args}; ## \type TclList
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaGrammar::constructor {args} {
  eval configure $args
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaGrammar::destructor {} {
  deleteWordgraphs
}

## ----------------------------------------------------------------------------
## deleteWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammar::deleteWordgraphs {} {
  foreach wgName [array names _wordgraphs] {
    itcl::delete object $_wordgraphs($wgName)
    unset _wordgraphs($wgName)
  }
}

## ----------------------------------------------------------------------------
## getWordgraphs
## ----------------------------------------------------------------------------
body YadaGrammar::getWordgraphs {} {
  set result ""
  foreach wgName [lsort -dictionary [array names _wordgraphs]] {
    lappend result $_wordgraphs($wgName)
  }

  return "$result"
}

## ----------------------------------------------------------------------------
## getWordgraph
## ----------------------------------------------------------------------------
body YadaGrammar::getWordgraph {wgName} {
  if {[info exists _wordgraphs($wgName)]} {
    return $_wordgraphs($wgName)
  } else {
    return ""
  }
}

## ----------------------------------------------------------------------------
## setWordgraph
## ----------------------------------------------------------------------------
body YadaGrammar::setWordgraph {wordgraph} {
  set wgName [$wordgraph cget -name]
  if {[info exists _wordgraphs($wgName)]} {
    itcl::delete $_wordgraphs($wgName)
  }

  set _wordgraphs($wgName) $wordgraph
}

## ----------------------------------------------------------------------------
## scan
## ----------------------------------------------------------------------------
body YadaGrammar::scan {} {
  if {!$_isUpToDate} {
    rescan
  }
}

## ----------------------------------------------------------------------------
## rescan
## ----------------------------------------------------------------------------
body YadaGrammar::rescan {} {
  global env

  .main printStatus "... rescanning grammar '$name'"

  # remove all contained wordgraphs from the grammar
  deleteWordgraphs

  # build a temporary xml log file
  set xmlFile [file join $env(YADA_TMP) _yada[pid].xml.gz]

  # build a cdg process which is to be asked about wordgraphs and annotations
  set machine [.main getMachine "localhost"]

  set process [CdgProcess ::#auto \
    -printCommand "" \
    -logFile "" \
    -xmlFile $xmlFile \
    -initFile none]

  # start a cdgp process
  set thisPid [$process start]
  if {$thisPid < 0} {
    .main printStatus ""
    printMessage "ERROR: cannot start cdgp process $env(YADA_CDGP)"
    return
  }

  # load the grammar files
  foreach file $_files {
    if {[regexp {.*\.gz} $file]} {
      $process tell "set preprocessor /usr/bin/zcat"
      $process tell "set prepsuffix .gz"
    } elseif {[regexp {.*\.m4} $file]} {
      $process tell "set preprocessor /usr/bin/m4 -s -P"
      $process tell "set prepsuffix .m4"
    }
    $process tell "load $file"
  }

  # give us the wordgraphs
  $process tell "wordgraph\nannotation\nquit\n"
  $process wait
  #itcl::delete object $process

  # get the document data
  set fp [open "|zcat $xmlFile" r]
  set data [read $fp]
  close $fp
  #file delete $xmlFile

  # parse it
  set document [dom parse $data]

  set node [$document getElement "cdgp"]
  if {$node == ""} {
    printMessage "ERROR: no cdgp logfile while recaning grammar." error
  }

  # parse all lattice nodes
  foreach latticeNode [$node getElementsByTagName "lattice"] {
    
    # build a new yada wordgraph
    set yadaWordgraph [YadaWordgraph ::#auto]
    $yadaWordgraph readXmlNode $latticeNode

    # assert it into the grammar
    setWordgraph $yadaWordgraph
  }
  
  # parse all annotations
  foreach annoNode [$node getElementsByTagName "annotation"] {
    set annoId [$annoNode getAttribute "id"]
    set yadaWordgraph [getWordgraph $annoId]
    $yadaWordgraph readXmlAnno $annoNode
  }

  # delete it
  $document delete
  set document ""
  set _isUpToDate 1
  .main printStatus "done rescanning"

  # reestablish the selection
  eval setSelection $_selection
}

## ----------------------------------------------------------------------------
## setSelection
## ----------------------------------------------------------------------------
body YadaGrammar::setSelection {args} {
  set _selection $args
}

## ----------------------------------------------------------------------------
## getSelection
## ----------------------------------------------------------------------------
body YadaGrammar::getSelection {} {
  return "$_selection"
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaGrammar::clone {args} {

  # build a new grammar
  set newGrammar [YadaGrammar ::#auto -name $name]
  eval $newGrammar setFiles $_files

  # clone all wordgraphs
  foreach wgName [array names _wordgraphs] {
    set newWordgraph [$_wordgraphs($wgName) clone]
    $newGrammar setWordgraph $newWordgraph
  }

  # clone the selection
  eval $newGrammar setSelection $_selection

  eval $newGrammar configure $args
 
  return $newGrammar
}

## ----------------------------------------------------------------------------
## getFiles
## ----------------------------------------------------------------------------
body YadaGrammar::getFiles {} {
  return $_files
}

## ----------------------------------------------------------------------------
## setFiles
## ----------------------------------------------------------------------------
body YadaGrammar::setFiles {args} {
  
  set _files ""
  foreach file $args {
    lappend _files [glob $file]
  }
}
