# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaMachine - bundled information about a yada machine
## \note Actually this class should be split into a YadaScheduler and a
## YadaMachine the latter of which only contains the configurational part of
## the current implementation. As things are we add this class to the YadaScheduler
## module \a and the YadaConfiguration module, a construction that might be a little
## wired.
## \ingroup YadaConfiguration
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## $Id: YadaMachine.tcl,v 1.9 2004/01/20 11:20:58 micha Exp $
## ----------------------------------------------------------------------------
class YadaMachine {
  inherit YadaConfigItem

  # variables ----------------------------------------------------------------

  ## name of the machine
  public variable name "localhost"

  ## list ip adresses host names making up this machine
  public variable nodes {127.0.0.1}

  ## overall sum of cpus in all nodes
  public variable cpus 1

  ## acess manor for this machine.
  ## possible values are "local" or any way of getting a shell on the nodes, i.e.
  ## \code ssh -x %node \endcode
  ## Notice, that the string \c %node is replaced with the actuall node name in
  ## YadaMachine::nodes to which a specific connection is to be established.
  public variable access "local"

  ## list of all YadaJob s queued up to be scheduled on all nodes
  private variable _jobQueue ""

  ## list of YadaJob s currently running on the nodes
  private variable _runQueue ""

  ## statistics about the average duration of one YadaJob.
  ## The machine updates this figure the more jobs have been seen.
  private variable _timePerJob 60000

  ## number of jobs run on this machine.
  ## This number is related to _timePerJob computing the average duration
  ## seen in _noComputedJobs.
  private variable _noComputedJobs 10
  
  ## used for profiling jobs on this machine.
  ## This hash mapps YadaJob objects to a clock tick.
  private variable _timerOfJob; ## \type TclArray

  ## hash mapping machine node names to the YadaJob s running on them
  private variable _jobsOnNode; ## \type TclArray

  # methods ------------------------------------------------------------------
  public method getLoad {}
  public method getTimePerJob {}
  public method startJob {}
  public method stopJob {job}; ## \type YadaJob
  public method wait {}
  public method queueJob {job}; ## \type YadaJob

  constructor {args} {}; ## \type TclList
  destructor {}

  private method _getNextNode {}
  private method _addTime {time}; ## \type TclNumber
  private method _startJob {}
};


## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaMachine::constructor {args} {
  eval configure $args
}

## ----------------------------------------------------------------------------
## wait
## ----------------------------------------------------------------------------
body YadaMachine::wait {} {
  while {1} {
    if {[llength $_runQueue] < $cpus} {
      return
    }
    vwait [scope _runQueue]
  }
}

## ----------------------------------------------------------------------------
## queueJob
## ----------------------------------------------------------------------------
body YadaMachine::queueJob {job} {
  lappend _jobQueue $job
  $job configure -machine $this
}

## ----------------------------------------------------------------------------
## _getNextNode
## ----------------------------------------------------------------------------
body YadaMachine::_getNextNode {} {
  set maxNodes [llength [array names _jobsOnNode]]
  
  if {$maxNodes < 1} {
    error "ERROR: invalid number $maxNodes of nodes in machine $name"
  }

  set minNode ""
  set minJobs 9999999
  foreach node [array names _jobsOnNode] {
    if {$_jobsOnNode($node) < $minJobs} {
      set minNode $node
      set minJobs $_jobsOnNode($node)
    }
  }

  return $minNode
}

## ----------------------------------------------------------------------------
## getLoad
## ----------------------------------------------------------------------------
body YadaMachine::getLoad {} {
  if {$cpus == 0} {
    return 0.0
  } else {
    return [expr [llength $_runQueue] / (0.0 + $cpus)]
  }
}

## ----------------------------------------------------------------------------
## startJob
## schedule jobs
## ----------------------------------------------------------------------------
body YadaMachine::startJob {} {
  set n 0

  if {$_jobQueue != ""} {
    for {set i [expr $cpus - [llength $_runQueue]]} {$i > 0} {incr i -1} {
      after idle [code $this _startJob]
      incr n
    }
    update
  }

  return $n
}

## ----------------------------------------------------------------------------
## _startJob
## remove the next YadaJob from the queue and run it
## ----------------------------------------------------------------------------
body YadaMachine::_startJob {} {
  if {$_jobQueue == ""} {
    return
  }

  # remove the next job
  set job [lindex $_jobQueue 0]
  $job setState "dequeuing"
  set _jobQueue [lreplace $_jobQueue 0 0]
  set _timerOfJob($job) [clock seconds]
  lappend _runQueue $job

  # build a process for that job
  set experiment [$job cget -experiment]
  set runner [$job cget -runner]
  set cdgp [$experiment cget -command]
  set wordgraphName [$job cget -wordgraphName]
  set logFile [$experiment getLogFileName $wordgraphName]
  set xmlFile [$experiment getXmlFileName $wordgraphName]
  set printCommand [code $runner print]
  set consumerCommand [$job cget -consumerCommand]

  # check file existence
  if {[file exists $xmlFile]} {
    stopJob $job
    return
  }

  # mark the wordgraph to be active 
  $runner markWordgraph $wordgraphName ;# unmarked in the consumer
  
  set process [CdgProcess ::#auto \
    -printCommand $printCommand \
    -cdgp $cdgp \
    -logFile $logFile \
    -xmlFile $xmlFile \
    -consumerCommand $consumerCommand \
    -job $job \
    -sync 1 \
    -initFile none ]
    
  $job configure -process [scope $process]

  # build the command with regard to the access experiment
  if {$access == "local"} {
    set command $cdgp
  } else {
    set newNode [_getNextNode]
    $job configure -node $newNode
    incr _jobsOnNode($newNode)
    
    if {[regsub -all -- "%node" $access $newNode tmpAccess]} {
      set command "$tmpAccess $cdgp"
    } else {
      set command "$access $cdgp"
    }

    # show scheduling
    if {0} {
      foreach node [lsort [array names _jobsOnNode]] {
	if {$node == $newNode} {
	  puts -nonewline "*$node $_jobsOnNode($node) "
	} else {
	  puts -nonewline " $node $_jobsOnNode($node) "
	}
      }
      puts ""
    }
  }

  
  $job setState "running"
  $process start $command
  $experiment start $job

  # CAUTION: no more code here

}

## ----------------------------------------------------------------------------
## stopJob
## ----------------------------------------------------------------------------
body YadaMachine::stopJob {job} {
  if {[$job hasPassedState "stopping"]} {
    return
  }
  $job setState "stopping"

  if {[info exists _timerOfJob($job)]} {
    set duration [expr ([clock seconds] - $_timerOfJob($job)) * 1000]
    _addTime $duration
    unset _timerOfJob($job)
  }
  
  # remove from job queue
  set index [lsearch -exact $_jobQueue $job]
  if {$index >= 0} {
    set _jobQueue [lreplace $_jobQueue $index $index]
  }

  # remove from run queue
  set index [lsearch -exact $_runQueue $job]
  if {$index >= 0} {
    set _runQueue [lreplace $_runQueue $index $index]
  }

  # count down jobs per node
  set node [$job cget -node]
  if {$node != ""} {
    catch {
      incr _jobsOnNode($node) -1
      if {$_jobsOnNode($node) < 0} {
	set _jobsOnNode($node) 0
	#puts "ERROR: negative job count on node $node"
      }
    }
  }
  
  itcl::delete object $job
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaMachine::destructor {} {
  foreach job "$_jobQueue $_runQueue" {
    stopJob $job
  }
}

## ----------------------------------------------------------------------------
## update machine statistics.
## This machine adds the given time to the internal timer statistics.
## Incremental averaging is computed like this:
## - \f$x1 \f$
## - \f$ x_1 * 1/2 + x_2/2 = (x_1 + x_2)/2 \f$
## - \f$ (x_1 * 1/2 + x_2/2) * 2/3 + x_3/3 = (x_1 + x_2 + x_3)/3 \f$
## - \f$ ((x_1 * 1/2 + x_2/2) * 2/3 + x_3/3) * 3/4 + x4/4 = (x_1 + x_2 + x_3 + x_4)/4 \f$
## - ...
## \param time a temporal span
## \returns the computed average time per job as calculated for this machine.
## ----------------------------------------------------------------------------
body YadaMachine::_addTime {time} {
  if {$time <= 0} {
    return $_timePerJob
  }

  incr _noComputedJobs

  set _timePerJob \
    [expr $_timePerJob * (($_noComputedJobs -1.0) / ($_noComputedJobs + 0.0)) \
	  + $time / ($_noComputedJobs + 0.0)]

  return $_timePerJob
}

## ----------------------------------------------------------------------------
## getTimePerJob
## ----------------------------------------------------------------------------
body YadaMachine::getTimePerJob {} {
  return $_timePerJob
}

## ----------------------------------------------------------------------------
## nodes
## ----------------------------------------------------------------------------
configbody YadaMachine::nodes {

  # delete removed nodes
  foreach node [array names _jobsOnNode] {
    if {[lsearch $nodes $node] < 0} {
      unset _jobsOnNode($node)
    }
  }


  # add new nodes
  set currentNodes [lsort $nodes]
  set nodes ""
  foreach node $currentNodes {
    # exclude disabled nodes
    if {[regexp {\(.*\)} $node]} {
      catch {unset _jobsOnNode($node)}
      lappend nodes $node
      continue
    }
    
    # test reachability
    if {[catch "exec fping -t100 -q $node >/dev/null 2>&1"]} {
      catch {unset _jobsOnNode($node)}
      lappend nodes "($node)"
      continue;
    } else {
      lappend nodes $node
    }

    if {![info exists _jobsOnNode($node)]} {
      set _jobsOnNode($node) 0
    }
  }
}
