# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaRunner - document to run YadaJobs.
## This document is the basic implementation of a YadaRunnableDocument. A more
## specific implementation of a YadaRunnableDocument is for example the YadaGlsStats.
## We implement here the basic functionality to select an schedule YadaJobs
## that run on YadaMachines.
## \ingroup YadaRunnableDocuments
##
## \author Michael Daum 
##
## $Id: YadaRunner.tcl,v 1.20 2003/11/19 09:39:07 micha Exp $
## ----------------------------------------------------------------------------
class YadaRunner {
  inherit YadaRunnableDocument

  # variables ----------------------------------------------------------------
  private variable _totalTime 0
  private variable _totalSolTime 0
  private variable _totalSoftTime 0
  private variable _totalScore 0
  private variable _noSolutions 0
  private variable _noErrors 0
  private variable _totalWordgraphs 0
  private variable _totalStructRecall 0
  private variable _totalLabelledRecall 0
  private variable _totalLexicalRecall 0
  private variable _totalStrictRecall 0
  private variable _typeOfColumn; ## \type TclArray
  private variable _scoreOf; ## \type TclArray
  private variable _structHitsOf; ## \type TclArray
  private variable _labelledHitsOf; ## \type TclArray
  private variable _lexicalHitsOf; ## \type TclArray
  private variable _strictHitsOf; ## \type TclArray
  private variable _triesOf; ## \type TclArray 

  # methods ------------------------------------------------------------------
  public method activationHandle {}
  public method save {{fileName ""}} ;## \type TclString
  public method clear {}

  constructor {args} {}; ## \type TclList

  private method _analyseDocument {document experiment} ;## \type DomDocument, YadaExperiment
  private method _showStatistics {}
  private method _computeStatistics {}
  private method _judgeRow {row}; ## \type TclNumber
  private method _getTypeOfColumn {colNo} ;## \type TclNumber
  private method _toHtml {}
  private method _resetDataOf {wordgraphs} ;## \type TclList
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaRunner::constructor {args} {

  # reconfigure table
  $itk_component(table) configure -rows 300 -cols 12 \
    -rowtagcommand [code $this _judgeRow]

  # set types of columns
  array set _typeOfColumn {
    0 "dictionary"
    1 "integer"
    2 "real"
    3 "integer"
    4 "integer"
    5 "integer"
    6 "integer"
    7 "real"
    8 "real"
    9 "real"
    10 "real"
  }


  itk_component add experimentBoxLabel {
    label $itk_component(heading).experimentBoxLabel \
	-text "Experiment:" \
	-anchor w
  }  {}

  itk_component add experimentBox {
    iwidgets::combobox $itk_component(heading).experimentBox \
	-completion false \
	-borderwidth 2 \
	-margin 4 \
	-selectioncommand [code $this selectExperiment]
  }  {}

  itk_component add totalScoreLabel {
    label $itk_component(heading).totalScoreLabel \
      -text "Score:" \
      -anchor w \
  } {}

  itk_component add totalScore {
    entry $itk_component(heading).totalScore \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add totalTimeLabel {
    label $itk_component(heading).totalTimeLabel \
      -text "Time:" \
      -anchor w 
  } {}

  itk_component add totalTime {
    entry $itk_component(heading).totalTime \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add noErrorsLabel {
    label $itk_component(heading).noErrorsLabel \
      -text "No.Errors:" \
      -anchor w 
  } {}

  itk_component add noErrors {
    entry $itk_component(heading).noErrors \
      -textvariable [scope _noErrors] \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add noSolutionsLabel {
    label $itk_component(heading).noSolutionsLabel \
      -text "No.Solutions:" \
      -anchor w 
  } {}

  itk_component add noSolutions {
    entry $itk_component(heading).noSolutions \
      -textvariable [scope _noSolutions] \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add totalSolTimeLabel {
    label $itk_component(heading).totalSolTimeLabel \
      -text "Sol.Time:" \
      -anchor w
  } {}

  itk_component add totalSolTime {
    entry $itk_component(heading).totalSolTime \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add totalSoftTimeLabel {
    label $itk_component(heading).totalSoftTimeLabel \
      -text "Soft.Time:" \
      -anchor w
  } {}

  itk_component add totalSoftTime {
    entry $itk_component(heading).totalSoftTime \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add totalStructRecallLabel {
    label $itk_component(heading).totalStructRecallLabel \
      -text "Struct.Recall:" \
      -anchor w
  } {}

  itk_component add totalStructRecall {
    entry $itk_component(heading).totalStructRecall \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add totalLabelledRecallLabel {
    label $itk_component(heading).totalLabelledRecallLabel \
      -text "Label.Recall:" \
      -anchor w
  } {}

  itk_component add totalLabelledRecall {
    entry $itk_component(heading).totalLabelledRecall \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add totalLexicalRecallLabel {
    label $itk_component(heading).totalLexicalRecallLabel \
      -text "Lex.Recall:" \
      -anchor w
  } {}

  itk_component add totalLexicalRecall {
    entry $itk_component(heading).totalLexicalRecall \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

  itk_component add totalStrictRecallLabel {
    label $itk_component(heading).totalStrictRecallLabel \
      -text "Strict.Recall:" \
      -anchor w
  } {}

  itk_component add totalStrictRecall {
    entry $itk_component(heading).totalStrictRecall \
      -borderwidth 2 \
      -state disabled \
      -width 20
  } {}

 # bindings
  set entry [$itk_component(experimentBox) component entry]
  bind $entry <Return> [code $this selectExperiment]
  bind $entry <Up> +[code $this selectExperiment]
  bind $entry <Down> +[code $this selectExperiment]

  # packing
  grid columnconfigure $itk_component(heading) 1 -weight 1
  grid columnconfigure $itk_component(heading) 2 -pad 5
  grid columnconfigure $itk_component(heading) 4 -weight 1
  grid columnconfigure $itk_component(heading) 5 -pad 5
  grid columnconfigure $itk_component(heading) 7 -weight 1
  grid columnconfigure $itk_component(heading) 8 -pad 5

  grid $itk_component(experimentBoxLabel)   -row 0 -column 0 -sticky ew -pady 2
  grid $itk_component(experimentBox)        -row 0 -column 1 -sticky ew -pady 2 -columnspan 4
  grid $itk_component(totalScoreLabel)  -row 1 -column 0 -sticky ew -pady 2
  grid $itk_component(totalScore)       -row 1 -column 1 -sticky ew -pady 2
  grid $itk_component(noErrorsLabel)    -row 2 -column 0 -sticky ew -pady 2
  grid $itk_component(noErrors)         -row 2 -column 1 -sticky ew -pady 2
  grid $itk_component(noSolutionsLabel) -row 3 -column 0 -sticky ew -pady 2
  grid $itk_component(noSolutions)      -row 3 -column 1 -sticky ew -pady 2

  grid $itk_component(totalTimeLabel)   -row 1 -column 3 -sticky ew -pady 2
  grid $itk_component(totalTime)        -row 1 -column 4 -sticky ew -pady 2
  grid $itk_component(totalSolTimeLabel) -row 2 -column 3 -sticky ew -pady 2
  grid $itk_component(totalSolTime)      -row 2 -column 4 -sticky ew -pady 2
  grid $itk_component(totalSoftTimeLabel) -row 3 -column 3 -sticky ew -pady 2
  grid $itk_component(totalSoftTime)      -row 3 -column 4 -sticky ew -pady 2

  grid $itk_component(totalStructRecallLabel) -row 0 -column 6 -sticky ew -pady 2
  grid $itk_component(totalStructRecall)      -row 0 -column 7 -sticky ew -pady 2
  grid $itk_component(totalLabelledRecallLabel) -row 1 -column 6 -sticky ew -pady 2
  grid $itk_component(totalLabelledRecall)      -row 1 -column 7 -sticky ew -pady 2
  grid $itk_component(totalLexicalRecallLabel)   -row 2 -column 6 -sticky ew -pady 2
  grid $itk_component(totalLexicalRecall)        -row 2 -column 7 -sticky ew -pady 2
  grid $itk_component(totalStrictRecallLabel) -row 3 -column 6 -sticky ew -pady 2
  grid $itk_component(totalStrictRecall)      -row 3 -column 7 -sticky ew -pady 2

  eval itk_initialize $args
  .main registerDocument $this
}

## ----------------------------------------------------------------------------
## activationHandle
## ----------------------------------------------------------------------------
body YadaRunner::activationHandle {} {
  chain ;# call YadaRunnableDocument::activationHandle

  # get the previous selection
  set box $itk_component(experimentBox)
  set experimentName [$box get]

  # default experiments to choose
  if {$experimentName == ""} {
    set experimentName "<none>"
  }

  set experimentNames [.main getActiveExperiments]

  set listOfEntries [[$box component list] get 0 end]
  if {$listOfEntries != $experimentNames} {
    $box clear list 
    eval $box insert list end $experimentNames
    selectExperiment $experimentName
  }
}

## ----------------------------------------------------------------------------
## clear
## ----------------------------------------------------------------------------
body YadaRunner::clear {} {
  chain ;# call YadaRunnableDocument::clear

  # reset the table 
  set table $itk_component(table)
  $table setCell \
       	0,0 "wordgraph" \
       	0,1 "values" \
       	0,2 "score" \
	0,3 "errors" \
       	0,4 "time" \
       	0,5 "sol.time" \
       	0,6 "soft time" \
	0,7 "struct" \
	0,8 "labelled" \
	0,9 "lexical" \
	0,10 "strict" 


  $table tag config default -anchor e
  $table tag config error -bg red3 -fg white -anchor e -relief flat
  $table tag config colored -bg gray80 -anchor e

  $table tag config sel -fg "" -bg "" -relief raised -bd 1
  $table tag config active -anchor w -fg black -bg gray -relief sunk -bd 1
  $table tag config title -font {-adobe-helvetica-medium-r-normal--12-*-*-*-*-*-*-*} -bg #a3a3a3 -fg white
  $table tag config runner -fg red3 -bg #a3a3a3
  $table tag raise runner title
  
  $table width 0 15 1 5 2 15 3 10 4 15 5 15 6 15 7 10 8 10 9 10 10 10

  # reset scores of selected wordgraphs
  _resetDataOf "[_getActiveWordgraphs]"

  _computeStatistics
}


## ----------------------------------------------------------------------------
## _analyseDocument
## ----------------------------------------------------------------------------
body YadaRunner::_analyseDocument {document experiment} {

  # default analysis
  YadaRunnableDocument::_analyseDocument $document $experiment

  # the wordgraph is needed at least
  if {$_analyseResult(wordgraph) == ""} {
    print "ERROR: no information available in this run\n"
    return
  }

  # save score and edge data
  set _scoreOf($_analyseResult(wordgraph)) $_analyseResult(score)
  set _structHitsOf($_analyseResult(wordgraph)) $_analyseResult(structHits)
  set _labelledHitsOf($_analyseResult(wordgraph)) $_analyseResult(labelledHits)
  set _lexicalHitsOf($_analyseResult(wordgraph)) $_analyseResult(lexicalHits)
  set _strictHitsOf($_analyseResult(wordgraph)) $_analyseResult(strictHits)
  set _triesOf($_analyseResult(wordgraph)) $_analyseResult(tries)

  set structRecall [expr $_analyseResult(structHits) * 100.0 / $_analyseResult(tries)] 
  set labelledRecall [expr $_analyseResult(labelledHits) * 100.0 / $_analyseResult(tries)] 
  set lexicalRecall [expr $_analyseResult(lexicalHits) * 100.0 / $_analyseResult(tries)]
  set strictRecall [expr $_analyseResult(strictHits) * 100.0 / $_analyseResult(tries)]   

  # add the collected data to the table
  _insertData \
    $_analyseResult(wordgraph) \
    $_analyseResult(noValues) \
    [format "%4.3e" $_analyseResult(score)] \
    [expr $_analyseResult(tries) - $_analyseResult(lexicalHits)] \
    $_analyseResult(totalTime) \
    $_analyseResult(solutionTime) \
    $_analyseResult(softTime) \
    [format "%3.2f" $structRecall ] \
    [format "%3.2f" $labelledRecall ] \
    [format "%3.2f" $lexicalRecall ] \
    [format "%3.2f" $strictRecall ]
}

## ----------------------------------------------------------------------------
## _computeStatistics
## ----------------------------------------------------------------------------
body YadaRunner::_computeStatistics {} {
  chain

  set table $itk_component(table)
  set rmax [$table cget -rows]
  set cmax [expr [$table cget -cols] -2]

  set _totalWordgraphs 0
  set _totalScore 0
  set _noSolutions 0
  set _noErrors 0
  set _totalSolTime 0
  set _totalSoftTime 0
  set _totalTime 0
  set _totalStructRecall 0
  set _totalLabelledRecall 0
  set _totalLexicalRecall 0
  set _totalStrictRecall 0

  set totalTries 0
  set totalStructHits 0
  set totalLabelledHits 0
  set totalLexicalHits 0
  set totalStrictHits 0

  for {set row 1} {$row < $rmax} {incr row} {

    set dataRow [$table getCell $row,0 $row,$cmax]
    set noValues	[lindex $dataRow 1]
    if {$noValues == ""} {
      continue
    }
    set wordgraph	[lindex $dataRow 0]
    set errors		[lindex $dataRow 3]
    set time		[lindex $dataRow 4]
    set solTime		[lindex $dataRow 5]
    set softTime	[lindex $dataRow 6]

    incr _totalWordgraphs
    incr _totalTime $time
    incr _totalSolTime $solTime
    incr _totalSoftTime $softTime
    incr _noErrors $errors

    set _totalScore [expr $_totalScore + $_scoreOf($wordgraph)]
    if {$_scoreOf($wordgraph) > 0.0} {
      incr _noSolutions
    }

    incr totalTries $_triesOf($wordgraph)
    incr totalStructHits $_structHitsOf($wordgraph)
    incr totalLabelledHits $_labelledHitsOf($wordgraph)
    incr totalLexicalHits $_lexicalHitsOf($wordgraph)
    incr totalStrictHits $_strictHitsOf($wordgraph)
  }

  if {$totalTries > 0} {
    set _totalStructRecall [expr $totalStructHits * 100.0 / $totalTries ]
    set _totalLabelledRecall [expr $totalLabelledHits * 100.0 / $totalTries ]
    set _totalLexicalRecall [expr $totalLexicalHits * 100.0 / $totalTries ]
    set _totalStrictRecall [expr $totalStrictHits * 100.0 / $totalTries ]
  }

  _showStatistics
}

## ----------------------------------------------------------------------------
## _showStatistics
## ----------------------------------------------------------------------------
body YadaRunner::_showStatistics {} {

  $itk_component(totalTime) configure -state normal
  $itk_component(totalSoftTime) configure -state normal
  $itk_component(totalSolTime) configure -state normal
  $itk_component(totalScore) configure -state normal
  $itk_component(totalStructRecall) configure -state normal
  $itk_component(totalLabelledRecall) configure -state normal
  $itk_component(totalLexicalRecall) configure -state normal
  $itk_component(totalStrictRecall) configure -state normal

  $itk_component(totalTime) delete 0 end
  $itk_component(totalSolTime) delete 0 end
  $itk_component(totalSoftTime) delete 0 end
  $itk_component(totalScore) delete 0 end
  $itk_component(totalStructRecall) delete 0 end
  $itk_component(totalLabelledRecall) delete 0 end
  $itk_component(totalLexicalRecall) delete 0 end
  $itk_component(totalStrictRecall) delete 0 end

  $itk_component(totalScore) insert 0 [format "%.4f" $_totalScore]
  $itk_component(totalTime) insert 0 "[timeToString $_totalTime]"
  $itk_component(totalSolTime) insert 0 "[timeToString $_totalSolTime]"
  $itk_component(totalSoftTime) insert 0 "[timeToString $_totalSoftTime]"
  $itk_component(totalStructRecall) insert 0 [format "%.4f%%" $_totalStructRecall]
  $itk_component(totalLabelledRecall) insert 0 [format "%.4f%%" $_totalLabelledRecall]
  $itk_component(totalLexicalRecall) insert 0 [format "%.4f%%" $_totalLexicalRecall]
  $itk_component(totalStrictRecall) insert 0 [format "%.4f%%" $_totalStrictRecall]

  $itk_component(totalTime) configure -state disabled
  $itk_component(totalSolTime) configure -state disabled
  $itk_component(totalSoftTime) configure -state disabled
  $itk_component(totalScore) configure -state disabled
  $itk_component(totalStructRecall) configure -state disabled
  $itk_component(totalLabelledRecall) configure -state disabled
  $itk_component(totalLexicalRecall) configure -state disabled
  $itk_component(totalStrictRecall) configure -state disabled
}

## ----------------------------------------------------------------------------
## _judgeRow
## ----------------------------------------------------------------------------
body YadaRunner::_judgeRow {row} {
  set errors [$itk_component(table) getCell $row,3]
  set score  [$itk_component(table) getCell $row,2]
  if {$errors > 0 || $score == 0.0} {
    return "error"
  } else {
    if {[expr $row % 2]} {
      return "default"
    } else {
      return "colored"
    }
  }
}

## ----------------------------------------------------------------------------
## _getTypeOfColumn
## ----------------------------------------------------------------------------
body YadaRunner::_getTypeOfColumn {colNo} {
  return $_typeOfColumn($colNo)
}

## ----------------------------------------------------------------------------
## _resetDataOf
## ----------------------------------------------------------------------------
body YadaRunner::_resetDataOf {wordgraphs} {
  chain $wordgraphs
  foreach wordgraph $wordgraphs {
    set _scoreOf($wordgraph) 0.0
    set _triesOf($wordgraph) 0
    set _structHitsOf($wordgraph) 0
    set _labelledHitsOf($wordgraph) 0
    set _lexicalHitsOf($wordgraph) 0
    set _strictHitsOf($wordgraph) 0
  }
}

## ----------------------------------------------------------------------------
## save
## ----------------------------------------------------------------------------
body YadaRunner::save {{fileName ""}} {

if {0} {

  if {$fileName == "" } {
    set fileName [.fileSelector activate \
      -title "Select a file" \
      -width 500 \
      -height 400 \
      -filter {"*.html"} \
      -mode "singlefile" \
    ]
    if {$fileName == ""} {
      return
    }
  }

  set fileId [open $fileName w]
  puts $fileId [_toHtml]
  close $fileId
} else {
  puts [_toHtml]
}

}

## ----------------------------------------------------------------------------
## toHtml
## ----------------------------------------------------------------------------
body YadaRunner::_toHtml {} {
  set experimentName [$itk_component(experimentBox) get]
  set table $itk_component(table)
  set titles [$table getCell 0,0 0,9]
  set cmax [expr [$table cget -cols] -2]
  set rmax [$table cget -rows]

  # head
  append result \
    "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Transitional//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd\">\n" \
    "<html xmlns=\"http://www.w3.org/1999/xhtml\" xml:lang=\"en\" lang=\"en\">\n" \
    "<head>\n" \
    " <title>YadaRunner $experimentName</title>\n" \
    "</head>\n" \
    "<body>\n" 

  # summary table
  append result \
    "<!-- begin summary  table -->\n" \
    "<table width=\"100%\" border=\"0\" cellpadding=\"0\" cellspacing=\"0\">\n" \
    "<tr>\n"

  append result "<td width=\"33%\"><table class=\"yadaTable\" >\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">Experiment:</td><td class=\"yadaCell\">$experimentName</td></tr>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">Score: </td><td class=\"yadaCell\">$_totalScore</td></tr>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">No.Solutions: </td><td class=\"yadaCell\">$_noSolutions</td></tr>\n"
  append result "</table></td>\n"
  append result "<td width=\"33%\"><table>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">Time: </td><td class=\"yadaCell\">$_totalTime</td></tr>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">Sol.Time: </td><td class=\"yadaCell\">$_totalSolTime</td></tr>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">Soft.Time: </td><td class=\"yadaCell\">$_totalSoftTime</td></tr>\n"
  append result "</table></td>\n"
  append result "<td width=\"33%\"><table>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">StructRecall: </td><td class=\"yadaCell\">$_totalStructRecall</td></tr>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">LexicalRecall: </td><td class=\"yadaCell\">$_totalLexicalRecall</td></tr>\n"
  append result "   <tr><td class=\"yadaHeader\" align=\"left\">StrictRecall: </td><td class=\"yadaCell\">$_totalStrictRecall</td></tr>\n"
  append result "</table></td>\n"

  # close summary table
  append result \
    "</tr>\n" \
    "</table>\n" \
    "<!-- end summary  table -->\n"

  # data title row
  append result \
    "<!-- begin data table -->\n" \
    "<table class=\"yadaTable\" width=\"100%\" border=\"0\">\n" \
    "<tr>\n"

  foreach title $titles {
    append result "   <th class=\"yadaHeader\" align=\"left\">$title</th>\n"
  }
  append result "</tr>\n"

  # date rows
  for {set row 1} {$row < $rmax} {incr row} {
    set dataRow [$table getCell $row,0 $row,$cmax]
    if {[lindex $dataRow 2] == ""} {
      continue
    }
    set isFirst 1
    append result "<tr>\n"
    foreach cellItem $dataRow {
      if {$isFirst} {
	set isFirst 0
	append result "<td class=\"yadaHeader\"><code>$cellItem</code></td>\n"
      } else {
	append result "<td class=\"yadaCell\"><code>$cellItem</code></td>\n"
      }
    }
    append result "</tr>\n"
  }

  # close data table
  append result \
    "</tr>\n" \
    "</table>\n" \
    "<!-- end data table -->\n" \
  
  # tail
  append result \
    "</body>\n" \
    "</html>\n" 

  return $result
}
