# Copyright (C) 1997-2003 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaWordgraph - bundled information about a wordgraph.
## This class represents a wordgraph in a YadaGrammar. Scanning a YadaGrammar
## (see YadaGrammar::scan()) results in bunch of YadaWordgraph s being constructed.
## \ingroup YadaConfiguration
##
## \author Michael Daum
##
## $Id: YadaWordgraph.tcl,v 1.8 2003/02/27 13:02:28 micha Exp $
## ----------------------------------------------------------------------------
class YadaWordgraph {

  # variables ----------------------------------------------------------------

  ## name of the wordgraph
  public variable name ""

  ## filename which this wordgraph has been extracted from
  public variable file ""

  ## linenumber in the file where this wordgraph has been defined
  public variable lineNo 0

  ## flag indicating whether the wordgraph is trivial (linear) or not
  public variable isTrivial false

  ## list of word arcs making up the wordgraph
  public variable arcs {}

  ## number of edges in the related annotation
  public variable noAnnoEdges 0

  ## number of words in the related annotation
  public variable noAnnoWords 0

  # methods ------------------------------------------------------------------
  public method readXmlNode {latticeNode}; ## \type DomNode
  public method writeXmlNode {file}; ## \type FileHandle
  public method readXmlAnno {annoNode}; ## \type DomNode
  public method getWords {}
  public method clone {args}; ## \type TclList

  constructor {args} {}; ## \type TclList
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaWordgraph::constructor {args} {
  eval configure $args
}

## ----------------------------------------------------------------------------
## readXmlAnno
## ----------------------------------------------------------------------------
body YadaWordgraph::readXmlAnno {annoNode} {
  if {[$annoNode nodeName] != "annotation"} {
    error "ERROR: `$annoNode' is not an annotation node"
  }

  set noAnnoEdges [$annoNode getAttribute "noedges"]
  set noAnnoWords [$annoNode getAttribute "nowords"]
}

## ----------------------------------------------------------------------------
## readXmlNode
## ----------------------------------------------------------------------------
body YadaWordgraph::readXmlNode {latticeNode} {
  if {[$latticeNode nodeName] != "lattice"} {
    error "ERROR: `$latticeNode' is not a lattice node"
  }

  # get the lattice id
  configure -name [$latticeNode getAttribute "id"]
  
  # get the file info
  set node [$latticeNode descendant 1 "file"]
  if {$node == ""} {
    error "ERROR: illegal xml lattice format"
  }
  set file [$node getAttribute "name"]
  set lineNo [$node getAttribute "lineNo"]

  # is it a trivial wordgraph
  set node [$latticeNode descendant 1 "boolean" "name" "trivial"]
  if {$node == ""} {
    error "ERROR: illegal xml lattice format"
  }
  set isTrivial [$node getAttribute "value"]
  
  # scan the arcs
  set arcNodes [$latticeNode getElementsByTagName "arc"]
  if {$arcNodes == ""} {
    error "ERROR: illegal xml lattice format"
  }
  set arcs ""
  foreach node $arcNodes {
    set from [$node getAttribute "from"]
    set to [$node getAttribute "to"]
    set word "[$node getAttribute word]"
    lappend arcs $from $to "$word"
  }
}

## ----------------------------------------------------------------------------
## writeXmlNode
## ----------------------------------------------------------------------------
body YadaWordgraph::writeXmlNode {file} {
  
  puts $file "<lattice id=\"$name\">"
  puts $file "  <file name=\"$file\" lineNo=\"$lineNo\"/>"
  puts $file "  <boolean name=\"trivial\" value=\"$isTrivial\"/>"
  foreach {from to word} $arcs {
    puts $file "    <arc from=\"$from\" to=\"$to\" word=\"$word\"/>"
  }
  puts $file "</lattice>"
}

## ----------------------------------------------------------------------------
## getWords
## ----------------------------------------------------------------------------
body YadaWordgraph::getWords {} {
  set words ""
  foreach {from to word} "$arcs" {
    lappend words "$word"
  }

  return "$words"
}

## ----------------------------------------------------------------------------
## clone
## ----------------------------------------------------------------------------
body YadaWordgraph::clone {args} {

  # build the clone
  set newWordgraph [YadaWordgraph ::#auto \
    -name $name \
    -file $file \
    -lineNo $lineNo \
    -isTrivial $isTrivial \
    -arcs $arcs]

  eval $newWordgraph configure $args

  return $newWordgraph
}
